#! /usr/bin/python

"""
SYNOPSIS
    General Ledger - Capturing Budgets.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

from TartanClasses import ASD, CCD, FileImport, ProgressBar, Sql, TartanDialog
from tartanFunctions import askQuestion, showError

class gl2020:
    def __init__(self, mf, conum, conam, period):
        self.mf = mf
        self.conum = conum
        self.conam = conam
        self.period = period
        self.setVariables()
        self.drawDialog()
        self.mf.startLoop()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, ["genmst", "genbud", "gentrn"],
            prog=__name__)
        self.s_cur = self.period[1][0] / 100
        self.e_cur = self.period[2][0] / 100
        self.s_old = self.s_cur - 100
        self.e_old = self.e_cur - 100

    def drawDialog(self):
        self.glm = {
            "stype": "R",
            "tables": ("genmst",),
            "cols": (("glm_acno", "", 0, "Acc-Num"),
                    ("glm_desc", "", 0, "Description", "Y")),
            "where": [("glm_cono", "=", self.conum)]}
        all = {
            "stype": "R",
            "tables": ("genbud","genmst"),
            "cols": (
                ("glb_acno", "", 0, "Acc-Num"),
                ("glm_desc", "", 0, "Description"),
                ("sum(glb_tramt) as tramt", "SI", 10, "Budget")),
            "where": [("glb_cono", "=", self.conum), ("glm_cono=glb_cono",),
                ("glb_acno=glm_acno",), ("glb_curdt", "between",
                self.period[1][0]/100, self.period[2][0]/100)],
            "group": "glb_cono, glb_acno, glm_desc",
            "order": "glb_acno"}
        bud = {
            "stype": "R",
            "tables": ("genbud",),
            "cols": (("glb_curdt", "", 0, "Period"),
                    ("glb_tramt", "", 0, "Budget")),
            "where": [("glb_cono", "=", self.conum), ("glb_curdt", "between",
                self.period[1][0]/100, self.period[2][0]/100)],
            "whera": (("C", "glb_acno", 0),)}
        tit = ("%-3d %s" % (self.conum, self.conam),
            "General Ledger Budgets Data Capture (%s)" % __name__)
        fld = (
            (("C",0,0,0),"IUI",7,"Acc-Num","Account Number",
                "r","N",self.doAcc,self.glm,None,None),
            (("C",0,0,1),"ONA",30,"Description"),
            (("C",0,0,2),"IUA",1,"F","(M)onth or (A)nnual Budget",
                "M","N",self.doFrq,None,None,None),
            (("C",0,0,3),"ID2",7,"Period","Month Period",
                "p","N",self.doPer,bud,None,None),
            (("C",0,0,4),"ISL",10,"Budget","Budget Value",
                "","N",self.doBud,None,None,None))
        row = (15,)
        but = (
            ("_Import File",None,self.doImport,0,("C", 0, 1),("C", 0, 2),
            """Import a File with the Correct Format i.e. Account Number and Annual Budget or Account Number and 12 Monthly Budgets"""),
            ("Auto _Populate",None,self.doPopulate,0,("C", 0, 1),("C", 0, 2),
            """Automatically populate all account budgets with the previous financial period actuals or budgets. This only applies to Profit and Loss accounts. Please Note that All Existing Budgets for the period will be Replaced."""),
            ("Show Budgets",all,None,0,("C", 0, 1),("C", 0, 2),
            "Display all annual budgets for all accounts"))
        self.df = TartanDialog(self.mf, title=tit, rows=row, eflds=fld,
            cend=((self.endData,"y"),), cxit=(self.exitData,), butt=but)

    def doImport(self, widget=None):
        state = self.df.disableButtons()
        tit = ("Budget File Import",)
        fle = {
            "stype":  "F",
            "types":  "fle",
            "ftype":  (("CSV & XLS Files", "*.[c,x][s,l][v,s]"),)}
        r1s = (("Annual", "A"), ("Monthly", "M"))
        fld = (
            (("T",0,0,0),("IRB",r1s),0,"Import Format","Import Format",
                "A","N",self.doImpTyp,None,None,None),)
        tnd = ((self.doImpEnd,"n"), )
        txt = (self.doImpExit, )
        self.ip = TartanDialog(self.mf, newscn=(0, 0), tops=True, title=tit,
            eflds=fld, tend=tnd, txit=txt)
        self.mf.startLoop()
        self.df.enableButtons(state=state)
        self.df.focusField(self.df.frt, self.df.pag, self.df.col)

    def doImpTyp(self, frt, pag, r, c, p, i, w):
        self.frq = w

    def doImpEnd(self):
        self.ip.closeProcess()
        impcol = [
            ["Account Number", 0, "UI", 7]]
        if self.frq == "A":
            impcol.append(["Annual Budget", 1, "SL", 10])
        else:
            for x in range(1, 13):
                impcol.append(["Month %2s" % x, x, "SL", 10])
        fi = FileImport(self.mf, impcol=impcol)
        err = None
        for num, line in enumerate(fi.impdat):
            if self.frq == "A":
                chk = 2
            else:
                chk = 13
            if len(line) != chk:
                err = "Line %s: Invalid Number of Fields (S/B %s is %s)" % \
                    (num + 1, chk, len(line))
                break
            self.acc = CCD(line[0], "UI", 7).work
            acc = self.sql.getRec(tables="genmst", where=[("glm_cono", "=",
                self.conum), ("glm_acno", "=", self.acc)], limit=1)
            if not acc:
                err = "Line %s: Invalid Account %s" % ((num + 1), self.acc)
                break
            if self.frq == "A":
                bud = CCD(line[1], "SD", 13.2)
                if bud.err:
                    err = "Line %s: %s" % ((num + 1), bud.err)
                    break
                self.bud = bud.work
                self.endData(imp=True)
                continue
            sy = self.s_cur / 100
            sm = self.s_cur % 100
            for b in line[1:]:
                self.per = (sy * 100) + sm
                bud = CCD(b, "SD", 13.2)
                if bud.err:
                    err = "Line %s: %s" % ((num + 1), bud.err)
                    break
                self.bud = bud.work
                self.endData(imp=True)
                sm += 1
                if sm > 12:
                    sm = 1
                    sy += 1
        if err:
           showError(self.mf.body, "Import Error", err)
        elif fi.impdat:
            self.mf.dbm.commitDbase()
        self.mf.closeLoop()

    def doImpExit(self, widget=None):
        self.ip.closeProcess()
        self.mf.closeLoop()

    def doPopulate(self, widget=None):
        answer = askQuestion(self.mf.body, "Are You Sure?", """Are You Sure This is what you want to do!!

Remember that all existing budgets for this financial period will be replaced!""")
        if answer == "no":
            return
        self.df.mstAlw.hide()
        self.rates = {}
        tit = ("Budget Variables",)
        r1s = (("Actuals", "A"), ("Budgets", "B"))
        r2s = (
            ("R1","0"),
            ("R10","1"),
            ("R50","2"),
            ("R100","3"),
            ("R1000","4"))
        fld = (
            (("T",0,1,0),("IRB",r1s),0,"Use Previous Year's","",
                "A","N",self.doUse,None,None,None),
            (("T",0,0,0),"ISD",7.2,"Standard Rate (+-)","",
                "","N",self.doRate,None,None,("efld",),None,"""The percentage by which the actual must be increased or decreased."""),
            (("T",0,1,0),("IRB",r2s),0,"Rounding to Nearest","",
                "0","N",self.doRound,None,None,None),
            (("C",0,0,0),"IUI",7,"Acc-Num","Account Number",
                "","N",self.doAccNo,self.glm,None,None),
            (("C",0,0,1),"ONA",30,"Description"),
            (("C",0,0,2),"ISD",7.2,"Rate","Rate (+-)",
                "","N",self.doRates,None,None,("efld",),None,"""The percentage by which the actual must be increased or decreased."""))
        tnd = ((self.doEnd,"n"),)
        txt = (self.doExit,)
        cnd = ((self.doEnd,"n"),)
        cxt = (self.doExit,)
        but = (
            ("Show",None,self.doShow,1,("C", 0, 1),("C", 0, 2)),
            ("Quit",None,self.doQuit,1,("C", 0, 1),None))
        state = self.df.disableButtons()
        self.rr = TartanDialog(self.mf, newscn=(0, 1), tops=True, title=tit,
            eflds=fld, tend=tnd, txit=txt, cend=cnd, cxit=cxt, butt=but)
        self.mf.startLoop()
        self.df.enableButtons(state=state)
        if not self.exit:
            self.doPopulation()
        self.df.mstAlw.show()
        self.df.focusField(self.df.frt, self.df.pag, self.df.col)

    def doUse(self, frt, pag, r, c, p, i, w):
        self.use = w

    def doRate(self, frt, pag, r, c, p, i, w):
        if not w:
            self.rate = 100
        else:
            self.rate = float(ASD(100) + ASD(w))

    def doRound(self, frt, pag, r, c, p, i, w):
        self.round = [0, 10, 50, 100, 1000][int(w)]

    def doExcept(self, frt, pag, r, c, p, i, w):
        if w == "N":
            self.rr.closeProcess()
            self.mf.closeLoop()
        else:
            self.rr.focusField("C", 0, 1)

    def doAccNo(self, frt, pag, r, c, p, i, w):
        desc = self.sql.getRec(tables="genmst", cols=["glm_desc"],
            where=[("glm_cono", "=", self.conum), ("glm_acno", "=", w)],
            limit=1)
        if not desc:
            return "Invalid Account Number"
        self.acno = w
        self.desc = desc[0]
        self.rr.loadEntry(frt, pag, p+1, data=self.desc)

    def doRates(self, frt, pag, r, c, p, i, w):
        self.rates[self.acno] = (self.desc, w)

    def doEnd(self):
        if self.rr.frt == "T":
            self.exit = False
            self.rr.focusField("C", 0, 1)
        else:
            self.rr.advanceLine(0)

    def doExit(self):
        if self.rr.frt == "T":
            self.exit = True
            self.rate = 0
        self.rr.closeProcess()
        self.mf.closeLoop()

    def doShow(self, widget=None):
        data = []
        keys = self.rates.keys()
        keys.sort()
        for key in keys:
            data.append((key, self.rates[key][0], self.rates[key][1]))
        titl = "Exceptions"
        head = ("Acc-Num", "Description", "Rate")
        line = {
            "stype": "C",
            "titl": titl,
            "head": head,
            "data": data}
        state = self.rr.disableButtons()
        self.mf.updateStatus("")
        self.rr.selChoice(line)
        self.rr.enableButtons(state=state)
        self.rr.focusField("C", 0, self.rr.col)

    def doQuit(self, widget=None):
        self.exit = True
        self.rate = 0
        self.rr.closeProcess()
        self.mf.closeLoop()

    def doPopulation(self):
        all = self.sql.getRec(tables="genmst", cols=["glm_acno"],
            where=[("glm_cono", "=", self.conum), ("glm_type", "=", "P")],
            order="glm_acno")
        if not all:
            return
        exc = self.rates.keys()
        p = ProgressBar(self.mf.body, typ="Populating Budgets", max=len(all),
            esc=False)
        for num, acc in enumerate(all):
            p.displayProgress(num)
            self.sql.delRec("genbud", where=[("glb_cono", "=", self.conum),
                ("glb_acno", "=", acc[0]), ("glb_curdt", "between", self.s_cur,
                self.e_cur)])
            if self.use == "A":
                bals = self.sql.getRec(tables="gentrn", cols=["glt_curdt",
                    "round(sum(glt_tramt), 2)"], where=[("glt_cono", "=",
                    self.conum), ("glt_acno", "=", acc[0]), ("glt_curdt",
                    "between", self.s_old, self.e_old)], group="glt_curdt",
                    order="glt_curdt")
            else:
                bals = self.sql.getRec(tables="genbud", cols=["glb_curdt",
                    "glb_tramt"], where=[("glb_cono", "=", self.conum),
                    ("glb_acno", "=", acc[0]), ("glb_curdt", "between",
                    self.s_old, self.e_old)], order="glb_curdt")
            for bal in bals:
                cur = bal[0] + 100
                if acc[0] in exc:
                    rate = self.rates
                    rate = float(ASD(100) + ASD(self.rates[acc[0]][1]))
                    amt = round(int(((bal[1] * rate) / 100)), 2)
                else:
                    amt = round(int(((bal[1] * self.rate) / 100)), 2)
                if self.round:
                    dif = amt % self.round
                    amt = amt - dif
                    if amt < 0 and dif > (self.round / 2):
                        amt = amt + self.round
                    elif amt > 0 and dif >= (self.round / 2):
                        amt = amt + self.round
                self.sql.insRec("genbud", data=[self.conum, acc[0], cur, amt])
        p.closeProgress()
        self.mf.dbm.commitDbase()

    def doAcc(self, frt, pag, r, c, p, i, w):
        desc = self.sql.getRec(tables="genmst", cols=["glm_desc"],
            where=[("glm_cono", "=", self.conum), ("glm_acno", "=", w)],
            limit=1)
        if not desc:
            return "Invalid Account Number"
        self.acc = w
        self.df.loadEntry(frt, pag, p+1, data=desc[0])

    def doFrq(self, frt, pag, r, c, p, i, w):
        if w not in ("M", "A"):
            return "Invalid Frequency (M/A)"
        self.frq = w
        if self.frq == "A":
            bud = self.sql.getRec(tables="genbud", cols=["sum(glb_tramt)"],
                where=[("glb_cono", "=", self.conum), ("glb_acno", "=",
                self.acc), ("glb_curdt", "between", self.s_cur, self.e_cur)],
                limit=1)
            if bud:
                self.df.loadEntry(frt, pag, p+2, data=bud[0])
            return "sk"

    def doPer(self, frt, pag, r, c, p, i, w):
        if w < self.period[1][0]/100 or w > self.period[2][0]/100:
             return "Invalid Period"
        self.per = w
        bud = self.sql.getRec(tables="genbud", cols=["glb_tramt"],
            where=[("glb_cono", "=", self.conum), ("glb_acno", "=", self.acc),
            ("glb_curdt", "=", self.per)], limit=1)
        if bud:
            self.df.loadEntry(frt, pag, p+1, data=bud[0])

    def doBud(self, frt, pag, r, c, p, i, w):
        self.bud = w

    def endData(self, imp=False):
        if self.frq == "A":
            mthamt = self.bud / 12
            p = self.period[1][0] / 100
            y = p / 100
            m = p % 100
            for x in range(12):
                per = (y * 100) + m
                whr = [("glb_cono", "=", self.conum), ("glb_acno", "=",
                    self.acc), ("glb_curdt", "=", per)]
                exst = self.sql.getRec(tables="genbud", cols=["count(*)"],
                    where=whr, limit=1)
                if exst[0]:
                    self.sql.updRec("genbud", cols=["glb_tramt"],
                    data=[mthamt], where=whr)
                else:
                    data = [self.conum, self.acc, per, mthamt]
                    self.sql.insRec("genbud", data=data)
                m = m + 1
                if m > 12:
                    y = y + 1
                    m = m - 12
        else:
            whr = [("glb_cono", "=", self.conum), ("glb_acno", "=", self.acc),
                ("glb_curdt", "=", self.per)]
            exst = self.sql.getRec(tables="genbud", cols=["count(*)"],
                where=whr, limit=1)
            if exst[0]:
                self.sql.updRec("genbud", cols=["glb_tramt"],
                data=[self.bud], where=whr)
            else:
                data = [self.conum, self.acc, self.per, self.bud]
                self.sql.insRec("genbud", data=data)
        if not imp:
            self.mf.dbm.commitDbase()
            self.df.advanceLine(0)

    def exitData(self):
        self.df.closeProcess()
        self.mf.closeLoop()

