#! /usr/bin/python

"""
SYNOPSIS
    General Ledger Journal Entries.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

import time
from TartanClasses import ASD, Batches, CCD, FileImport, GetCtl, ProgressBar
from TartanClasses import PwdConfirm, Sql, TartanDialog
from tartanFunctions import askQuestion, chkCtlAcc, getSingleRecords
from tartanFunctions import mthendDate, showError
from tartanWork import gltrtp

class gl2040:
    def __init__(self, mf, conum, conam, period, rtn, usr):
        self.mf = mf
        self.conum = conum
        self.conam = conam
        self.period = period
        if rtn == 1:
            self.jtyp = "manual"
        else:
            self.jtyp = "standard"
        self.rtn = 4
        self.capnm = usr
        no = self.checkControls()
        if not no:
            self.batchHeader()
            if self.bh.batno:
                no = self.setVariables()
                if not no:
                    if self.jtyp == "manual":
                        self.manualHeader()
                    else:
                        self.standardHeader()
                    self.mf.startLoop()

    def checkControls(self):
        # Check for Company Record
        self.gc = GetCtl(self.mf)
        ctlmst = self.gc.getCtl("ctlmst", self.conum)
        if not ctlmst:
            return "no"
        # Check for Intercompany Facility
        sql = Sql(self.mf.dbm, ["ctlmst", "genint"], prog=__name__)
        coy = sql.getRec(tables="ctlmst", cols=["count(*)"],
            where=[("ctm_cono", "<>", self.conum)], limit=1)
        if not coy[0]:
            self.incoac = False
        else:
            int = sql.getRec(tables="genint", cols=["cti_inco"],
                where=[("cti_cono", "=", self.conum)])
            if int:
                self.incoac = [self.conum]
                [self.incoac.append(coy[0]) for coy in int]
            else:
                self.incoac = False
        # Create SQL Object
        tab = ["ctlctl", "ctlmst", "ctlvmf", "ctlvtf", "genint", "genjlm",
            "genmst", "gentrn", "genrcc", "genrct"]
        self.sql = Sql(self.mf.dbm, tab, prog=__name__)
        if not self.incoac:
            self.allcoy = self.conum
            return self.doChkLoadCtls()

    def batchHeader(self):
        if self.jtyp == "standard":
            multi = "Y"
        else:
            multi = None
        self.bh = Batches(self.mf, self.conum, self.conam, self.period,
            "GEN", self.rtn, multi=multi)
        self.bh.doBatchHeader()

    def setVariables(self):
        t = time.localtime()
        self.sysdtw = ((t[0] * 10000) + (t[1] * 100) + t[2])
        self.s_per = self.period[1][0] / 100
        self.e_per = self.period[1][0] / 100

    def manualHeader(self):
        self.crspag = 2
        self.drspag = 3
        self.asspag = 4
        tit = ("%03i %s" % (self.conum, self.conam), "General Ledger " + \
            gltrtp[self.rtn - 1][1] + " Data Capture (%s)" % __name__)
        lst = {
            "stype": "R",
            "tables": ("gentrn",),
            "cols": (
                ("glt_acno", "", 0, "Acc-Num"),
                ("glt_refno", "", 0, "Ref-Num"),
                ("glt_trdt", "", 0, "   Date"),
                ("glt_tramt", "", 0, "      Amount")),
            "where": [("glt_cono", "=", self.conum), ("glt_type", "=",
                self.rtn), ("glt_batch", "=", self.bh.batno)],
            "order": "glt_refno"}
        coy = {
            "stype": "R",
            "tables": ("ctlmst",),
            "cols": (
                ("ctm_cono", "", 0, "Coy"),
                ("ctm_name", "", 0, "Name", "Y"))}
        glm = {
            "stype": "R",
            "tables": ("genmst",),
            "cols": (
                ("glm_acno", "", 0, "Acc-Num"),
                ("glm_desc", "", 0, "Description", "Y")),
            "whera": [["C", "glm_cono", 2, 0]]}
        vat = {
            "stype": "R",
            "tables": ("ctlvmf",),
            "cols": (
                ("vtm_code", "", 0, "C"),
                ("vtm_desc", "", 0, "Description", "Y")),
            "whera": [["C", "vtm_cono", 2, 0]]}
        self.fld = [
            (("T",0,0,0),"OUI",9,"Batch %s Quantity" % self.bh.batno),
            (("T",0,0,0),"OSD",13.2,"Value"),
            (("C",0,0,0),"INa",9,"Ref-Num","Reference Number",
                "i","N",self.doTrnRef,None,None,("notblank",)),
            (("C",0,0,1),"ID1",10,"Date","Transaction Date",
                "r","N",self.doTrnDat,None,None,("efld",)),
            [("C",0,0,2),"IUI",3,"Coy","Company Number",
                self.conum,"N",self.doCoyNum,coy,None,("notzero",)],
            (("C",0,0,3),"IUI",7,"Acc-Num","Account Number",
                "","N",self.doAccNum,glm,None,None),
            (("C",0,0,4),"ONA",20,"Description"),
            (("C",0,0,5),"ISD",13.2,"Amount","Transaction Amount",
                "","N",self.doTrnAmt,None,None,None),
            (("C",0,0,6),"IUA",1,"V","V.A.T. Code",
                "","N",self.doVatCod,vat,None,("notblank",)),
            (("C",0,0,7),"ISD",13.2,"VAT-Amt","V.A.T. Amount",
                "","N",self.doVatAmt,None,None,("efld",)),
            (("C",0,0,8),"INA",(20,30),"Details","Details",
                "r","N",self.doTrnDet,None,None,("notblank",))]
        if not self.incoac:
            self.fld[4][1] = "OUI"
            self.allcoy = self.conum
        row = (20,)
        but = (
            ("_Import File",None,self.doImport,0,("C",0,1),("C",0,2),
            "Import a CSV or XLS File with the Correct Format i.e. Reference "\
            "Number, Transaction Date, Company Number, Account Number, "\
            "Transaction Amount, V.A.T. Code, V.A.T. Amount, Details. "\
            "If there is a VAT Code with no VAT Amount, it will be "\
            "Calculated."),
            ("View Entries",lst,None,0,("C",0,1),("C",0,2)),
            ("End Batch",None,self.exitPage,0,("C",0,1),("C",0,2)),
            ("Abort Current",None,self.doAbort,1,None,None))
        cnd = ((self.endPage,"y"),)
        cxt = (self.exitPage,)
        self.df = TartanDialog(self.mf, title=tit, eflds=self.fld, butt=but,
            rows=row, cend=cnd, cxit=cxt)
        self.df.loadEntry("T", 0, 0, data=self.bh.batqty)
        self.df.loadEntry("T", 0, 1, data=self.bh.batval)
        self.df.focusField("C", 0, 1)

    def doImport(self, widget=None):
        self.df.closeProcess()
        impcol = []
        pos = 0
        for num, fld in enumerate(self.fld[2:]):
            if num == 4:
                continue
            if type(fld[2]) in (tuple, list):
                size = fld[2][1]
            else:
                size = fld[2]
            impcol.append([fld[4], pos, fld[1][1:], size])
            pos += 1
        r1s = (("Yes", "Y"), ("No", "N"))
        if self.incoac:
            fld = [(("T",0,0,0),("IRB",(("Yes","Y"),("No","N"))),0,
                "Allow Intercompany","","N","N",self.doAllowIC,None,None,None)]
        else:
            fld = []
            self.impint = "N"
        fi = FileImport(self.mf, impcol=impcol, impfld=fld)
        p = ProgressBar(self.mf.body, typ="Importing Journals",
            max=len(fi.impdat), esc=False)
        trans = []
        tot = 0
        err = False
        for num, line in enumerate(fi.impdat):
            p.displayProgress(num)
            if len(line) != 8:
                err = "Line %s: Invalid Number of Fields (S/B 8 is %s)" % \
                    (num + 1, len(line))
                break
            ref = CCD(line[0], "Na", 9)
            err = ref.err
            if err:
                err = "Line %s: %s" % (num + 1, err)
                break
            dte = CCD(line[1], "D1", 10)
            err = dte.err
            if not err:
                # Check if date in period
                if dte.work < self.period[1][0] or dte.work > self.period[2][0]:
                    err = "Invalid Date, Not in Financial Period"
            if err:
                err = "Line %s: %s" % (num + 1, err)
                break
            coy = CCD(line[2], "UI", 3)
            err = coy.err
            if not err:
                # Check if valid company number
                if self.impint == "N" and coy.work != self.conum:
                    err = "Invalid Company (S/B %s is %s)" % (self.conum,
                        coy.disp)
                elif self.incoac and coy.work not in self.incoac:
                    err = "Invalid Company, No Intercompany Record 1"
                else:
                    self.allcoy = coy.work
                    err = self.doChkLoadCtls()
                    if err == "rf":
                        err = "Invalid Company, Control Record Error"
            if err:
                err = "Line %s: %s" % (num + 1, err)
                break
            acc = CCD(line[3], "UI", 7)
            err = acc.err
            if not err:
                # Check if valid account number
                chk = self.sql.getRec(tables="genmst", where=[("glm_cono",
                    "=", coy.work), ("glm_acno", "=", acc.work)], limit=1)
                if not chk:
                    err = "Invalid Account Number %s" % acc.disp
                if not err:
                    err = chkCtlAcc(self.mf, coy.work, acc.work)
            if err:
                err = "Line %s: %s" % (num + 1, err)
                break
            val = CCD(line[4], "SD", 13.2)
            err = val.err
            if err:
                err = "Line %s: %s" % (num + 1, err)
                break
            elif not val.work:
                continue
            tot = float(ASD(tot) + ASD(val.work))
            cod = CCD(line[5], "UA", 1)
            err = cod.err
            if not err:
                # Check if valid vat code
                chk = self.sql.getRec(tables="ctlvmf", where=[("vtm_cono",
                    "=", coy.work), ("vtm_code", "=", cod.work)], limit=1)
                if not chk:
                    err = "Invalid VAT Code"
            if err:
                err = "Line %s: %s" % (num + 1, err)
                break
            vat = CCD(line[6], "SD", 13.2)
            err = vat.err
            if err:
                err = "Line %s: %s" % (num + 1, err)
                break
            des = CCD(line[7], "NA", 30)
            err = des.err
            if err:
                err = "Line %s: %s" % (num + 1, err)
                break
            trans.append([ref.work, dte.work, coy.work, acc.work, val.work,
                cod.work, vat.work, des.work])
        p.closeProgress()
        if not err and tot:
            err = "Debits Do Not Equal Credits"
        if err:
            showError(self.mf.body, "Import Error", err)
        else:
            for ref, dte, self.allcoy, acc, val, cod, vat, des in trans:
                self.trnref = ref
                self.trndat = dte
                if self.bh.multi == "N":
                    self.curdt = self.bh.curdt
                else:
                    self.curdt = self.trndat / 100
                self.doChkLoadCtls()
                self.acno = acc
                self.trnamt = val
                self.vatcod = cod
                if self.vatcod and not vat:
                    vatr = self.sql.getRec(tables="ctlvmf", cols=["vtm_new"],
                        where=[("vtm_cono", "=", self.conum), ("vtm_code", "=",
                        self.vatcod)], limit=1)
                    if not vatr:
                        rate = 0
                    else:
                        rate = float(vatr[0])
                    self.vatamt = round((self.trnamt * rate / (rate + 100)), 2)
                else:
                    self.vatamt = vat
                self.trndet = des
                self.updateTables()
                self.updateBatch()
            self.bh.doBatchTotal(imp=True)
        self.mf.closeLoop()

    def doAllowIC(self, frt, pag, r, c, p, i, w):
        self.impint = w
        if self.impint == "N":
            self.batind = "N"

    def doAbort(self, widget=None):
        ok = askQuestion(self.mf.body, head="Abort",
            mess="Are You Certain You Want to Cancel These Batch Postings?")
        if ok == "yes":
            self.mf.dbm.rollbackDbase()
            self.df.closeProcess()
            self.mf.closeLoop()
        else:
            self.df.focusField(self.df.frt, self.df.pag, self.df.col)

    def doTrnRef(self, frt, pag, r, c, p, i, w):
        self.trnref = w

    def doTrnDat(self, frt, pag, r, c, p, i, w):
        if w < self.period[1][0] or w > self.period[2][0]:
            return "Invalid Date, Not in Financial Period"
        self.trndat = w
        if self.bh.multi == "Y":
            self.curdt = self.trndat / 100
        else:
            self.curdt = self.bh.curdt

    def doCoyNum(self, frt, pag, r, c, p, i, w):
        acc = self.sql.getRec(tables="ctlmst", where=[("ctm_cono", "=", w)])
        if not acc:
            return "Invalid Company Number"
        if self.incoac and w not in self.incoac:
            return "Invalid Company, No Intercompany Record 1"
        self.allcoy = w
        return self.doChkLoadCtls()

    def doChkLoadCtls(self):
        # Check for Intercompany Records
        if self.allcoy != self.conum:
            acc = self.sql.getRec(tables="genint", where=[("cti_cono", "=",
                self.allcoy), ("cti_inco", "=", self.conum)], limit=1)
            if not acc:
                return "Invalid Company, No Intercompany Record 2"
        # Get Company Details
        ctlmst = self.gc.getCtl("ctlmst", self.allcoy)
        if not ctlmst:
            return "rf"
        # Set Company VAT Default
        self.taxdf = ctlmst["ctm_taxdf"]
        if not self.taxdf:
            self.taxdf = "N"
        # Check for Integrated Systems
        mod = []
        for x in range(0, 21, 2):
            mod.append(ctlmst["ctm_modules"][x:x+2])
        # Load Ass Integration
        if mod.count("AR"):
            ctl = self.gc.getCtl("assctl", self.allcoy)
            if not ctl:
                return "rf"
            self.ass_gl = ctl["cta_glint"]
        else:
            self.ass_gl = "N"
        # Load Crs Integration
        if mod.count("CR"):
            crsctl = self.gc.getCtl("crsctl", self.allcoy)
            if not crsctl:
                return "rf"
            self.crs_gl = crsctl["ctc_glint"]
        else:
            self.crs_gl = "N"
        # Load Drs Integration and Chain Store Flag
        if mod.count("DR"):
            drsctl = self.gc.getCtl("drsctl", self.allcoy)
            if not drsctl:
                return "rf"
            self.drs_gl = drsctl["ctd_glint"]
            self.drs_ch = drsctl["ctd_chain"]
        else:
            self.drs_gl = "N"
            self.drs_ch = "N"
        # Load Rtl Integration
        if mod.count("RT"):
            ctl = self.gc.getCtl("rtlctl", self.allcoy)
            if not ctl:
                return "rf"
            self.rtl_gl = ctl["ctr_glint"]
        else:
            self.rtl_gl = "N"
        # Check for Control Records e.g. VAT, Crs and Drs
        self.ctlctl = self.gc.getCtl("ctlctl", self.allcoy)
        if not self.ctlctl:
            return "rf"
        ctls = ["vat_ctl"]
        if self.crs_gl == "Y":
            ctls.append("crs_ctl")
        if self.drs_gl == "Y":
            ctls.append("drs_ctl")
        if self.gc.chkRec(self.allcoy, self.ctlctl, ctls):
            return "rf"
        self.convat = self.ctlctl["vat_ctl"]
        # Get for Asset Accounts
        if self.ass_gl == "Y":
            self.assctl = []
            ass = self.sql.getRec(tables="assgrp", cols=["asg_assacc"],
                where=[("asg_cono", "=", self.allcoy)])
            for a in ass:
                self.assctl.append(a[0])
        # Get for Rental Accounts
        if self.rtl_gl == "Y":
            self.rtlctl = []
            rtl = self.sql.getRec(tables="rtlprm", cols=["rtp_rtlacc"],
                where=[("rtp_cono", "=", self.allcoy)])
            for a in rtl:
                self.rtlctl.append(a[0])
        # Set Batch Indicator
        if self.allcoy == self.conum:
            self.batind = "N"
        else:
            self.batind = ""

    def doAccNum(self, frt, pag, r, c, p, i, w):
        acc = self.sql.getRec(tables="genmst", where=[("glm_cono", "=",
            self.allcoy), ("glm_acno", "=", w)], limit=1)
        if not acc:
            return "Invalid Account"
        self.vattot = 0
        dic = self.sql.genmst_dic
        if self.ass_gl == "Y" and w in self.assctl:
            ctl = True
        elif self.crs_gl == "Y" and w == self.ctlctl["crs_ctl"]:
            ctl = True
        elif self.drs_gl == "Y" and w == self.ctlctl["drs_ctl"]:
            ctl = True
        elif self.rtl_gl == "Y" and w in self.rtlctl:
            ctl = True
        else:
            ctl = False
        if ctl:
            cf = PwdConfirm(self.mf, conum=self.allcoy, system="GEN",
                code="CtlAcc")
            if not cf.pwd or cf.flag == "no":
                return "This Is An Integrated Control Account"
        elif acc[dic["glm_ind"][1]] == "N":
            cf = PwdConfirm(self.mf, conum=self.allcoy, system="GEN",
                code="NoPost")
            if not cf.pwd or cf.flag == "no":
                return "Postings Disallowed"
        self.acno = w
        if not acc[dic["glm_vat"][1]]:
            if acc[dic["glm_type"][1]] == "P":
                self.taxgl = self.taxdf
            else:
                self.taxgl = "N"
        else:
            self.taxgl = acc[dic["glm_vat"][1]]
        self.df.loadEntry("C", 0, p+1, acc[dic["glm_desc"][1]])

    def doTrnAmt(self, frt, pag, r, c, p, i, w):
        if w == 0:
            return "Invalid Amount"
        self.trnamt = w
        self.df.loadEntry("C", 0, p+1, self.taxgl)

    def doVatCod(self, frt, pag, r, c, p, i, w):
        vat = self.sql.getRec(tables="ctlvmf", cols=["vtm_new"],
            where=[("vtm_cono", "=", self.conum), ("vtm_code", "=", w)],
            limit=1)
        if not vat:
            return "Invalid V.A.T. Code"
        self.vatcod = w
        rate = float(vat[0])
        vat = str(round((self.trnamt * rate / (rate + 100)), 2))
        vat = self.df.loadEntry("C", 0, p+1, data=vat)
        self.vatamt = vat.work
        if not self.vatamt:
            return "sk"

    def doVatAmt(self, frt, pag, r, c, p, i, w):
        if self.trnamt < 0 and w > 0:
            self.vatamt = float(ASD(0) - ASD(w))
        elif self.trnamt > 0 and w < 0:
            self.vatamt = float(ASD(0) - ASD(w))
        else:
            self.vatamt = w
        self.df.loadEntry(frt, pag, p, data=self.vatamt)

    def doTrnDet(self, frt, pag, r, c, p, i, w):
        self.trndet = w

    def standardHeader(self):
        tit = ("%03i %s" % (self.conum, self.conam), "General Ledger " + \
            gltrtp[self.rtn - 1][1] + " Data Capture (%s)" % __name__)
        r1s = (
            ("Monthly", "M"),
            ("Quarterly", "3"),
            ("Bi-Annually", "6"),
            ("Annually", "Y"))
        r2s = (("Yes", "Y"), ("No", "N"))
        fld = (
            (("T",0,0,0),("IRB",r1s),0,"Frequency","Frequency (M/3/6/Y)",
                "M","N",self.doFreq,None,None,None),
            (("T",0,1,0),("IRB",r2s),9,"All Journals","All Journals",
                "N","N",self.doAllJnl,None,None,None),
            (("T",0,2,0),("IRB",r2s),1,"All Periods","All Periods",
                "N","N",self.doAllPer,None,None,None))
        tnd = ((self.endPage,"y"),)
        txt = (self.exitPage,)
        self.df = TartanDialog(self.mf, title=tit, eflds=fld,
            tend=tnd, txit=txt)

    def doFreq(self, frt, pag, r, c, p, i, w):
        self.freq = w
        self.wher = [("gjm_cono", "=", self.conum), ("gjm_freq", "=",
            self.freq), ("gjm_start","<=",self.bh.curdt), ("gjm_end", ">=",
            self.bh.curdt), ("gjm_last", "<", self.bh.curdt)]
        data = self.sql.getRec(tables="genjlm", cols=["gjm_num", "gjm_desc"],
            where=self.wher)
        if not data:
            return "No Valid Standard Journals"
        if self.freq == "M":
            self.mths = 1
        elif self.freq == "3":
            self.mths = 3
        elif self.freq == "6":
            self.mths = 6
        else:
            self.mths = 12

    def doAllJnl(self, frt, pag, r, c, p, i, w):
        self.alljnl = w

    def doAllPer(self, frt, pag, r, c, p, i, w):
        self.allper = w

    def endPage(self):
        if self.jtyp == "manual":
            self.updateTables()
            self.updateBatch()
            self.cancel = False
            self.agecan = False
            self.df.advanceLine(0)
        else:
            self.df.closeProcess()
            if self.alljnl == "N":
                all = getSingleRecords(self.mf, "genjlm", ("gjm_num",
                    "gjm_desc"), where=self.wher)
            else:
                all = self.sql.getRec(tables="genjlm", where=self.wher)
            if all:
                for gjm in all:
                    self.batind = "N"
                    self.trnref = gjm[self.sql.genjlm_col.index("gjm_num")]
                    self.trndet = gjm[self.sql.genjlm_col.index("gjm_desc")]
                    start = gjm[self.sql.genjlm_col.index("gjm_start")]
                    last = gjm[self.sql.genjlm_col.index("gjm_last")]
                    dates = []
                    while not start > self.bh.curdt:
                        if start < self.s_per or start <= last:
                            pass
                        elif self.allper == "N" and start == self.bh.curdt:
                            dates.append(start)
                        elif self.allper == "Y" and not start > self.bh.curdt:
                            dates.append(start)
                        start = self.nextPeriod(start)
                    for self.curdt in dates:
                        self.trndat = mthendDate((self.curdt * 100))
                        data = self.sql.getRec(tables="genjlt",
                            cols=["gjt_acno", "gjt_amnt", "gjt_vatc"],
                            where=[("gjt_cono", "=", self.conum),
                                ("gjt_num", "=", self.trnref),
                                ("gjt_amnt", "<>", 0)])
                        if not data:
                            continue
                        for tran in data:
                            self.acno = tran[0]
                            self.trnamt = tran[1]
                            self.vatcod = tran[2]
                            vat = self.sql.getRec(tables="ctlvmf",
                                cols=["vtm_new"], where=[("vtm_cono", "=",
                                self.conum), ("vtm_code", "=", self.vatcod)],
                                limit=1)
                            if not vat:
                                rate = 0
                            else:
                                rate = float(vat[0])
                            self.vatamt = round((self.trnamt*rate/(rate+100)),2)
                            self.updateTables()
                            self.updateBatch()
                        self.sql.updRec("genjlm", cols=["gjm_last"],
                            data=[self.curdt], where=[("gjm_cono", "=",
                            self.conum), ("gjm_num", "=", self.trnref),
                            ("gjm_freq", "=", self.freq)])
            self.closeBatch()
            self.mf.closeLoop()

    def nextPeriod(self, period):
        yy = period / 100
        mm = (period % 100) + self.mths
        if mm > 12:
            yy += 1
            mm -= 12
        return (yy * 100) + mm

    def updateTables(self):
        if self.allcoy != self.conum:
            inter = "y"
            ac1 = self.sql.getRec(tables="genint", cols=["cti_acno"],
                where=[("cti_cono", "=", self.conum), ("cti_inco", "=",
                self.allcoy)], limit=1)[0]
            ac2 = self.sql.getRec(tables="genint", cols=["cti_acno"],
                where=[("cti_cono", "=", self.allcoy), ("cti_inco", "=",
                self.conum)], limit=1)[0]
        else:
            inter = "n"
        amt = float(ASD(self.trnamt) - ASD(self.vatamt))
        data = (self.allcoy, self.acno, self.curdt, self.trndat, self.rtn,
            self.trnref, self.bh.batno, amt, self.vatamt, self.trndet,
            self.vatcod, self.batind, 0, self.sysdtw, 0)
        self.sql.insRec("gentrn", data=data)
        if self.vatcod:
            if self.vatamt:
                data = (self.allcoy, self.convat, self.curdt, self.trndat,
                    self.rtn, self.trnref, self.bh.batno, self.vatamt, 0.00,
                    self.trndet, self.vatcod, self.batind, 0, self.sysdtw, 0)
                self.sql.insRec("gentrn", data=data)
            if amt < 0:
                vtyp = "O"
            else:
                vtyp = "I"
            data = (self.allcoy, self.vatcod, vtyp, self.curdt, "G", self.rtn,
                self.bh.batno, self.trnref, self.trndat, self.acno, self.trndet,
                amt, self.vatamt, self.sysdtw, 0, 0)
            self.sql.insRec("ctlvtf", data=data)
        if inter == "y":
            data = (self.conum, ac1, self.curdt, self.trndat, self.rtn,
                self.trnref, self.bh.batno, self.trnamt, 0.00, self.trndet,
                self.vatcod, "N", 0, self.sysdtw, 0)
            self.sql.insRec("gentrn", data=data)
            val = float(ASD(0) - ASD(self.trnamt))
            data = (self.allcoy, ac2, self.curdt, self.trndat, self.rtn,
                self.trnref, self.bh.batno, val, 0.00, self.trndet,
                self.vatcod, "", 0, self.sysdtw, 0)
            self.sql.insRec("gentrn", data=data)

    def updateBatch(self, rev=False):
        if rev:
            self.bh.batqty = self.bh.batqty - 1
            self.bh.batval = float(ASD(self.bh.batval) - ASD(self.trnamt))
        else:
            self.bh.batqty = self.bh.batqty + 1
            self.bh.batval = float(ASD(self.bh.batval) + ASD(self.trnamt))
        if self.jtyp == "manual":
            self.df.loadEntry("T", 0, 0, data=self.bh.batqty)
            self.df.loadEntry("T", 0, 1, data=self.bh.batval)

    def exitPage(self, widget=None):
        if self.jtyp == "standard":
            self.df.closeProcess()
            self.mf.closeLoop()
        elif self.bh.batval != 0:
            err = "Debits Do Not Equal Credits"
            self.mf.updateStatus(err)
            showError(self.mf.body, "Total Error", err)
            self.df.focusField(self.df.frt, self.df.pag, self.df.col)
        else:
            self.df.closeProcess()
            self.closeBatch()
            self.mf.closeLoop()

    def closeBatch(self):
        if self.jtyp == "manual":
            self.bh.doBatchTotal()
        else:
            self.sql.updRec("gentrn", cols=["glt_batind"], data=["Y"],
                where=[("glt_cono", "=", self.conum), ("glt_type", "=",
                self.rtn), ("glt_batch", "=", self.bh.batno), ("glt_curdt",
                "between", self.s_per, self.e_per), ("glt_batind", "=", "N")])
            self.mf.dbm.commitDbase()
