#! /usr/bin/python

"""
SYNOPSIS
    General Ledger Trial Balance.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

import time
from TartanClasses import ASD, CCD, ProgressBar, Sql, TartanDialog
from tartanFunctions import doPrinter, doReportlab, doWriteExport, getModName
from tartanFunctions import showError

class gl3040:
    def __init__(self, mf, coy, nam, per):
        self.mf = mf
        self.conum = coy
        self.conam = nam
        self.period = per
        self.setVariables()
        self.mainProcess()
        self.mf.startLoop()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, ["genbal", "genmst", "gentrn"],
            prog=__name__)
        t = time.localtime()
        self.sysdttm = "(Printed on: %i/%02i/%02i at %02i:%02i) %6s" % (t[0],
            t[1], t[2], t[3], t[4], __name__)
        self.head = ("%03u %-30s %s" % (self.conum, self.conam, "%s"))
        self.colsh = ["Acc-Num", "T", "Description", "Debit", "Credit"]
        self.forms = [("UI", 7), ("UA", 1), ("NA", 40)] + [("SD", 13.2)] * 2
        self.s_per = self.period[1][0] / 100
        self.e_per = self.period[2][0] / 100
        self.dtot = 0.00
        self.ctot = 0.00
        self.gp = 0.00
        self.t = None

    def mainProcess(self):
        self.tit = ("%03i %s" % (self.conum, self.conam),
            "General Ledger Trial Balance (%s)" % __name__)
        r1s = (("Yes","Y"), ("No","N"))
        fld = (
            (("T",0,0,0),("IRB",r1s),0,"Opening Balances Only",
                "Opening Balances Only","N","Y",self.doOpeBal1,None,
                None,None),
            (("T",0,1,0),("IRB",r1s),0,"Include Opening Balances",
                "Include Opening Balances","Y","Y",self.doOpeBal2,None,
                None,None),
            (("T",0,2,0),"Id2",7,"Starting Period","Starting Period",
                self.s_per,"Y",self.doStartPer,None,None,("efld",)),
            (("T",0,3,0),"Id2",7,"Ending Period","Ending Period",
                self.e_per,"Y",self.doEndPer,None,None,("efld",)),
            (("T",0,4,0),("IRB",r1s),0,"Ignore Zero Balances",
                "Ignore Zero Balances","Y","Y",self.doZerBal,None,None,None))
        tnd = ((self.doEnd,"y"), )
        txt = (self.doExit, )
        self.df = TartanDialog(self.mf, title=self.tit, eflds=fld, tend=tnd,
            txit=txt, view=("Y","V"), mail=("Y","N"))

    def doOpeBal1(self, frt, pag, r, c, p, i, w):
        self.opebal = w
        if self.opebal == "Y":
            self.start = None
            self.end = None
            self.incope = None
            return "sk3"

    def doOpeBal2(self, frt, pag, r, c, p, i, w):
        self.incope = w
        if self.incope == "Y":
            self.start = self.s_per
            self.df.loadEntry("T", 0, 2, data=self.start)
            return "sk1"

    def doStartPer(self, frt, pag, r, c, p, i, w):
        if w < self.s_per or w > self.e_per:
            return "Invalid Period, Outside Financial Period"
        self.start = w

    def doEndPer(self, frt, pag, r, c, p, i, w):
        if w < self.start:
            return "Invalid Period, Before Start Period"
        if w > self.e_per:
            return "Invalid Period, Outside Financial Period"
        self.end = w

    def doZerBal(self, frt, pag, r, c, p, i, w):
        self.zerbal = w

    def doEnd(self):
        self.df.closeProcess()
        all = self.sql.getRec(tables="genmst", cols=["glm_acno",
            "glm_desc", "glm_type"], where=[("glm_cono", "=", self.conum)],
            order="glm_type desc, glm_acno")
        if not all:
            showError(self.mf.body, "Selection Error", "No Accounts")
        elif self.df.repprt[0][2] == "export":
            self.exportReport(all)
        else:
            self.printReport(all)
        self.closeProcess()

    def exportReport(self, all):
        p = ProgressBar(self.mf.body, max=len(all), esc=True)
        expnam = getModName(self.mf.wrkdir, __name__, self.conum)
        self.expheads = [self.head % self.sysdttm]
        if self.opebal == "Y":
            date = CCD(self.period[1][0], "D1", 10.0)
            self.expheads.append("General Ledger Opening Balances as at %s" % \
                date.disp)
        else:
            self.expheads.append("General Ledger Trial Balance for Period "\
                "%s to %s" % (self.df.t_disp[0][0][2], self.df.t_disp[0][0][3]))
            self.expheads.append("(Options: Opening Balances Included %s)" % \
                self.incope)
        self.expcolsh = [self.colsh]
        self.expforms = self.forms
        self.expdatas = []
        for num, dat in enumerate(all):
            can = p.displayProgress(num)
            if can == "yes":
                p.closeProgress()
                return
            vals = self.getValues(dat)
            if not vals:
                continue
            self.expdatas.append(["BODY", [vals[0].work, vals[1].work,
                vals[2].work, vals[3].work, vals[4].work]])
        p.closeProgress()
        self.grandTotal()
        doWriteExport(xtype=self.df.repprt[0][1], name=expnam, sheet=__name__,
            heads=self.expheads, colsh=self.expcolsh, forms=self.expforms,
            datas=self.expdatas, rcdic=self.mf.rcdic)

    def printReport(self, all):
        chrs = len(self.colsh)
        for f in self.forms:
            chrs += int(f[1])
        if f[0][0].lower() == "s":
            chrs -= 2
        else:
            chrs -= 1
        pad = chrs - 35 - len(self.sysdttm)
        self.head1 = self.head % (" " * pad + self.sysdttm)
        if self.opebal == "Y":
            date = CCD(self.period[1][0], "D1", 10.0)
            self.head2 = "General Ledger Opening Balances as at %s%s" % \
                (date.disp, "%s%s")
        else:
            self.head2 = "General Ledger Trial Balance for Period %s to "\
                "%s%s" % (self.df.t_disp[0][0][2], self.df.t_disp[0][0][3],
                "%s%s")
        pad = chrs - len(self.head2) + 4 - 11 # %s%s and ' Page     1'
        self.head2 = self.head2 % (" " * pad, " Page %5s")
        p = ProgressBar(self.mf.body, max=len(all), esc=True)
        pdfnam = getModName(self.mf.wrkdir, __name__, self.conum, ext="pdf")
        self.c, self.ogn, self.cpi, self.ldg, self.lpp, self.suc, self.duc = \
            doReportlab(self.mf.prgdir, pdfnam, self.head1)
        self.pgnum = 0
        self.pglin = 99
        for num, rec in enumerate(all):
            can = p.displayProgress(num)
            if can == "yes":
                if self.t:
                    self.t.textLine()
                    self.t.textLine("Print Job ABORTED")
                break
            vals = self.getValues(rec)
            if not vals:
                continue
            if self.pglin > self.lpp:
                self.doNewPage()
            self.t.textLine("%s %1s %-40s %s %s" % (vals[0].disp, vals[1].disp,
                vals[2].disp, vals[3].disp, vals[4].disp))
            self.pglin = self.pglin + 1
        p.closeProgress()
        if self.t:
            if can != "yes":
                self.grandTotal()
            self.c.drawText(self.t)
        self.c.showPage()
        self.c.save()
        doPrinter(mf=self.mf, conum=self.conum, rep=pdfnam, head=self.tit,
            view=self.df.repprt[0][1], prtnam=self.df.repprt[0][2],
            mail=self.df.repprt[1][2])
        self.closeProcess()

    def getValues(self, data):
        acno = CCD(data[0], "UI", 7)
        desc = CCD(data[1], "NA", 40)
        atyp = CCD(data[2], "UA", 1)
        bal = 0.0
        if self.opebal == "Y" or self.incope == "Y":
            o = self.sql.getRec(tables="genbal", cols=["glo_cyr"],
                where=[("glo_cono", "=", self.conum), ("glo_acno", "=",
                acno.work), ("glo_trdt", "=", self.period[1][0])], limit=1)
            if o:
                b = CCD(o[0], "SD", 13.2)
            else:
                b = CCD(0, "SD", 13.2)
            bal = float(ASD(bal) + ASD(b.work))
        if self.opebal == "N" and self.incope == "Y":
            o = self.sql.getRec(tables="gentrn",
                cols=["round(sum(glt_tramt),2)"], where=[("glt_cono", "=",
                self.conum), ("glt_acno", "=", acno.work), ("glt_curdt",
                ">", self.s_per), ("glt_curdt", "<", self.start)], limit=1)
            if o and o[0]:
                b = CCD(float(o[0]), "SD", 13.2)
            else:
                b = CCD(0, "SD", 13.2)
            bal = float(ASD(bal) + ASD(b.work))
        if self.opebal == "N":
            o = self.sql.getRec(tables="gentrn",
                cols=["round(sum(glt_tramt),2)"], where=[("glt_cono", "=",
                self.conum), ("glt_acno", "=", acno.work), ("glt_curdt",
                "between", self.start, self.end)], limit=1)
            if o and o[0]:
                b = CCD(float(o[0]), "SD", 13.2)
            else:
                b = CCD(0, "SD", 13.2)
            bal = float(ASD(bal) + ASD(b.work))
        if bal == 0 and self.zerbal == "Y":
            return
        if bal < 0:
            cr = CCD(bal, "SD", 13.2)
            dr = CCD(0.00, "SD", 13.2)
        else:
            dr = CCD(bal, "SD", 13.2)
            cr = CCD(0.00, "SD", 13.2)
        self.dtot = float(ASD(self.dtot) + ASD(dr.work))
        self.ctot = float(ASD(self.ctot) + ASD(cr.work))
        if atyp.work == "P":
            self.gp = float(ASD(self.gp) + ASD(dr.work) + ASD(cr.work))
        return (acno, atyp, desc, dr, cr)

    def pageHeading(self):
        self.pgnum = self.pgnum + 1
        self.t = self.c.beginText(self.ogn[0], self.ogn[1])
        self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
        self.t.textLine(self.head1)
        self.t.textLine()
        self.t.textLine(self.head2 % self.pgnum)
        if self.opebal == "N":
            self.t.textLine()
            self.t.textLine("(Options: Opening Balances Included %s)" % \
                (self.incope))
        self.t.textLine()
        self.t.textLine("%-7s %-1s %-40s %13s %13s" % \
            ("Acc-Num", "T", "Description", "Debit ", "Credit "))
        self.t.textLine("%-86s" % (self.suc * len(self.head1)))
        self.t.setFont("Tartan", self.cpi, self.ldg)
        if self.opebal == "Y":
            self.pglin = 6
        else:
            self.pglin = 8
    
    def grandTotal(self):
        if self.df.repprt[0][2] == "export":
            self.expdatas.append(["ULINES"])
            self.expdatas.append(["TOTAL", ["", "", "Grand Totals", self.dtot,
                self.ctot]])
            self.expdatas.append(["ULINED"])
            return
        d = CCD(self.dtot, "SD", 13.2)
        c = CCD(self.ctot, "SD", 13.2)
        self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
        self.t.textLine("%-84s" % (self.suc * len(self.head1)))
        self.t.textLine("%9s %-40s %13s %13s" % \
            ("", "Grand-Totals", d.disp, c.disp))
        self.pglin += 2
        if self.pglin > self.lpp:
            self.doNewPage()
        gp = CCD(self.gp, "SD", 13.2)
        if gp.work > 0:
            self.t.textLine()
            self.t.textLine("%9s %-40s %13s %13s" % \
                ("", "Gross-Loss", gp.disp, ""))
        else:
            self.t.textLine()
            self.t.textLine("%9s %-40s %13s %13s" % \
                ("", "Gross-Profit", "", gp.disp))
        self.pglin += 2
        diff = CCD(float(ASD(d.work) + ASD(c.work)), "SD", 13.2)
        if diff.work:
            if self.pglin > self.lpp:
                self.doNewPage()
            self.t.textLine()
            self.t.textLine("%9s %-40s %13s %13s" % \
                ("", "Difference", "", diff.disp))
        self.t.setFont("Tartan", self.cpi, self.ldg)

    def doNewPage(self):
        if self.pgnum != 0:
            self.c.drawText(self.t)
            self.c.showPage()
        self.pageHeading()

    def doExit(self):
        self.df.closeProcess()
        self.closeProcess()

    def closeProcess(self):
        self.mf.closeLoop()
