#! /usr/bin/python

"""
SYNOPSIS
    General Ledger Bank Reconciliation.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

import time
from TartanClasses import ASD, CCD, ProgressBar, Sql, TartanDialog
from tartanFunctions import doReportlab, getModName, doPrinter, mthendDate

class gl3080:
    def __init__(self, mf, coy, nam, per):
        self.mf = mf
        self.conum = coy
        self.conam = nam
        self.period = per
        self.setVariables()
        self.mainProcess()
        self.mf.startLoop()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, ["ctlctl", "genbal", "genmst", "gentrn"],
            prog=__name__)
        self.pdfnam = getModName(self.mf.wrkdir, __name__, self.conum,
            ext="pdf")
        t = time.localtime()
        self.sysdttm = "(Printed on: %i/%02i/%02i at %02i:%02i)" % \
            (t[0], t[1], t[2], t[3], t[4])
        self.s_per = self.period[1][0] / 100
        self.e_per = self.period[2][0] / 100
        self.c_per = (t[0] * 100) + t[1]

    def mainProcess(self):
        self.tit = ("%03i %s" % (self.conum, self.conam),
            "General Ledger Bank Reconciliation (%s)" % __name__)
        glm = {
            "stype": "R",
            "tables": ("ctlctl", "genmst"),
            "cols": (
                ("ctl_conacc", "", 0, "Acc-Num"),
                ("glm_desc", "", 0, "Description", "Y")),
            "where": [("ctl_cono", "=", self.conum), ("ctl_code", "like",
                "bank_%"), ("glm_cono=ctl_cono",), ("glm_acno=ctl_conacc",)]}
        fld = (
            (("T",0,0,0),"IUI",7,"Bank Account","Bank Account",
                "","Y",self.doBankAcc,glm,None,("efld",)),
            (("T",0,1,0),"Id2",7,"Accounting Period","Accounting Period",
                self.c_per,"Y",self.doPer,None,None,("efld",)))
        tnd = ((self.doEnd,"y"), )
        txt = (self.doExit, )
        self.df = TartanDialog(self.mf, title=self.tit, eflds=fld, tend=tnd,
            txit=txt, view=("N","V"), mail=("Y","N"))

    def doBankAcc(self, frt, pag, r, c, p, i, w):
        acc = self.sql.getRec(tables=["ctlctl", "genmst"], cols=["glm_desc"],
            where=[("glm_cono", "=", self.conum), ("glm_acno", "=", w),
            ("ctl_cono=glm_cono",), ("ctl_conacc=glm_acno",), ("ctl_code",
            "like", "bank_%")], limit=1)
        if not acc:
            return "Invalid Bank Account Number"
        self.acnow = w
        self.acnod = self.df.t_disp[pag][r][p]

    def doPer(self, frt, pag, r, c, p, i, w):
        self.perw = w
        self.perd = self.df.t_disp[pag][r][p]

    def doEnd(self):
        self.df.closeProcess()
        self.extractBalance()
        self.sortPayments()
        self.sortReceipts()
        self.sortImports()
        if self.bal != 0 or self.pays or self.recs or self.imps:
            self.printReport()
        self.closeProcess()

    def sortPayments(self):
        col = ["glt_trdt", "glt_refno", "glt_desc", "glt_tramt"]
        whr = [("glt_cono", "=", self.conum), ("glt_acno", "=", self.acnow),
            ("glt_curdt", "<=", self.perw), ("(", "glt_recon", "=", 0, "or",
            "glt_recon", ">", self.perw, ")"), ("glt_type", "in", (2, 4))]
        ord = "glt_curdt, glt_trdt, glt_type, glt_refno, glt_batch"
        self.pays = self.sql.getRec(tables="gentrn", cols=col, where=whr,
            order=ord)
        if not self.pays:
            self.pays = []

    def sortReceipts(self):
        col = ["glt_trdt", "glt_refno", "glt_desc", "glt_tramt"]
        whr = [("glt_cono", "=", self.conum), ("glt_acno", "=", self.acnow),
            ("glt_curdt", "<=", self.perw), ("(", "glt_recon", "=", 0, "or",
            "glt_recon", ">", self.perw, ")"), ("glt_type", "=", 6)]
        ord = "glt_curdt, glt_trdt, glt_type, glt_refno, glt_batch"
        self.recs = self.sql.getRec(tables="gentrn", cols=col, where=whr,
            order=ord)
        if not self.recs:
            self.recs = []

    def sortImports(self):
        dte = mthendDate((self.perw * 100) + 1)
        col = ["grt_date", "grt_refno", "grt_memo", "grt_amount"]
        whr = [("grt_cono", "=", self.conum), ("grt_acno", "=", self.acnow),
            ("grt_date", "<=", dte), ("grt_flag", "=", "N")]
        ord = "grt_date, grt_refno"
        self.imps = self.sql.getRec(tables="genrct", cols=col, where=whr,
            order=ord)
        if not self.imps:
            self.imps = []
    
    def printReport(self):
        p = ProgressBar(self.mf.body,
            max=(len(self.pays) + len(self.recs) + len(self.imps)), esc=True)
        self.head = ("%03u %-30s %39s %10s" % \
            (self.conum, self.conam, self.sysdttm, __name__))
        self.c, self.ogn, self.cpi, self.ldg, self.lpp, self.suc, self.duc = \
            doReportlab(self.mf.prgdir, self.pdfnam, self.head)
        self.pgnum = 0
        self.pglin = 99
        self.tot_chq = 0
        self.tot_dep = 0
        self.tot_imp = 0
        self.pageHeading()
        can = "no"
        bal = CCD(self.bal, "SD", 13.2)
        self.t.textLine("%-71s %-13s" % ("Balance as per Cash Book", bal.disp))
        if self.pays:
            self.t.textLine()
            self.t.textLine("%-70s" % ("Add: Cheques Not Presented",))
            self.t.textLine()
            self.pglin = self.pglin + 4
            max = len(self.pays) - 1
            for num, pay in enumerate(self.pays):
                can = p.displayProgress(num)
                if can == "yes":
                    self.t.textLine()
                    self.t.textLine("Print Job ABORTED")
                    break
                self.printLine(num, max, pay, "C")
            self.pglin = self.pglin + 1
        if self.recs and can != "yes":
            self.t.textLine()
            self.t.textLine("%-70s" % ("Less: Deposits Not Presented",))
            self.t.textLine()
            self.pglin = self.pglin + 3
            max = len(self.recs) - 1
            for num, rec in enumerate(self.recs):
                can = p.displayProgress(len(self.pays) + num)
                if can == "yes":
                    self.t.textLine()
                    self.t.textLine("Print Job ABORTED")
                    break
                self.printLine(num, max, rec, "D")
        # Bank Imports
        if self.imps and can != "yes":
            self.t.textLine()
            self.t.textLine("%-70s" % ("+-: Imports Not Captured",))
            self.t.textLine()
            self.pglin = self.pglin + 3
            max = len(self.imps) - 1
            for num, rec in enumerate(self.imps):
                can = p.displayProgress(len(self.pays) + len(self.imps) + num)
                if can == "yes":
                    self.t.textLine()
                    self.t.textLine("Print Job ABORTED")
                    break
                self.printLine(num, max, rec, "I")
        p.closeProgress()
        self.t.textLine("%72s%12s" % ("", 12 * self.suc))
        b = float(ASD(self.bal) + ASD(self.tot_chq) - ASD(self.tot_dep) + \
            ASD(self.tot_imp))
        bal = CCD(b, "SD", 13.2)
        self.t.textLine("%-71s %-13s" % ("Balance as per Bank Statement",
            bal.disp))
        self.c.drawText(self.t)
        self.c.showPage()
        self.c.save()
        doPrinter(mf=self.mf, conum=self.conum, rep=self.pdfnam,
            head=self.tit, view=self.df.repprt[0][1],
            prtnam=self.df.repprt[0][2], mail=self.df.repprt[1][2])
    
    def printLine(self, num, max, dat, typ):
        trdt = CCD(dat[0], "D1", 10)
        refno = CCD(dat[1], "Na", 9)
        desc = CCD(dat[2], "NA", 30)
        if typ in ("D", "I"):
            amt = dat[3]
        else:
            amt = float(ASD(0) - ASD(dat[3]))
        tramt = CCD(amt, "SD", 13.2)
        if self.pglin > self.lpp:
            if self.pgnum != 0:
                self.c.drawText(self.t)
                self.c.showPage()
            self.pageHeading()
        if typ == "C":
            self.tot_chq = float(ASD(self.tot_chq) + ASD(tramt.work))
            if num == max:
                bal = CCD(self.tot_chq, "SD", 13.2)
        elif typ == "D":
            self.tot_dep = float(ASD(self.tot_dep) + ASD(tramt.work))
            if num == max:
                bal = CCD(self.tot_dep, "SD", 13.2)
        else:
            self.tot_imp = float(ASD(self.tot_imp) + ASD(tramt.work))
            if num == max:
                bal = CCD(self.tot_imp, "SD", 13.2)
        if num == max:
            self.t.textLine("%-5s %-9s %-10s %-30s %-13s %-13s" % \
                ("", refno.disp, trdt.disp, desc.disp, tramt.disp, bal.disp))
        else:
            self.t.textLine("%-5s %-9s %-10s %-30s %-13s" % \
                ("", refno.disp, trdt.disp, desc.disp, tramt.disp))
        self.pglin = self.pglin + 1
    
    def extractBalance(self):
        o = self.sql.getRec(tables="genbal", cols=["glo_cyr"],
            where=[("glo_cono", "=", self.conum), ("glo_acno", "=",
            self.acnow), ("glo_trdt", "=", self.period[1][0])], limit=1)
        if o:
            b = CCD(o[0], "SD", 13.2)
        else:
            b = CCD(0, "SD", 13.2)
        self.bal = b.work
        o = self.sql.getRec(tables="gentrn", cols=["round(sum(glt_tramt), 2)"],
            where=[("glt_cono", "=", self.conum), ("glt_acno", "=", self.acnow),
            ("glt_curdt", ">=", self.s_per), ("glt_curdt", "<=", self.perw)],
            limit=1)
        if o[0]:
            b = CCD(o[0], "SD", 13.2)
        else:
            b = CCD(0, "SD", 13.2)
        self.bal = float(ASD(self.bal) + ASD(b.work))
    
    def pageHeading(self):
        self.t = self.c.beginText(self.ogn[0], self.ogn[1])
        self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
        self.pgnum = self.pgnum + 1
        self.t.textLine(self.head)
        self.t.textLine()
        self.t.textLine("%-38s %-7s %13s %-7s %10s %5s" % \
            ("General Ledger Bank Reconciliation for", self.acnod,
            "up to Period", self.perd, "Page", self.pgnum))
        self.t.textLine("%s" % (self.suc * len(self.head)))
        self.t.setFont("Tartan", self.cpi, self.ldg)
        self.pglin = 4

    def doExit(self):
        self.df.closeProcess()
        self.closeProcess()

    def closeProcess(self):
        self.mf.closeLoop()

