#! /usr/bin/env python

"""
SYNOPSIS
    Value Added Tax Report.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

import time
from TartanClasses import ASD, CCD, GetCtl, ProgressBar, Sql, TartanDialog
from tartanFunctions import doPrinter, doReportlab, doWriteExport, getModName
from tartanFunctions import showError, mthendDate

class ms3010:
    def __init__(self, mf, coy, nam):
        self.mf = mf
        self.conum = coy
        self.conam = nam
        no = self.setVariables()
        if not no:
            self.mainProcess()
            self.mf.startLoop()

    def setVariables(self):
        gc = GetCtl(self.mf)
        ctlmst = gc.getCtl("ctlmst", self.conum)
        if not ctlmst:
            return "no"
        self.genleg = False
        for x in range(0, 21, 2):
            if ctlmst["ctm_modules"][x:x+2] == "GL":
                self.genleg = True
                break
        if self.genleg:
            ctlctl = gc.getCtl("ctlctl", self.conum)
            if not ctlctl:
                return "no"
            self.convat = ctlctl["vat_ctl"]
        tab = ["ctlvmf", "ctlvtf"]
        if self.genleg:
            tab.extend(["ctlynd", "genbal", "gentrn"])
        self.sql = Sql(self.mf.dbm, tables=tab, prog=__name__)
        t = time.localtime()
        self.sysdtw = (t[0] * 10000) + (t[1] * 100) + t[2]
        self.sysdttm = "(Printed on: %i/%02i/%02i at %02i:%02i)" % \
            (t[0], t[1], t[2], t[3], t[4])
        self.head = ("%03u %-30s %s" % (self.conum, self.conam, "%s"))
        self.colsh = ["C", "T", "Curr-Dt", "S", "TT", "Acc-Num", "Reference",
            "Date", "Remarks", "Input-Value", "Input-Tax ", "Output-Value",
            "Output-Tax"]
        self.forms = [("UA", 2), ("UA", 2), ("D2", 7), ("UA", 2),
            ("UI", 2, False, False, True), ("NA", 7), ("Na", 9),
            ("D1", 10), ("NA", 30), ("SD", 13.2), ("SD", 13.2),
            ("SD", 13.2), ("SD",13.2)]
        self.ctopv = 0
        self.ctopt = 0
        self.ctipv = 0
        self.ctipt = 0
        self.ttopv = {"C": 0, "N": 0, "S": 0, "X": 0, "Z": 0}
        self.ttopt = {"C": 0, "N": 0, "S": 0, "X": 0, "Z": 0}
        self.ttipv = {"C": 0, "N": 0, "S": 0, "X": 0, "Z": 0}
        self.ttipt = {"C": 0, "N": 0, "S": 0, "X": 0, "Z": 0}
        self.gtopv = 0
        self.gtopt = 0
        self.gtipv = 0
        self.gtipt = 0
        self.t = None

    def mainProcess(self):
        self.tit = ("%03i %s" % (self.conum, self.conam),
            "V.A.T. Statement (%s)" % __name__)
        data = self.sql.getRec(tables="ctlvtf", cols=["vtt_paid"],
            where=[("vtt_cono", "=", self.conum)], group="vtt_paid",
            order="vtt_paid")
        self.cats = [
            ["A", "All"],
            ["B", "All Except (N)one"],
            ["S", "Standard"],
            ["Z", "Zero Rated"],
            ["C", "Capital Item"],
            ["X", "Excluded Item"],
            ["N", "None"]]
        vtc = {
            "stype": "C",
            "titl": "VAT Types",
            "data": self.cats}
        vtm = {
            "stype": "R",
            "tables": ("ctlvmf",),
            "cols": (
                ("vtm_code", "", 0, "C"),
                ("vtm_desc", "", 0, "Description", "Y"),
                ("vtm_cat",  "", 0, "C")),
            "where": [("vtm_cono", "=", self.conum)]}
        pdt = {
            "stype": "C",
            "titl": "Payment Date",
            "data": data}
        r1s = (("A/R","A"),("Crs","A"),("Drs","A"),("G/L","A"),("Rtl","R"))
        r1s = (("Yes", "Y"), ("No", "N"))
        r2s = (("Yes", "Y"), ("No", "N"), ("Reprint", "R"))
        fld = (
            (("T",0,0,0),"Id2",7,"Starting Period","Starting Period",
                "","Y",self.doStartPeriod,None,None,("efld",)),
            (("T",0,1,0),"ID2",7,"Ending Period","Ending Period",
                "","Y",self.doEndPeriod,None,None,("efld",)),
            (("T",0,2,0),"IUA",1,"Vat Category","Vat Category",
                "B","Y",self.doCat,vtc,None,("efld",)),
            (("T",0,3,0),"IUA",1,"Vat Code","Vat Code",
                "","Y",self.doCode,vtm,None,("efld",)),
            (("T",0,4,0),"IUA",1,"System","Originating System",
                "","Y",self.doSystem,None,None,None),
            (("T",0,5,0),("IRB",r1s),0,"Totals Only","Totals Only",
                "Y","Y",self.doTotsOnly,None,None,None),
            (("T",0,6,0),("IRB",r2s),0,"Flag Items as Paid",
                "Flag Items as Paid","N","Y",self.doFlagPaid,None,None,None),
            (("T",0,7,0),"Id1",10,"Payment Date","Payment Date",
                "","Y",self.doPayDate,pdt,None,("notzero",)))
        tnd = ((self.doEnd,"y"), )
        txt = (self.doExit, )
        self.df = TartanDialog(self.mf, title=self.tit, eflds=fld, tend=tnd,
            txit=txt, view=("Y","V"), mail=("Y","N"))

    def doStartPeriod(self, frt, pag, r, c, p, i, w):
        self.sper = w
        self.sperd = self.df.t_disp[pag][r][p]

    def doEndPeriod(self, frt, pag, r, c, p, i, w):
        self.eper = w
        self.eperd = self.df.t_disp[pag][r][p]
        if not self.genleg:
            return
        chk = self.sql.getRec(tables="ctlynd", cols=["cye_start", "cye_end"],
            where=[("cye_cono", "=", self.conum)])
        self.start = 0
        for c in chk:
            if self.eper >= c[0] / 100 and self.eper <= c[1] / 100:
                self.start = c[0]
        if not self.start:
            return "Invalid Ending Period"

    def doCat(self, frt, pag, r, c, p, i, w):
        if w and w not in ("A", "B", "N","S","Z","C","X"):
            return "Invalid V.A.T Category"
        if not w:
            self.cat = "A"
        else:
            self.cat = w
        if not self.cat:
            self.code = ""
            return "sk1"

    def doCode(self, frt, pag, r, c, p, i, w):
        if w:
            acc = self.sql.getRec(tables="ctlvmf", cols=["vtm_desc"],
                where=[("vtm_cono", "=", self.conum), ("vtm_code", "=", w)],
                limit=1)
            if not acc:
                return "Invalid V.A.T Code"
        self.code = w

    def doSystem(self, frt, pag, r, c, p, i, w):
        self.system = w

    def doTotsOnly(self, frt, pag, r, c, p, i, w):
        self.totsonly = w
        if self.totsonly == "Y":
            self.df.topEntry[0][8][2][0].hide()
            self.df.topEntry[0][8][3][0].hide()
        else:
            self.df.topEntry[0][8][2][0].show()
            self.df.topEntry[0][8][3][0].show()
        if self.sper or \
           not self.cat in ("A", "B") or \
           self.code or \
           self.system or \
           self.totsonly == "Y":
            self.flag = "N"
            return "sk2"

    def doFlagPaid(self, frt, pag, r, c, p, i, w):
        self.flag = w
        if self.flag == "N":
            self.paid = 0
            return "sk1"
        self.df.topf[pag][7][8]["where"] = [("vtt_cono", "=", self.conum),
            ("vtt_paid", ">=", mthendDate(self.eper*100))]

    def doPayDate(self, frt, pag, r, c, p, i, w):
        self.paid = w
        self.paidd = self.df.t_disp[pag][r][p]

    def doEnd(self):
        self.df.closeProcess()
        vtm = [("vtm_cono", "=", self.conum)]
        if self.cat == "B":
            vtm.append(("vtm_cat", "<>", "N"))
        elif self.cat != "A":
            vtm.append(("vtm_cat", "=", self.cat))
        if self.code:
            vtm.append(("vtm_code", "=", self.code))
        all = self.sql.getRec(tables="ctlvmf", where=vtm)
        if not all:
            showError(self.mf.body, "Selection Error", "No Records Selected")
        else:
            if self.df.repprt[0][2] == "export":
                self.exportReport(all)
            else:
                self.printReport(all)
            if self.flag == "Y":
                self.sql.updRec("ctlvtf", where=[("vtt_cono", "=", self.conum),
                    ("vtt_curdt", "<=", self.eper), ("vtt_paid", "=", 0)],
                    cols=["vtt_paid"], data=[self.paid])
                self.mf.dbm.commitDbase()
        self.closeProcess()

    def exportReport(self, all):
        p = ProgressBar(self.mf.body, max=len(all), esc=True)
        expnam = getModName(self.mf.wrkdir, __name__, self.conum)
        self.expheads = [self.head % self.sysdttm]
        self.expheads.append("VAT Statements for Period %s to %s" % \
            (self.sperd, self.eperd))
        self.expcolsh = [self.colsh]
        self.expforms = self.forms
        self.expdatas = []
        self.pcat = None
        self.pcode = None
        for num, dat in enumerate(all):
            can = p.displayProgress(num)
            if can == "yes":
                p.closeProgress()
                return
            vals = self.getValues1(dat)
            if not vals:
                continue
            code, self.pdesc, cat, vtf = vals
            if not self.pcat:
                self.pcat = cat
            if not self.pcode:
                self.pcode = code
            if cat != self.pcat:
                self.doPrintCodeTotal(ttype="X")
                self.doPrintCatTotal(ttype="X")
                self.pcat = cat
                self.pcode = code
            if code != self.pcode:
                self.doPrintCodeTotal(ttype="X")
                self.pcode = code
            for dat in vtf:
                vtyp, curdt,styp, ttyp, acno, refno, refdt, desc, exc, \
                    tax, inc = self.getValues2(dat)
                line = ["BODY", [self.pcode, vtyp.work, curdt.work, styp.work,
                    ttyp.work, acno.work, refno.work, refdt.work, desc.work]]
                if vtyp.work == "I":
                    line[1].extend([inc.work, tax.work, 0, 0])
                    self.ctipv = float(ASD(self.ctipv) + ASD(inc.work))
                    self.ctipt = float(ASD(self.ctipt) + ASD(tax.work))
                    self.ttipv[self.pcat] = float(ASD(self.ttipv[self.pcat]) \
                        + ASD(inc.work))
                    self.ttipt[self.pcat] = float(ASD(self.ttipt[self.pcat]) \
                        + ASD(tax.work))
                    self.gtipv = float(ASD(self.gtipv) + ASD(inc.work))
                    self.gtipt = float(ASD(self.gtipt) + ASD(tax.work))
                else:
                    line[1].extend([0, 0, exc.work, tax.work])
                    self.ctopv = float(ASD(self.ctopv) + ASD(inc.work))
                    self.ctopt = float(ASD(self.ctopt) + ASD(tax.work))
                    self.ttopv[self.pcat] = float(ASD(self.ttopv[self.pcat]) \
                        + ASD(inc.work))
                    self.ttopt[self.pcat] = float(ASD(self.ttopt[self.pcat]) \
                        + ASD(tax.work))
                    self.gtopv = float(ASD(self.gtopv) + ASD(inc.work))
                    self.gtopt = float(ASD(self.gtopt) + ASD(tax.work))
                self.expdatas.append(line)
        p.closeProgress()
        self.doPrintCodeTotal(ttype="X")
        self.doPrintCatTotal(ttype="X", page=False)
        self.doPrintGrandTotal(ttype="X")
        doWriteExport(xtype=self.df.repprt[0][1], name=expnam, sheet=__name__,
            heads=self.expheads, colsh=self.expcolsh, forms=self.expforms,
            datas=self.expdatas, rcdic=self.mf.rcdic)

    def printReport(self, all):
        p = ProgressBar(self.mf.body, max=len(all), esc=True)
        self.head = ("%03u %-30s %23s %18s %27s %6s" % \
            (self.conum, self.conam, "", self.sysdttm, "", __name__))
        self.pdfnam = getModName(self.mf.wrkdir, __name__, self.conum,
            ext="pdf")
        self.c, self.ogn, self.cpi, self.ldg, self.lpp, self.suc, self.duc = \
            doReportlab(self.mf.prgdir, self.pdfnam, self.head)
        self.pgnum = 0
        self.pglin = 99
        self.pcat = None
        self.pcode = None
        for num, dat in enumerate(all):
            can = p.displayProgress(num)
            if can == "yes":
                if self.t:
                    self.t.textLine()
                    self.t.textLine("Print Job ABORTED")
                break
            vals = self.getValues1(dat)
            if not vals:
                continue
            code, self.pdesc, cat, vtf = vals
            if not self.pcat:
                self.pcat = cat
            if not self.pcode:
                self.pcode = code
            if cat != self.pcat:
                self.doPrintCodeTotal()
                self.doPrintCatTotal()
                self.c.drawText(self.t)
                self.c.showPage()
                self.pcat = cat
                self.pcode = code
                self.doPrintHeading("A")
            if code != self.pcode:
                self.doPrintCodeTotal()
                self.pcode = code
                if self.totsonly == "N":
                    self.doPrintHeading("C")
            for dat in vtf:
                vtyp, curdt,styp, ttyp, acno, refno, refdt, desc, exc, \
                    tax, inc = self.getValues2(dat)
                if self.pglin > self.lpp:
                    if self.pgnum != 0:
                        self.c.drawText(self.t)
                        self.c.showPage()
                    self.doPrintHeading("A")
                    self.pglin = self.pglin + 1
                if vtyp.work == "I":
                    if self.totsonly != "Y":
                        self.t.textLine("%s %s %s %s %s %s %s %s %s %-13s "\
                            "%-13s" % (curdt.disp, styp.disp, ttyp.disp,
                            acno.disp, refno.disp, refdt.disp, desc.disp,
                            inc.disp, tax.disp, " ", " "))
                        self.pglin = self.pglin + 1
                    self.ctipv = float(ASD(self.ctipv) + ASD(inc.work))
                    self.ctipt = float(ASD(self.ctipt) + ASD(tax.work))
                    self.ttipv[self.pcat] = float(ASD(self.ttipv[self.pcat]) \
                        + ASD(inc.work))
                    self.ttipt[self.pcat] = float(ASD(self.ttipt[self.pcat]) \
                        + ASD(tax.work))
                    self.gtipv = float(ASD(self.gtipv) + ASD(inc.work))
                    self.gtipt = float(ASD(self.gtipt) + ASD(tax.work))
                elif vtyp.work == "O":
                    if self.totsonly != "Y":
                        self.t.textLine("%s %s %s %s %s %s %s %-13s %-13s "\
                            "%s %s" % (curdt.disp, styp.disp, ttyp.disp,
                            acno.disp, refno.disp, refdt.disp, desc.disp,
                            " ", " ", inc.disp, tax.disp))
                        self.pglin = self.pglin + 1
                    self.ctopv = float(ASD(self.ctopv) + ASD(inc.work))
                    self.ctopt = float(ASD(self.ctopt) + ASD(tax.work))
                    self.ttopv[self.pcat] = float(ASD(self.ttopv[self.pcat]) \
                        + ASD(inc.work))
                    self.ttopt[self.pcat] = float(ASD(self.ttopt[self.pcat]) \
                        + ASD(tax.work))
                    self.gtopv = float(ASD(self.gtopv) + ASD(inc.work))
                    self.gtopt = float(ASD(self.gtopt) + ASD(tax.work))
        p.closeProgress()
        if self.t:
            if can != "yes":
                self.doPrintCodeTotal()
                self.doPrintCatTotal()
                if self.cat in ("A", "B"):
                    self.c.drawText(self.t)
                    self.c.showPage()
                    self.doPrintHeading("S")
                    self.doPrintSummary()
                self.c.drawText(self.t)
        self.c.showPage()
        self.c.save()
        doPrinter(mf=self.mf, conum=self.conum, rep=self.pdfnam, head=self.tit,
            view=self.df.repprt[0][1], prtnam=self.df.repprt[0][2],
            mail=self.df.repprt[1][2])

    def getValues1(self, data):
        code = data[self.sql.ctlvmf_col.index("vtm_code")]
        pdesc = data[self.sql.ctlvmf_col.index("vtm_desc")]
        cat = data[self.sql.ctlvmf_col.index("vtm_cat")]
        vtt = [("vtt_cono", "=", self.conum), ("vtt_code", "=", code)]
        if self.sper:
            vtt.append(("vtt_curdt", ">=", self.sper))
        if self.eper:
            vtt.append(("vtt_curdt", "<=", self.eper))
        if self.system:
            vtt.append(("vtt_styp", "=", self.system))
        if self.flag == "R":
            vtt.append(("vtt_paid", "=", self.paid))
        else:
            vtt.append(("vtt_paid", "=", 0))
        vtf = self.sql.getRec(tables="ctlvtf", where=vtt,
            order="vtt_styp, vtt_ttyp, vtt_refdt, vtt_refno")
        if not vtf:
            return
        return (code, pdesc, cat, vtf)

    def getValues2(self, data):
        vtyp = CCD(data[self.sql.ctlvtf_col.index("vtt_vtyp")],"UA",1)
        curdt = CCD(data[self.sql.ctlvtf_col.index("vtt_curdt")],"D2",7)
        styp = CCD(data[self.sql.ctlvtf_col.index("vtt_styp")],"UA",1)
        ttyp = CCD(data[self.sql.ctlvtf_col.index("vtt_ttyp")],"UI",2)
        acno = CCD(data[self.sql.ctlvtf_col.index("vtt_acno")],"NA",7)
        refno = CCD(data[self.sql.ctlvtf_col.index("vtt_refno")],"Na",9)
        refdt = CCD(data[self.sql.ctlvtf_col.index("vtt_refdt")],"D1",10)
        desc = CCD(data[self.sql.ctlvtf_col.index("vtt_desc")],"NA",30)
        exc = CCD(data[self.sql.ctlvtf_col.index("vtt_exc")],"SD",13.2)
        tax = CCD(data[self.sql.ctlvtf_col.index("vtt_tax")],"SD",13.2)
        inc = CCD(((exc.work * 100) + (tax.work * 100)) / 100, "SD", 13.2)
        return (vtyp, curdt,styp, ttyp, acno, refno, refdt, desc, exc, tax, inc)

    def doPrintHeading(self, htype):
        if htype in ("A", "C"):
            for c in self.cats:
                if c[0] == self.pcat:
                    desc = c[1]
                    break
        if htype in ("A", "S"):
            self.t = self.c.beginText(self.ogn[0], self.ogn[1])
            self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
            self.pgnum = self.pgnum + 1
            self.t.textLine(self.head)
            self.t.textLine()
            if self.flag == "Y":
                flag = "(Transactions Flagged on %s)" % self.paidd
            elif self.flag == "N":
                flag = "(Transactions Not Flagged)"
            elif self.flag == "R":
                flag = "(Transactions Paid on %s)" % self.paidd
            self.t.textLine("%-25s %-7s %-2s %-7s %-71s %4s %5s" % \
                ("VAT Statements for Period",
                self.sperd, "to", self.eperd, flag, "Page", self.pgnum))
            self.t.textLine()
            if htype == "A":
                if self.totsonly != "Y":
                    self.t.textLine("Category: %1s %s Code: %s %s" % \
                        (self.pcat, desc, self.pcode, self.pdesc))
                else:
                    self.t.textLine("Category: %1s %s - Totals Only" % \
                        (self.pcat, desc))
                self.pglin += 1
            else:
                self.t.textLine("V.A.T. Summary")
            self.pglin = 5
        elif htype == "C":
            self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
            self.t.textLine()
            self.t.textLine("Category: %1s %s - Code: %s %s" % \
                (self.pcat, desc, self.pcode, self.pdesc))
            self.pglin += 2
        self.t.textLine()
        if htype == "S":
            self.t.textLine("%-30s %13s %13s %13s %13s" % ("Category", \
                "Input-Value ", "Input-Tax ", "Output-Value ", " Output-Tax "))
        else:
            self.t.textLine("%-7s %-1s %-2s %-7s %-9s %-10s %-30s "\
                "%13s %13s %13s %13s" % ("Curr-Dt", "S", "TT", "Acc-Num",
                "Reference", "   Date", "Remarks", "Input-Value ",
                "Input-Tax ", "Output-Value ", " Output-Tax "))
        self.t.textLine("%s" % (self.suc * len(self.head)))
        self.t.setFont("Tartan", self.cpi, self.ldg)
        self.pglin += 3

    def doPrintCodeTotal(self, ttype="P"):
        j = CCD(self.ctipv, "SD", 13.2)
        k = CCD(self.ctipt, "SD", 13.2)
        l = CCD(self.ctopv, "SD", 13.2)
        m = CCD(self.ctopt, "SD", 13.2)
        if ttype == "P":
            if self.totsonly != "Y":
                self.t.textLine("%-125s" % (self.suc * len(self.head)))
            self.t.textLine("%-41s %-14s %-15s %-13s %-13s %-13s %-13s" % ("",
                "Code-Total", self.pcode, j.disp, k.disp, l.disp, m.disp))
        else:
            self.expdatas.append(["ULINES"])
            desc = "Code-Total %s" % self.pcode
            line = ["TOTAL", ["", "", "", "", "", "", "", "", desc,
                j.work, k.work, l.work, m.work]]
            self.expdatas.append(line)
        self.ctipv = 0
        self.ctipt = 0
        self.ctopv = 0
        self.ctopt = 0

    def doPrintCatTotal(self, ttype="P", page=True):
        j = CCD(self.ttipv[self.pcat], "SD", 13.2)
        k = CCD(self.ttipt[self.pcat], "SD", 13.2)
        l = CCD(self.ttopv[self.pcat], "SD", 13.2)
        m = CCD(self.ttopt[self.pcat], "SD", 13.2)
        if ttype == "P":
            self.t.textLine("%-125s" % (self.suc * len(self.head)))
            self.t.textLine("%-41s %-14s %-15s %-13s %-13s %-13s %-13s" % ("",
                "Category-Total", self.pcat, j.disp, k.disp, l.disp, m.disp))
        else:
            self.expdatas.append(["ULINES"])
            desc = "Category-Total %s" % self.pcat
            line = ["TOTAL", ["", "", "", "", "", "", "", "", desc,
                j.work, k.work, l.work, m.work]]
            self.expdatas.append(line)
            if page:
                self.expdatas.append(["PAGE", [self.expheads, self.expcolsh]])

    def doPrintGrandTotal(self, ttype="P"):
        j = CCD(self.gtipv, "SD", 13.2)
        k = CCD(self.gtipt, "SD", 13.2)
        l = CCD(self.gtopv, "SD", 13.2)
        m = CCD(self.gtopt, "SD", 13.2)
        if ttype == "P":
            self.t.textLine("%-125s" % (self.suc * len(self.head)))
            self.t.textLine("%-41s %-14s %-15s %-13s %-13s %-13s %-13s" % ("",
                "Grand-Total", "", j.disp, k.disp, l.disp, m.disp))
        else:
            self.expdatas.append(["ULINES"])
            desc = "Grand-Total"
            line = ["TOTAL", ["", "", "", "", "", "", "", "", desc,
                j.work, k.work, l.work, m.work]]
            self.expdatas.append(line)
            self.expdatas.append(["ULINED"])

    def doPrintSummary(self):
        tots = [0, 0, 0, 0]
        for c in self.cats:
            if c[0] in ("A", "B"):
                continue
            if self.cat == "B" and c[0] == "N":
                continue
            d = CCD(c[1], "NA", 30)
            j = CCD(self.ttipv[c[0]], "SD", 13.2)
            tots[0] = float(ASD(tots[0]) + ASD(j.work))
            k = CCD(self.ttipt[c[0]], "SD", 13.2)
            tots[1] = float(ASD(tots[1]) + ASD(k.work))
            l = CCD(self.ttopv[c[0]], "SD", 13.2)
            tots[2] = float(ASD(tots[2]) + ASD(l.work))
            m = CCD(self.ttopt[c[0]], "SD", 13.2)
            tots[3] = float(ASD(tots[3]) + ASD(m.work))
            self.t.textLine("%s %s %s %s %s" % (d.disp, j.disp, k.disp,
                l.disp, m.disp))
        self.t.textLine("%-125s" % (self.suc * len(self.head)))
        d = CCD("Grand Totals", "NA", 30)
        j = CCD(tots[0], "SD", 13.2)
        k = CCD(tots[1], "SD", 13.2)
        l = CCD(tots[2], "SD", 13.2)
        m = CCD(tots[3], "SD", 13.2)
        self.t.textLine("%s %s %s %s %s" % (d.disp, j.disp, k.disp,
            l.disp, m.disp))
        self.t.textLine()
        self.t.textLine()
        tvat = CCD(float(ASD(tots[1]) + ASD(tots[3])), "SD", 13.2)
        self.t.textLine("%-72s %13s" % ("Total Net Tax Due/Owed", tvat.disp))
        if not self.genleg:
            return
        bal = 0.0
        o = self.sql.getRec(tables="genbal", cols=["glo_cyr"],
            where=[("glo_cono", "=", self.conum), ("glo_acno", "=",
            self.convat), ("glo_trdt", "=", self.start)], limit=1)
        if o:
            b = CCD(o[0], "SD", 13.2)
        else:
            b = CCD(0, "SD", 13.2)
        bal = float(ASD(bal) + ASD(b.work))
        o = self.sql.getRec(tables="gentrn",
            cols=["round(sum(glt_tramt), 2)"], where=[("glt_cono", "=",
            self.conum), ("glt_acno", "=", self.convat), ("glt_curdt",
            ">=", self.start / 100), ("glt_curdt", "<=", self.eper)], limit=1)
        if o and o[0]:
            b = CCD(float(o[0]), "SD", 13.2)
        else:
            b = CCD(0, "SD", 13.2)
        tbal = CCD(float(ASD(bal) + ASD(b.work)), "SD", 13.2)
        self.t.textLine()
        self.t.textLine("%-72s %13s" % ("Balance of VAT Control", tbal.disp))

    def doExit(self):
        self.df.closeProcess()
        self.closeProcess()

    def closeProcess(self):
        self.mf.closeLoop()

