#! /usr/bin/python

"""
SYNOPSIS
    Change Financial Year End.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt

It must be executed at the end of every financial period and it creates
the necessary records in the ctlynd table and also generates the correct
opening balances in the general ledger system if applicable.

It must also be executed whenever alterations to a previous financial period,
have been made, to update the opening balances of later financial periods.
"""

from TartanClasses import SplashScreen, Sql, TartanDialog
from tartanFunctions import getPeriods, mthendDate, showError

class msy020:
    def __init__(self, mf, conum, conam, period):
        self.mf = mf
        self.conum = conum
        self.conam = conam
        self.period = period
        no = self.setVariables()
        if not no:
            self.drawScreen()
            self.mf.startLoop()

    def setVariables(self):
        if self.period[0] == 0:
            showError(self.mf.body, "Year-End Error",
                "Period 0 Cannot be Changed, Change Period 1 Instead!")
            return "no"
        self.sql = Sql(self.mf.dbm, "ctlynd", prog=__name__)
        if self.sql.error:
            return self.sql.error
        self.pers = self.sql.getRec(tables="ctlynd", where=[("cye_cono", "=",
            self.conum), ("cye_period", ">=", self.period[0])],
            order="cye_period")
        if self.pers[0][self.sql.ctlynd_col.index("cye_final")] == "Y":
            showError(self.mf.body, "Year-End Error",
                "This Financial Period Has Already Been Finalised!")
            return "no"
        self.s_per = self.pers[0][self.sql.ctlynd_col.index("cye_start")]
        self.e_per = self.pers[0][self.sql.ctlynd_col.index("cye_end")]
        self.m_per = self.getEndDate(self.s_per)
        self.l_per = self.pers[-1][self.sql.ctlynd_col.index("cye_period")]
        if self.period[0] == 1 and self.l_per > 1:
            showError(self.mf.body, "Year-End Error",
                "Period 1 Cannot be Changed as Periods > 1 Exist!")
            return "no"

    def getEndDate(self, date):
        yr = (date / 10000) + 1
        mt = ((date / 100) % 100) - 1
        if not mt:
            yr -= 1
            mt = 12
        return mthendDate(((yr * 10000) + (mt * 100) + 1))

    def getStartDate(self, date):
        yr = date / 10000
        mt = ((date / 100) % 100) + 1
        if mt > 12:
            yr += 1
            mt = 1
        return ((yr * 10000) + (mt * 100) + 1)

    def drawScreen(self):
        tit = ("%03i %s" % (self.conum, self.conam), "Year End Change",)
        fld = [
            [("T",0,0,0),"I@cye_start",0,"","",
                self.s_per,"N",self.doStartPer,None,None,("efld",)],
            (("T",0,1,0),"I@cye_end",0,"","",
                self.e_per,"N",self.doEndPer,None,None,("efld",))]
        if self.period[0] != 1:
            fld[0][1] = "O@cye_start"
            fld[0][7] = None
        tnd = ((self.doEnd,"y"),)
        txt = (self.doExit,)
        self.df = TartanDialog(self.mf, tops=False, title=tit, eflds=fld,
            tend=tnd, txit=txt, focus=True)
        if self.period[0] != 1:
            self.df.loadEntry("T", 0, 0, data=self.s_per)

    def doStartPer(self, frt, pag, r, c, p, i, w):
        y = (w / 10000) - 1
        m = (w % 10000) / 100
        if m == 2:
            if (y % 4) == 0:
                d = 29
            else:
                d = 28
        else:
            d = w % 100
        self.s0 = (y*10000) + (m*100) + d
        self.s_per = w

    def doEndPer(self, frt, pag, r, c, p, i, w):
        if w <= self.s_per:
            return "Invalid End Period"
        y = (w / 10000) - 1
        m = (w % 10000) / 100
        if m == 2:
            if (y % 4) == 0:
                d = 29
            else:
                d = 28
        else:
            d = w % 100
        self.e0 = (y*10000) + (m*100) + d
        self.e_per = w

    def doEnd(self):
        self.df.closeProcess()
        if self.period[0] == 1:
            self.sql.updRec("ctlynd", cols=["cye_start", "cye_end"],
                data=[self.s0, self.e0], where=[("cye_cono", "=", self.conum),
                ("cye_period", "=", 0)])
        from msy010 import msy010
        for per in self.pers:
            num = per[self.sql.ctlynd_col.index("cye_period")]
            if num == self.period[0]:
                self.start = self.s_per
                self.end = self.e_per
            else:
                self.start = self.getStartDate(self.end)
                self.end = self.getEndDate(self.start)
            self.sql.updRec("ctlynd", cols=["cye_start", "cye_end"],
                data=[self.start, self.end], where=[("cye_cono", "=",
                self.conum), ("cye_period", "=", num)])
        spl = SplashScreen(self.mf.body,
            text="Running Year End Routine for Period %s" % num)
        per = getPeriods(self.mf, self.conum, num - 1)
        per = (num - 1, (per[0].work, per[0].disp), (per[1].work, per[1].disp))
        msy010(self.mf, self.conum, self.conam, per, args=("N", False))
        spl.closeSplash()
        self.mf.dbm.commitDbase(ask=True)
        self.mf.closeLoop()

    def doExit(self):
        self.df.closeProcess()
        self.mf.closeLoop()

if __name__ == "__main__":
    from TartanClasses import MainFrame
    mf = MainFrame()
    mf.dbm.openDbase()
    per = getPeriods(mf, 1, 5)
    if per[1] and per[2]:
        per = (5, (per[0].work, per[0].disp), (per[1].work, per[1].disp))
        ex = msy020(mf, 1, "", per)
    else:
        print "Invalid Period %s for Company %s" % (5, 1)
    mf.dbm.closeDbase()
