#! /usr/bin/env python

"""
SYNOPSIS
    Stores Stock Take Variance Report.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

import time
from TartanClasses import ASD, Balances, CCD, GetCtl, ProgressBar, Sql
from TartanClasses import TartanDialog
from tartanFunctions import doReportlab, getModName, doPrinter, showError
from TartanClasses import Balances

class st5030:
    def __init__(self, mf, coy, nam, period):
        self.mf = mf
        self.conum = coy
        self.conam = nam
        self.period = period
        no = self.setVariables()
        if not no:
            self.mainProcess()
            self.mf.startLoop()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, ["strmf1", "strtrn", "strvar"],
            prog=__name__)
        gc = GetCtl(self.mf)
        strctl = gc.getCtl("strctl", self.conum)
        if not strctl:
            return "no"
        self.locs = strctl["cts_locs"]
        self.pdfnam = getModName(self.mf.wrkdir, __name__, self.conum,
            ext="pdf")
        t = time.localtime()
        self.sysdtw = (t[0] * 10000) + (t[1] * 100) + t[2]
        self.sysdtd = "%i/%02i/%02i" % (t[0], t[1], t[2])
        self.sysdttm = "(Printed on: %i/%02i/%02i at %02i:%02i)" % \
            (t[0], t[1], t[2], t[3], t[4])
        self.t = None

    def mainProcess(self):
        self.tit = ("%03i %s" % (self.conum, self.conam),
                "Stock Take Variance Report (%s)" % __name__)
        fld = (
            (("T",0,0,0),"ID1",10,"Reporting Date","Reporting Date",
                self.sysdtw,"Y",self.doDate,None,None,("efld",)),)
        tnd = ((self.doEnd,"Y"), )
        txt = (self.closeProcess, )
        self.df = TartanDialog(self.mf, title=self.tit, eflds=fld, tend=tnd,
            txit=txt, view=("N","V"), mail=("Y","N"))

    def doDate(self, frt, pag, r, c, p, i, w):
        if w < self.period[1][0] or w > self.period[2][0]:
            return "Invalid Date, Not in Financial Period"
        self.date = w

    def doEnd(self):
        self.df.closeProcess()
        all = self.sql.getRec(tables="strvar", where=[("stv_cono", "=",
            self.conum)], order="stv_group, stv_code, stv_loc")
        if not all:
            showError(self.mf.body, "Processing Error",
            "No Records Selected")
        else:
            self.printReport(all)
        self.mf.closeLoop()

    def printReport(self, all):
        p = ProgressBar(self.mf.body, max=len(all), esc=True)
        self.head = ("%03u %-30s %31s %33s %31s %6s" % \
            (self.conum, self.conam, "", self.sysdttm, "", __name__))
        self.c, self.ogn, self.cpi, self.ldg, self.lpp, self.suc, self.duc = \
            doReportlab(self.mf.prgdir, self.pdfnam, self.head)
        count = 0
        old_loc = ""
        self.pgnum = 0
        self.pglin = 99
        for num, dat in enumerate(all):
            can = p.displayProgress(num)
            if can == "yes":
                if self.t:
                    self.t.textLine()
                    self.t.textLine("Print Job ABORTED")
                break
            grp = CCD(dat[self.sql.strvar_col.index("stv_group")], "UA", 3)
            code = CCD(dat[self.sql.strvar_col.index("stv_code")], "NA", 20)
            loc = CCD(dat[self.sql.strvar_col.index("stv_loc")], "UA", 1)
            bin = CCD(dat[self.sql.strvar_col.index("stv_bin")], "UA", 8)
            qty1 = CCD(dat[self.sql.strvar_col.index("stv_qty")], "SD", 12.2)
            prc1 = CCD(dat[self.sql.strvar_col.index("stv_ucost")], "UD", 12.2)
            self.loc = loc.work
            self.locd = loc.disp
            bals = Balances(self.mf, "STR", self.conum, (self.sysdtw/100),
                keys=(grp.work, code.work, loc.work, ("P", self.period[0])))
            m_ob, m_mv, m_cb, y_ob, y_mv, y_cb, ac, lc = bals.doStrBals()
            fqbal = CCD(y_cb[0], "SD", 12.2)
            fvbal = CCD(y_cb[1], "SD", 12.2)
            if ac:
                fprice = CCD(ac, "SD", 12.2)
            else:
                fprice = CCD(lc, "SD", 12.2)
            diff = CCD(float(ASD(fqbal.work) - ASD(qty1.work)), "SD", 12.2)
            rslt = self.sql.getRec("strmf1", cols=["st1_desc", "st1_uoi"],
                where=[("st1_cono", "=", self.conum), ("st1_group", "=",
                grp.work), ("st1_code", "=", code.work)], limit=1)
            desc = CCD(rslt[0], "NA", 30)
            uoi = CCD(rslt[1], "NA", 10)
            if self.pglin > self.lpp:
                if self.pgnum:
                    self.c.drawText(self.t)
                    self.c.showPage()
                self.pageHeading()
            self.t.textLine("%s %s %s %s %s %s %s %s %s %s" % \
                (grp.disp, code.disp, desc.disp, uoi.disp, bin.disp,
                fprice.disp, prc1.disp, fqbal.disp, qty1.disp, diff.disp))
            self.pglin = self.pglin + 1
        p.closeProgress()
        if can != "yes" and self.pgnum:
            if self.t:
                self.c.drawText(self.t)
            self.c.showPage()
        self.c.save()
        doPrinter(mf=self.mf, conum=self.conum, rep=self.pdfnam,
            head=self.tit, view=self.df.repprt[0][1],
            prtnam=self.df.repprt[0][2], mail=self.df.repprt[1][2])
        self.closeProcess()
            
    def pageHeading(self):
        self.t = self.c.beginText(self.ogn[0], self.ogn[1])
        self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
        self.pgnum = self.pgnum + 1
        self.t.textLine(self.head)
        self.t.textLine()
        self.t.textLine("%-32s %-10s %89s %5s" % \
            ("Stock Take Variance Report as at", self.sysdtd, "Page",
                self.pgnum))
        self.t.textLine()
        self.t.textLine("%-8s %s  %s" % ("Location", self.loc, self.locd))
        self.t.textLine()
        self.t.textLine("%-3s %-20s %-30s %-10s %-8s   %-12s  %-12s %-12s " \
            "%-10s %-12s" % \
            ("Grp", "Product-Code", "Description", "U.O.I", "Bin-Loc",
            "File-Cost", "Stkt-Cost", "File-Qty", "Stkt-Qty", "Difference"))
        self.t.textLine("%s" % (self.suc * len(self.head)))
        self.t.setFont("Tartan", self.cpi, self.ldg)
        self.pglin = 8

    def closeProcess(self):
        self.df.closeProcess()
        self.mf.closeLoop()

    def extractBals(self, grp, code, loc):
        qbal = 0
        vbal = 0
        rslt = self.sql.getRec("strtrn", cols=["stt_qty", "stt_cost"],
            where=[("stt_cono", "=", self.conum), ("stt_group", "=", grp),
            ("stt_code", "=", code), ("stt_loc", "=", loc)])
        if rslt:
            for dat in rslt:
                qty = CCD(dat[1], "SD", 12.2)
                cst = CCD(dat[2], "SD", 12.2)
                qbal = float(ASD(qbal) + ASD(qty.work))
                vbal = float(ASD(vbal) + ASD(cst.work))
        if vbal and qbal:
            fprc = round((vbal / qbal),2)
        else:
            fprc = 0
        return qbal, vbal, fprc

