#! /usr/bin/python

"""
SYNOPSIS
    Stores Stock Take Merge.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

import time
from TartanClasses import ASD, Balances, CCD, GetCtl, Sql, TartanDialog
from tartanFunctions import showError

class st5040:
    def __init__(self, mf, coy, nam, period):
        self.mf = mf
        self.conum = coy
        self.conam = nam
        self.period = period
        no = self.setVariables()
        if not no:
            self.dataHeader()
            self.mf.startLoop()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, ["gentrn", "strmf1", "strtrn", "strvar",
            "strprc"], prog=__name__)
        gc = GetCtl(self.mf)
        strctl = gc.getCtl("strctl", self.conum)
        if not strctl:
            return "no"
        self.glint = strctl["cts_glint"]
        self.locs = strctl["cts_locs"]
        if self.glint == "Y":
            ctlctl = gc.getCtl("ctlctl", self.conum)
            if not ctlctl:
                return "no"
            ctls = ["stk_soh", "stk_susp"]
            if gc.chkRec(self.conum, ctlctl, ctls):
                return "no"
            self.stk_soh = ctlctl["stk_soh"]
            self.stk_susp = ctlctl["stk_susp"]
        t = time.localtime()
        self.sysdtw = (t[0] * 10000) + (t[1] * 100) + t[2]

    def dataHeader(self):
        tit = ("%03i %s" % (self.conum, self.conam), "Stores Ledger " + \
            " Stock Take Merge (%s)" % __name__)
        fld = [
            (("T",0,0,0),"ID1",10,"Merging Date","Merging Date",
                self.sysdtw,"N",self.doDate,None,None,("efld",))]
        tnd = ((self.endPage,"y"),)
        txt = (self.exitPage,)
        self.df = TartanDialog(self.mf, title=tit, eflds=fld, tend=tnd,
            txit=txt)

    def doDate(self, frt, pag, r, c, p, i, w):
        if w < self.period[1][0] or w > self.period[2][0]:
            return "Invalid Date, Not in Financial Period"
        self.date = w
        self.curdt = w / 100

    def endPage(self):
        col = ["stv_group", "stv_code", "stv_loc", "stv_bin", "stv_qty",
            "stv_ucost", "stv_usell"]
        all = self.sql.getRec(tables=["strmf1", "strvar"], cols=col,
            where=[("stv_cono", "=", self.conum), ("stv_cono=st1_cono",),
            ("stv_group=st1_group",), ("stv_code=st1_code",)],
            order="stv_group, stv_code, stv_loc")
        if not all:
            showError(self.mf.body, "Processing Error",
            "No Records Selected")
        else:
            for num, dat in enumerate(all):
                self.updateTables(num+1, col, dat)
                self.sql.delRec("strvar", where=[("stv_cono", "=", self.conum),
                    ("stv_group", "=", dat[col.index("stv_group")]),
                    ("stv_code", "=", dat[col.index("stv_code")]),
                    ("stv_loc", "=", dat[col.index("stv_loc")])])
                self.mf.dbm.commitDbase()
        self.exitPage()

    def exitPage(self):
        self.df.closeProcess()
        self.mf.closeLoop()

    def updateTables(self, cnt, col, rec):
        grp = rec[col.index("stv_group")]
        code = rec[col.index("stv_code")]
        loc = rec[col.index("stv_loc")]
        vqty = CCD(rec[col.index("stv_qty")], "SD", 12.2)
        vprc = CCD(rec[col.index("stv_ucost")], "UD", 12.2)
        sell = CCD(rec[col.index("stv_usell")], "UD", 12.2)
        # Selling Price Records
        if sell.work:
            whr = [("stp_cono", "=", self.conum), ("stp_group", "=", grp),
                ("stp_code", "=", code), ("stp_loc", "=", loc), ("stp_level",
                "=", 1)]
            prc = self.sql.getRec(tables="strprc", where=whr, limit=1)
            if prc:
                self.sql.updRec("strprc", cols=["stp_price"], data=[sell.work],
                    where=whr)
            else:
                self.sql.insRec("strprc", data=[self.conum, grp, code, loc, 1,
                    sell.work])
        # Test for Variances
        bals = Balances(self.mf, "STR", self.conum, (self.sysdtw/100),
            keys=(grp, code, loc, ("P", self.period[0])))
        m_ob, m_mv, m_cb, y_ob, y_mv, y_cb, ac, lc = bals.doStrBals()
        fqty = CCD(y_cb[0], "SD", 12.2)
        fval = CCD(y_cb[1], "SD", 12.2)
        if fval.work and fqty.work:
            c = round((fval.work / fqty.work), 2)
        else:
            c = 0
        fprc = CCD(c, "SD", 12.2)
        if fprc.work != vprc.work and vprc.work:
            prc = vprc.work
        else:
            prc = fprc.work
        prc = CCD(prc, "SD", 12.2)
        val = round((prc.work * vqty.work),2)
        vval = CCD(val, "SD", 12.2)
        qdif = CCD(float(ASD(vqty.work) - ASD(fqty.work)), "SD", 12.2)
        vdif = CCD(float(ASD(vval.work) - ASD(fval.work)), "SD", 12.2)
        if not qdif.work and not vdif.work:
            return
        # Stores Ledger Transaction
        if  qdif.work >= 0:
            rtn = 5
        else:
            rtn = 6
        ref = CCD(cnt, "Na", 9).work
        self.sql.insRec("strtrn", data=[self.conum, rec[col.index("stv_group")],
            rec[col.index("stv_code")], rec[col.index("stv_loc")], self.date,
            rtn, ref, "ST-MERG", "", qdif.work, vdif.work, 0, self.curdt,
            "Stock Take Adjustment", 0, "", "", "STR", 0, 0, 0, "",
            self.sysdtw, 0])
        # General Ledger Control Transaction (Stock On Hand)
        col = self.sql.gentrn_col
        acc = self.sql.getRec(tables="gentrn", where=[("glt_cono", "=",
            self.conum), ("glt_acno", "=", self.stk_soh), ("glt_curdt", "=",
            self.curdt), ("glt_trdt", "=", self.date), ("glt_type", "=", 4),
            ("glt_refno", "=", "STOCK-ADJ"), ("glt_batch", "=", "ST-MERG")],
            limit=1)
        if acc:
            amnt = float(ASD(acc[col.index("glt_tramt")]) + ASD(vdif.work))
            self.sql.updRec("gentrn", cols=["glt_tramt"], data=[amnt],
                where=[("glt_seq", "=", acc[col.index("glt_seq")])])
        else:
            self.sql.insRec("gentrn", data=[self.conum, self.stk_soh,
            self.curdt, self.date, 4, "STOCK-ADJ", "ST-MERG", vdif.work,
            0, "Stock Take Adjustment", "N", "", 0, self.sysdtw, 0])
        # General Ledger Control Transaction (Stock Suspense)
        val = float(ASD(0) - ASD(vdif.work))
        acc = self.sql.getRec(tables="gentrn", where=[("glt_cono", "=",
            self.conum), ("glt_acno", "=", self.stk_susp), ("glt_curdt", "=",
            self.curdt), ("glt_trdt", "=", self.date), ("glt_type", "=", 4),
            ("glt_refno", "=", "STOCK-ADJ"), ("glt_batch", "=", "ST-MERG")],
            limit=1)
        if acc:
            amnt = float(ASD(acc[col.index("glt_tramt")]) + ASD(val))
            self.sql.updRec("gentrn", cols=["glt_tramt"], data=[amnt],
                where=[("glt_seq", "=", acc[col.index("glt_seq")])])
        else:
            self.sql.insRec("gentrn", data=[self.conum, self.stk_susp,
            self.curdt, self.date, 4, "STOCK-ADJ", "ST-MERG", val,
            0, "Stock Take Adjustment", "N", "", 0, self.sysdtw, 0])

