#! /usr/bin/python

"""
SYNOPSIS
    Salaries and Wages Payslips.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""

import os, time
from TartanClasses import ASD, CCD, GetCtl, LoanInterest, PrintPayslip, Sql
from TartanClasses import TartanDialog
from tartanFunctions import askQuestion, dateDiff, getModName, getSingleRecords
from tartanFunctions import showError
from tartanWork import payeRates

class wg2020:
    def __init__(self, mf, conum, conam, period):
        self.mf = mf
        self.conum = conum
        self.conam = conam
        self.period = period
        no = self.checkControls()
        if not no:
            self.setVariables()
            self.mainProcess()
            self.mf.startLoop()

    def checkControls(self):
        gc = GetCtl(self.mf)
        wagctl = gc.getCtl("wagctl", self.conum)
        if not wagctl:
            return "no"
        self.hrsmt = float(wagctl["ctw_m_hrs"])
        self.hrswk = float(wagctl["ctw_w_hrs"])
        self.hrsdy = float(wagctl["ctw_d_hrs"])
        self.glint = wagctl["ctw_glint"]
        self.bestac = wagctl["ctw_bestac"]
        self.besttp = wagctl["ctw_besttp"]
        tabs = ["wagcod", "wagedc", "wagmst", "wagcap", "wagbal", "wagtf1",
            "wagtf2", "waglmf", "wagltf"]
        if self.glint == "Y":
            ctlctl = gc.getCtl("ctlctl", self.conum)
            if not ctlctl:
                return "no"
            ctls = ["wag_slc", "wag_sli"]
            if gc.chkRec(self.conum, ctlctl, ctls):
                return "no"
            self.glctls = (ctlctl["wag_slc"], ctlctl["wag_sli"])
            tabs.extend(["ctlbra", "ctldep", "ctlmst", "genmst", "genint",
                "gentrn"])
        else:
            self.glctls = None
        self.sql = Sql(self.mf.dbm, tables=tabs, prog=__name__)
        if self.sql.error:
            return self.sql.error

    def setVariables(self):
        self.pdfnam = getModName(self.mf.wrkdir, __name__, self.conum,
            ext="pdf")
        t = time.localtime()
        self.sysdtw = (t[0] * 10000) + (t[1] * 100) + t[2]
        self.sysdtd = CCD(self.sysdtw, "d1", 10).disp
        self.s_per = self.period[1][0] / 100
        self.e_per = self.period[2][0] / 100
        self.head = ("%03u %-30s %40s %10s" % \
            (self.conum, self.conam, self.sysdtd, "wg2020"))
        self.etotal = 0
        self.pgnum = 0
        self.empnos = []

    def mainProcess(self):
        self.tit = ("%03i %s" % (self.conum, self.conam),
            "Salaries and Wages Payslips (%s)" % __name__)
        mes = {
            "stype": "R",
            "tables": ("ctlmes",),
            "cols": [
                ("mss_message", "",  0, "Cod"),
                ("mss_detail",  "", 60, "Details")],
            "where": [("mss_system", "=", "WAG")],
            "screen": self.mf.body}
        r1s = (("Weekly","W"),("Fortnightly","F"),("Monthly","M"))
        r2s = (("Yes","Y"),("Singles","S"))
        r3s = (("Yes","Y"),("No","N"))
        fld = (
            (("T",0,0,0),"ID1",10,"Pay-Run Date","Pay-Run Date",
                "","N",self.doRunDate,None,None,("efld",)),
            (("T",0,1,0),"ID1",10,"Payment Date","Payment Date",
                self.sysdtw,"N",self.doPayDate,None,None,("efld",)),
            (("T",0,2,0),"IUI",3,"Message Code","Message Code",
                "","N",self.doMess,mes,None,None),
            (("T",0,3,0),("IRB",r1s),0,"Frequency","Payment Frequency",
                "M","N",self.doFreq,None,None,None),
            (("T",0,4,0),("IRB",r2s),0,"Whole File","Whole File",
                "S","N",self.doWhole,None,None,None),
            (("T",0,5,0),"INA",3,"Branch","Branch",
                "","N",self.doBranch,None,None,None),
            (("T",0,6,0),"IUI",1,"Department","Department",
                "","N",self.doDept,None,None,None),
            (("T",0,7,0),("IRB",r3s),0,"Ignore Standards","Ignore Standards",
                "N","N",self.doIgnore,None,None,None),
            (("T",0,8,0),("IRB",r3s),0,"Include Monthly","Include Monthly",
                "Y","N",self.doMthly,None,None,None),
            (("T",0,9,0),("IRB",r3s),0,"Exclude Minus Balances",
                "Exclude Minus Balances","N","N",self.doMinus,None,None,None),
            (("T",0,10,0),("IRB",r3s),0,"Preview Only","Preview Only",
                "Y","N",self.doPreview,None,None,None))
        tnd = ((self.doEnd,"y"), )
        txt = (self.doExit, )
        self.df = TartanDialog(self.mf, title=self.tit, eflds=fld, tend=tnd,
            txit=txt, view=("N","V"), mail=("B","Y"))

    def doRunDate(self, frt, pag, r, c, p, i, w):
        self.rundt = w
        self.curdt = w / 100
        self.taxyr = w / 10000
        mth = self.curdt % 100
        if mth > 2:
            self.taxyr = self.taxyr + 1
        if self.glint == "Y":
            if self.curdt < self.s_per or self.curdt > self.e_per:
                return "Invalid Date, Not In Financial Period"
        y = w / 10000
        m = (w % 10000) / 100
        if m < 3:
            self.sdate = ((y - 1) * 10000) + 301
        else:
            self.sdate = (y * 10000) + 301

    def doPayDate(self, frt, pag, r, c, p, i, w):
        self.paydt = w

    def doMess(self, frt, pag, r, c, p, i, w):
        if not w:
            self.mess = None
        else:
            self.mess = self.sql.getRec(tables="ctlmes", cols=["mss_detail"],
            where=[("mss_system", "=", "WAG"), ("mss_message", "=", w)],
            limit=1)

    def doFreq(self, frt, pag, r, c, p, i, w):
        old = self.sql.getRec(tables="wagtf1", cols=["count(*)"],
            where=[("wt1_cono", "=", self.conum), ("wt1_date", "=",
            self.rundt), ("wt1_freq", "=", w)], limit=1)
        if old[0]:
            ok = askQuestion(screen=self.mf.body, head="Duplicate Payrun",
                mess="A Payrun for this date has already been run!\n\n"\
                "Do you want to continue?", default="no")
            if ok == "no":
                return "Duplicate Payrun"
        self.freq = w

    def doWhole(self, frt, pag, r, c, p, i, w):
        self.whole = w
        if self.whole == "Y":
            self.branch = ""
            self.dept = ""
            return "sk2"

    def doBranch(self, frt, pag, r, c, p, i, w):
        self.branch = w

    def doDept(self, frt, pag, r, c, p, i, w):
        self.dept = w

    def doIgnore(self, frt, pag, r, c, p, i, w):
        self.ignore = w

    def doMthly(self, frt, pag, r, c, p, i, w):
        self.mthly = w

    def doMinus(self, frt, pag, r, c, p, i, w):
        self.minus = w

    def doPreview(self, frt, pag, r, c, p, i, w):
        self.preview = w

    def doEnd(self):
        self.df.closeProcess()
        self.prtnam = self.df.repprt[0][2]
        if self.df.repprt[1][2]:
            self.email = self.df.repprt[1][2]
        else:
            self.email = self.df.repprt[1][1]
        self.printm = self.df.repprt[1][3]
        if self.bestac:
            self.export = open(os.path.join(self.mf.wrkdir,
                "best%03d_%s.txt" % (self.conum, self.paydt)), "w")
            # Header for BEST
            self.export.write("%1s%4s%-40s%8s%1s%8s%-15s%1s%2s%1s%9s%2s%4s"\
                "\r\n" % ("*", self.bestac, self.conam, self.paydt, "Y", "",
                "SALARIES EFT", "+", self.besttp, 0, "", "01", "LIVE"))
        else:
            self.export = None
        whr = [("wgm_cono", "=", self.conum), ("wgm_freq", "=", self.freq),
            ("wgm_status", "=", "C")]
        if self.branch:
            whr.append(("wgm_branch", "=", self.branch))
        if self.dept:
            whr.append(("wgm_dept", "=", self.dept))
        if self.whole == "S":
            whr.extend(
                [("wcp_cono", "=", self.conum),
                ("wcp_empno=wgm_empno",),
                ("wcp_paid", "=", "N")])
            col = ["wgm_empno", "wgm_sname", "wgm_fname"]
            all = getSingleRecords(self.mf, ["wagmst", "wagcap"], col,
                where=whr, group=True)
        else:
            all = self.sql.getRec(tables="wagmst", where=whr, order="wgm_empno")
            if not all:
                showError(self.mf.body, "Error", "No Records Available")
        if all:
            for emp in all:
                self.doProcess(emp)
            if self.glint == "Y":
                chk = self.doIntegration()
            else:
                chk = True
            if chk:
                self.doPrint()
                if self.export:
                    # Trailer for BEST
                    value = long(round((self.etotal * 100), 0))
                    self.export.write("%1s%4s%1s%30s%013u%47s\r\n" % \
                        (2, self.bestac, "T", "", value, ""))
                    self.export.close()
                if self.preview == "Y":
                    self.mf.dbm.rollbackDbase()
                else:
                    self.mf.dbm.commitDbase()
        self.mf.closeLoop()

    def doExit(self):
        self.df.closeProcess()
        self.mf.closeLoop()

    def doProcess(self, emp):
        self.empno = int(emp[self.sql.wagmst_col.index("wgm_empno")])
        self.dob = emp[self.sql.wagmst_col.index("wgm_dob")]
        self.begin = emp[self.sql.wagmst_col.index("wgm_start")]
        if self.begin > self.rundt:
            return
        self.ebranch = emp[self.sql.wagmst_col.index("wgm_branch")]
        self.edept = emp[self.sql.wagmst_col.index("wgm_dept")]
        self.efreq = emp[self.sql.wagmst_col.index("wgm_freq")]
        self.etype = emp[self.sql.wagmst_col.index("wgm_ptype")]
        self.erate = float(emp[self.sql.wagmst_col.index("wgm_payamt")])
        self.paye = emp[self.sql.wagmst_col.index("wgm_paye")]
        self.fix_rate = emp[self.sql.wagmst_col.index("wgm_fix_rate")]
        self.direct = emp[self.sql.wagmst_col.index("wgm_direct")]
        self.doRates(self.erate)
        self.taxamt = 0
        self.doCodes()
        no = self.doCapture()
        if not no:
            self.empnos.append(self.empno)
            self.nhrs = 0
            self.npay = 0
            self.taxbl = 0
            self.notax = 0
            self.rtpay = 0
            self.anpay = 0
            self.pspay = 0
            self.uifpay = 0
            self.taxdd = 0
            self.nondd = 0
            self.psded = 0
            self.shift = 0
            self.drbal = 0
            self.totern = 0
            self.totded = 0
            self.pgnum = self.pgnum + 1
            self.doEarnings()
            self.doDeductions()
            dat = (self.conum, self.empno, self.pgnum, self.rundt, self.freq,
                self.etype, self.ebranch, self.edept, self.erate, self.nhrs,
                self.npay, self.taxbl, self.notax, self.rtpay, self.anpay,
                self.pspay, self.uifpay, self.taxdd, self.nondd, self.psded,
                self.shift, self.drbal, self.sysdtw, 0)
            self.sql.insRec("wagtf1", data=dat)
            self.doPAYE()
            self.doStaffLoans()

    def doRates(self, rate):
        if self.freq == "M":
            self.mrate = rate
            self.wrate = round((self.mrate * self.hrswk / self.hrsmt), 2)
            self.drate = round((self.mrate * self.hrsdy / self.hrsmt), 2)
            self.hrate = round((self.mrate / self.hrsmt), 2)
        else:
            self.hrate = rate
            self.drate = round(self.hrate * self.hrsdy, 2)
            self.wrate = round(self.hrate * self.hrswk, 2)
            self.mrate = round(self.hrate * self.hrsmt, 2)
        self.frate = round((self.wrate * 2), 2)

    def doCodes(self):
        self.ern = {}
        self.ded = {}
        codes = self.sql.getRec(tables="wagcod", where=[("wcd_cono", "=",
            self.conum), ("wcd_empno", "=", self.empno)])
        if not codes:
            return
        for c in codes:
            ctype = c[self.sql.wagcod_col.index("wcd_type")]
            code = int(c[self.sql.wagcod_col.index("wcd_code")])
            eamt = float(c[self.sql.wagcod_col.index("wcd_eamt")])
            ramt = float(c[self.sql.wagcod_col.index("wcd_ramt")])
            det = self.sql.getRec(tables="wagedc", where=[("ced_cono", "=",
                self.conum), ("ced_type", "=", ctype), ("ced_code", "=", code)],
                limit=1)
            if not det:
                continue
            if ctype == "D" and det[
                    self.sql.wagedc_col.index("ced_ror")] == 4102:
                self.taxamt = float(ASD(self.taxamt) + ASD(eamt))
            elif self.ignore == "Y":
                continue
            elif ctype == "E":
                if code in self.ern:
                    self.ern[code][0] = float(ASD(self.ern[code][0])+ASD(eamt))
                    self.ern[code][1] = float(ASD(self.ern[code][0])+ASD(ramt))
                else:
                    self.ern[code] = [eamt, ramt]
            elif ctype == "D":
                if code in self.ded:
                    self.ded[code][0] = float(ASD(self.ded[code][0])+ASD(eamt))
                    self.ded[code][1] = float(ASD(self.ded[code][0])+ASD(ramt))
                else:
                    self.ded[code] = [eamt, ramt]

    def doCapture(self):
        yes = self.sql.getRec(tables="wagcap", cols=["wcp_type", "wcp_code",
            "round(sum(wcp_amt),2)"], where=[("wcp_cono", "=", self.conum),
            ("wcp_empno", "=", self.empno), ("wcp_ind", "=", "Y"),
            ("wcp_paid", "=", "N")], group="wcp_type, wcp_code")
        no = self.sql.getRec(tables="wagcap", cols=["wcp_type", "wcp_code"],
            where=[("wcp_cono", "=", self.conum), ("wcp_empno", "=",
            self.empno), ("wcp_ind", "=", "N"), ("wcp_paid", "=", "N")])
        if not yes:
            return "no"
        if not no:
            no = []
        for y in yes:
            typ = y[0]
            cod = int(y[1])
            amt = float(y[2])
            ind = self.sql.getRec(tables="wagedc", cols=["ced_eind", "ced_ror"],
                where=[("ced_cono", "=", self.conum), ("ced_type", "=", typ),
                ("ced_code", "=", cod)], limit=1)
            if not ind:
                continue
            if typ == "E":
                if cod in self.ern:
                    if ind[0] == "R":
                        self.ern[cod][0] = amt
                    else:
                        self.ern[cod][0] = float(ASD(self.ern[cod][0]) + \
                        ASD(amt))
                else:
                    self.ern[cod] = [amt, 0]
            else:
                if ind[1] == 4102:
                    self.taxamt = float(ASD(self.taxamt) + ASD(amt))
                elif cod in self.ded:
                    if ind[0] == "R":
                        self.ded[cod] = amt
                    else:
                        self.ded[cod][0] = float(ASD(self.ded[cod][0]) + \
                        ASD(amt))
                else:
                    self.ded[cod] = [amt, 0]
        for n in no:
            typ = n[0]
            cod = int(n[1])
            if typ == "E":
                if cod in self.ern:
                    del self.ern[cod]
            else:
                if cod in self.ded:
                    del self.ded[cod]
        if self.preview == "N":
            self.sql.updRec("wagcap", cols=["wcp_paid"],
            data=["Y"], where=[("wcp_cono", "=", self.conum), ("wcp_empno",
            "=", self.empno)])

    def doEarnings(self):
        if self.freq == "W":
            rate = self.wrate
        elif self.freq == "F":
            rate = self.frate
        elif self.freq == "M":
            rate = self.mrate
        idx = self.ern.keys()
        idx.sort()
        for cod in idx:
            ced = self.sql.getRec(tables="wagedc", where=[("ced_cono", "=",
                self.conum), ("ced_type", "=", "E"), ("ced_code", "=", cod)],
                limit=1)
            if not ced:
                continue
            desc = ced[self.sql.wagedc_col.index("ced_desc")]
            eind = ced[self.sql.wagedc_col.index("ced_eind")]
            ebase = ced[self.sql.wagedc_col.index("ced_ebase")]
            eamt = ced[self.sql.wagedc_col.index("ced_eamt")]
            if cod == 1 and eamt == 0:
                # Invalid wgexp0 conversion/imported from old system
                eamt = 1
            rind = ced[self.sql.wagedc_col.index("ced_rind")]
            ramt = ced[self.sql.wagedc_col.index("ced_ramt")]
            tax = ced[self.sql.wagedc_col.index("ced_taxcode")]
            taxportion = ced[self.sql.wagedc_col.index("ced_taxportion")]
            earnt = ced[self.sql.wagedc_col.index("ced_earntype")]
            eglco = long(ced[self.sql.wagedc_col.index("ced_eglco")])
            eglno = long(ced[self.sql.wagedc_col.index("ced_eglno")])
            uifportion = ced[self.sql.wagedc_col.index("ced_uifportion")]
            if eind == "F" and not self.ern[cod][0]:
                self.ern[cod][0] = eamt
            if cod != 1 and not self.ern[cod][0]:
                continue
            if cod < 6:
                if cod == 5:
                    hrs = self.ern[cod][0] * self.hrsdy
                else:
                    hrs = self.ern[cod][0]
                if cod == 1 and hrs == 0:
                    if self.freq == "W":
                        hrs = self.hrswk
                    elif self.freq == "F":
                        hrs = self.hrswk * 2
                    elif self.freq == "M":
                        hrs = self.hrsmt
                if hrs == self.hrsmt:
                    emt = self.mrate
                elif hrs == self.hrswk:
                    emt = self.wrate
                elif hrs == self.hrsdy:
                    emt = self.drate
                else:
                    emt = round((hrs * self.hrate * eamt), 2)
                self.nhrs = float(ASD(self.nhrs) + ASD(hrs))
                self.npay = float(ASD(self.npay) + ASD(emt))
                if not uifportion:
                    uifportion = 100.00
            elif eind == "R":
                hrs = 0
                ct1 = self.ern[cod][0]
                if ebase == 1:
                    emt = round((ct1 * eamt * self.erate), 2)
                elif ebase == 2:
                    emt = round((eamt * rate), 2)
                elif ebase == 3:
                    emt = round((eamt * rate / 100), 2)
                elif ebase == 4:
                    emt = round((eamt * self.drate * ct1), 2)
                elif ebase == 5:
                    emt = round((eamt * self.hrate * ct1), 2)
                else:
                    emt = 0
            else:
                hrs = 0
                if self.ern[cod][0]:
                    emt = self.ern[cod][0]
                else:
                    emt = eamt
            if tax == "N":
                self.notax = float(ASD(self.notax) + ASD(emt))
            elif taxportion:
                amt = round((emt * taxportion / 100), 2)
                self.notax = float(ASD(self.notax) + ASD(emt) - ASD(amt))
                self.taxbl = float(ASD(self.taxbl) + ASD(amt))
            if tax == "R":
                self.rtpay = float(ASD(self.rtpay) + ASD(emt))
            if tax == "O":
                self.anpay = float(ASD(self.anpay) + ASD(emt))
            if tax == "T":
                self.pspay = float(ASD(self.pspay) + ASD(emt))
            else:
                self.totern = float(ASD(self.totern) + ASD(emt))
            if uifportion:
                amt = round((emt * uifportion / 100), 2)
                self.uifpay = float(ASD(self.uifpay) + ASD(amt))
            # Write Transaction Detail
            wagtf2 = [self.conum, self.empno, self.pgnum, self.rundt, "E",
                cod, hrs, emt, 0, tax, taxportion, earnt, eglco, eglno, 0, 0]
            self.doWriteWagtf2(wagtf2)

    def doDeductions(self):
        if self.freq == "W":
            rate = self.wrate
        elif self.freq == "F":
            rate = self.frate
        elif self.freq == "M":
            rate = self.mrate
        idx = self.ded.keys()
        idx.sort()
        for cod in idx:
            ced = self.sql.getRec(tables="wagedc", where=[("ced_cono", "=",
                self.conum), ("ced_type", "=", "D"), ("ced_code", "=", cod)],
                limit=1)
            if not ced:
                continue
            desc = ced[self.sql.wagedc_col.index("ced_desc")]
            eind = ced[self.sql.wagedc_col.index("ced_eind")]
            ebase = ced[self.sql.wagedc_col.index("ced_ebase")]
            eamt = ced[self.sql.wagedc_col.index("ced_eamt")]
            elim = ced[self.sql.wagedc_col.index("ced_elimit")]
            rind = ced[self.sql.wagedc_col.index("ced_rind")]
            rbase = ced[self.sql.wagedc_col.index("ced_rbase")]
            ramt = ced[self.sql.wagedc_col.index("ced_ramt")]
            rlim = ced[self.sql.wagedc_col.index("ced_rlimit")]
            must = ced[self.sql.wagedc_col.index("ced_must")]
            tax = ced[self.sql.wagedc_col.index("ced_taxcode")]
            taxportion = ced[self.sql.wagedc_col.index("ced_taxportion")]
            hlim = ced[self.sql.wagedc_col.index("ced_hr_limit")]
            mthly = ced[self.sql.wagedc_col.index("ced_monthly")]
            balno = ced[self.sql.wagedc_col.index("ced_balno")]
            report = ced[self.sql.wagedc_col.index("ced_report")]
            earnt = ced[self.sql.wagedc_col.index("ced_earntype")]
            eglco = long(ced[self.sql.wagedc_col.index("ced_eglco")])
            eglno = long(ced[self.sql.wagedc_col.index("ced_eglno")])
            if not eglco and eglno:
                eglco = self.conum
            rglco = long(ced[self.sql.wagedc_col.index("ced_rglco")])
            rglno = long(ced[self.sql.wagedc_col.index("ced_rglno")])
            if not rglco and rglno:
                rglco = self.conum
            if mthly == "Y" and self.mthly == "N":
                continue
            if hlim and self.nhrs < hlim:
                continue
            if eind == "R":
                ct1 = self.ded[cod][0]
                if ebase == 1:
                    emt = round((ct1 * eamt * self.erate), 2)
                elif ebase == 2:
                    emt = round((eamt * rate), 2)
                elif ebase == 3:
                    emt = round((eamt * rate / 100), 2)
                elif ebase == 4:
                    emt = round((eamt * self.drate * ct1), 2)
                elif ebase == 5:
                    emt = round((eamt * self.hrate * ct1), 2)
                elif ebase == 6:
                    emt = round((eamt * self.uifpay / 100), 2)
                else:
                    emt = 0
            elif self.ded[cod][0]:
                emt = self.ded[cod][0]
            else:
                emt = eamt
            # Check Balance Records
            if balno:
                bwh = [("wbl_cono", "=", self.conum), ("wbl_empno", "=",
                    self.empno), ("wbl_balno", "=", balno)]
                bal = self.sql.getRec(tables="wagbal", where=bwh, limit=1)
                if bal:
                    if bal[3] < emt:
                        emt = bal[3]
                    bal[3] = float(ASD(bal[3]) - ASD(emt))
            else:
                bal = None
            if elim and emt > elim:
                emt = elim
            if tax == "N":
                self.nondd = float(ASD(self.nondd) + ASD(emt))
            else:
                self.taxdd = float(ASD(self.taxdd) + ASD(emt))
            if must == "N" and self.taxdd > self.totern:
                emt = 0
            if not emt:
                continue
            if tax == "T":
                self.psded = float(ASD(self.psded) + ASD(emt))
            else:
                self.totded = float(ASD(self.totded) + ASD(emt))
            if tax != "N" and taxportion:
                amt = round((emt * taxportion / 100), 2)
                self.taxbl = float(ASD(self.taxbl) - ASD(amt))
            # Update Balance Records
            if bal:
                self.sql.updRec("wagbal", data=bal, where=bwh)
            # Update Union Records
#            if report == "Y":
#                uni = self.sql.getRec(tables="waguni", cols=["wgu_eamt"],
#                    where=[("wgu_cono", "=", self.conum), ("wgu_empno", "=",
#                   self.empno)], limit=1)
#                if not uni:
#                    dat = [self.conum, self.empno, emt, 0]
#                    self.sql.insRec("waguni", data=dat)
#                else:
#                    amt = float(ASD(uni[0]) + ASD(emt))
#                    self.sql.updRec("waguni", cols=["wgu_eamt"], data=[amt],
#                    where=[("wgu_cono", "=", self.conum), ("wgu_empno", "=",
#                    self.empno)])
            if rind == "N":
                rmt = 0
            elif eind == "R":
                ct1 = self.ded[cod][1]
                if rbase == 1:
                    rmt = round((ct1 * ramt * self.erate), 2)
                elif rbase == 2:
                    rmt = round((ramt * rate), 2)
                elif rbase == 3:
                    rmt = round((ramt * rate / 100), 2)
                elif rbase == 4:
                    rmt = round((ramt * self.drate * ct1), 2)
                elif rbase == 5:
                    rmt = round((ramt * self.hrate * ct1), 2)
                elif ebase == 6:
                    rmt = round((ramt * self.uifpay / 100), 2)
                else:
                    rmt = 0
            elif self.ded[cod][1]:
                rmt = self.ded[cod][1]
            else:
                rmt = ramt
            if rlim and rmt > rlim:
                rmt = rlim
            # Write Transaction Detail
            wagtf2 = [self.conum, self.empno, self.pgnum, self.rundt, "D",
                cod, 0, emt, rmt, tax, taxportion, earnt, eglco, eglno,
                rglco, rglno]
            self.doWriteWagtf2(wagtf2)

    def doPAYE(self):
        if self.paye == "N":
            return
        if not self.taxamt:
            # Not manually captured PAYE
            if self.fix_rate and self.direct:
                self.doFlatRate()
            else:
                self.doFormulae()
            if self.taxamt < 0:
                self.taxamt = 0
            else:
                tx = self.sql.getRec(tables="wagtf2",
                    cols=["round(sum(wt2_eamt),2)"], where=[("wt2_cono", "=",
                    self.conum), ("wt2_empno", "=", self.empno), ("wt2_date",
                    "between", self.sdate, self.rundt), ("wt2_type", "=", "D"),
                    ("wt2_code", "=", 1)])
                if tx == [[None]]:
                    tx = 0
                else:
                    tx = tx[0][0]
                self.taxamt = float(ASD(self.taxamt) - ASD(tx))
        if self.taxamt <= 0:
            return
        ced = self.sql.getRec(tables="wagedc", where=[("ced_cono", "=",
            self.conum), ("ced_type", "=", "D"), ("ced_code", "=", 1)],
            limit=1)
        if not ced:
            return
        desc = ced[self.sql.wagedc_col.index("ced_desc")]
        eglco = long(ced[self.sql.wagedc_col.index("ced_eglco")])
        if not eglco:
            eglco = self.conum
        eglno = long(ced[self.sql.wagedc_col.index("ced_eglno")])
        self.totded = float(ASD(self.totded) + ASD(self.taxamt))
        # Write PAYE Transaction Detail
        wagtf2 = [self.conum, self.empno, self.pgnum, self.rundt, "D",
            1, 0, self.taxamt, 0, "N", 0, "", eglco, eglno, 0, 0]
        self.doWriteWagtf2(wagtf2)

    def doFlatRate(self):
        inc, ret, tot = self.doTaxableIncome()
        if inc <= 0:
            self.taxamt = 0
            return
        self.taxamt = round((inc * self.fix_rate / 100), 2)
        if self.taxamt <= 0:
            self.taxamt = 0
            return
        if self.rtpay and self.rtpay > 30000:
            bas = float(ASD(self.rtpay) - ASD(30000))
            rt = rounded((bas * self.fix_rate / 100), 2)
            self.taxamt = round((w-taxamt * self.pers / self.base), 2)
            self.taxamt = float(ASD(self.taxamt) + ASD(rt))
        else:
            self.taxamt = round((self.taxamt * self.pers / self.base), 2)

    def doFormulae(self):
        taxinc, retinc, totinc = self.doTaxableIncome()
        if taxinc <= 0:
            self.taxamt = 0
            return
        ann = 0
        ret = 0
        self.taxamt, prb, arb, nrb, rmx = self.doCompute(taxinc)
        if taxinc != totinc:
            ann, x, x, x, x = self.doCompute(totinc)
            ann = float(ASD(ann) - ASD(self.taxamt))
        age = dateDiff(self.dob, self.rundt, "years")
        if age > 64:
            prb = float(ASD(prb) + ASD(arb))
        if age > 74:
            prb = float(ASD(prb) + ASD(nrb))
        self.taxamt = float(ASD(self.taxamt) - ASD(prb))
        if retinc > 0 and retinc > rmx:
            avg = round(((self.taxamt / taxinc) * 100), 2)
            ret = round(((float(ASD(retinc) - ASD(rmx))) * avg / 100), 2)
        self.taxamt = round((self.taxamt * self.pers / self.base), 2)
        self.taxamt = float(ASD(self.taxamt) + ASD(ann) + ASD(ret))

    def doCompute(self, inc):
        rates = payeRates[str(self.taxyr)]["rates"]
        allow = payeRates[str(self.taxyr)]["allow"]
        tax = 0
        for lvl in rates:
            if inc > lvl[0]:
                tax = round(lvl[1] + (((inc - lvl[0] + 1) * lvl[2]) / 100.0), 2)
            if tax:
                break
        prb = allow[0]                # Primary Rebate
        arb = allow[1]                # Over 65 Rebate Extra
        if len(allow) == 4:
            nrb = allow[2]            # Over 75 Rebate Extra
            rmx = allow[3]            # Retrenchment Allowance
        else:
            nrb = 0                   # Over 75 Rebate Extra
            rmx = allow[2]            # Retrenchment Allowance
        return tax, prb, arb, nrb, rmx

    def doTaxableIncome(self):
        ed = self.sql.getRec(tables="wagtf1", cols=["round(sum(wt1_taxbl), 2)",
            "round(sum(wt1_rtpay), 2)", "round(sum(wt1_anpay), 2)"],
            where=[("wt1_cono", "=", self.conum), ("wt1_empno", "=",
            self.empno), ("wt1_date", "between", self.sdate, self.rundt)])
        inc = 0
        anp = 0
        if ed:
            rtp = ed[0][1]
            anp = ed[0][2]
            inc = float(ASD(ed[0][0]) - ASD(anp))
        if self.begin < self.sdate:
            begin = self.sdate
        else:
            begin = self.begin
        if self.efreq == "M":
            self.base = 12
            y1 = int(self.rundt / 10000)
            y2 = int(begin / 10000)
            m1 = int((self.rundt - (y1 * 10000)) / 100)
            m2 = int((begin - (y2 * 10000)) / 100)
            self.pers = 1 + ((y1 - y2) * self.base) + (m1 - m2)
            while self.pers > self.base:
                self.pers = self.pers - self.base
        else:
            days = dateDiff(begin, self.rundt, "days")
            if self.efreq == "W":
                per = 7
                self.base = 52
            else:
                per = 14
                self.base = 26
            self.pers = int(days / per)
            if days % per > 0:
                self.pers = self.pers + 1
            if self.pers > self.base:
                self.pers = self.base
        inc = round((inc * self.base / self.pers), 2)
        tot = float(ASD(inc) + ASD(anp))
        return inc, float(ed[0][1]), tot

    def doStaffLoans(self):
        if self.ignore == "Y":
            return
        loan = 0
        self.lonbl1 = 0
        self.lonbl2 = 0
        while loan != 99:
            totpay = float(ASD(self.totern) - ASD(self.totded))
            if totpay <= 0:
                return
            loan = loan + 1
            lmf = self.sql.getRec(tables="waglmf", where=[("wlm_cono", "=",
                self.conum), ("wlm_empno", "=", self.empno), ("wlm_loan",
                "=", loan)], limit=1)
            if not lmf:
                continue
            dcode = lmf[self.sql.waglmf_col.index("wlm_ded_code")]
            irate = lmf[self.sql.waglmf_col.index("wlm_int_per")]
            sdate = lmf[self.sql.waglmf_col.index("wlm_start_date")]
            damnt = lmf[self.sql.waglmf_col.index("wlm_ded_amt")]
            ced = self.sql.getRec(tables="wagedc", where=[("ced_cono", "=",
                self.conum), ("ced_type", "=", "D"), ("ced_code", "=", dcode)],
                limit=1)
            if not ced:
                return # Create Loan wagedc
            mthly = ced[self.sql.wagedc_col.index("ced_monthly")]
            earnt = ced[self.sql.wagedc_col.index("ced_earntype")]
            eglco = long(ced[self.sql.wagedc_col.index("ced_eglco")])
            if not eglco:
                eglco = self.conum
            eglno = long(ced[self.sql.wagedc_col.index("ced_eglno")])
            if mthly == "Y" and self.mthly == "N":
                return
            if self.preview == "Y":
                upd = "N"
            else:
                upd = "Y"
            lint = LoanInterest(self.mf, self.period, self.conum,
                self.empno, loan, irate, sdate, self.rundt, "Payslip",
                self.curdt, update=upd, glctls=self.glctls)
            if not lint.bal:
                continue
            if damnt > lint.bal:
                ded = lint.bal
            else:
                ded = damnt
            if totpay <= ded:
                ded = totpay
            # Write Salary Transaction
            wagtf2 = [self.conum, self.empno, self.pgnum, self.rundt, "D",
                dcode, 0, ded, 0, "N", 0, earnt, eglco, eglno, 0, 0]
            self.doWriteWagtf2(wagtf2)
            # Write Loan Transaction
            ded = float(ASD(0) - ASD(ded))
            dat = [self.conum, self.empno, loan, "Payslip", 4, self.rundt,
                self.pgnum, ded, 0, 0, 0, self.curdt, "Salary Deduction",
                "N", self.sysdtw, 0]
            self.sql.insRec("wagltf", data=dat)

    def doWriteWagtf2(self, dat):
        old = self.sql.getRec(tables="wagtf2", cols=["wt2_hrs", "wt2_eamt",
            "wt2_ramt"], where=[("wt2_cono", "=", dat[0]), ("wt2_empno", "=",
            dat[1]), ("wt2_page", "=", dat[2]), ("wt2_date", "=", dat[3]),
            ("wt2_type", "=", dat[4]), ("wt2_code", "=", dat[5])], limit=1)
        if old:
            old[0] = float(ASD(old[0]) + ASD(dat[6]))
            old[1] = float(ASD(old[1]) + ASD(dat[7]))
            old[2] = float(ASD(old[2]) + ASD(dat[8]))
            self.sql.updRec("wagtf2",
                cols=["wt2_hrs", "wt2_eamt", "wt2_ramt"], data=[old[0],
                old[1], old[2]], where=[("wt2_cono", "=", dat[0]),
                ("wt2_empno", "=", dat[1]), ("wt2_date", "=", dat[3]),
                ("wt2_type", "=", dat[4]), ("wt2_code", "=", dat[5])])
        else:
            self.sql.insRec("wagtf2", data=dat)

    def doPrint(self):
        if self.preview == "Y":
            runtp = "p"
        else:
            runtp = "o"
        psl = PrintPayslip(self.mf, self.conum, self.empnos, self.rundt,
            progs="y", runtp=runtp, prtnam=self.prtnam, email=self.email,
            printm=self.printm, message=self.mess, export=self.export,
            bestac=self.bestac)
        self.etotal = float(ASD(self.etotal) + ASD(psl.etotal))

    def doIntegration(self):
        tr1 = self.sql.getRec(tables="wagtf1", where=[("wt1_cono", "=",
            self.conum), ("wt1_date", "=", self.rundt), ("wt1_freq", "=",
            self.freq)], order="wt1_empno")
        if not tr1:
            return
        for t1 in tr1:
            empno = int(t1[self.sql.wagtf1_col.index("wt1_empno")])
            self.branch = t1[self.sql.wagtf1_col.index("wt1_branch")]
            self.dept = int(t1[self.sql.wagtf1_col.index("wt1_dept")])
            cols = ["dep_dr_coy", "dep_dr_sal", "dep_dr_com", "dep_cr_coy",
                "dep_cr_sal", "dep_cr_com"]
            self.glacno = self.sql.getRec(tables="ctldep", cols=cols,
                where=[("dep_cono", "=", self.conum), ("dep_branch", "=",
                self.branch), ("dep_code", "=", self.dept)], limit=1)
            # Check that all G/L Accounts are Valid
            err = None
            if not self.glacno:
                self.glacno = self.doGetControls()
            if not self.glacno:
                err = "y"
            else:
                if not self.glacno[0]:
                    self.glacno[0] = self.conum
                desc = self.doReadGL(self.glacno[0], self.glacno[1])
                if not desc:
                    err = "y"
                if not self.glacno[2]:
                    self.glacno[2] = self.glacno[1]
                desc = self.doReadGL(self.glacno[0], self.glacno[2])
                if not desc:
                    err = "y"
                if not self.glacno[3]:
                    self.glacno[3] = self.conum
                if not self.glacno[4]:
                    self.glacno[4] = self.wagctl
                desc = self.doReadGL(self.glacno[3], self.glacno[4])
                if not desc:
                    err = "y"
                if not self.glacno[5]:
                    self.glacno[5] = self.glacno[4]
                desc = self.doReadGL(self.glacno[3], self.glacno[5])
                if not desc:
                    err = "y"
            if err:
                showError(self.mf.body, "Department Error", "Department "\
                    "%s %s %s Has an Invalid GL Account Number. Please "\
                    "Correct it and Re-Run Payslips." % (self.conum,
                    self.branch, self.dept))
                return
            tr2 = self.sql.getRec(tables="wagtf2", where=[("wt2_cono", "=",
                self.conum), ("wt2_empno", "=", empno), ("wt2_date", "=",
                self.rundt)])
            if not tr2:
                continue
            #p = ProgressBar(self.mf.body, max=len(tr2),
            #    typ="Integrating %s" % self.rundp)
            for num, t2 in enumerate(tr2):
                #p.displayProgress(num)
                # Check if Notional
                if t2[self.sql.wagtf2_col.index("wt2_taxcode")] == "T":
                    continue
                code = t2[self.sql.wagtf2_col.index("wt2_code")]
                wgtp = t2[self.sql.wagtf2_col.index("wt2_type")]
                gltp = t2[self.sql.wagtf2_col.index("wt2_gl_type")]
                coy1 = t2[self.sql.wagtf2_col.index("wt2_gl_econo")]
                acc1 = int(t2[self.sql.wagtf2_col.index("wt2_gl_eacno")])
                coy2 = self.glacno[3]
                acc2 = self.glacno[4]
                edc = self.sql.getRec(tables="wagedc", cols=["ced_desc",
                    "ced_eglco", "ced_eglno"], where=[("ced_cono", "=",
                    self.conum), ("ced_type", "=", wgtp), ("ced_code", "=",
                    code)], limit=1)
                edes = edc[0]
                if edc and not coy1:
                    coy1 = edc[1]
                if edc and not acc1:
                    acc1 = edc[2]
                if wgtp == "E":
                    if not coy1:
                        coy1 = self.glacno[0]
                    if gltp == "S" and not acc1: acc1 = self.glacno[1]  # Salary
                    if gltp == "C" and not acc1: acc1 = self.glacno[2]  # Comm
                elif not coy1:
                    coy1 = self.conum
                desc = self.doReadGL(coy1, acc1)
                if not desc:
                    showError(self.mf.body, "Code Error",
                        "Invalid GL Account for Code %s %s " % (wgtp, code))
                    #p.closeProgress()
                    return
                # Expense Record
                eamt = float(t2[self.sql.wagtf2_col.index("wt2_eamt")])
                if not eamt:
                    continue
                if wgtp == "D":
                    eamt = float(ASD(0) - ASD(eamt))
                edes = "Sal/Wag - %s" % edes
                self.doGenTrn(coy1, acc1, eamt, edes)
                if coy1 != self.conum:
                    err = self.doGenInt(coy1, eamt)
                    if err:
                        #p.closeProgress()
                        return
                if wgtp == "E":
                    if not coy2:
                        coy2 = self.glacno[3]
                    if gltp == "S" and not acc2: acc2 = self.glacno[4]  # Salary
                    if gltp == "C" and not acc2: acc2 = self.glacno[5]  # Comm
                else:
                    if not coy2:
                        coy2 = self.conum
                    if not acc2:
                        acc2 = self.wagctl
                desc = self.doReadGL(coy1, acc1)
                if not desc:
                    showError(self.mf.body, "Code Error",
                        "Invalid GL Account for Code %s %s " % (wgtp, code))
                    #p.closeProgress()
                    return
                # Control Record
                eamt = float(ASD(0) - ASD(eamt))
                self.doGenTrn(coy2, acc2, eamt)
                if coy2 != self.conum:
                    err = self.doGenInt(coy2, eamt)
                    if err:
                        #p.closeProgress()
                        return
            #p.closeProgress()
        return True

    def doGetControls(self):
        tit = ("Control Accounts",)
        glm = {
            "stype": "R",
            "tables": ("genmst",),
            "cols": (
                ("glm_acno", "", 0, "Acc-Num"),
                ("glm_desc", "", 0, "Description", "Y")),
            "where": [("glm_cono", "=", self.conum)]}
        fld = (
            (("T",0,0,0),"ONA",3,"Branch"),
            (("T",0,0,25),"ONA",30,""),
            (("T",0,1,0),"OUI",1,"Department"),
            (("T",0,1,25),"ONA",30,""),
            (("T",0,2,0),"IUI",3,"Debit  Company","",
                self.conum,"N",self.doGLCoy,coy,None,("notzero",)),
            (("T",0,2,25),"ONA",30,""),
            (("T",0,3,0),"IUI",7,"       Salary","Debit Salary",
                "","N",self.doGLAcc,glm,None,("notzero",)),
            (("T",0,3,25),"ONA",30,""),
            (("T",0,4,0),"IUI",7,"       Commission","Debit Commission",
                "","N",self.doGLAcc,glm,None,("efld",)),
            (("T",0,4,25),"ONA",30,""),
            (("T",0,5,0),"IUI",3,"Credit Company","",
                self.conum,"N",self.doGLCoy,coy,None,("notzero",)),
            (("T",0,5,25),"ONA",30,""),
            (("T",0,6,0),"IUI",7,"       Salary","Credit Salary",
                "","N",self.doGLAcc,glm,None,("notzero",)),
            (("T",0,6,25),"ONA",30,""),
            (("T",0,7,0),"IUI",7,"       Commission","Credit Commission",
                "","N",self.doGLAcc,glm,None,("efld",)),
            (("T",0,7,25),"ONA",30,""))
        tnd = ((self.doCtlEnd,"n"), )
        txt = (self.doCtlExit, )
        state = self.df.disableButtons()
        self.tf = TartanDialog(self.mf, tops=True, title=tit, eflds=fld,
            tend=tnd, txit=txt, screen=self.mf.body)
        bch = self.sql.getRec(tables="ctlbra", cols=["bra_name"],
            where=[("bra_cono", "=", self.conum), ("bra_code", "=",
            self.branch)], limit=1)
        if not bch:
            bch = "Unknown"
        else:
            bch = bch[0]
        dep = self.sql.getRec(tables="ctldep", cols=["dep_name"],
            where=[("dep_cono", "=", self.conum), ("dep_branch", "=",
            self.branch), ("dep_code", "=", self.dept)], limit=1)
        if not dep:
            dep = "Unknown"
        else:
            dep = dep[0]
        self.glacno = [0, 0, 0, 0, 0, 0]
        self.tf.loadEntry("T", 0, 0, data=self.branch)
        self.tf.loadEntry("T", 0, 1, data=bch)
        self.tf.loadEntry("T", 0, 2, data=self.dept)
        self.tf.loadEntry("T", 0, 3, data=dep)
        self.doLoadGL()
        self.tf.focusField("T", 0, 4)
        self.mf.startLoop()
        self.df.enableButtons(state=state)

    def doGLCoy(self, frt, pag, r, c, p, i, w):
        name = self.sql.getRec(tables="ctlmst", cols=["ctm_name"],
            where=[("ctm_cono", "=", w)], limit=1)
        if not name:
            return "Invalid Company Number"
        self.glcoy = w
        if c == 6:
            self.glacno[0] = self.glcoy
        else:
            self.glacno[3] = self.glcoy
        self.tf.loadEntry(frt, pag, p+1, data=name[0])

    def doGLAcc(self, frt, pag, r, c, p, i, w):
        self.glacc = w
        desc = self.doReadGL(self.glcoy, self.glacc)
        if not desc:
            return "Invalid Account Number"
        chk = chkCtlAcc(self.mf, self.glcoy, self.glacc)
        if chk:
            return chk
        if c == 7:
            self.glacno[1] = self.glacc
        elif c == 9:
            self.glacno[2] = self.glacc
        elif c == 13:
            self.glacno[4] = self.glacc
        elif c == 15:
            self.glacno[5] = self.glacc
        self.doLoadGL()

    def doLoadGL(self):
        if not self.glacno:
            return
        if self.glacno[1]:
            desc = self.doReadGL(self.glacno[0], self.glacno[1])
            if not desc:
                desc = ""
            else:
                desc = desc[0]
            self.tf.loadEntry("T", 0, 7, data=desc)
        if self.glacno[2]:
            desc = self.doReadGL(self.glacno[0], self.glacno[2])
            if not desc:
                desc = ""
            else:
                desc = desc[0]
            self.tf.loadEntry("T", 0, 9, data=desc)
        if self.glacno[4]:
            desc = self.doReadGL(self.glacno[3], self.glacno[4])
            if not desc:
                desc = ""
            else:
                desc = desc[0]
            self.tf.loadEntry("T", 0, 13, data=desc)
        if self.glacno[5]:
            desc = self.doReadGL(self.glacno[3], self.glacno[5])
            if not desc:
                desc = ""
            else:
                desc = desc[0]
            self.tf.loadEntry("T", 0, 15, data=desc)

    def doReadGL(self, cono, acno):
        desc = self.sql.getRec(tables="genmst", cols=["glm_desc"],
            where=[("glm_cono", "=", cono), ("glm_acno", "=", acno)], limit=1)
        return desc

    def doCtlExit(self):
        self.glacno = None
        self.doCtlClose()

    def doCtlEnd(self):
        self.doCtlClose()

    def doCtlClose(self):
        self.tf.closeProcess()
        self.mf.closeLoop()

    def doGenTrn(self, cono, acno, amnt, edes=None):
        batno = CCD("%03i%04i" % (self.conum, (self.curdt % 10000)), "Na", 7)
        refno = CCD(1, "Na", 9)
        whr = [("glt_cono", "=", cono), ("glt_acno", "=", acno),
            ("glt_curdt", "=", self.curdt), ("glt_trdt", "=", self.rundt),
            ("glt_type", "=", 4), ("glt_refno", "=", refno.work),
            ("glt_batch", "=", batno.work)]
        if edes:
            whr.append(("glt_desc", "=", edes))
        else:
            edes = "Salaries and Wages"
        trn = self.sql.getRec(tables="gentrn", cols=["glt_seq", "glt_tramt"],
            where=whr, limit=1)
        if not trn:
            data = [cono, acno, self.curdt, self.rundt, 4, refno.work,
                batno.work, amnt, 0.00, edes, "", "", 0, self.sysdtw, 0]
            self.sql.insRec("gentrn", data=data)
        else:
            newamt = float(ASD(trn[1]) + ASD(amnt))
            self.sql.updRec("gentrn", cols=["glt_tramt"], data=[newamt],
                where=[("glt_seq", "=", int(trn[0]))])

    def doGenInt(self, cono, amnt):
        # General Ledger Transaction (Intercompany From)
        acc = self.sql.getRec(tables="genint", cols=["cti_acno"],
            where=[("cti_cono", "=", self.conum), ("cti_inco", "=", cono)],
            limit=1)
        if not acc:
            showError(self.mf.body, "Inter Company Error", "Invalid "\
                "Account %s for Company %s in %s" % (acc, cono, self.conum))
            return "err"
        self.doGenTrn(self.conum, acc[0], amnt)
        # General Ledger Transaction (Intercompany To)
        acc = self.sql.getRec(tables="genint", cols=["cti_acno"],
            where=[("cti_cono", "=", cono), ("cti_inco", "=", self.conum)],
            limit=1)
        if not acc:
            showError(self.mf.body, "Inter Company Error", "Invalid "\
                "Account %s for Company %s in %s" % (acc, self.conum, cono))
            return "err"
        amnt = float(ASD(0) - ASD(amnt))
        self.doGenTrn(cono, acc[0], amnt)

