"""
SYNOPSIS
    These are various classes used in TARTAN.

    This file is part of Tartan Systems (TARTAN).

AUTHOR
    Written by Paul Malherbe, <paul@tartan.co.za>

COPYING
    Copyright (C) 2004-2012 Paul Malherbe. Free use of this software is
    granted under the terms of the GNU General Public License (GPL) as
    published by the Free Software Foundation, either version 2 of the
    License, or (at your option) any later version.

    TARTAN is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY, without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License along
    with TARTAN, if not, write to the Free Software Foundation, Inc.,
    59 Temple Place, Suite 330, Boston, MA  02111-1307  USA or view/download
    it at http://www.gnu.org/licenses/gpl.txt
"""
#================================
# Standard Python imports e.g. os
#================================
import base64, copy, getpass, glob, gzip, os, re, shutil, subprocess, sys, tarfile, time
try:
    from PIL import Image as PILImage
except:
    PILImage = False
#========================================================
# TARTAN Standard Functions and Variables e.g. showError
#========================================================
from tartanFunctions import *
from tartanWork import *
#========================================================
# PyFPDF
#========================================================
from fpdf import FPDF

def rgb(col):
    return (col // 65536), (col // 256 % 256), (col% 256)
#========================================================
# Excel import and export modules
#========================================================
try:
    import xlrd
    from xlrd.xldate import xldate_as_tuple
    XLRD = True
except:
    XLRD = False
try:
    import xlwt
    XLWT = True
except:
    XLWT = False
#================
# PyChart imports
#================
try:
    from pychart import area, arrow, axis, bar_plot, category_coord
    from pychart import chart_object, line_plot, pie_plot
    from pychart import theme as pychart_theme
    from pychart import canvas as pychart_canvas
    PYCHART = True
except:
    PYCHART = False
#============
# GTK imports
#============
gui, gtk, gobject, pango = doGtkImport()
if gui:
    PYGTK_VERSION = gtk.pygtk_version
    #=======================================================================
    # Subclassing Window, Entry, TextView and CheckButton for custom signals
    #=======================================================================
    class MyWindow(gtk.Window):
        __gsignals__ = dict(
            mywindo=(gobject.SIGNAL_RUN_LAST|gobject.SIGNAL_ACTION,None,(str,)),
            fullscn=(gobject.SIGNAL_RUN_LAST|gobject.SIGNAL_ACTION,None,(str,)))
    if gobject.pygtk_version < (2, 8):
        gobject.type_register(MyWindow)
    gtk.binding_entry_add_signal(
        MyWindow, gtk.keysyms.F11, 0, "fullscn", str, "F11")
    gtk.binding_entry_add_signal(
        MyWindow, gtk.keysyms.Escape, 0, "mywindo", str, "Escape")

    class MyEntry(gtk.Entry):
        __gsignals__ = dict(
            myentry=(gobject.SIGNAL_RUN_LAST|gobject.SIGNAL_ACTION,None,(str,)))
    if gobject.pygtk_version < (2, 8):
        gobject.type_register(MyEntry)
    gtk.binding_entry_add_signal(
        MyEntry, gtk.keysyms.F1, 0, "myentry", str, "F1")
    gtk.binding_entry_add_signal(
        MyEntry, gtk.keysyms.F5, 0, "myentry", str, "F5")
    gtk.binding_entry_add_signal(
        MyEntry, gtk.keysyms.Escape, 0, "myentry", str, "Escape")

    class MyTextView_v(gtk.TextView):
        __gsignals__ = dict(
            mytxtvv=(gobject.SIGNAL_RUN_LAST|gobject.SIGNAL_ACTION,None,(str,)))
    if gobject.pygtk_version < (2, 8):
        gobject.type_register(MyTextView_v)
    gtk.binding_entry_add_signal(
        MyTextView_v, gtk.keysyms.F1, 0, "mytxtvv", str, "F1")
    gtk.binding_entry_add_signal(
        MyTextView_v, gtk.keysyms.F5, 0, "mytxtvv", str, "F5")
    gtk.binding_entry_add_signal(
        MyTextView_v, gtk.keysyms.Escape, 0, "mytxtvv", str, "Escape")
    gtk.binding_entry_add_signal(
        MyTextView_v, gtk.keysyms.Return, 0, "mytxtvv", str, "Return")
    gtk.binding_entry_add_signal(
        MyTextView_v, gtk.keysyms.KP_Enter, 0, "mytxtvv", str, "KP_Enter")

    class MyTextView_V(gtk.TextView):
        __gsignals__ = dict(
            mytxtvV=(gobject.SIGNAL_RUN_LAST|gobject.SIGNAL_ACTION,None,(str,)))
    if gobject.pygtk_version < (2, 8):
        gobject.type_register(MyTextView_V)
    gtk.binding_entry_add_signal(
        MyTextView_V, gtk.keysyms.F1, 0, "mytxtvV", str, "F1")
    gtk.binding_entry_add_signal(
        MyTextView_V, gtk.keysyms.F5, 0, "mytxtvV", str, "F5")
    gtk.binding_entry_add_signal(
        MyTextView_V, gtk.keysyms.F9, 0, "mytxtvV", str, "F9")
    gtk.binding_entry_add_signal(
        MyTextView_V, gtk.keysyms.Escape, 0, "mytxtvV", str, "Escape")

    class MyButton(gtk.Button):
        __gsignals__ = dict(
            mybuttn=(gobject.SIGNAL_RUN_LAST|gobject.SIGNAL_ACTION,None,(str,)))
    if gobject.pygtk_version < (2, 8):
        gobject.type_register(MyButton)

    class MyCheckButton(gtk.CheckButton):
        __gsignals__ = dict(
            mychkbx=(gobject.SIGNAL_RUN_LAST|gobject.SIGNAL_ACTION,None,(str,)))
    if gobject.pygtk_version < (2, 8):
        gobject.type_register(MyCheckButton)
    gtk.binding_entry_add_signal(
        MyCheckButton, gtk.keysyms.F5, 0, "mychkbx", str, "F5")
    gtk.binding_entry_add_signal(
        MyCheckButton, gtk.keysyms.Escape, 0, "mychkbx", str, "Escape")

class MainFrame:
    """
    Use this class to draw the Main Window with the following widgets:

        head = Label for the heading of the applicatio.
        body = VBox for the body of the applicatio.
        mess = Entry for status messages and prompt.
        conf = Entry for confirmation of input.

    Options:
        rcdic   = dictionary of the tartanrc fil.
        display = whether or not to display widgets or else just create the
                  object without displaying any widgets. This is to be able
                  to use the database objects without gui. Remember that the
                  'gui' variable must be True for widgets to be created.
        fixed   = whether the window is a fixed siz.
        maxi    = whether to start the window maximized.
        dbobj   = whether to create a database object.
    """
    def __init__(self, rcdic=None, display=True, fixed=True, maxi=False, dbobj=True):
        if rcdic:
            self.rcdic = rcdic
        else:
            self.rcdic = loadRcFile()
        self.display = display
        self.fixed = fixed
        self.maxi = maxi
        self.dbobj = dbobj
        self.loop = 0
        for key in self.rcdic:
            if   key == "ft":
                self.dft = self.mft = self.rcdic[key]
            elif key == "fs":
                self.dfs = self.mfs = self.rcdic[key]
            elif key == "t1":
                self.dft = self.rcdic[key]
            elif key == "s1":
                self.dfs = self.rcdic[key]
            elif key == "t2":
                self.mft = self.rcdic[key]
            elif key == "s2":
                self.mfs = self.rcdic[key]
            else:
                exec "self.%s = self.rcdic[key]" % key
        self.override = "natrat"
        if display:
            try:
                if not gui:
                    raise Exception("GUI Error")
                self.drawMainFrame()
            except:
                showException(None, self.rcdic["wrkdir"], "It appears "\
                    "that pygtk has NOT been installed!!!", maxTBlevel=0)
                self.body = None
        else:
            self.body = None
        if self.dbobj:
            # Create database class and sqlRec function
            self.dbm = Dbase(self.rcdic, screen=self.body)
            if self.dbm.err:
                sys.exit()
            if self.dbm.checkDbase():
                self.dbm.openDbase()
                self.sqlRec = Sql(self.dbm, prog=__name__).sqlRec
                self.dbm.closeDbase()
            else:
                self.sqlRec = Sql(self.dbm, prog=__name__).sqlRec

    def drawMainFrame(self):
        # Set the theme
        self.fontm = "%s %s" % (self.mft, self.mfs)
        self.fontd = "%s %s" % (self.dft, self.dfs)
        setTheme(gtk, self.theme, self.fontm, self.fontd)
        # Create a new toplevel window
        title = "%s %s %s" % ("Tartan Systems",
            "Copyright (C) 2004-2012 Paul Malherbe",
            "(%s, %s)" % (self.dbase, self.dbname))
        self.geom = self.geo.split("x")
        for x in range(2):
            self.geom[x] = int(self.geom[x])
        ipath = os.path.join(self.prgdir, "img")
        icon = os.path.join(ipath, "tartan.bmp")
        if not os.path.isfile(icon):
            icon = None
        if not self.fixed:
            self.mkw = MkWindow(name="Main Window", icon=icon,
                title=title, decor=True, remov=True, ipath=ipath)
        else:
            self.mkw = MkWindow(name="Main Window", icon=icon, title=title,
                decor=True, remov=True, place=self.rcdic["plc"], ipath=ipath,
                size=(int(self.geom[0]), int(self.geom[1]), "fixed"))
        #self.mkw.newwin.connect("size-request", self.sizeChanged)
        # Make copy of window
        self.window = self.mkw.newwin
        # Set full screen toggle facility
        self.window.connect("fullscn", self.screenState)
        self.window.connect("window_state_event", self.lastState)
        # Populate the window using a table
        self.table = gtk.Table(1, 3, False)
        self.window.add(self.table)
        # Heading
        event, self.head = doLabel(gtk, self.window, style="out", color=True,
            font=pango.FontDescription(self.fontm))
        self.table.attach(event, 0, 1, 0, 1, gtk.EXPAND|gtk.FILL, 0)
        # Body
        alw = gtk.Alignment(0.5, 0.5, 0, 0)
        self.table.attach(alw, 0, 1, 1, 2)
        self.body = gtk.VBox(False, 0)
        alw.add(self.body)
        # Status
        self.mess = gtk.Entry()
        self.mess.set_has_frame(False)
        self.mess.unset_flags(gtk.CAN_FOCUS)
        self.mess.modify_font(pango.FontDescription(self.fontm))
        self.table.attach(self.mess, 0, 1, 2, 3, gtk.EXPAND|gtk.FILL, 0)
        # Maximize the window
        if self.maxi:
            self.window.maximize()

        # Background Image
        #back = os.path.join(ipath, "tartan.gif")
        #if os.path.isfile(back):
        #    pixbuf = gtk.gdk.pixbuf_new_from_file(back)
        #    pixbuf = pixbuf.scale_simple(self.geom[0],
        #        self.geom[1], gtk.gdk.INTERP_BILINEAR)
        #    pixmap, mask = pixbuf.render_pixmap_and_mask()
        #    del pixbuf
        #    self.window.set_app_paintable(True)
        #    self.window.realize()
        #    self.window.window.set_back_pixmap(pixmap, False)
        #    del pixmap

        # Show all widgets: head, body, mess and window
        self.window.show_all()

    def sizeChanged(self, *opts):
        width, height = self.window.get_size()
        self.window.move((gtk.gdk.screen_width() - width) / 2, 0)

    def updateStatus(self, mess=None, bg="white", fg="black"):
        # clear any previous message
        self.mess.set_text("")
        # display new message
        if mess:
            self.mess.set_text(mess)
        # change color
        st = self.mess.get_style().copy()
        st.base[gtk.STATE_NORMAL] = self.mess.get_colormap().alloc_color(bg)
        st.text[gtk.STATE_NORMAL] = self.mess.get_colormap().alloc_color(fg)
        self.mess.set_style(st)
        self.gtkDispatch()

    def quitProcess(self, widget=None, text=None):
        return True

    def startLoop(self):
        self.loop += 1
        try:
            gtk.main()
        except:
            print "startLoop Error"

    def closeLoop(self):
        self.loop -= 1
        try:
            gtk.main_quit()
        except:
            print "closeLoop Error"

    def gtkDispatch(self):
        # To facilitate dispatching events in routines without gtk
        while gtk.events_pending():
            gtk.main_iteration()

    def screenState(self, widget, key):
        if self.wstate == 0:
            self.window.fullscreen()
        else:
            self.window.unfullscreen()

    def lastState(self, widget, event):
        self.wstate = event.new_window_state

class TartanMenu:
    """
    Use this to class to draw menus as follows:

    Create a list of lists of MenuBar Items for example:

        men = [["MM","Control", "Control Routines and exit"],
               ["MM","General", "General Ledger Menu"],
               ["MM","Creditors", "Creditor's Ledger Menu"],
               ["MM","Debtors", "Debtor's Ledger Menu"],
               ["MM","Stores", "Stores Ledger Menu"],
               ["MM","Help", "User's Manual and About","right"]]

        Where the fields are as follows:

        MM = Main Menu:
             Menu Name
             Detail
        CM = Cascade Menu:
             Parent Name
             Menu Name
             Detail
        
    Create a list of lists of Menu Items for example:

        mod = [["F","Control","tarCfg","Configure Tartan"],
               ["PNN","Control","co1010","Company Records"],
               ["F","Control","sysEnd","Exit Program"],
               ["CY","Control","l -l","Directory"]]

        Where the fields are as follows:

            Type        F =  Function
                        P123 where  1 = Requiring Company Number (Y/N)
                                    2 = Requiring Period Selection (Y/N)
                                    3 = Pass the User Name to the Module
                        CY = System Command requiring a terminal session
                        CN = System Command not requiring a terminal session
                        S  = A standard callable python class
            Parent Menu Name
            Program, Function or Command
            Security Level
            Prompt
            Routine number if required for type P above

    Draw the menus as follows:

        menu = TartanMenu(mf, usr, men, mod, ign, lvl, cmd, img)

        men = Menues
        mod = Modules
        lvl = Security Level
        cmd = Command to Execute on Selection
        img = Image to Display
    """
    def __init__(self, mf, usr, men, mod, lvl=None, cmd=None, img=None):
        self.mf = mf
        self.usr = usr
        self.men = men
        self.mod = mod
        self.lvl = lvl
        self.cmd = cmd
        self.img = img
        self.menmodCreate()

    def menmodCreate(self):
        # Create a list of menues and a dictionary of all the modules
        menues = []
        self.moddic = {}
        for m in self.mod:
            if self.lvl != None and m[3] > self.lvl:
                continue
            if m[1] not in menues:
                menues.append(m[1])
            if m[0][0] == "P":
                mm = "mm_%s" % m[2][:2]
                if mm not in menues:
                    menues.append(mm)
                if len(m) == 6:
                    mod = "%s%s" % (m[2][:-1], m[5])
                    self.moddic[mod] = [m[0], m[4], m[5]]
                else:
                    mod = "%s%s" % (m[2][:-1], 0)
                    self.moddic[mod] = [m[0], m[4], None]
            else:
                mod = m[2]
                self.moddic[mod] = [m[0], m[4], None]
        self.menues = []
        for men in self.men:
            if men[0] == "MM" and men[1] in menues:
                self.menues.append(men)
            elif men[2] in menues:
                self.menues.append(men)

    def drawMenu(self):
        text = "Tartan Systems - Main Menu (%s)" % self.usr
        self.mf.head.set_text(text)
        self.menu_bar = gtk.MenuBar()
        self.mf.body.pack_start(self.menu_bar, True, False, 0)
        self.labs = {}
        for m in self.menues:
            if m[0] == "MM":
                if "mbar" in self.labs:
                    pass
                else:
                    self.labs["mbar"] = []
                self.labs, labl = getUnderline(self.labs, "mbar", m[2])
                #===============================================================
                # Attempt to use stock labels and icons
                #===============================================================
                nam = gtk.stock_lookup("gtk-%s" % m[2].lower())
                if nam:
                    gtk.stock_add([(nam[0], labl, nam[2], nam[3], nam[4])])
                    menu = gtk.ImageMenuItem(stock_id=nam[0])
                else:
                    menu = gtk.MenuItem(label=labl)
                #===============================================================
                if len(m) == 4 and m[3] == "right":
                    menu.set_right_justified(1)
                exec "%s = gtk.Menu()" % m[1]
                exec "menu.set_submenu(%s)" % m[1]
                self.menu_bar.append(menu)
            elif m[0] == "CM":
                if m[1] in self.labs:
                    pass
                else:
                    self.labs[m[1]] = []
                self.labs, labl = getUnderline(self.labs, m[1], m[3])
                #===============================================================
                # Attempt to use stock labels and icons
                #===============================================================
                nam = gtk.stock_lookup("gtk-%s" % m[3].lower())
                if nam:
                    gtk.stock_add([(nam[0], labl, nam[2], nam[3], nam[4])])
                    cas = gtk.ImageMenuItem(stock_id=nam[0])
                else:
                    cas = gtk.MenuItem(label=labl)
                #===============================================================
                exec "%s.append(cas)" % m[1]
                exec "%s = gtk.Menu()" % m[2]
                exec "cas.set_submenu(%s)" % m[2]
        for m in self.mod:
            if m[3] > self.lvl:
                continue
            if m[1] in self.labs:
                pass
            else:
                self.labs[m[1]] = []
            self.labs, labl = getUnderline(self.labs, m[1], m[4])
            #===================================================================
            # Attempt to use stock labels and icons
            #===================================================================
            if m[4].lower() == "change user":
                nam = gtk.stock_lookup("gtk-users")
            elif m[4].lower() == "change password":
                nam = gtk.stock_lookup("gtk-users")
            elif m[4].lower() == "user's maintenance":
                nam = gtk.stock_lookup("gtk-users")
            elif m[4].lower() == "upgrade system":
                nam = gtk.stock_lookup("gtk-network")
            elif m[4].lower() == "update file formats":
                nam = gtk.stock_lookup("gtk-convert")
            elif m[4].lower() == "fix/alter table data":
                nam = gtk.stock_lookup("gtk-edit")
            elif m[4].lower() == "copy tables":
                nam = gtk.stock_lookup("gtk-copy")
            elif m[4].lower() == "backup database":
                nam = gtk.stock_lookup("gtk-backup")
            elif m[4].lower() == "restore database":
                nam = gtk.stock_lookup("gtk-restore")
            else:
                nam = gtk.stock_lookup("gtk-%s" % m[4].lower())
            if nam:
                gtk.stock_add([(nam[0], labl, nam[2], nam[3], nam[4])])
                item = gtk.ImageMenuItem(stock_id=nam[0])
            else:
                item = gtk.MenuItem(label=labl)
            #===================================================================
            try:
                exec "%s.append(item)" % m[1]
                if m[0][0] == "P":
                    if len(m) == 6:
                        mod = "%s%s" % (m[2][:-1], m[5])
                    else:
                        mod = "%s%s" % (m[2][:-1], 0)
                else:
                    mod = m[2]
                if self.cmd:
                    item.connect("activate", self.cmd, mod)
            except:
                continue
        self.menu_bar.show_all()
        iw, ih = self.mf.geom
        ih = int(ih * .90)
        if self.img and os.path.isfile(self.img):
            self.tar_image = gtk.Image()
            pixbuf = gtk.gdk.pixbuf_new_from_file(self.img)
            scaled_buf = pixbuf.scale_simple(iw, ih, gtk.gdk.INTERP_BILINEAR)
            self.tar_image.set_from_pixbuf(scaled_buf)
        else:
            self.tar_image = gtk.Frame()
            self.tar_image.set_size_request(iw, ih)
        self.mf.body.pack_start(self.tar_image)
        self.tar_image.show()
        self.mf.updateStatus("Select a Routine from the Menu")

    def checkNotes(self, user):
        t = time.localtime()
        self.cdate = ((t[0] * 10000) + (t[1] * 100) + t[2])
        self.mf.dbm.openDbase()
        self.sql = Sql(self.mf.dbm, ["ctlnot", "ctlpwu"], prog=__name__)
        notes = self.sql.getRec(tables="ctlnot", where=[("(", "not_user", "=",
            user, "or", "not_auser", "=", user, ")"), ("not_adate", "<=",
            self.cdate), ("not_aflag", "<>", "C")])
        if notes:
            ok = askQuestion(self.mf.body, "Notes", "There are Notes "\
                "Flagged for Action Still Outstanding, do You want to "\
                "View them?")
            if ok == "yes":
                self.showNotes(user)
        self.mf.dbm.closeDbase()

    def showNotes(self, user):
        notes = self.sql.getRec(tables="ctlnot", where=[("(", "not_user", "=",
            user, "or", "not_auser", "=", user, ")"), ("not_adate", "<=",
            self.cdate), ("not_aflag", "<>", "C")], order="not_sys, not_seq")
        if not notes:
            return
        data = []
        for note in notes:
            nte = note[:4]
            nte.append(note[6])
            nte.append(note[7])
            nte.append(note[8])
            nte.append(note[5])
            nte.append(note[9])
            data.append(nte)
        tits = "Today's Notes"
        cols = [
            ("not_cono", "Coy", 3, "UI", "N"),
            ("not_sys", "Sys", 3, "UA", "N"),
            ("not_key", "Key", 30, "NA", "N"),
            ("not_date", "Entry-Date", 10, "D1", "N"),
            ("not_aflag", "F", 1, "UA", "N"),
            ("not_adate", "Action-Dte", 10, "D1", "N"),
            ("not_auser", "Action-User", 20, "NA", "N"),
            ("not_desc", "Description", 50, "TX", "N"),
            ("not_seq", "Sequence", 10, "US", "N")]
        sr = SelectChoice(self.mf.body, tits, cols, data)
        if sr.selection:
            self.changeNotes(sr.selection[1:])
            self.showNotes(user)

    def changeNotes(self, note):
        self.nseq = note[-1:][0]
        tit = ("Notes Editing",)
        usr = {
            "stype": "R",
            "tables": ("ctlpwu",),
            "cols": (
                ("usr_name", "", 0, "User Name"),
                ("usr_fnam", "", 0, "Full Name", "Y"))}
        r1s = (("Normal","N"),("Urgent","U"),("Completed","C"))
        fld = (
            (("T",0,0,0),"O@not_cono",0,""),
            (("T",0,1,0),"O@not_sys",0,""),
            (("T",0,2,0),"O@not_key",0,""),
            (("T",0,3,0),"O@not_date",0,""),
            (("T",0,4,0),"OTv",50,"Details","Details",
                "","N",None,None,None,("notblank",)),
            (("T",0,5,0),("IRB",r1s),0,"Action Flag (C/N/U)","Action Flag",
                "N","N",self.doActFlag,None,None,None),
            (("T",0,6,0),"I@not_adate",0,"","",
                "","N",self.doActDate,None,None,("efld",)),
            (("T",0,7,0),"I@not_auser",0,"","",
                "","N",self.doActUser,usr,None,("efld",)))
        but = (
            ("Accept",None,self.doNEnd,1,None,None),
            ("Cancel",None,self.doNExit,1,None, None))
        tnd = ((self.doNEnd, "n"), )
        txt = (self.doNExit, )
        self.nf = TartanDialog(self.mf, newscn=(0,1), tops=True, title=tit,
            eflds=fld, butt=but, tend=tnd, txit=txt, focus=False)
        for x in range(5):
            if x < 4:
                self.nf.loadEntry("T", 0, x, data=note[x])
            else:
                self.nf.loadEntry("T", 0, x, data=note[x+3])
                self.nf.topEntry[0][x].set_wrap_mode(gtk.WRAP_WORD)
                self.actflag = note[x]
                self.nf.loadEntry("T", 0, x+1, data=self.actflag)
                self.actdate = note[x+1]
                self.nf.loadEntry("T", 0, x+2, data=self.actdate)
                self.actuser = note[x+2]
                self.nf.loadEntry("T", 0, x+3, data=self.actuser)
        self.nf.focusField("T", 0, 5)
        self.mf.startLoop()

    def doActFlag(self, frt, pag, r, c, p, i, w):
        self.actflag = w
        if self.actflag == "C":
            return "nd"

    def doActDate(self, frt, pag, r, c, p, i, w):
        self.actdate = w

    def doActUser(self, frt, pag, r, c, p, i, w):
        usr = self.sql.getRec(tables="ctlpwu", where=[("usr_name", "=", w)],
            limit=1)
        if not usr:
            return "Invalid User Name"
        self.actuser = w

    def doNEnd(self, widget=None):
        self.sql.updRec("ctlnot", cols=["not_aflag", "not_adate", "not_auser"],
            data=[self.actflag, self.actdate, self.actuser], where=[("not_seq",
            "=", self.nseq)])
        self.mf.dbm.commitDbase()
        self.doNExit()

    def doNExit(self, widget=None):
        self.nf.closeProcess()
        self.mf.closeLoop()

    def removeMenu(self):
        self.menu_bar.destroy()
        self.tar_image.destroy()

class TartanDialog:
    """
    Dialog for TARTAN Financial Systems

    Use this class to create a Dialog Screen by passing the folowing
    parameters to it:

    mf     =  The mainframe object

    newscn =  A tuple indicating that a new window must be created as follows:
                1st item: Decorate the window,
                    0=No, 1=Yes allowing destroy, 2=Yes preventing destroy
                2nd item: Place a frame in the window, 0=No or 1=Yes
                3rd item: Make the window modal, 0=No or 1=Yes

    screen =  The frame to place the widget on, defaults to mf.body

    tops   =  True to display a Top floating window, defaults to False

    title  =  The tuple of titles of the capture screen, None for None

    tags   =  A tuple of tuples of tags on the notebook pages, if applicable,
              having:
                name    = the tag name
                comd    = the command to execute when enabling else None
                on      = Tuple showing when to enable the tag (frt, pag, pos),
                          this can also be tuples of tuples,
                          for the end of a Frame use (frt, pag, 0)
                off     = Tuple showing when to disable the tag (frt, pag, pos),
                          this can also be tuples of tuples,
                          for the end of a Frame use (frt, pag, 0)

    eflds  =  A tuple of tuples of entry fields to place in a frame on the
              top of the normal rows and columns containing the following:
                place   = (where,page,row,col) to place the field as follows:
                            where   = T for top and C for Array of Columns
                            page    = 0   = On the Base Frame or
                                      1-9 = On a notebook page 1-9
                            row     = The row
                            col     = The column
                            ent     = The entry field start pos (Only top)
                type    = 3 fields as follows:
                            first   = (I)nput, (O)utput
                            second  = as per TartanClasses CCD class
                                      e.g. N,U,S,U,H for normal, upper,
                                      signed, unsigned or Hidden
                                                       or
                                      @ to use a column from the ffield table
                            third   = as per TartanClasses class e.g. A,I,L,D
                                      for alpha, integer, long or decimal
                                                       or
                                      the column name in the ffield table
                            Note: If a field from ffield is used then the type
                                  will be taken from the ffield data and the
                                  following fields will default to the ffield
                                  values:
                                      size
                                      text
                                      prompt
                          OR
                            ("[I,O]RB",xyz) for Radiobutton where:
                                I,O = (I)nput or (O)utput
                                RB  = Radiobutton indicator
                                xyz = the list or tuple of options
                          OR
                            [I,O][TV,Tv] for TextViews where:
                                I,O = (I)nput or (O)utput
                                TV= Textview F9 terminating
                                Tv= Textview Return terminating
                size    = The size of the entry field in characters as follows:
                            Either a tuple of (display size, entry size) or
                            a single value denoting both sizes. The sizes
                            are either an integer or float where the quotient
                            is the total display size e.g. 13.2 means the
                            display size, including the sign, would be 13 chrs.
                            i.e. 999999999.99-
                text    = Heading or Label
                prompt  = The entry field prompt. Append '(noesc)' to ignore
                          the Escape key i.e. No Backup Field or Exit Frame.
                default = Only if 'where' above = C:
                            r to repeat the previous line's column
                            i to increment the previous line's Integer
                            p to increment the previous line's Period
                          Else if 'where' above = T:
                            Normal value, '' for no default
                          Else if type above = RNV:
                            The String or Integer
                clear   = Clear the field when escaping to back up Y/N
                cr      = The function to perform for the <CR> key else None
                          It must return one of the following (default = None):
                            'bu'          = Back up to the previous field
                            'cl'          = Clear line and focus on 1st column
                            'ff[1-99]'    = Force focus to fieldnumber [1-99]
                            'nf'          = Do not auto focus on next field
                            'nd'          = End the frame normally
                            'nc'          = End the frame with no confirm
                            'ok' or None  = All fine
                            'rf'          = Re-focus on same field
                            'rt'          = Return True
                            'sk[1-99]'    = Skip the next [1-99] fields
                            'xt'          = Exit the frame
                            'text'        = An error message
                f1      = The record selection function to perform as:
                            a_name = {
                                 'stype':  "R",
                                 'tables': ('genbal',),
                                 'cols':   name, type size heading e.g.
                                           (('glo_acno', '', 0, 'Acc-Num'),
                                            ('glo_trdt', '', 0, 'Date'),
                                            ('glo_cyr', '', 0, 'Balance')),
                                           if type is a tuple ("XX", gltrtp)
                                           then display XXX type from list.
                                 'extra':  A tuple of tuples of additional
                                           data e,g. ((0, "All Companies"),)
                                 'where':  'glm_cono=%s' % self.conum,
                                 'whera':  [['T', 'glo_acno', 0, 0]],
                                              (the type of frame T or C,
                                               the field name,
                                               the column number,
                                               the column page)
                                 'autoc':  True/False to override default
                                 'group':  The fields to group on
                                 'order':  The fields to order on
                                 'wtype':  The where is a select statement
                                 'index':  The index of the item to return
                                 'comnd':  The command to execute after
                                           the select statement exits
                                 'sort':   If the columns can be sorted,
                                 'size':   The default screen size,
                                 'butt':   Additional Buttons and functions}
                     OR = The choice selection function to perform as:
                            a_name = {
                                 'stype':  "C",
                                 'titl':   "Select the Required Type",
                                 'head':   Column Headings, None for None
                                 'data':   (('aaaa','aaa'),('bbbb','bbb')),
                                 'retn':   "I"=index, "D"=data or None
                                 'index':  The index of the item to return,
                                 'butt':   Additional Buttons and functions}
                     OR = The file/directory selection dialog to perform as:
                            a_name = {
                                 'stype':  "F",
                                 'types':  "fle or dir",
                                 'initd':  The base directory,
                                 'ftype':  (('Zip Files', '*.zip'),),
                                 'multi':  True/False for multiples}
                            Note: ftype is a tuple of tuples for file selection
                                  and a string for directory selection
                     OR = The calendar date selection:
                            a_name = {
                                 'stype':  "D"}
                     OR = A function in the root module to perform:
                            a_name = {
                                 'stype':  "M",
                                 'func':   the function to execute}
                f5      = The function to perform to delete a record
                            returns a tuple indicating the next focus position
                            i.e. (frt, pag, col) or None for the default next
                            focus position
                valid   = A tuple of tuples of type and data for validation e.g.
                            ('efld',)
                            ('notblank',)
                            ('notzero',)
                            ('between', (5,2))
                            ('in', ("Y",'N'))
                            ('dir',)
                            ('file',)
                            ('path',)
                label   = A list only where place[0] above = C, this will
                            be a label on the left of the first column. The
                            list contains the following: (heading, rows) else
                            None
                tooltip = A string to display as a tooltip.

    rows   =  A tuple giving the number of rows per page e.g. (5,5) (max = 15)

    butt   =  A tuple of tuples of buttons to create at the bottom of the frame
              as follows:
                titl = The buttons title
                srec = The record selection function as for cols
                comm = The command to execute
                       Note: Only one of srec or comm can be used not both
                norm = The normal state of the button 0=off, 1=on, 2=ignore
                on   = Tuple showing when to enable button (frt, pag, pos),
                       this can also be tuples of tuples,
                       for the end of a Frame use (frt, pag, 0)
                off  = Tuple showing when to disable button (frt, pag, pos),
                       this can also be tuples of tuples,
                       for the end of a Frame use (frt, pag, 0)
                tip  = The tooltip
                row  = A row to place the button on.

    tend   =  A list of lists of functions to execute on finishing top frames
                by page having: (function, cnf)

    txit   =  A list of functions to execute on exiting top frames by page

    cend   =  A list of lists of functions to execute on finishing a data line
                by page having: (function, cnf)

    cxit   =  A list of functions to execute on exiting a data frame by page

    view   =  An entry field asking whether you would like to view the report
              It is a tuple (csv, "D") where:
                csv is whether to have an export facility
                "D" is the default e.g. ("View", "Print" or "Xport")
              A list self.repprt[0] is created having 3 fields as follows:
                1) "Y" or "N" view widget created
                2) query widget contents
                3) printer widget contents

    mail   =  An entry field asking whether you would like to email the report
              It is a tuple (flag, prnt, desc) where:
                flag is one of the following:
                   "Y" = Address Field (noblank)
                   "B" = Address Field (efld)
                   "N" = No Address Field (default N)
                   Lower case (y,b,n) indicates a Y default else N
                prnt is "Y" or "N" to choose to view/print emailed docs or not
              A list self.repprt[1] is created having 3 fields as follows:
                1) "Y" or "N" mail widget created
                2) query widget contents
                3) address widget contents
                4) view/print emailed document
    save   =  An entry field asking whether you would like to archive the report
              It is a character X where X is the default Y or N
              A list self.repprt[2] is created having 2 fields as follows:
                1) "Y" or "N" view widget created
                2) query widget contents

    focus  =  Whether to automatically focus on the first screen's first column.
                 The default is True
    eframe =  True or False if Entry widget must be framed, Defaults to False
    vport  =  Create a label in a viewport to display the message
    clicks =  A function to execute when an Entry Label field label is clicked

    Notes:  The 'self.pos' and 'pos' variable is the colEntry field number (0-?)
            The 'self.idx' and 'idx' variable is the pos in the row (0-?)
            The 'self.col' and 'col' variable is the column number (1-?)
            The mail and view options only apply to the ('T',0) frame

    In the function executed by the widget you can do the following:
            Verify the data
            Update tables
            Quit the program
            Continue by returning a value as per above
    """
    def __init__(self, mf, **args):
        self.mf = mf
        options = (
            "newscn", "screen", "tops", "title", "tags", "eflds", "rows",
            "butt", "tend", "txit", "cend", "cxit", "view", "mail", "save",
            "focus", "eframe", "vport", "clicks")
        # Set defaults
        for option in options:
            if option == "focus":
                self.focus = True
            elif option == "eframe":
                if mf.rcdic["efrm"] == "Y":
                    self.eframe = True
                else:
                    self.eframe = False
            elif option == "vport":
                self.vport = False
            elif option == "clicks":
                self.clicks = False
            else:
                exec "self.%s = None" % option
        # Set passed options
        for arg in args:
            if arg == "butt" and args["butt"]:
                self.butt = []
                for but in args["butt"]:
                    self.butt.append(list(but))
            else:
                exec "self.%s = args[arg]" % arg
        if self.screen and self.newscn:
            showError(self.screen, "Screen Error",
                "You cannot have a 'screen' and 'newscn' together, only One!")
            sys.exit()
            return
        if self.newscn:
            """
            This is a tuple having (0,0,0) where:
                First digit:
                    0 = No decoration
                    1 = Decorated with the destroy button
                    2 = Decorated without the destroy button
                Second digit:
                    0 = No Frame
                    1 = With a Frame
                Third digit:
                    0 = Not Modal
                    1 = Modal
                Fourth digit:
                    0 = Not Transient
                    1 = Transient
            """
            if len(self.newscn) == 2:
                self.newscn = list(self.newscn)
                self.newscn.append(1)
                self.newscn.append(1)
            elif len(self.newscn) == 3:
                self.newscn = list(self.newscn)
                self.newscn.append(1)
            if self.newscn[0] in (1, 2):
                decor = True
                if self.newscn[0] == 1:
                    remov = False
                else:
                    remov = True
            else:
                decor = False
                remov = False
            if self.newscn[1] == 1:
                frame = True
            else:
                frame = None
            if self.newscn[2] == 1:
                modal = True            # Default
            else:
                modal = False
            if self.newscn[3] == 1:
                trans = self.mf.window  # Default
            else:
                trans = False
            self.mkw = MkWindow(trans=trans, decor=decor, frame=frame,
                modal=modal, remov=remov)
            self.window = self.mkw.newwin
            self.screen = gtk.VBox(False, 0)
            if frame:
                self.mkw.newfrm.add(self.screen)
            else:
                self.window.add(self.screen)
        else:
            self.window = self.mf.window
            if not self.screen:
                self.screen = self.mf.body
        # Set all variables
        self.setVariables()
        # Draw all widgets
        self.drawScreens()
        # Wait for all widgets to be dispatched
        mf.gtkDispatch()
        # Focus to the first field
        if self.focus and self.first:
            self.focusField(self.first[0], self.first[1], 1)
        # Set Notebook Page Switching
        if self.nbf:
            self.nb.connect_after("switch-page", self.swiPage)

    def doKeys(self, ent, key):
        """
        ent = The widget
        key = The key pressed
        """
        self.mf.updateStatus("")
        if self.frt == "T":
            fld = self.topf[self.pag][self.idx]
        else:
            fld = self.colf[self.pag][self.idx]
        if type(fld[1]) in (list, tuple):
            typ = fld[1][0][0]
        else:
            typ = fld[1][0]
        if typ == "O":
            # Output field only
            return True
        if key == "F1" and fld[8]:
            self.selectData()
        elif key == "F5" and fld[9]:
            self.deleteRecord(ent)
        elif key == "F9" and fld[1][1:] == "TV":
            self.execCommand(self.frt, self.pag, fld[7])
        elif key in ("KP_Enter", "Return") and fld[1][1:] == "Tv":
            self.execCommand(self.frt, self.pag, fld[7])
        elif key == "Escape" and self.esc:
            if self.idx == 0:
                self.doExitFrame(self.frt, self.pag)
            else:
                self.backUp(self.frt, self.pag)
        else:
            return
        return True

    def setVariables(self):
        if "acnf" in self.mf.rcdic:
            self.acnf = self.mf.rcdic["acnf"].lower()
        else:
            self.acnf = "y"
        if "ttip" in self.mf.rcdic:
            self.ttip = self.mf.rcdic["ttip"].lower()
        else:
            self.ttip = "y"
        #=======================================================================
        # New amendments to display extra characters (outside widget width)
        #=======================================================================
        if "txtv" in self.mf.rcdic:
            self.txtv = self.mf.rcdic["txtv"].lower()
        else:
            self.txtv = "y"
        self.txtidx = None
        self.txtvwr = None
        #=======================================================================
        # Create the Tooltips widget
        #=======================================================================
        if PYGTK_VERSION[1] < 12 and self.ttip == "y":
            self.ttips = gtk.Tooltips()
        #=======================================================================
        self.tagbut = {"chrs": []}  # the dictionary of underline characters
        if self.tags:
            for x in self.tags:
                i = x[0].find("_")
                if i != -1:
                    self.tagbut["chrs"].append(x[0][i+1])
        if self.butt:
            for x in self.butt:
                i = x[0].find("_")
                if i != -1:
                    self.tagbut["chrs"].append(x[0][i+1])
        self.first = []             # the first field details
                                    # i.e. (frt, page, row, col)
        self.last = [[0,0]]         # last position visited by frt/page
        self.skip = [[]]            # columns skipped by frame 'sk'
        self.topq = [0]             # top number of columns by page
        self.colq = [0]             # col number of columns by page
        self.topf = [[]]            # eflds tops data by page
        self.colf = [[]]            # eflds cols data by page
        self.topEntry = [{}]        # dic of top entry fields by page/col
        self.topCompl = [{}]        # dic of top entry compl widget by page/col
        self.topListl = [{}]        # dic of top entry compl list by page/col
        self.colLabel = [{}]        # dic of col label fields by page/row
        self.colEntry = [{}]        # dic of col entry fields by page/col
        self.colCompl = [{}]        # dic of col entry compl widget by page/col
        self.colListl = [{}]        # dic of col entry compl list by page/col
        self.t_work = []            # list of arrays of top work by page/0/idx
        self.t_disp = []            # list of arrays of top disp by page/0/idx
        self.c_work = []            # list of arrays of col work by page/row/idx
        self.c_disp = []            # list of arrays of col disp by page/row/idx
        self.tsiz = [[0,0,0]]       # largest label, entry field and also the
                                    # maximum width by top page
        self.pag = None             # Clearing for doKeys() problem
        self.pgs = 0                # Last page number
        self.topz = False           # Whether or not there is a top page 0
        self.colz = False           # Whether or not there is a col page 0
        self.topv = [[]]            # top validators by page/col
        self.colv = [[]]            # col validators by page/col
        self.rpad = 3               # Default padding between Radiobuttons
        self.has_focus = []         # List of widgets which can focus
        #=======================================================================
        # Checking and Adding Ouput and Mail Widget Requirements
        #=======================================================================
        if not self.view:
            self.repprt = [["N", "N", ""]]
        else:
            self.repprt = [["Y", "", ""]]
        smtp = None
        if self.mf.dbobj and self.mf.dbm.dbopen:
            try:
                gc = GetCtl(self.mf)
                ctlsys = gc.getCtl("ctlsys", error=False)
                if ctlsys:
                    smtp = ctlsys["sys_smtp"]
            except:
                pass
        if not self.mail or not smtp:
            self.repprt.append(["N", "N", "", "Y"])
        else:
            self.repprt.append(["Y", "", "", "Y"])
        if not self.save:
            self.repprt.append(["N", "N"])
        else:
            self.repprt.append(["Y", ""])
        if self.view or self.mail or self.save:
            if self.eflds and tuple(self.eflds[0][0][:2]) != ("T", 0):
                showError(self.screen, "View or Mail Option",
                    "These Options Require a (T)op Zero Page")
                sys.exit()
            row = 0
            col = 0
            for fld in self.eflds:
                if fld[0][0] == "T" and fld[0][1] == 0:
                    row = fld[0][2]
                    col = fld[0][3]
                else:
                    break
            self.eflds = list(self.eflds)
            if self.repprt[0][0] == "Y":
                printers = showPrinters()
                prt = {
                    "stype": "C",
                    "titl": "Available Printers",
                    "head": ("Name","Description"),
                    "data": printers}
                row += 1
                rvs = [("View","V"),("Print","P")]
                mes = """Select what to do with the generated report.
                    
View   - The generated report will be opened in the default
         pdf viewer.

Print  - The report will be printed on the selected printer."""
                if self.view[0].lower() == "y":
                    rvs.append(("Export CSV","C"))
                    if XLWT:
                        rvs.append(("Export XLS","X"))
                    mes = """%s

Export - The report in the selected format will be opened
         in the default spreadsheet application.""" % mes
                self.eflds.append((("T",0,row,col),("IRB",rvs),0,"Output",
                    "",self.view[1].upper(),"Y",self.setView,None,None,
                    None,None,mes))
                row += 1
                self.eflds.append((("T",0,row,col),"INA",(30,50),"Printer Name",
                    "","Default","Y",self.setPrtr,prt,None,("in", printers)))
            if self.repprt[1][0] == "Y":
                row += 1
                if self.mail[0] in ("b", "n", "y"):
                    deflt = "Y"
                else:
                    deflt = "N"
                yns = (("Yes","Y"),("No","N"))
                if len(self.mail) == 3:
                    desc = self.mail[2]
                else:
                    desc = "E-Mail Report"
                self.eflds.append((("T",0,row,col),("IRB",yns),0,
                    desc,"",deflt,"Y",self.setMail,None,None,None))
                if self.mail[0].lower() in ("b", "y"):
                    sql = Sql(self.mf.dbm, ["telmst", "telcon"], prog=__name__)
                    data = []
                    tdm = sql.getRec(tables="telmst", cols=["tdm_name",
                        "tdm_email"], where=[("tdm_email", "<>", "")],
                        order="tdm_name")
                    for t in tdm:
                        data.append([t[0], "", "", t[1]])
                    tdc = sql.getRec(tables="telcon", cols=["tdc_name",
                        "tdc_contact", "tdc_desig", "tdc_email"],
                        where=[("tdc_email", "<>", "")], order="tdc_name, "\
                        "tdc_contact")
                    if tdc:
                        data.extend(tdc)
                    adr = {
                        "stype": "C",
                        "titl": "Available Addresses",
                        "head": ("Name","Contact","Designation","Address"),
                        "typs": (("NA",0,"Y"),("NA",0),("NA",0),("NA",0)),
                        "data": data,
                        "index": 3}
                    row += 1
                    a = [("T",0,row,col),"ITX",50,"E-Mail Address","","",
                        "Y",self.setAddr,adr,None]
                    if self.mail[0] == "Y":
                        a.append(("email", False))
                    else:
                        a.append(("email", True))
                    a.extend([None,"E-Mail Addresses, Comma Separated"])
                    self.eflds.append(tuple(a))
                if self.mail[1].lower() == "y":
                    row += 1
                    if self.mail[1] == "y":
                        deflt = "Y"
                    else:
                        deflt = "N"
                    self.eflds.append((("T",0,row,col),("IRB",yns),0,
                        "View/Print Emailed Document","",deflt,"Y",
                        self.setPrint,None,None,None))
            if self.repprt[2][0] == "Y":
                row += 1
                yns = (("Yes","Y"),("No","N"))
                self.eflds.append((("T",0,row,col),("IRB",yns),0,
                    "Archive Report","",self.save.upper(),"Y",
                    self.setArchive,None,None,None))
        #=======================================================================
        frt = ""
        pag = 999
        tot = 0
        num = 1
        ctxt = self.screen.get_pango_context()
        layt = pango.Layout(ctxt)
        layt.set_text("X")
        self.wdth, self.hght = layt.get_pixel_size()
        maxcol = int(self.mf.geom[0] / self.wdth) - 5
        for col in self.eflds:
            col = list(col)
            if col[1][1] == "@":
                sql = Sql(self.mf.dbm, "ffield", prog=__name__)
                seq = sql.ffield_col
                ffld = sql.getRec(tables="ffield",
                    where=[("ff_name", "=", col[1][2:])], limit=1)
                if not ffld:
                    showError(self.screen, "Field Error",
                        "The Column Name %s Is Invalid" % col[1][2:])
                    sys.exit()
                typ = ffld[seq.index("ff_type")]
                if typ == "TX":
                    typ = "Tv"
                col[1] = "%s%s" % (col[1][0], typ)
                if not col[2]:
                    col[2] = float(ffld[seq.index("ff_size")])
                if col[0][0] == "T":
                    if not col[3]:
                        col[3] = ffld[seq.index("ff_desc")]
                    if len(col) == 4:
                        pass
                    elif not col[4]:
                        col[4] = ffld[seq.index("ff_desc")]
                else:
                    if not col[3]:
                        col[3] = ffld[seq.index("ff_head")]
                    if len(col) == 4:
                        pass
                    elif not col[4]:
                        col[4] = ffld[seq.index("ff_desc")]
            elif len(col) > 4 and not col[4]:
                col[4] = col[3]
            if type(col[2]) in (int, float):
                if col[1][1:] in ("TV", "Tv"):
                    col[2] = (col[2], -1)
                else:
                    col[2] = (col[2], col[2])
            elif col[1][1:] in ("TV", "Tv"):
                col[2] = (col[2][0], self.hght*col[2][1])
            if frt != col[0][0] or pag != col[0][1]:
                frt = col[0][0]
                pag = col[0][1]
                tot = 0
                row = 999
            if frt == "C":
                tot = tot + int(col[2][0])
            elif col[1][1:] != "RB":
                if row != col[0][2]:
                    row = col[0][2]
                    tot = 0
                if len(col[0]) == 5:
                    tot = int(col[0][4]) + len(col[3]) + int(col[2][0])
                elif col[0][3] == 0:
                    tot = tot + len(col[3]) + int(col[2][0])
                else:
                    tot = int(col[0][3]) + len(col[3]) + int(col[2][0])
            #===================================================================
            # Check to see if Columns exceed the maximum allowed i.e. maxcol
            # which is based on the screen and font sizes.
            #===================================================================
            if tot > maxcol:
                showWarning(screen=self.screen,
                    head="Maximum Exceeded", mess="This Page (%s %s) "\
                    "Exceeds %s Cols, (Using %s Cols), Either:\n\n"\
                    "a) Decrease Your Font Family and Size (%s) or\n"\
                    "b) Increase Your Screen Geometry (%s)." % (frt, pag,
                    maxcol, tot, self.mf.fontd, self.mf.rcdic["geo"]))
            #===================================================================
            if not self.first:
                self.first = col[0]
            if col[0][1] < self.pgs:
                showError(self.screen, "Page Number",
                    "The Page Number is Out of Sequence")
                sys.exit()
            elif col[1][0] == "R":
                showError(self.screen, "Old Radio Button %s" % str(col[0]),
                    "Radio Buttons MUST Have a Tuple of Choices")
                sys.exit()
            elif type(col[1]) in (list, tuple) and len(col) > 9 and col[10]:
                showError(self.screen, "Radio Button %s" % str(col[0]),
                    "Radio Buttons CAN NOT Have a Validation")
                sys.exit()
            elif len(col) not in (4, 11, 12, 13):
                showError(self.screen, "Field Details",
                    "This Field Has Invalid Details %s" % str(col[0]))
                sys.exit()
            elif col[0][1] > self.pgs:
                while col[0][1] > self.pgs:            # Added for Blank Page
                    num = 1
                    chr = 0
                    self.pgs += 1                      # Changed for Blank Page
                    self.last.append([0,0])
                    self.skip.append([])
                    self.topq.append(0)
                    self.topf.append([])
                    self.topEntry.append({})
                    self.topCompl.append({})
                    self.topListl.append({})
                    self.colq.append(0)
                    self.colf.append([])
                    self.colLabel.append({})
                    self.colEntry.append({})
                    self.colCompl.append({})
                    self.colListl.append({})
                    self.tsiz.append([0,0,0])
                    self.topv.append([])
                    self.colv.append([])
            if len(col) > 4 and not col[8] and col[1][1:] in ("D1", "d1"):
                col[8] = {"stype": "D"}
            if col[0][0] == "T":
                if col[0][1] == 0:
                    self.topz = True
                if col[0][3] == 0:
                    if len(col[0]) == 5:
                        t = col[0][4]
                    else:
                        t = len(col[3])
                    if type(col[1]) in (list, tuple):
                        d = 0
                        for x in range(0, len(col[1][1])):
                            d = d + self.rpad + len(col[1][1][x][0])
                    else:
                        d = int(col[2][0] / 1)
                    if t > self.tsiz[self.pgs][0]:
                        self.tsiz[self.pgs][0] = t
                    if d > self.tsiz[self.pgs][1]:
                        self.tsiz[self.pgs][1] = d
                try:
                    self.topq[self.pgs] += 1
                except:
                    self.topq[self.pgs] = 1
                self.topf[self.pgs].append(col)
                if len(col) > 10:
                    self.topv[self.pgs].append(col[10])
                else:
                    self.topv[self.pgs].append(None)
            elif col[0][0] == "C":
                if col[0][1] == 0:
                    self.colz = True
                try:
                    self.colq[self.pgs] += 1
                except:
                    self.colq[self.pgs] = 1
                self.colf[self.pgs].append(col)
                if len(col) > 10:
                    self.colv[self.pgs].append(col[10])
                else:
                    self.colv[self.pgs].append(None)
                num += 1
        if self.colz and self.pgs:
            showError(self.screen, "Col Frame",
                "Cannot Have a Zero Column Page With Notebook")
            sys.exit()
        if self.pgs > 0 and self.pgs > len(self.tags):
            showError(self.screen, "Notebook",
                "Not All Notebook Tags Have Been Set")
            sys.exit()
        #=======================================================================
        # Set Number of Rows, Defaulting to 15
        #=======================================================================
        if not self.rows:
            self.rows = []
            for r in range(0, self.pgs+1):
                self.rows.append(15)
        else:
            self.rows = list(self.rows)
        for p in self.colf:
            for c in p:
                if len(c) > 11 and c[11]:
                    self.rows[c[0][1]] = c[11][1]
        #=======================================================================
        # Create t_work, t_disp, c_work and c_disp arrays for all pages
        #=======================================================================
        for x in range(0, self.pgs+1):
            q = self.topq[x]
            a = makeArray(q,1,1)
            self.t_work.append(a)
            a = makeArray(q,1,1,"S")
            self.t_disp.append(a)
            q = self.colq[x]
            r = self.rows[x]
            a = makeArray(q,r,1)
            self.c_work.append(a)
            a = makeArray(q,r,1,"S")
            self.c_disp.append(a)
        #=======================================================================
        # Calculate width of the widest row on the page
        #=======================================================================
        for p in range(0, self.pgs+1):
            row = None
            wid = 0
            for x in self.topf[p]:
                if x[0][3] == 0:
                    txt = self.tsiz[p][0]
                else:
                    if wid < x[0][3]:
                        wid = x[0][3]
                    if len(x[0]) == 5:
                        txt = x[0][4] - x[0][3]
                    else:
                        txt = len(x[3])
                if type(x[1]) in (list, tuple):
                    siz = 0
                    for y in range(0, len(x[1][1])):
                        siz = siz + self.rpad + len(x[1][1][y][0])
                else:
                    siz = int(x[2][0] / 1)
                if row == x[0][2]:
                    wid = wid + txt + siz
                else:
                    wid = txt + siz
                if wid > self.tsiz[p][2]:
                    self.tsiz[p][2] = wid*1
                row = x[0][2]
        #=======================================================================
        if self.butt != None:
            self.bsiz = 0
            self.brow = 0
            for but in self.butt:
                if but[1] and but[2]:
                    showError(self.screen, "Buttons",
                        "Buttons Cannot Have Both 'srec' and 'comm' Commands")
                    sys.exit()
                if self.bsiz < len(but[0]):
                    self.bsiz = len(but[0])

    def setView(self, frt, pag, r, c, p, i, w):
        self.repprt[0][1] = w
        if w == "V":
            self.repprt[0][2] = "view"
            self.loadEntry(frt, pag, p+1, data="")
            return "sk1"
        if w in ("C", "X"):
            self.repprt[0][2] = "export"
            self.loadEntry(frt, pag, p+1, data="")
            if self.repprt[1][0] == "N":
                return "sk1"
            elif self.repprt[1][0] == "Y":
                self.repprt[1][1] = "N"
                self.loadEntry(frt, pag, p+2, data="N")
                if self.mail[0].lower() in ("b", "y"):
                    self.repprt[1][2] = ""
                    self.loadEntry(frt, pag, p+3, data="")
                    if self.repprt[2][0] == "Y":
                        return "sk4"
                    else:
                        return "sk3"
                elif self.repprt[2][0] == "Y":
                    return "sk3"
                else:
                    return "sk2"

    def setPrtr(self, frt, pag, r, c, p, i, w):
        self.repprt[0][2] = w

    def setMail(self, frt, pag, r, c, p, i, w):
        self.repprt[1][1] = w
        if w == "N":
            self.repprt[1][2] = ""
            self.repprt[1][3] = ""
            skp = 0
            if self.mail[0].lower() in ("b", "y"):
                skp += 1
            if self.mail[1].lower() == "y":
                skp += 1
            for x in range(1, skp + 1):
                self.loadEntry(frt, pag, p+x, data="")
            return "sk%s" % skp

    def setAddr(self, frt, pag, r, c, p, i, w):
        self.repprt[1][2] = w

    def setPrint(self, frt, pag, r, c, p, i, w):
        self.repprt[1][3] = w

    def setArchive(self, frt, pag, r, c, p, i, w):
        self.repprt[2][1] = w

    def drawScreens(self):
        self.mstAlw = gtk.Alignment(0.5, 0.5, 0, 0)
        self.mstFrame = gtk.VBox(False, 0)
        self.mstAlw.add(self.mstFrame)
        self.screen.pack_start(self.mstAlw, True, True, 0)
        if self.tops:
            # Display the Title
            if self.title:
                if type(self.title) == str:
                    text = "%s" % self.title
                elif len(self.title) == 1:
                    text = "%s" % self.title[0]
                else:
                    text = "%-s - %s" % (self.title[0], self.title[1])
            else:
                text = self.mf.head.get_text()
            if not self.mf.fixed:
                self.mf.head.hide()
            evt, lab = doLabel(gtk, self.window, text=text,
                font=pango.FontDescription(self.mf.fontm))
            self.mstFrame.pack_start(evt, False, False, 0)
        self.nbf = False
        self.frt = ""
        for pag in range((self.pgs + 1)):
            if pag and not self.nbf:
                self.drawNoteBook()
                self.nbf = True
            self.drawTopFrame(pag)
            self.drawColFrame(pag)
        if self.butt:
            self.drawButtons()
        if self.newscn or self.vport:
            self.window.show_all()
        else:
            self.mstAlw.show_all()

    def drawNoteBook(self):
        # Create a new notebook and place the position of the tabs
        self.nb = gtk.Notebook()
        self.nb.set_tab_pos(gtk.POS_TOP)
        self.nb.unset_flags(gtk.CAN_FOCUS)
        self.mstFrame.pack_start(self.nb, False, False, 0)
        #=======================================================================
        # Append pages to the notebook
        #=======================================================================
        for num, tag in enumerate(self.tags):
            # Create buton for notebook tag
            if tag[2] or tag[3]:
                nam = getUnderline(self.tagbut, "chrs", tag[0], sav=True)
                if nam:
                    self.tagbut, nam = nam
                else:
                    nam = tag[0]
            else:
                nam = tag[0]
            but = MyButton(nam)
            but.set_sensitive(False)
            but.unset_flags(gtk.CAN_FOCUS)
            but.set_relief(gtk.RELIEF_NONE)
            but.connect_after("clicked", self.chgPage, num)
            # Create alignment
            alw = gtk.Alignment(0.5, 0.5, 0, 0)
            alw.set_padding(0, 0, 5, 5)
            # Create vbox
            vbx = gtk.VBox(False, 0)
            alw.add(vbx)
            # Add alignment and vbox to notebook
            self.nb.append_page(alw, tab_label=but)
            # Rename widgets for future access
            exec "self.nb.Tag%s = but" % num
            exec "self.nb.Page%s = vbx" % (num + 1)

    def swiPage(self, *args):
        self.pageChg(args[2], self.tags[args[2]][1])

    def chgPage(self, wid, num):
        if num != self.nb.get_current_page():
            self.nb.set_current_page(num)

    def pageChg(self, num, cmd):
        self.pag = num + 1
        if cmd:
            cmd()
        elif self.last[self.pag][1]:
            m = self.colq[self.pag]
            l = self.last[self.pag][1]
            while l > m:
                l -= m
            p = (l - 1) % m
            if l == m and self.colf[self.pag][p][1][0] == "O":
                dir = "B"
            else:
                dir = "F"
            self.focusField("C", self.pag, self.last[self.pag][1], dir=dir)
        elif self.last[self.pag][0]:
            l = self.last[self.pag][0]
            m = self.topq[self.pag]
            p = l - 1
            if l == m and self.topf[self.pag][p][1][0] in ("O", "ORB"):
                dir = "B"
            else:
                dir = "F"
            self.focusField("T", self.pag, self.last[self.pag][0], dir=dir)
        elif self.topq[self.pag]:
            self.focusField("T", self.pag, 1)
        else:
            self.focusField("C", self.pag, 1)

    def selPage(self, label):
        # Select notebook page by label text
        found = False
        for num, tag in enumerate(self.tags):
            if tag[0] == label:
                found = True
                break
        if not found:
            print "Invalid Notebook Tag %s" % label
        else:
            exec "wid = self.nb.Tag%s" % num
            wid.emit("clicked")

    def drawTopFrame(self, pag):
        if self.topf[pag] == []:
            return
        frame = "self.topPage%s" % pag
        if pag:
            exec "mast = self.nb.Page%s" % pag
        else:
            mast = self.mstFrame
        frm = gtk.Frame()
        frm.set_shadow_type(gtk.SHADOW_ETCHED_OUT)
        mast.pack_start(frm, False, False, 2)
        alw = gtk.Alignment(0.5, 0.5, 0, 0)
        if self.vport:
            hbx = gtk.HBox(False, 0)
            hbx.pack_start(alw, False, False, 0)
            self.vport = doLabel(gtk, self.window, text="")
            hbx.pack_end(self.vport[0], False, False, 0)
            frm.add(hbx)
            self.setViewPort("I", 0)
        else:
            frm.add(alw)
        exec "%s = gtk.VBox(False, 0)" % frame
        exec "alw.add(%s)" % frame
        self.lst = None
        for num, fld in enumerate(self.topf[pag]):
            if num == (self.topq[pag] - 1):
                nxt = None
            else:
                nxt = self.topf[pag][num+1]
            self.drawTopFlds(num, fld, nxt)

    def setViewPort(self, typ, total):
        bg = "white"
        if (typ == "C" and total > 0) or (typ != "C" and total < 0):
            fg = "red"
        else:
            fg = "blue"
        self.vport[0].modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse(bg))
        self.vport[1].modify_fg(gtk.STATE_NORMAL, gtk.gdk.color_parse(fg))
        self.vport[1].set_markup("<b><span size='%s'>%s</span></b>" % \
            (int(self.mf.dfs) * 2000, CCD(total, "SD", 13.2).disp))

    def drawTopFlds(self, col, fld, nxt=None):
        pag = fld[0][1]
        row = fld[0][2]
        exec "frame = self.topPage%s" % pag
        if row != self.lst:
            self.tabl = gtk.Table(1, 1, False)
            self.tabl.set_border_width(0)
            frame.pack_start(self.tabl, False, False, 1)
            self.acc = 0
            self.lst = row
        if type(fld[1]) in (list, tuple):
            typ = fld[1][0]
        else:
            typ = fld[1]
        if typ[1:] == "RB":
            siz = self.tsiz[pag][1]
        elif typ[1:] in ("TV", "Tv"):
            siz = int(fld[2][0] / 1)
            hgt = fld[2][1]
        else:
            siz = int(fld[2][0] / 1)
            if typ[1:] in ("FF", "TX"):
                lim = 0
            else:
                lim = int(fld[2][1] / 1)
        dsc = fld[0][3]                         # the text starting column
        if not dsc:
            dsc = self.acc
        if len(fld[0]) == 5:
            dse = fld[0][4]                     # the entry starting column
        elif fld[0][3] == 0 and self.acc == 0:
            dse = self.tsiz[pag][0]
        else:
            dse = dsc + len(fld[3])
        if dse < dsc:
            showError(self.screen, "Attach Error",
                "Label '%s' Has Invalid attach paramaters (%s %s)" % \
                (fld[3], dsc, dse))
            sys.exit()
        if typ[0] == "O":                       # Output Field Only
            cr = f1 = f5 = None                 # Not Used with Output Field
        else:
            cr = fld[7]                         # Enter Key
            f1 = fld[8]                         # F1 Key
            f5 = fld[9]                         # F5 Key

        #=======================================================================
        ## Padding for autocompletion scrollbar
        #if f1 and "stype" in f1 and f1["stype"] in ("C", "R") and \
        #                                           self.acnf == "y":
        #    siz = siz + 3
        #=======================================================================

        if dsc != self.acc:
            lpad = dsc - self.acc
        else:
            lpad = 0
        if lpad and lpad > 0:
            txt = " " * lpad
        else:
            txt = ""
        if fld[3]:
            txt = txt + fld[3]
        if dse == dsc:
            dse = dsc + len(txt)
        rpad = dse - self.acc - len(txt)
        if rpad and rpad > 0:
            txt = txt + " " * rpad
        if len(txt) > 0:
            xpad = 0
            evt, lab = doLabel(gtk, self.window, text=txt, color=False,
                style=None, font=pango.FontDescription(self.mf.fontd))
            if self.clicks and typ[0] == "I":
                evt.connect("button_press_event", self.clicks, (pag, col))
            self.tabl.attach(evt, dsc, dse, 0, 1, 0, gtk.EXPAND|gtk.FILL, 5, 0)
        else:
            xpad = 1
        e_dic = self.topEntry[pag]    # Entry Fields
        c_dic = self.topCompl[pag]    # Completion Widget
        l_dic = self.topListl[pag]    # Completion List
        if typ[1:] == "RB":
            e_dic[col] = []
            hbox = gtk.HBox(False, 0)
            self.tabl.attach(hbox, dse, dse+siz, 0, 1, 0, 0, xpad, 0)
            for num, val in enumerate(fld[1][1]):
                b = MyCheckButton(label=val[0])
                b.connect("mychkbx", self.doKeys)
                s = b.connect_after("toggled", self.execRB, num)
                hbox.pack_start(b, False, False, 0)
                e_dic[col].append([b, s, (val[1], cr)])
                #===============================================================
                # Tooltips
                #===============================================================
                if self.ttip == "y" and len(fld) == 13 and fld[12]:
                    if PYGTK_VERSION[1] < 12:
                        self.ttips.set_tip(b, fld[12])
                    else:
                        b.set_tooltip_text(fld[12])
                #===============================================================
            self.doCheckButton(e_dic[col], col, "disable")
        elif typ[1:] in ("TV", "Tv"):
            if typ[1:] == "Tv":
                e_dic[col] = MyTextView_v()
            else:
                e_dic[col] = MyTextView_V()
            e_dic[col].set_wrap_mode(gtk.WRAP_WORD)
            e_dic[col].set_size_request(self.wdth*siz, hgt)
            if typ[0] == "O":
                e_dic[col].set_editable(False)
            if typ[1:] == "Tv":
                e_dic[col].connect("mytxtvv", self.doKeys)
            else:
                e_dic[col].connect("mytxtvV", self.doKeys)
            if typ[1:] == "TV":
                sw = gtk.ScrolledWindow()
                sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
                sw.add(e_dic[col])
            if self.eframe:
                efrm = gtk.Frame()
                efrm.set_shadow_type(gtk.SHADOW_IN)
                if typ[1:] == "TV":
                    efrm.add(sw)
                else:
                    efrm.add(e_dic[col])
                self.tabl.attach(efrm, dse, dse+siz, 0, 5, 0, 0, xpad, 0)
            elif typ[1:] == "TV":
                self.tabl.attach(sw, dse, dse+siz, 0, 5, 0, 0, xpad, 0)
            else:
                self.tabl.attach(e_dic[col], dse, dse+siz, 0, 5, 0, 0, xpad, 0)
            c_dic[col] = None
            l_dic[col] = None
            #===================================================================
            # Tooltips
            #===================================================================
            if self.ttip == "y" and len(fld) == 13 and fld[12]:
                if PYGTK_VERSION[1] < 12:
                    self.ttips.set_tip(e_dic[col], fld[12])
                else:
                    e_dic[col].set_tooltip_text(fld[12])
            #===================================================================
            self.doTextView("T", pag, col, e_dic[col], "disable")
        else:
            e_dic[col] = MyEntry(max=lim)
            e_dic[col].set_width_chars(siz)
            if not self.eframe:
                e_dic[col].set_has_frame(False)
            if typ[0] == "O":
                e_dic[col].set_editable(False)
            if typ[1] == "H":
                e_dic[col].set_visibility(False)
            e_dic[col].connect("activate", self.execEntry, ("T", pag, cr))
            e_dic[col].connect("myentry", self.doKeys)
            self.tabl.attach(e_dic[col], dse, dse+siz, 0, 1, 0, 0, xpad, 0)
            #===================================================================
            # Auto completion
            #===================================================================
            if self.acnf == "y" and f1 and self.checkAutoc(f1, typ):
                c_dic[col] = gtk.EntryCompletion()
                l_dic[col] = gtk.ListStore(str)
                c_dic[col].set_model(l_dic[col])
                e_dic[col].set_completion(c_dic[col])
                c_dic[col].set_text_column(0)
                c_dic[col].connect("match-selected", self.execAutoc,
                    ("T", pag, col))
            else:
                c_dic[col] = None
                l_dic[col] = None
            #===================================================================
            # Tooltips
            #===================================================================
            if self.ttip == "y" and len(fld) == 13 and fld[12]:
                if PYGTK_VERSION[1] < 12:
                    self.ttips.set_tip(e_dic[col], fld[12])
                else:
                    e_dic[col].set_tooltip_text(fld[12])
            #===================================================================
            self.doEntryField("T", pag, col, e_dic[col], None, "disable")
        self.acc = dse + siz

    def drawColFrame(self, pag):
        frame = "self.colPage%s" % pag
        if pag == 0:
            alw = gtk.Alignment(0.5, 0.5, 0, 0)
            mast = gtk.VBox(False, 0)
            alw.add(mast)
            self.mstFrame.pack_start(alw, False, False, 0)
        else:
            exec "mast = self.nb.Page%s" % pag
        if self.colf[pag] == []:
            return
        cols = self.colq[pag]
        if len(self.colf[pag][0]) > 11 and self.colf[pag][0][11]:
            self.dolab = True
            cols += 1
            rows = self.colf[pag][0][11][1]
        else:
            self.dolab = False
            rows = self.rows[pag]
        frm = gtk.Frame()
        frm.set_shadow_type(gtk.SHADOW_ETCHED_OUT)
        exec "%s = gtk.Table(rows+1, cols, False)" % frame
        exec "frm.add(%s)" % frame
        mast.pack_start(frm, False, False, 2)
        # Create left label heading
        if self.dolab:
            text=self.colf[pag][0][11][0]
            evt, lab = doLabel(gtk, self.window, text=text, style="out")
            exec "%s.attach(evt, 0, 1, 0, 1)" % frame
        # Create headings for each column
        for x, y in enumerate(self.colf[pag]):
            text = str(y[3])
            evt, lab = doLabel(gtk, self.window, text=text, style="out")
            exec "%s.attach(evt, x+1, x+2, 0, 1)" % frame
        for row in range(rows):
            for col, fld in enumerate(self.colf[pag]):
                self.drawColFlds(pag, row, col, fld)

    def drawColFlds(self, pag, row, col, fld):
        frame = "self.colPage%s" % pag
        typ = fld[1]
        siz = int(fld[2][0] / 1)
        lim = int(fld[2][1] / 1)
        txt = fld[3]
        seq = ((row * self.colq[pag]) + col)
        if typ[0] == "O":                   # Output Field Only
            cr = f1 = f5 = None             # Not Used with Output Field
        else:
            cr = fld[7]                     # Enter Key
            f1 = fld[8]                     # F1 Key
            f5 = fld[9]                     # F5 Key
        if self.dolab and col == 0:
            t_dic = self.colLabel[pag]
            t_dic[row] = doLabel(gtk, self.window, text=txt, color=False,
                style=None, font=pango.FontDescription(self.mf.fontd))[1]
            exec "%s.attach(t_dic[row], col, col+1, row+1, row+2, "\
                "gtk.EXPAND|gtk.FILL, gtk.EXPAND|gtk.FILL, 5, 0)" % frame
        e_dic = self.colEntry[pag]
        c_dic = self.colCompl[pag]
        l_dic = self.colListl[pag]
        if typ[1:] in ("TV", "Tv"):
            if typ[1:] == "Tv":
                e_dic[seq] = MyTextView_v()
            else:
                e_dic[seq] = MyTextView_V()
            e_dic[seq].set_wrap_mode(gtk.WRAP_WORD)
            e_dic[seq].set_size_request(self.wdth*siz, -1)
            if typ[1:] == "Tv":
                e_dic[seq].connect("mytxtvv", self.doKeys)
            else:
                e_dic[seq].connect("mytxtvV", self.doKeys)
            if self.eframe:
                efrm = gtk.Frame()
                efrm.set_shadow_type(gtk.SHADOW_IN)
                efrm.add(e_dic[seq])
                exec "%s.attach(efrm, col+1, col+2, row+1, row+2, "\
                    "gtk.EXPAND|gtk.FILL, gtk.EXPAND|gtk.FILL, 1, 1)" % frame
            else:
                exec "%s.attach(e_dic[seq], col+1, col+2, row+1, row+2, "\
                    "gtk.EXPAND|gtk.FILL, gtk.EXPAND|gtk.FILL, 1, 1)" % frame
            c_dic[seq] = None
            l_dic[seq] = None
            #===================================================================
            # Tooltips
            #===================================================================
            if self.ttip == "y" and len(fld) == 13 and fld[12]:
                if PYGTK_VERSION[1] < 12:
                    self.ttips.set_tip(e_dic[col], fld[12])
                else:
                    e_dic[col].set_tooltip_text(fld[12])
            #===================================================================
            self.doTextView("C", pag, col, e_dic[seq], "disable")
        else:
            e_dic[seq] = MyEntry(max=lim)
            e_dic[seq].set_width_chars(siz)
            if not self.eframe:
                e_dic[seq].set_has_frame(False)
            if typ[0] == "O":
                e_dic[seq].set_editable(False)
            if typ[1] == "H":
                e_dic[seq].set_visibility(False)
            e_dic[seq].connect("activate", self.execEntry, ("C", pag, cr))
            e_dic[seq].connect("myentry", self.doKeys)
            exec "%s.attach(e_dic[seq], col+1, col+2, row+1, row+2, "\
                "gtk.EXPAND|gtk.FILL, gtk.EXPAND|gtk.FILL, 1, 1)" % frame
            #===================================================================
            # Auto completion
            #===================================================================
            if self.acnf == "y" and f1 and self.checkAutoc(f1, typ):
                c_dic[seq] = gtk.EntryCompletion()
                l_dic[seq] = gtk.ListStore(str)
                c_dic[seq].set_model(l_dic[seq])
                e_dic[seq].set_completion(c_dic[seq])
                c_dic[seq].set_text_column(0)
                c_dic[seq].connect("match-selected", self.execAutoc,
                    ("C", pag, seq))
            else:
                c_dic[seq] = None
                l_dic[seq] = None
            #===================================================================
            # Tooltips
            #===================================================================
            if self.ttip == "y" and len(fld) == 13 and fld[12]:
                if PYGTK_VERSION[1] < 12:
                    self.ttips.set_tip(e_dic[seq], fld[12])
                else:
                    e_dic[seq].set_tooltip_text(fld[12])
            #===================================================================
            self.doEntryField("C", pag, col, e_dic[seq], None, "disable")

    def checkAutoc(self, f1, typ):
        # If 'autoc' field in f1 dictionary then that is the default else if
        # alphanumeric data field then True else False
        if "stype" in f1 and f1["stype"] in ("C", "R"):
            if "autoc" in f1:
                return f1["autoc"]
            if typ in ("INA", "INa", "IUA"):
                return True

    def chgFieldSize(self, frt, pag, pos, siz):
        if frt == "T":
            self.topf[pag][pos][2] = siz
            self.topEntry[pag][pos].set_max_length(siz)
        else:
            self.colf[pag][pos][2] = siz
            for r in range(self.rows[pag]):
                seq = ((r * self.colq[pag]) + pos)
                self.colEntry[pag][seq].set_max_length(siz)

    #===========================================================================
    # New amendments to display extra characters (outside widget width)
    #===========================================================================
    def doShowExtraText(self, widget):
        dat = widget.get_text()
        if len(dat.strip()) > widget.get_width_chars():
            if not self.txtvwr:
                self.txtvwr = gtk.TextView()
                self.txtvwr.set_editable(False)
                self.txtvwr.set_sensitive(False)
                self.txtvwr.set_wrap_mode(gtk.WRAP_WORD)
                self.txtvwr.set_size_request(widget.get_width_chars(), -1)
                #self.txtvwr.set_justification(gtk.JUSTIFY_FILL)
                exec "self.colPage%s.attach(self.txtvwr, self.idx+1, "\
                    "self.idx+2, self.row+2, self.row+3)" % self.pag
                self.txtvwr.show()
            buf = self.txtvwr.get_buffer()
            buf.set_text(dat)

    def doEraseExtraText(self):
        if self.txtidx:
            self.txtidx[0].disconnect(self.txtidx[1])
            self.txtidx = None
        if self.txtvwr:
            self.txtvwr.destroy()
            self.txtvwr = None
    #===========================================================================

    def drawButtons(self):
        if self.tops:
            pad = 0
        else:
            pad = 5
        butframe = gtk.VBox(False, 0)
        self.mstFrame.pack_start(butframe, False, False, pad)
        hmg = True
        for but in self.butt:
            if len(but) == 8:
                hmg = False
        if hmg:
            tot = len(self.butt)
            row = tot / 6
            if tot % 6:
                row += 1
            col = tot / row
            if col % row:
                col += 1
            tab = gtk.Table(row, col, hmg)
            butframe.pack_start(tab, False, False, 0)
        else:
            r = 0
        for num, but in enumerate(self.butt):
            txt = getUnderline(self.tagbut, "chrs", but[0], sav=True)[1]
            #===================================================================
            # Attempt to use stock labels and icons for the buttons
            #===================================================================
            nam = gtk.stock_lookup("gtk-%s" % but[0].lower())
            if nam:
                gtk.stock_add([(nam[0], txt, nam[2], nam[3], nam[4])])
                stk = nam[0]
            else:
                stk = None
            #===================================================================
            button = MyButton(label=txt, stock=stk)
            exec "self.B%s = button" % num
            if self.ttip == "y" and len(but) > 6 and but[6]:
                if PYGTK_VERSION[1] < 12:
                    exec "self.ttips.set_tip(self.B%s, '%s')" % (num, but[6])
                else:
                    exec "self.B%s.set_tooltip_text('%s')" % (num, but[6])
            exec "self.B%s.set_sensitive(False)" % num
            exec "self.B%s.unset_flags(gtk.CAN_FOCUS)" % num
            if but[1]:
                exec "self.B%s.connect('clicked',self.selectBut,num)" % num
            elif but[2]:
                exec "self.B%s.connect('clicked', but[2])" % num
            if hmg:
                r = num / col
                p = num - (r * col)
                exec "tab.attach(self.B%s, p, p+1, r, r+1)" % num
            else:
                if len(but) == 8:
                    r = but[7]
                try:
                    exec "h%s.pack_start(self.B%s, True, True, 0)" % (r, num)
                except:
                    exec "h%s = gtk.HBox(True, 0)" % r
                    exec "butframe.pack_start(h%s, True, False, 0)" % r
                    exec "h%s.pack_start(self.B%s, True, True, 0)" % (r, num)

    def selectData(self):
        if self.frt == "T":
            opts = self.topf[self.pag][self.idx][8]
        else:
            opts = self.colf[self.pag][self.idx][8]
        state = self.disableButtons()
        rs = self.selectItem(self.pag, opts)
        self.enableButtons(state=state)
        if rs != None:
            self.doInvoke(None, rs)
        else:
            self.focusField(self.frt, self.pag, self.col)

    def selectBut(self, but, num):
        but.set_sensitive(False)
        opts = self.butt[num][1]
        state = self.disableButtons()
        rs = self.selectItem(self.pag, opts)
        self.enableButtons(state=state)
        if rs != None:
            self.doInvoke(None, rs)
        else:
            self.focusField(self.frt, self.pag, self.col)

    def selectItem(self, pag, opts):
        if opts["stype"] == "C":
            self.rs = self.selChoice(opts)
        elif opts["stype"] == "D":
            if self.frt == "T":
                fwid = self.topEntry[pag][self.pos]
            else:
                fwid = self.colEntry[pag][self.pos]
            return TartanCalendar(screen=fwid, select=True, quit=False).date
        elif opts["stype"] == "F":
            return self.selFile(opts)
        elif opts["stype"] == "M" and "func" in opts:
            try:
                self.rs = opts["func"]()
            except:
                self.rs = None
            if self.rs == None:
                return None
            if type(self.rs) == str:
                return self.rs
        elif opts["stype"] == "R":
            if not "screen" in opts:
                opts["screen"] = self.screen
            self.rs = self.selRecord(pag, opts)
        if self.rs.selection:
            if "comnd" in opts and opts["comnd"]:
                ok = opts["comnd"](self.frt, pag, self.row, self.col,
                    self.pos, self.idx, self.rs.selection)
                return ok
            if type(self.rs.selection) == int or type(self.rs.selection) == str:
                return self.rs.selection
            if "mode" in opts and opts["mode"] == "M":
                d = []
                for s in self.rs.selection:
                    if "index" in opts:
                        d.append(s[opts["index"]])
                    else:
                        d.append(s[0])
                return d
            if "index" in opts:
                return self.rs.selection[opts["index"]]
            return self.rs.selection[0]

    def selChoice(self, opts):
        if "deco" in opts:
            deco = opts["deco"]
        else:
            deco = None
        if "titl" in opts:
            titl = opts["titl"]
        else:
            titl = None
        if "head" in opts:
            head = opts["head"]
        else:
            head = None
        data = opts["data"]
        if "typs" in opts:
            typs = opts["typs"]
        else:
            typs = None
        if "mode" in opts:
            mode = opts["mode"]
        else:
            mode = "S"
        if "retn" in opts:
            retn = opts["retn"]
        else:
            retn = "D"
        if "butt" in opts:
            butt = opts["butt"]
        else:
            butt = None
        if "size" in opts:
            size = opts["size"]
        else:
            size = None
        sc = SChoice(self.mf, deco=deco, titl=titl, head=head, data=data,
            typs=typs, mode=mode, retn=retn, butt=butt, scrn=self.screen,
            size=size)
        self.mf.updateStatus("")
        return sc

    def selRecord(self, pag, opts):
        # Set the Defaults
        for key in ("screen", "stype", "tables", "cols", "extra", "where", "group", "title", "order", "wtype", "zero", "butt", "sort", "size"):
            if key in opts:
                exec "%s = opts['%s']" % (key, key)
            elif key == "screen":
                screen = self.screen
            elif key == "tables":
                tables = []
            elif key == "sort":
                sort = True
            elif key == "zero":
                zero = "N"
            else:
                exec "%s = None" % key
        whr = self.setWhere(pag, opts)
        # Default the order to the first column
        if not order:
            order = cols[0][0]
        sr = SRec(self.mf, screen=screen, tables=tables, title=title,
            cols=cols, where=whr, group=group, order=order, extra=extra,
            wtype=wtype, butt=butt, sort=sort, size=size, zero=zero)
        return sr

    def setWhere(self, pag, opts):
        if "where" in opts and opts["where"]:
            where = opts["where"][:]
        else:
            where = None
        if "whera" in opts and opts["whera"]:
            for o in opts["whera"]:
                s = o[0]
                f = o[1]
                c = int(o[2])
                if len(o) == 4:
                    pag = o[3]
                if s == "T":
                    if type(self.topf[pag][c][1]) in (list, tuple):
                        t = "A"
                    else:
                        t = self.topf[pag][c][1][2:]
                    d = self.t_work[pag][0][c]
                else:
                    r = int((self.last[pag][1] - 1) / self.colq[pag])
                    t = self.colf[pag][c][1][2:]
                    d = self.c_work[pag][r][c]
                if not where:
                    where = [(f, "=", d)]
                else:
                    where.append((f, "=", d))
        return where

    def selFile(self, opts):
        if opts["types"] == "dir":
            dialog = FileDialog(initd=opts["initd"])
            sel = dialog.askdirectory()
        else:
            if not "initd" in opts:
                opts["initd"] = None
            if not "multi" in opts:
                opts["multi"] = False
            dialog = FileDialog(initd=opts["initd"], ftype=opts["ftype"],
                multi=opts["multi"])
            sel = dialog.askopenfilename()
        return sel

    def focusField(self, frt, pag, col, clr=True, err="", dir="F"):
        """
        This routine is used to focus on a widget:
            frt = The frame type, T = Top, C = Columnar
            pag = The frame number
            col = The column number
            clr = Clear the frame if first field, True or False
            err = An error message to display on the status line
            dir = Direction, F = Forwards, B = Backwards when skipping fields
        """
        if frt == "T" and not self.topq[pag]:
            return
        elif frt == "C" and not self.colq[pag]:
            return
        for widget in self.has_focus:
            widget.set_sensitive(False)
            widget.unset_flags(gtk.CAN_FOCUS)
        if frt == self.first[0] and pag == self.first[1] and col == 1 and clr:
            for x in range(0, self.pgs+1):
                self.clearFrame("T", x)
                self.clearFrame("C", x)
        self.frt = frt
        self.pag = int(pag)
        self.col = int(col)
        self.pos = self.col - 1
        self.checkButtonsTags(self.frt, self.pag, self.col)
        if self.frt == "T":
            self.row = self.pos / self.topq[self.pag]
            self.idx = self.pos
            flds = self.topf[self.pag][self.idx]
            fldd = self.topEntry[self.pag][self.pos]
            if flds[1][0] == "I" and self.topListl[self.pag][self.pos]:
                faut = [self.topCompl[self.pag][self.pos],
                    self.topListl[self.pag][self.pos]]
            else:
                faut = None
            #===================================================================
            # Setting the default data in top entry field
            #===================================================================
            if len(flds) > 4 and flds[6] == "Y":
                self.clearEntry(self.frt, self.pag, self.col)
                work = ""
            else:
                work = self.t_work[self.pag][self.row][self.pos]
            if work:
                dflt = work
            elif len(flds) == 4:
                dflt = ""
            else:
                dflt = flds[5]
            #if not dflt:
            #    dflt = ""
            #===================================================================
        else:
            self.row = self.pos / self.colq[self.pag]
            self.idx = self.pos % self.colq[self.pag]
            flds = self.colf[self.pag][self.idx]
            fldd = self.colEntry[self.pag][self.pos]
            if flds[1][0] == "I" and self.colListl[self.pag][self.pos]:
                faut = [self.colCompl[self.pag][self.pos],
                    self.colListl[self.pag][self.pos]]
            else:
                faut = None
            #===================================================================
            # Setting the default data in col entry field
            #===================================================================
            if len(flds) > 4 and flds[6] == "Y":
                if self.idx == 0:
                    self.clearLine(self.pag, self.row)
                work = ""
            else:
                work = self.c_work[self.pag][self.pos / self.colq[
                    self.pag]][self.idx]
            if work:
                dflt = work
            elif len(flds) == 4:
                dflt = ""
            elif flds[5] not in ("r","i","p"):
                dflt = flds[5]
            elif self.col > self.colq[self.pag]:
                dflt = self.c_work[self.pag][(self.pos / self.colq[
                    self.pag]) - 1][self.idx]
                if not dflt or flds[5] == "r":
                    # Repeat previous value
                    pass
                elif flds[5] == "i":
                    # Increment previous value
                    try:
                        dflt = int(dflt) + 1
                    except:
                        dflt = ""
                elif flds[5] == "p":
                    # Increment previous period
                    y = dflt / 100
                    m = (dflt % 100) + 1
                    if m > 12:
                        y += 1
                        m = m - 12
                    dflt = (y * 100) + m
                    if dflt == 1:
                        dflt = ""
            else:
                dflt = ""
            #if not dflt:
            #    dflt = ""
        #=======================================================================
        # Store the last col position by page
        #=======================================================================
        if self.frt == "T":
            self.last[self.pag][0] = self.col
        else:
            self.last[self.pag][1] = self.col
        #=======================================================================
        # Check if Output Field Only and if so, skip it and return
        #=======================================================================
        sk = "N"
        if flds[1][0] in ("O", "ORB"):
            self.doSkipFld(dir.upper(), dflt)
            return
        #=======================================================================
        # Setting the status line help message
        #=======================================================================
        if flds[8] and flds[1][0] == "I":
            sufx = ", <F1> to Select"
        else:
            sufx = ""
        if flds[9]:
            sufx = sufx + ", <F5> to Delete"
        if flds[1][1:] == "TV":
            sufx = sufx + ", <F9> to Continue"
        if flds[4][-7:] == "(noesc)":
            self.esc = False
            text = flds[4][:-7]
            cnt = sufx.rfind(",")
            if cnt != -1:
                sufx = "%s or %s" % (sufx[:cnt], sufx[cnt+1:])
        else:
            self.esc = True
            text = flds[4]
            if self.idx == 0:
                sufx = sufx + " or <Esc> to Exit"
            else:
                sufx = sufx + " or <Esc> to Back Up"
        if err and err != "ok":
            if self.mf.rcdic["errs"].lower() == "m":
                flenam = os.path.join(self.mf.prgdir, "snd", "ding.wav")
                try:
                    if sys.platform == "win32":
                        from winsound import PlaySound, SND_FILENAME, SND_ASYNC
                        PlaySound(flenam, SND_FILENAME|SND_ASYNC)
                    else:
                        from wave import open as waveOpen
                        from ossaudiodev import open as ossOpen
                        s = waveOpen(flenam, "rb")
                        (nc,sw,fr,nf,comptype, compname) = s.getparams( )
                        dsp = ossOpen("/dev/dsp", "w")
                        try:
                            from ossaudiodev import AFMT_S16_NE
                        except ImportError:
                            if byteorder == "little":
                                AFMT_S16_NE = ossaudiodev.AFMT_S16_LE
                            else:
                                AFMT_S16_NE = ossaudiodev.AFMT_S16_BE
                        dsp.setparameters(AFMT_S16_NE, nc, fr)
                        data = s.readframes(nf)
                        s.close()
                        dsp.write(data)
                        dsp.close()
                except:
                    gtk.gdk.beep()
            else:
                gtk.gdk.beep()
            self.mf.updateStatus("%s, Retry%s" % (err, sufx), "white", "red")
        else:
            self.mf.updateStatus("Enter %s%s" % (text, sufx))
        #=======================================================================
        if type(flds[1]) in (list, tuple):
            self.doCheckButton(fldd, self.idx, "focus", dflt)
        elif flds[1][1:] in ("TV", "Tv"):
            self.doTextView(self.frt, self.pag, self.idx, fldd, "focus", dflt)
        else:
            self.doEntryField(self.frt,self.pag,self.idx,fldd,faut,"focus",dflt)

    def doSkipFld(self, dir, dflt=""):
        if dir == "F":
            self.doInvoke(None, dflt)
        else:
            self.loadEntry(self.frt, self.pag, self.pos, data=dflt)
            if self.col == 1:
                if self.txit and self.txit[self.pag]:
                    self.txit[self.pag]()
                elif self.tend and self.tend[self.pag]:
                    if self.tend[self.pag][0]:
                        self.tend[self.pag][0]()
            else:
                self.col = self.col - 1
                self.focusField(self.frt, self.pag, self.col, dir=dir)

    def deleteRecord(self, widget=None):
        # Disable the widget
        if widget:
            self.setWidget(widget, state="disabled")
        ok = self.entryConfirm(mess="Delete This Record (Y/N)")
        if ok == "n":
            self.focusField(self.frt, self.pag, self.col)
        else:
            if self.frt == "C":
                com = self.colf[self.pag][self.idx][9]
            else:
                com = self.topf[self.pag][self.idx][9]
            if com:
                ok = com()
            else:
                ok = "Invalid or Missing Deletion Command"
            if ok == "nf":
                return
            if ok and ok != "ok":
                showError(self.screen, "Deletion Error", ok)
            if self.frt == "C":
                col = (self.row * self.colq[self.pag]) + 1
                self.focusField(self.frt, self.pag, col)
            else:
                self.focusField(self.first[0], self.first[1], 1)

    def backUp(self, frt, pag):
        #=======================================================================
        # New amendments to display extra characters (outside widget width)
        #=======================================================================
        self.doEraseExtraText()
        #=======================================================================
        if frt == "T":
            flds = self.topf[pag]
            fldd = self.topEntry[pag][self.pos]
        else:
            flds = self.colf[pag]
            fldd = self.colEntry[pag][self.pos]
        last = (frt, pag, self.pos)
        if self.doGetClrFlag(flds):
            self.clearEntry(frt, pag, self.col)
        if (frt == "T" and self.col == 1) or (frt == "C" and self.idx == 0):
            self.doExitFrame(frt, pag)
            return
        if type(flds[self.idx][1]) in (list, tuple):
            self.doCheckButton(fldd, self.idx, "disable")
        elif flds[self.idx][1][1:] in ("TV", "Tv"):
            self.doTextView(frt, pag, self.idx, fldd, "disable")
        else:
            self.doEntryField(frt, pag, self.idx, fldd, None, "disable")
        self.idx = self.idx - 1
        self.pos = self.pos - 1
        self.col = self.col - 1
        clr = self.doGetClrFlag(flds)
        for s in range(len(self.skip[self.pag]) - 1, -1, -1):
            if self.skip[self.pag][s] == (self.pos + 1):
                if clr:
                    self.clearEntry(frt, pag, self.col)
                if frt == "T" and self.col == 1:
                    self.doExitFrame(frt, pag)
                    return
                elif frt == "C" and self.idx == 0:
                    self.doExitFrame(frt, pag)
                    return
                else:
                    self.idx = self.idx - 1
                    self.pos = self.pos - 1
                    self.col = self.col - 1
                    clr = self.doGetClrFlag(flds)
        while flds[self.idx][1][0] in ("O", "ORB"):
            if clr:
                self.clearEntry(frt, pag, self.col)
            if (frt == "T" and self.col == 1) or (frt == "C" and self.idx == 0):
                self.doExitFrame(frt, pag)
                return
            else:
                self.idx = self.idx - 1
                self.pos = self.pos - 1
                self.col = self.col - 1
                clr = self.doGetClrFlag(flds)
        if last[0] == "T":
            wrk = self.t_work[last[1]][0][last[2]]
        else:
            row = last[2] / self.colq[last[1]]
            idx = last[2] % self.colq[last[1]]
            wrk = self.c_work[last[1]][row][idx]
        self.loadEntry(last[0], last[1], last[2], data=wrk)
        self.focusField(frt, pag, self.col, clr=clr)

    def doGetClrFlag(self, flds):
        if len(flds[self.idx]) > 4 and flds[self.idx][6] == "Y":
            return True

    def clearFrame(self, frt, pag):
        self.skip[pag] = []
        if frt == "T":
            beg = 0
            end = self.topq[pag]
            fldd = self.topEntry[pag]
        else:
            beg = self.colq[pag]
            end = (self.rows[pag]+1) * self.colq[pag]
            fldd = self.colEntry[pag]
        if pag > 0:
            if frt == "T":
                self.last[pag][0] = 0
            else:
                self.last[pag][1] = 0
        for x in range(0, len(fldd)):
            self.clearEntry(frt, pag, (x+1))

    def clearLine(self, pag, row=None, focus=False):
        self.skip[pag] = []
        if row:
            pos = row * self.colq[pag]
        else:
            pos = int(self.pos / self.colq[pag]) * self.colq[pag]
        for y in range(pos, (pos + self.colq[pag])):
            r = int(y / self.colq[pag])
            c = y - (r*self.colq[pag])
            self.c_work[pag][r][c] = ""
            self.c_disp[pag][r][c] = ""
            if type(self.colEntry[pag][y]) in (MyTextView_v, MyTextView_V):
                buf = self.colEntry[pag][y].get_buffer()
                buf.set_text("")
            else:
                self.colEntry[pag][y].set_text("")
        if focus:
            self.focusField("C", pag, pos+1)

    def clearEntry(self, frt, pag, col):
        if frt == "T":
            idx = col - 1
            pos = idx = col - 1
            row = int(pos / self.topq[pag])
            if self.topf[pag][idx][1][2:] in ("I","D","d","0","1","2","3"):
                self.t_work[pag][row][idx] = 0
            else:
                self.t_work[pag][row][idx] = ""
            self.t_disp[pag][row][idx] = ""
            fldd = self.topEntry[pag][pos]
            if type(self.topf[pag][idx][1]) in (list, tuple):
                self.doCheckButton(fldd, pos, "clear")
            elif self.topf[pag][idx][1][1:] in ("TV", "Tv"):
                buf = fldd.get_buffer()
                buf.set_text("")
            else:
                fldd.set_text("")
        else:
            pos = col - 1
            idx = pos % self.colq[pag]
            row = int(pos / self.colq[pag])
            if self.colf[pag][idx][1][2:] in ("I","D","d","0","1","2","3"):
                self.c_work[pag][row][idx] = 0
            else:
                self.c_work[pag][row][idx] = ""
            self.c_disp[pag][row][idx] = ""
            fldd = self.colEntry[pag][pos]
            if self.colf[pag][idx][1][1:] in ("TV", "Tv"):
                buf = fldd.get_buffer()
                buf.set_text("")
            else:
                fldd.set_text("")

    def loadEntry(self, frt, pag, pos, data=None):
        """
        where
            frt  =    Frame Type ('T', 'C')
            pag  =    Page Number
            pos  =    Position (0 to (self.rows * self.[t,c]olq) - 1)
            data =    Data to be loaded in work and disp
        """
        if frt == "T":
            fld = self.topEntry[pag][pos]
            idx = pos % self.topq[pag]
            if data != None:
                if type(self.topf[pag][idx][1]) in (list, tuple):
                    d = CCD(data, "NA", 1)
                    if d.work == "0":
                        d.disp = ""
                else:
                    d = CCD(data, self.topf[pag][idx][1],
                        self.topf[pag][idx][2][1])
                    if d.work == "0":
                        d.disp = ""
                self.t_work[pag][0][idx] = d.work
                self.t_disp[pag][0][idx] = d.disp
            if type(self.topf[pag][idx][1]) in (list, tuple):
                self.doCheckButton(fld, idx, "load", data)
            else:
                disp = self.t_disp[pag][0][idx]
                if self.topf[pag][idx][1][1:] in ("TV", "Tv"):
                    fld.set_wrap_mode(gtk.WRAP_NONE)
                    buf = fld.get_buffer()
                    buf.set_text(disp)
                elif not disp:
                    fld.set_text("")
                else:
                    fld.set_text(disp)
        else:
            fld = self.colEntry[pag][pos]
            row = pos / self.colq[pag]
            idx = pos % self.colq[pag]
            if data != None:
                d = CCD(data, self.colf[pag][idx][1],
                    self.colf[pag][idx][2][1])
                self.c_work[pag][row][idx] = d.work
                self.c_disp[pag][row][idx] = d.disp
            disp = self.c_disp[pag][row][idx]
            if self.colf[pag][idx][1][1:] in ("TV", "Tv"):
                fld.set_wrap_mode(gtk.WRAP_NONE)
                buf = fld.get_buffer()
                buf.set_text(disp)
            elif not disp:
                fld.set_text("")
            else:
                fld.set_text(disp)
        if data != None:
            return d

    def getEntry(self, frt, pag, pos, wid=True, txt=True):
        """
        where
            frt  =  Frame Type ('T', 'C')
            pag  =  Page Number
            pos  =  Position (0 to (self.rows * self.[t,c]olq) - 1)
            wid  =  Return the widget
            txt  =  Return the widget contents
        """
        if frt == "T":
            fwid = self.topEntry[pag][pos]
        else:
            fwid = self.colEntry[pag][pos]
        if type(fwid) in (MyTextView_v, MyTextView_V):
            buf = fwid.get_buffer()
            ftxt = buf.get_text(buf.get_start_iter(), buf.get_end_iter())
        else:
            ftxt = fwid.get_text()
        if wid and txt:
            return (fwid, ftxt)
        elif wid:
            return fwid
        elif txt:
            return ftxt

    def doInvoke(self, fldd, data):
        if fldd:
            self.frt, self.pag, self.col, cmd = fldd
            self.pos = self.col - 1
            if self.frt == "T":
                self.row = self.pos / self.topq[self.pag]
                self.idx = self.pos
            else:
                self.row = self.pos / self.colq[self.pag]
                self.idx = self.pos % self.colq[self.pag]
        elif self.frt == "T" and len(self.topf[self.pag][self.idx]) > 4:
            cmd = self.topf[self.pag][self.idx][7]
        elif self.frt == "C" and len(self.colf[self.pag][self.idx]) > 4:
            cmd = self.colf[self.pag][self.idx][7]
        else:
            cmd = None
        self.loadEntry(self.frt, self.pag, self.pos, data=data)
        self.execEntry(None, (self.frt, self.pag, cmd))
        return True

    def doKeyPressed(self, frt, pag,  pos, data):
        self.loadEntry(frt, pag, pos, data=data)
        if frt == "T":
            fldd = self.topEntry[pag][pos]
        else:
            fldd = self.colEntry[pag][pos]
        if type(fldd) == MyEntry:
            fldd.emit("activate")
        elif type(fldd) == MyTextView_v:
            fldd.emit("mytxtvv", "KP_Enter")

    def execRB(self, wid, num):
        self.mf.updateStatus("")
        try:
            txt, cmd = self.topEntry[self.pag][self.pos][num][2]
            self.loadEntry("T", self.pag, self.pos, data=txt)
            self.execCommand("T", self.pag, cmd)
        except:
            pass

    def execAutoc(self, completion, model, iter, opts):
        self.mf.updateStatus("")
        self.loadEntry(opts[0], opts[1], opts[2], data=model[iter][0])
        completion.get_entry().emit("activate")
        return True

    def execEntry(self, wid, *opt):
        self.mf.updateStatus("")
        #=======================================================================
        # New amendments to display extra characters (outside widget width)
        #=======================================================================
        self.doEraseExtraText()
        #=======================================================================
        frt, pag, cmd = opt[0]
        self.execCommand(frt, pag, cmd)

    def execCommand(self, frt, pag, cmd):
        if frt == "T":
            flds = self.topf[pag][self.idx]
            fldd = self.topEntry[pag][self.pos]
        else:
            flds = self.colf[pag][self.idx]
            fldd = self.colEntry[pag][self.pos]
        if type(flds[1]) in (list, tuple):
            fld = self.doCheckButton(fldd, self.idx, "active")
            fld = CCD(fld, "NA", 1)
        elif flds[1][1:] in ("TV", "Tv"):
            buf = fldd.get_buffer()
            txt = buf.get_text(buf.get_start_iter(), buf.get_end_iter())
            fld = CCD(txt, "TV", len(txt))
        else:
            fld = CCD(fldd.get_text(), flds[1], flds[2][1])
        if fld.err:
            ok = ("no", fld.err, fld.data)
        elif len(flds) > 4:
            ok = self.doValidation(frt, pag, self.idx, fld.work)
        else:
            ok = ("ok",)
        if ok[0] == "no":
            if flds[1][0] in ("O", "ORB"):
                print "Invalid Output Data %s for (%s)" % (ok[1:], flds)
                sys.exit()
            else:
                self.focusField(frt, pag, self.col, err=ok[1])
        else:
            if ok[0] == "ex":
                fld.work = fld.disp = ok[1]
            if type(flds[1]) in (list, tuple):
                self.doCheckButton(fldd, self.idx, "disable")
            elif flds[1][1:] in ("TV", "Tv"):
                self.doTextView(frt, pag, self.idx, fldd, "disable")
            else:
                if flds[1][1] != "H":
                    fldd.set_text(fld.disp)
                self.doEntryField(frt, pag, self.idx, fldd, None, "disable")
            if frt == "T":
                self.t_work[pag][self.row][self.idx] = fld.work
                self.t_disp[pag][self.row][self.idx] = fld.disp
                if cmd:
                    if cmd == "rt":                 # Return True
                        return True
                    ok = cmd(frt, pag, self.row, self.col, self.pos, self.idx,
                        fld.work)
                    if type(ok) == tuple:
                        col, ok = ok
                    else:
                        col = None
                    if ok in (None, "ok"):          # Continue normally
                        if self.col == self.topq[pag]:
                            self.doEndFrame(frt, pag)
                        else:
                            self.focusField(frt, pag, (self.col + 1))
                    elif ok[:2] == "bu":            # Invoke Esc to backup
                        if len(ok) == 2:
                            tm = 1
                        else:
                            tm = int(ok[2:])
                        for x in range(tm):
                            self.backUp(frt=frt, pag=pag)
                    elif ok[:2] == "sk":            # Skip one or more columns
                        if len(ok) == 2:
                            tm = 1
                        else:
                            tm = int(ok[2:])
                        for x in range(0, tm):
                            self.col += 1
                            self.skip[self.pag].append(self.col)
                        if self.col == self.topq[pag]:
                            self.doEndFrame(frt, pag)
                        else:
                            self.focusField(frt, pag, self.col + 1)
                    elif ok[:2] == "ff":            # Force new column focus
                        ok = ok.split("|")          # ff[col]|Message|Clear
                        col = int(ok[0][2:])
                        if col == 0:
                            col = 1
                        if len(ok) == 1:
                            self.focusField(frt, pag, col)
                        elif len(ok) == 2:
                            self.focusField(frt, pag, col, err=ok[1])
                        else:
                            self.focusField(frt, pag, col, err=ok[1], clr=ok[2])
                    elif ok == "xt":                # Force frame exit
                        self.doExitFrame(frt, pag, xit="y")
                    elif ok == "nc":                # End frame, no confirm
                        self.doEndFrame(frt, pag, "N")
                    elif ok == "nd":                # End frame normally
                        self.doEndFrame(frt, pag)
                    elif ok == "rf":                # Refocus same field
                        self.focusField(frt, pag, self.col)
                    elif ok == "nf":                # No focus to next field
                        pass
                    elif col:
                        self.focusField(frt, pag, col, err=ok)
                    else:
                        self.focusField(frt, pag, self.col, err=ok)
                elif self.col == self.topq[pag]:    # End of frame
                    self.doEndFrame(frt, pag)
                else:                               # Focus to next field
                    self.focusField(frt, pag, (self.col + 1))
            else:
                self.c_work[pag][self.row][self.idx] = fld.work
                self.c_disp[pag][self.row][self.idx] = fld.disp
                if cmd:
                    if cmd in ("nd", "nf", "rt"):
                        return True                 # Return True
                    ok = cmd(frt, pag, self.row, self.col, self.pos,
                        self.idx, fld.work)
                    if type(ok) == tuple:
                        col, ok = ok
                    else:
                        col = None
                    if ok in (None, "ok"):          # Continue normally
                        if self.col % self.colq[pag] == 0:
                            self.doEndFrame(frt, pag)
                        else:
                            self.focusField(frt, pag, (self.col + 1))
                    elif ok[:2] == "bu":            # Invoke Esc to backup
                        if len(ok) == 2:
                            tm = 1
                        else:
                            tm = int(ok[2:])
                        for x in range(tm):
                            self.backUp(frt=frt, pag=pag)
                    elif ok == "cl":                # Clear line
                        self.clearLine(pag, row=self.row)
                        col = 1 + (self.row * self.colq[pag])
                        self.focusField(frt, pag, col)
                    elif ok[:2] == "sk":            # Skip one or more columns
                        if len(ok) == 2:
                            tm = 1
                        else:
                            tm = int(ok[2:])
                        for x in range(0, tm):
                            self.col += 1
                            self.skip[self.pag].append(self.col)
                        if (self.col) % self.colq[pag] == 0:
                            self.doEndFrame(frt, pag)
                        else:
                            self.focusField(frt, pag, self.col + 1)
                    elif ok[:2] == "ff":            # Force new column focus
                        col = int(ok[2:])
                        if col == 0:
                            col = 1
                        self.focusField(frt, pag, col)
                    elif ok == "xt":                # Force frame exit
                        self.doExitFrame(frt, pag)
                    elif ok == "nc":                # End frame, no confirm
                        self.doEndFrame(frt, pag, "N")
                    elif ok == "nd":                # End frame normally
                        self.doEndFrame(frt, pag)
                    elif ok == "rf":                # Refocus same field
                        self.focusField(frt, pag, self.col)
                    elif ok == "nf":                # No focus to next field
                        pass
                    elif col:
                        self.focusField(frt, pag, col, err=ok)
                    else:
                        self.focusField(frt, pag, self.col, err=ok)
                elif self.col % self.colq[pag] == 0:# End of frame
                    self.doEndFrame(frt, pag)
                else:                               # Focus to next field
                    self.focusField(frt, pag, (self.col + 1))

    def doCheckButton(self, grp, num, action=None, data=None, opt=None):
        """
        Check Button actions as follows:

        grp         = The group
        num         = The number in the group
        action      = The action to be taken as follows:
                        clear   = Clear all buttons
                        disable = Disable the buttons
                        focus   = Set and focus the button
                        load    = Set the button
                        active  = Return the active button
        data        = The data to be loaded if action is load
        """
        if action == "active":
            for n, b in enumerate(grp):
                if b[0].get_active() == True:
                    return b[2][0]
            return
        if action == "disable":
            for b in grp:
                try:
                    self.has_focus.remove(b[0])
                except:
                    pass
                b[0].set_sensitive(False)
                b[0].unset_flags(gtk.CAN_FOCUS)
            return
        for b in grp:
            try:
                b[0].disconnect(b[1])
                b[0].set_active(False)
            except:
                pass
        if action == "load":
            for b in grp:
                if b[2][0] == data:
                    b[0].set_active(True)
                    break
        elif action == "focus":
            for n, b in enumerate(grp):
                self.has_focus.append(b[0])
                b[0].set_active(False)
                b[0].set_sensitive(True)
                b[0].set_flags(gtk.CAN_FOCUS)
                if b[2][0] == data:
                    b[0].set_active(True)
                    self.window.present()
                    b[0].grab_focus()
        for n, b in enumerate(grp):
            s = b[0].connect_after("toggled", self.execRB, n)
            grp[n] = [b[0], s, b[2]]

    def doTextView(self, frt, pag, idx, vwr, act, data=None):
        """
        vwr:      The textview widget
        act:      The action to take as follows:
                    disable = disable the field
                    focus   = set and focus the field
        data:     The default data
        """
        if act == "focus":
            self.has_focus.append(vwr)
            vwr.set_wrap_mode(gtk.WRAP_WORD)
            vwr.set_editable(True)
            vwr.set_sensitive(True)
            vwr.set_flags(gtk.CAN_FOCUS)
            if data != None:
                buf = vwr.get_buffer()
                buf.set_text(data)
            vwr.emit("select_all", True)
            self.window.present()
            vwr.grab_focus()
        elif act == "disable":
            try:
                self.has_focus.remove(vwr)
            except:
                pass
            if frt == "T":
                hgt = self.topf[pag][idx][2][1]
            else:
                hgt = self.colf[pag][idx][2][1]
            if hgt == -1:
                vwr.set_wrap_mode(gtk.WRAP_NONE)
            vwr.set_editable(False)
            vwr.set_sensitive(False)
            vwr.unset_flags(gtk.CAN_FOCUS)

    def doEntryField(self, frt, pag, idx, ent, aut, act, data=None):
        """
        ent:      The entry widget
        aut:      Autocompletion data comprising the following or None:
                      The completion widget
                      The completion list
        act:      The action to take as follows:
                    disable = disable the field
                    focus   = set and focus the field
        data:     The default data
        """
        if frt == "T":
            flds = self.topf[pag][idx]
        else:
            flds = self.colf[pag][idx]
        if act == "focus":
            self.has_focus.append(ent)
            ent.set_sensitive(True)
            ent.set_flags(gtk.CAN_FOCUS)
            ent.select_region(0, -1)
            if data:
                ent.set_text(str(data))
            if aut and self.acnf == "y":
                items = []
                aut[1].clear()
                if "index" in flds[8]:
                    pos = flds[8]["index"]
                else:
                    pos = 0
                if flds[8]["stype"] == "C" and flds[8]["data"]:
                    for item in flds[8]["data"]:
                        if type(item) in (tuple, list):
                            aut[1].append([item[pos]])
                        else:
                            aut[1].append([item])
                elif flds[8]["stype"] == "R":
                    if "wtype" in flds[8] and flds[8]["wtype"] == "D":
                        dat = flds[8]["where"]
                    else:
                        col = [flds[8]["cols"][pos][0]]
                        tab = []
                        for t in flds[8]["tables"]:
                            tab.append(t)
                        whr = self.setWhere(pag, flds[8])
                        if "group" in flds[8]:
                            grp = flds[8]["group"]
                        else:
                            grp = None
                        if "order" in flds[8]:
                            ord = flds[8]["order"]
                        else:
                            ord = None
                        sql = Sql(self.mf.dbm, tab, prog=__name__)
                        dat = sql.getRec(tables=tab, cols=col, where=whr,
                            group=grp, order=ord)
                    for d in dat:
                        if type(d[0]) == str:
                            aut[1].append([d[0].strip()])
                        else:
                            aut[1].append([d[0]])
            #===================================================================
            # New amendments to display extra characters (outside widget width)
            #===================================================================
            self.txtvwr = None        # Indicates if a TextView exists
            self.txtidx = None        # Indicates the connection has been made
            if self.txtv == "y" and frt == "C" and ent.get_width_chars() != ent.get_max_length():
                idx = ent.connect("changed", self.doShowExtraText)
                self.txtidx = (ent, idx)
            #===================================================================
            self.window.present()
            ent.grab_focus()
        elif act == "disable":
            try:
                self.has_focus.remove(ent)
            except:
                pass
            ent.select_region(0, 0)
            ent.set_sensitive(False)
            ent.unset_flags(gtk.CAN_FOCUS)

    def doValidation(self, frt, pag, idx, wrk):
        if frt == "T":
            des = self.topf[pag][idx][4]
            val = self.topv[pag][idx]
            if type(self.topf[pag][idx][1]) in (list, tuple):
                # Radio Button
                if wrk:
                    return ("ok", )
                else:
                    return ("no", "Invalid %s" % des)
        else:
            des = self.colf[pag][idx][4]
            val = self.colv[pag][idx]
        if not val:
            return ("ok", )
        if val[0] == "efld":
            if frt == "T":
                typ = self.topf[pag][idx][1]
                siz = self.topf[pag][idx][2][1]
            else:
                typ = self.colf[pag][idx][1]
                siz = self.colf[pag][idx][2][1]
            chk = CCD(wrk, typ, siz)
            if chk.err:
                return ("no", chk.err)
            else:
                return ("ok", )
        if val[0] == "notblank":
            if not wrk:
                return ("no", "Invalid %s" % des)
            elif len(val) == 2 and val[1] == "nospaces" and wrk.count(" "):
                return ("no", "Invalid, Spaces Not Allowed")
            else:
                return ("ok", )
        elif val[0] == "notzero":
            if wrk == 0:
                return ("no", "Invalid %s" % des)
            else:
                return ("ok", )
        elif val[0] == "in":
            if wrk in val[1]:
                return ("ok", )
            else:
                return ("no", "Invalid %s" % des)
        elif val[0] == "notin":
            if wrk in val[1]:
                return ("no", "Invalid %s" % des)
            else:
                return ("ok", )
        elif val[0] == "between":
            if wrk < val[1] or wrk > val[2]:
                return ("no", "Invalid %s (%s-%s)" % (des,val[1], val[2]))
            else:
                return ("ok", )
        elif val[0] == "dir":
            if os.path.isdir(wrk):
                return ("ok", )
            else:
                return ("no", "Invalid %s, Not Found" % des)
        elif val[0] == "fle":
            if not wrk and len(val) == 2 and val[1] == "blank":
                return ("ok", )
            wrk = wrk.split(",")
            err = False
            for w in wrk:
                if not os.path.isfile(w):
                    err = True
                    break
            if not err:
                return ("ok", )
            else:
                return ("no", "Invalid %s, Not Found" % w)
        elif val[0] == "cmd":
            if not wrk:
                return ("no", "Invalid %s" % des)
            return ("ex", wrk)
        elif val[0] == "file":
            if os.path.isfile(wrk):
                return ("ok", )
            else:
                return ("no", "Invalid %s, Not Found" % des)
        elif val[0] == "path":
            if os.path.exists(wrk):
                return ("ok", )
            else:
                return ("no", "Invalid %s, Not Found" % des)
        elif val[0] == "email":
            eml = wrk.split(",")
            for e in eml:
                if len(val) == 2:
                    no = self.doCheckEmail(e, val[1])
                else:
                    no = self.doCheckEmail(e, True)
                if no:
                    return ("no", "Invalid %s, %s" % (des, no))
            return ("ok", )
        elif val[0] == "chars":
            if len(wrk) > val[1]:
                return ("no", "Invalid Size, Only %s Allowed" % val[1])
            return ("ok", )

    def doCheckEmail(self, address, blank=False):
        """
        Check for problems with the supplied e-mail address.
        """
        if not address:
            if blank:
                return
            return "Address is empty"
        by_at_sign = address.split("@")
        if len(by_at_sign) != 2:
            return "Address must have one and only one @"
        name, hostname = by_at_sign
        if not name:
            return "Name of account is missing"
        if "." not in hostname:
            return "Address is missing . after @"
        for component in hostname.split("."):
            if not component:
                return "Domain name (%s) not fully qualified" % hostname

    def doCheckFields(self, fields=None):
        """
        fields is either None for all or a tuple having:
            Page Type i.e. T or C
            Page Number i.e. 0-9
            Range of Column Numbers on the Page i.e. (0,1,2,3,4,8,9) or
                None for all columns
        """
        if fields == None:
            for pag in range(0, self.pgs+1):
                for fld in range(0, len(self.topf[pag])):
                    ok = self.doCheckField("T", pag, fld)
                    if ok:
                        return ok
                for fld in range(0, len(self.colf[pag])):
                    ok = self.doCheckField("C", pag, fld)
                    if ok:
                        return ok
        elif fields[2] == None:
            if fields[0] == "T":
                for fld in range(0, len(self.topf[fields[1]])):
                    ok = self.doCheckField("T", fields[1], fld)
                    if ok:
                        return ok
            elif fields[0] == "C":
                for fld in range(0, len(self.colf[fields[1]])):
                    ok = self.doCheckField("C", fields[1], fld)
                    if ok:
                        return ok
        else:
            for col in fields[2]:
                ok = self.doCheckField(fields[0], fields[1], col)
                if ok:
                    return ok
        return ("", 0, 0, "")

    def doCheckField(self, frt, pag, fld):
        if frt == "T":
            if len(self.topf[pag][fld]) == 4:
                pass
            else:
                wrk = self.t_work[pag][0][fld]
                ok = self.doValidation(frt, pag, fld, wrk)
                if ok[0] == "no":
                    return (frt, pag, fld, ok[1])
                if ok[0] == "ex":
                    self.loadEntry(frt, pag, fld, data=ok[1])
        elif frt == "C":
            if len(self.colf[pag][fld]) == 4:
                pass
            else:
                for row in range(0, self.colq[pag]):
                    wrk = self.c_work[pag][row][fld]
                    ok = self.doValidation(frt, pag, fld, wrk)
                    if ok[0] == "no":
                        return (frt, pag, fld, ok[1])
                    if ok[0] == "ex":
                        self.loadEntry(frt, pag, fld, data=ok[1])

    def doExitFrame(self, frt, pag, xit="n"):
        self.mf.updateStatus("")
        if frt == "T" and self.txit and self.txit[pag]:
            self.txit[pag]()
        elif frt == "C" and self.cxit and self.cxit[pag]:
            self.cxit[pag]()
        elif xit == "y":
            sys.exit()

    def doEndFrame(self, frt, pag, cnf=None):
        """
        frt = The frame type, T or C
        pag = The frame number, 0 - 9
        cnf = None, "y", "n" or "d"
                  None or "n" = No confirmation required
                  "y"         = Confirmation required, either a Y or N
                  "d"         = Same as "y" but Enter defaults to Y
        """
        self.mf.updateStatus("")
        self.checkButtonsTags(frt, pag, 0)
        if frt == "T":
            if not self.tend or not self.tend[pag]:
                cmd = None
            else:
                cmd = self.tend[pag][0]
                if not cnf:
                    cnf = self.tend[pag][1]
        elif not self.cend or not self.cend[pag]:
            cmd = None
        else:
            cmd = self.cend[pag][0]
            if cnf == None:
                cnf = self.cend[pag][1]
        if cnf:
            cnf = cnf.lower()
        if cnf in (None, "n"):
            ok = "y"
        else:
            ok = self.entryConfirm(mess="", dflt=cnf)
        if ok == "n":
            for x in range(len(self.skip[self.pag])-1, -1, -1):
                if self.col == self.skip[self.pag][x]:
                    self.checkButtonsTags(frt, pag, self.col)
                    self.col = self.col - 1
            self.focusField(frt, pag, self.col, dir="B")
        elif cmd:
            cmd()
        elif frt == "C":
            self.advanceLine(pag)

    def advanceLine(self, pag):
        self.skip[pag] = []
        row = int((self.last[pag][1] - 1) / self.colq[pag]) + 1
        if row == self.rows[pag]:
            self.scrollScreen(pag)
        else:
            col = (row * self.colq[pag]) + 1
            self.focusField("C", pag, col)

    def scrollScreen(self, pag):
        for x in range(1, self.rows[pag]):
            col = self.colEntry[pag]
            for y in range(0, self.colq[pag]):
                self.c_work[pag][x-1][y] = self.c_work[pag][x][y]
                self.c_disp[pag][x-1][y] = self.c_disp[pag][x][y]
                p1 = ((x-1) * self.colq[pag]) + y
                p2 = p1 + self.colq[pag]
                if type(col[p1]) in (MyTextView_v, MyTextView_V):
                    bf1 = col[p1].get_buffer()
                    bf2 = col[p2].get_buffer()
                    txt = bf2.get_text(bf2.get_start_iter(), bf2.get_end_iter())
                    bf1.set_text(txt)
                else:
                    col[p1].set_text(col[p2].get_text())
        self.clearLine(pag, row=(self.rows[pag]-1))
        col = (self.rows[pag] * self.colq[pag]) - (self.colq[pag] - 1)
        self.focusField("C", pag, col)

    def entryConfirm(self, mess="", dflt=""):
        state = self.disableButtons()
        if self.mf.rcdic["cnf"].lower() == "n":
            self.answer = "y"
        else:
            if not mess:
                mess = "Confirm Entry (Y/N)"
            self.mf.updateStatus(mess, bg="red", fg="white")
            self.conf = gtk.Entry(max=1)
            self.conf.set_size_request(0, 0)
            self.conf.set_flags(gtk.CAN_FOCUS)
            if dflt == "d":
                self.conf.set_text("Y")
                self.conf.connect("activate", self.checkYesNo, "Y")
            self.conf.connect("changed", self.checkYesNo, "Y")
            self.mstFrame.pack_start(self.conf, False, False, 0)
            self.conf.grab_focus()
            self.mf.startLoop()
            self.conf.destroy()
            self.mf.updateStatus("")
        self.enableButtons(state=state)
        return self.answer.lower()

    def checkYesNo(self, wid, mess):
        self.answer = self.conf.get_text()
        if self.answer.lower() not in ("y", "n"):
            self.conf.set_text("")
            return
        self.mf.closeLoop()

    def checkButtonsTags(self, frt, pag, col):
        if self.butt:
            if col == 0:
                pass
            elif frt == "T" and self.topq[pag] and col != self.topq[pag]:
                col = col % self.topq[pag]
            elif frt == "C" and self.colq[pag] and col != self.colq[pag]:
                col = col % self.colq[pag]
            for num, but in enumerate(self.butt):
                if but[3] == 2:
                    continue
                state = None
                if not but[4] or not but[5]:
                    if but[3] == 0:
                        state = False
                    elif but[3] == 1:
                        state = True
                if but[4]:
                    if type(but[4][0]) in (list, tuple):
                        for chk in but[4]:
                            if chk == (frt, pag, col):
                                state = True
                    elif tuple(but[4]) == (frt, pag, col):
                        state = True
                if but[5]:
                    if type(but[5][0]) in (list, tuple):
                        for chk in but[5]:
                            if chk == (frt, pag, col):
                                state = False
                    elif tuple(but[5]) == (frt, pag, col):
                        state = False
                if not state == None:
                    exec "self.B%s.set_sensitive(%s)" % (num, state)
        if self.tags:
            for num, tag in enumerate(self.tags):
                if not tag[2] and not tag[3]:
                    continue
                if tag[2]:
                    if type(tag[2][0]) in (list, tuple):
                        on = []
                        for l in tag[2]:
                            on.append(list(l))
                    else:
                        on = [list(tag[2])]
                else:
                    on = []
                if tag[3]:
                    if type(tag[3][0]) in (list, tuple):
                        off = []
                        for l in tag[3]:
                            off.append(list(l))
                    else:
                        off = [list(tag[3])]
                else:
                    off = []
                for flag in on:
                    if flag[1] == None:
                        flag[1] = pag
                    if tuple(flag) == (frt, pag, col):
                        self.enableTag(num)
                for flag in off:
                    if flag[1] == None:
                        flag[1] = pag
                    if tuple(flag) == (frt, pag, col):
                        self.disableTag(num)

    def disableTag(self, num):
        exec "self.nb.Tag%s.set_sensitive(False)" % num
        exec "self.nb.Tag%s.unset_flags(gtk.CAN_FOCUS)" % num

    def enableTag(self, num):
        exec "self.nb.Tag%s.set_sensitive(True)" % num
        exec "self.nb.Tag%s.set_flags(gtk.CAN_FOCUS)" % num

    def disableButtons(self):
        if self.butt:
            state = []
            for num, but in enumerate(self.butt):
                # Save current state for reinstating
                exec "st = self.B%s.state" % num
                state.append(st)
                exec "self.B%s.set_sensitive(False)" % num
                exec "self.B%s.unset_flags(gtk.CAN_FOCUS)" % num
        else:
            state = None
        return state

    def enableButtons(self, state=None):
        # Reinstate buttons to saved state
        if state and self.butt:
            for num, but in enumerate(self.butt):
                if state[num] == 4:
                    exec "self.B%s.set_sensitive(False)" % num
                    exec "self.B%s.unset_flags(gtk.CAN_FOCUS)" % num
                else:
                    exec "self.B%s.set_sensitive(True)" % num
                    exec "self.B%s.set_flags(gtk.CAN_FOCUS)" % num

    def setWidget(self, wid, state=None, bg=None, fg=None):
        if state == "normal":
            wid.set_sensitive(True)
        elif state == "disabled":
            wid.set_sensitive(False)
        elif state == "focus":
            wid.set_flags(gtk.CAN_FOCUS)
            wid.grab_focus()
            return
        if not bg and not fg:
            return
        style = wid.get_style().copy()
        if bg:
            style.base[gtk.STATE_NORMAL] = wid.get_colormap().alloc_color(bg)
        if fg:
            style.text[gtk.STATE_NORMAL] = wid.get_colormap().alloc_color(fg)
        wid.set_style(style)

    def closeProcess(self):
        self.mstAlw.destroy()
        if self.newscn:
            self.window.destroy()
        if self.tops and not self.mf.fixed:
            self.mf.head.show()
        self.mf.updateStatus("")

class SRec:
    """
    This class is used for selection of a range of valid records.
    The following parameters are allowed:

        screen  :   Screen to build Enquiry On or None for new window
        title   :   A title to display else None for the default
        tables  :   A list of relevant tables e.g. ["genmst","gentrn"]
        cols    :   A list of columns to be printed, all functions MUST be
                        renamed using the 'as' clause containing:
                            column or 'as' name
                            display code
                            display size
                            heading
                            Function key selection
                    e.g. [["omb_state","UI", 5.0,"Stmnt"]],
                          ["omb_date", "D1", 10.0,"Date"],
                          ["sum(omb_amount) as value","SD",13.2,"Value"],
                          ["omb_recon", "NA", 1.0, "R"]]
                    None = All columns
                    SPECIAL:
                        If display code = ("XX",systp), it uses the values in
                            systp subscripted by the value of 'column'  e.g.
                            ('crt_type',('XX',crtrtp),3,'Typ') produces 'Pur'
                        If display code = ("xx",systp), it uses the desc in
                            systp subscripted by the value of 'column' e.g.
                            ('crt_type',('xx',crtrtp),10,'Description')
                            produces 'Purchases'
        where   :   A list of tuples of where conditions e.g.
                    [("omb_recon", "=", "Y")]
                    None = No where statement
        group   :   A string for the 'group' statement
                    e.g. "omb_date, omb_state, omb_recon"
                    None = No group statement
        order   :   A string of the 'order by' statement e.g.
                    "omb_state, omb_date"
                    None = No order statement
        wtype   :   'D' = Use the 'where' option as the data
                    'S' = Use the 'where' option as the select statement
                    None = Normal i.e. Create select statement
        extra   :   Additional data for type 'S'
        zero    :   0, 1 or N
                    Return without display if there are '0' or '1' records to
                    select from else "N" to display irrespectively.
        butt    :   A list of tuples of buttons, in addition to Quit, to draw
                        with their respective commands
        sort    :   Whether the columns can be sorted
        size    :   The default size of the display screen else None
        deco    :   Whether the window must be decorated or not, default True
    """
    def __init__(self, mf, **args):
        self.mf = mf
        defaults = {
            "screen": None,
            "title": None,
            "tables": [],
            "cols": None,
            "where": None,
            "group": None,
            "order": None,
            "wtype": None,
            "extra": None,
            "size": None,
            "sort": True,
            "zero": "N",
            "butt": None,
            "deco": True}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        self.doProcess()

    def doProcess(self):
        fltr = False
        if not self.cols:
            self.cols = []
        self.coll = []
        for col in self.cols:
            col = list(col)
            if len(col) == 4:
                col.append("N")
            if col[4].lower() == "f":
                fltr = self.mf
            self.coll.append(col)
        self.setTableFields()
        if self.wtype == "D":
            self.data = self.where
        else:
            self.data = self.getSelectStatement()
            if self.extra:
                for num, exr in enumerate(self.extra):
                    self.data.insert(num, exr)
        if not self.data:
            self.gt = 0
        else:
            self.gt = len(self.data)
        if not self.gt:
            self.selection = None
        elif self.zero == "1" and self.gt == 1:
            self.selection = self.data
        else:
            sd = SelectChoice(self.screen, self.title, self.cols, self.data,
                sort=self.sort, font=self.mf.fontd, butt=self.butt,
                size=self.size, deco=self.deco, fltr=fltr)
            if sd.selection:
                self.selection = sd.selection[1:]
            else:
                self.selection = None

    def setTableFields(self):
        self.l1 = ""                    # string of all columns
        self.l3 = []                    # list of all columns
        self.cols = []
        if not self.coll:
            for tab in self.tables:
                sql = Sql(self.mf.dbm, tab, prog=__name__)
                exec "col = sql.%s_col" % tab
                exec "dic = sql.%s_dic" % tab
                exec "fld = sql.%s_fld" % tab
                self.l1 = "%s%s," % (self.l1, fld)
                self.l3.extend(col)
                for c in col:
                    self.cols.append([c, dic[c][5], dic[c][3], dic[c][2], None])
        else:
            sql = Sql(self.mf.dbm, "ffield", prog=__name__)
            for col in self.coll:
                if col[0][:4].lower() == "sum(" and col[1][1:] == "D":
                    col[0] = "Round(" + col[0]
                    b = col[0].find(")") + 1
                    d = int(round(col[2] * 10) % 10)
                    col[0] = col[0][:b] + ",%s)" % d + col[0][b:]
                else:
                    d = 0
                self.l1 = "%s%s," % (self.l1, col[0])
                nam = removeFunctions(col[0], d)
                c = nam.split(" as ")
                self.l3.append(c[-1:])
                if not col[1] or not col[2] or not col[3]:
                    dat = sql.getRec(tables="ffield", where=[("ff_name",
                        "=", col[0])], limit=1)
                    if dat:
                        if not col[1]:
                            col[1] = dat[3]
                        if not col[2]:
                            col[2] = dat[4]
                        if not col[3]:
                            col[3] = dat[6]
                if len(col) == 5:
                    sch = col[4]
                else:
                    sch = None
                self.cols.append([col[0], col[3], col[2], col[1], sch])
        self.l1 = self.l1[:-1]

    def getSelectStatement(self):
        sql = Sql(self.mf.dbm, self.tables, prog=__name__)
        cols = []
        for col in self.cols:
            cols.append(col[0])
        if self.group:
            # Check that all columns are in the group statement
            chk = self.group.split(",")
            if len(chk) != len(self.cols):
                self.group = ""
                for col in self.cols:
                    if col[0].count("("):
                        continue
                    self.group = "%s%s, " % (self.group, col[0])
                self.group = self.group[:-2]
        # If no order statement make first non function column the order col
        if not self.order:
            for col in self.cols:
                if not col[0].count("("):
                    self.order = col[0]
        return sql.getRec(tables=self.tables, cols=cols, where=self.where,
            group=self.group, order=self.order)

class SChoice:
    """
    This class is used for selection of a range of valid choices.
    The following parameters are allowed:

    deco = Whether the window must be decorated or not, default False
    titl = The widget title, None = No Title
    head = The column headings, None = No Headings
    data = Either a tuple/list or a tuple/list of tuples/lists
    typs = A list of data types for formatting and searchable column i.e.
               [("UI",2), ("NA",30,"Y"), ("SD",13.2)]
    scrn = A widget to display the choices on
    retn = D to return the data selected or I to return the index position of
           the data selected or None to return nothing. Defaults to Data
    mode = S, M or N for Single, Multiple or No (Display Only) select mode
    butt = A list of tuples of buttons, in addition to Quit, to draw
           with their respective commands.
    selc = The column to use for Selection if mode = 'M'
    size = The default of the screen else None
    """
    def __init__(self, mf, **args):
        self.mf = mf
        defaults = {
            "scrn": None,
            "titl": None,
            "head": None,
            "data": None,
            "typs": None,
            "retn": "D",
            "mode": "S",
            "butt": None,
            "selc": None,
            "size": None,
            "deco": False}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        if not self.mf.fixed:
            self.scrn = None
        elif self.scrn == None:
            self.scrn = self.mf.body
        if self.data:
            if type(self.data[0]) == tuple:
                pass
            elif type(self.data[0]) == list:
                pass
            else:
                data = []
                for txt in self.data:
                    data.append((txt,))
                self.data = tuple(data)
            if self.mode.upper() == "M":
                data = []
                for dat in self.data:
                    dat = list(dat)
                    dat.insert(0, None)
                    data.append(dat)
                self.data = tuple(data)
            else:
                self.data = tuple(self.data)
            size = [0] * len(self.data[0])
            for d in self.data:
                for x in range(len(d)):
                    if len(str(d[x])) > size[x]:
                        size[x] = len(str(d[x]))
            self.cols = []
            srch = "N"
            for n, c in enumerate(self.data[0]):
                if self.head:
                    desc = self.head[n]
                else:
                    desc = ""
                if self.typs:
                    if len(self.typs[n]) == 3:
                        typ, siz, srch = self.typs[n]
                    else:
                        typ, siz = self.typs[n]
                    if not siz:
                        siz = size[n]
                else:
                    typ, siz = ("NA", size[n])
                if self.mode.upper() == "M" and n == 0:
                    self.cols.append([str(n), desc, siz, "CB", None])
                else:
                    self.cols.append([str(n), desc, siz, typ, None])
                if self.mode.upper() != "M":
                    if type(c) == tuple or type(c) == list:
                        if len(c) > 3:
                            srch = c[4].upper()
            # Make the last column searchable if no searchable columns set
            if self.mode.upper() == "M":
                if self.selc == None:
                    self.cols[n][4] = "Y"
                else:
                    self.cols[self.selc+1][4] = "Y"
            elif srch == "N":
                self.cols[n][4] = "Y"
            if len(self.data) > 50:
                # Set decoration mode to True
                self.deco = True
            sd = SelectChoice(self.scrn, self.titl, self.cols, self.data,
                sort=False, font=self.mf.fontd, butt=self.butt, size=self.size,
                deco=self.deco)
            if sd.selection:
                if self.mode == "S":
                    if self.retn == "D":
                        self.selection = sd.selection[1:]
                    elif self.retn == "I":
                        self.selection = sd.selection[:1]
                    else:
                        self.selection = None
                elif self.mode == "M":
                    self.selection = []
                    for sel in sd.selection:
                        if self.retn == "D":
                            self.selection.append(sel[2])
                        elif self.retn == "I":
                            self.selection.append(sel[0])
                        else:
                            self.selection = None
                else:
                    self.selection = None
            else:
                self.selection = None
        else:
            self.selection = None

class SelectChoice:
    """
    This class displays a list of available choices.

    scrn   - The widget to display this widget on else None.
    titl   - The title of the widget else None.
    cols   - A list of tuples of columns to display as follows:
                Name       - The column name from ffield
                Heading    - The column heading
                Size       - As per CCD e.g. 7.2
                Type       - Valid CCD type or else:
                                "CB" for a check box
                                ("XX",systp) to use the code in systp
                                    subscripted by the value of 'data' for
                                    this column e.g.  ('crt_type','XX',3,'Typ')
                                    produces 'Pur'
                                ("xx",systp) to use the desc in systp
                                    subscripted by the value of 'data' for this
                                    column e.g. ('crt_type','xx',10,'Desc')
                                    produces 'Purchases'
                Searchable
                    Y or N to allow '<Ctl> F' search
                    F      to allow filtering of data (auto includes Y)
    data   - A list of lists of data as per the cols.
    sort   - Whether the columns can be sorted, defaults to True
    font   - The font being used by the widget in 'scrn' above.
    loop   - Whether to create a main loop, defaults to True.
    comm   - The command to execute with the selected column(s).
    butt   - A list of tuples of buttons, in addition to Quit, to draw,
                with their respective commands and if a selection must
                be returned. A Quit button will override the default one.
    deco   - Whether or not the window must be decorated, default True.
    neww   - Whether to create a new window, defaults to True
    live   - Whether the treeview can have focus, defaults to True
    size   - The size of the display screen or None to calculate
    posn   - The Initial Row to highlight
    fltr   - If filtering is available. This must be a mf class
    """
    def __init__(self, scrn, titl, cols, data, sort=True, font=None, loop=True, comm=None, butt=None, deco=True, neww=True, live=True, size=None, posn=0, fltr=False):
        if scrn:
            self.scrn = scrn
            self.top = scrn.get_toplevel()
        else:
            self.scrn = None
            self.top = None
        if titl:
            self.titl = titl
        else:
            self.titl = "Valid Choices"
        self.cols = list(cols)
        self.data = data
        if not self.cols[0][1]:
            # No column headings so no sort
            self.sort = False
        else:
            self.sort = sort
        self.font = font
        self.loop = loop
        self.comm = comm
        self.butt = butt
        self.deco = deco
        if not self.scrn:
            self.neww = True
        else:
            self.neww = neww
        if self.cols[0][3] == "CB":
            self.live = True
        else:
            self.live = live
        self.size = size
        self.posn = posn
        self.fltr = fltr
        self.selection = []
        self.drawWindow()
        self.showData()
        if self.loop:
            gtk.main()

    def drawWindow(self):
        if self.neww:
            if self.top and self.top.window:
                # Change the cursor of the toplevel to a watch
                self.top.window.set_cursor(gtk.gdk.Cursor(gtk.gdk.WATCH))
            # Create a new window
            mkw = MkWindow(trans=self.top, decor=self.deco, frame=True,
                modal=True, title=self.titl, arrow=True, resiz=True)
            self.window = mkw.newwin
            # create a vbox
            self.vbox = gtk.VBox(False, 0)
            if self.titl and not self.deco:
                # create a label for the title
                if self.top:
                    win = self.top
                else:
                    win = self.window
                font = pango.FontDescription("sans 12")
                evt, lab = doLabel(gtk, win, text=self.titl, font=font)
                self.vbox.pack_start(evt, False, False, 0)
            mkw.newfrm.add(self.vbox)
        else:
            # create a VBox
            self.vbox = gtk.VBox(False, 0)
            self.scrn.pack_start(self.vbox, True, True, 2)
        # Get character width and height and set vbox size
        if self.neww:
            ctxt = self.window.get_pango_context()
        else:
            ctxt = self.scrn.get_pango_context()
        layt = pango.Layout(ctxt)
        layt.set_text("X")
        cwdth, chght = layt.get_pixel_size()
        maxw = int(gtk.gdk.screen_width() * .75)
        maxh = int(gtk.gdk.screen_height() * .75)
        if not self.size:
            width = 0
            for col in self.cols:
                width += int(col[2])
                width += 3
            width = width * cwdth
            if width > maxw:
                width = maxw
            if len(self.data) <= 10:
                height = int(gtk.gdk.screen_height() * .30)
            elif len(self.data) <= 20:
                height = int(gtk.gdk.screen_height() * .60)
            else:
                height = maxh
            self.size = [width, height]
        else:
            if self.size[0] > maxw:
                self.size[0] = maxw
            if self.size[1] > maxh:
                self.size[1] = maxh
        self.vbox.set_size_request(self.size[0], self.size[1])
        # create a scrolled window
        self.sw = gtk.ScrolledWindow()
        self.sw.set_shadow_type(gtk.SHADOW_ETCHED_IN)
        if self.size[0] == -1:
            self.sw.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        else:
            self.sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        self.vbox.pack_start(self.sw, True, True, 0)
        if self.neww and self.loop:
            self.window.connect("mywindo", self.doQuit, "quit")
            # create continue and quit buttons
            bb = gtk.HBox()
            self.vbox.pack_start(bb, False, False, 0)
            if self.butt:
                butt = self.butt
            else:
                butt = []
                if self.cols[0][3] == "CB":
                    butt.append(("Accept", None))
            # Add a Filter Button
            if self.fltr:
                butt.append(("Filter",))
            # Check if a Quit button has been declared
            quit = False
            for but in butt:
                if but[0] in ("Quit", "Exit"):
                    quit = True
            if not quit:
                butt.append(("Quit", self.doQuit))
            tagbut = {"buts": []}    # the dictionary of underline characters
            for num, but in enumerate(butt):
                txt = getUnderline(tagbut, "buts", but[0])[1]
                #===============================================================
                # Attempt to use stock labels and icons for the buttons
                #===============================================================
                nam = gtk.stock_lookup("gtk-%s" % but[0].lower())
                if nam:
                    gtk.stock_add([(nam[0], txt, nam[2], nam[3], nam[4])])
                    stk = nam[0]
                else:
                    stk = None
                #===============================================================
                bt = MyButton(label=txt, stock=stk)
                if but[0] == "Filter":
                    bt.connect("clicked", self.doFilter)
                else:
                    bt.connect("clicked", self.doButton, but[1:])
                bb.pack_start(bt, True, True, 0)

    def showData(self):
        #=======================================================================
        # Create and populate a ListStore
        #=======================================================================
        dtypes = []
        for col in self.cols:
            if col[3] == "CB":
                dtypes.append(bool)
            else:
                dtypes.append(str)
        lstore = gtk.ListStore(*dtypes)
        for idx, row in enumerate(self.data):
            row = list(row)
            for num, dat in enumerate(row):
                row[num] = self.doFormat(dat, self.cols[num])
            lstore.append(row)
        #=======================================================================
        # Create the TreeView
        #=======================================================================
        self.treeview = gtk.TreeView(lstore)
        if self.live and self.cols[0][3] != "CB":
            self.treeview.connect("row-activated", self.doSelect)
        # Check for headers
        if self.cols[0][1]:
            self.treeview.set_headers_visible(True)
        else:
            self.treeview.set_headers_visible(False)
        # Display alternate light and dark rows and vertical grid lines
        self.treeview.set_rules_hint(True)
        try:
            self.treeview.set_grid_lines(gtk.TREE_VIEW_GRID_LINES_VERTICAL)
        except:
            pass
        # Create the TreeViewColumns to display the data
        col = [None] * len(self.cols)
        for n, c in enumerate(self.cols):
            if c[3] == "CB":
                cell = gtk.CellRendererToggle()
                cell.set_property("activatable", True)
                cell.connect_after("toggled", self.colToggled, lstore)
            else:
                cell = gtk.CellRendererText()
            col[n] = gtk.TreeViewColumn(c[1], cell)
            if c[3] == "CB":
                col[n].add_attribute(cell, "active", n)
            else:
                # Format the data
                col[n].add_attribute(cell, "text", n)
                if c[4] and c[4].lower() == "y":
                    # Make treeview searchable on the default column
                    self.treeview.set_search_column(n)
                if self.sort:
                    # Allow sorting on the column
                    col[n].set_sort_column_id(n)
                else:
                    # Allow facility to change the searchable column
                    col[n].set_clickable(True)
                    col[n].connect("clicked", self.doSearch, n)
            self.treeview.append_column(col[n])
        self.sw.add(self.treeview)
        # Check if Live
        if self.live:
            # Focus on the required column
            col = self.treeview.get_column(0)
            self.treeview.scroll_to_cell(self.posn, col,
                use_align=True, row_align=0.2)
            self.treeview.set_cursor(self.posn, col, False)
            self.treeview.grab_focus()
        else:
            # Focus on the last column
            if self.data:
                self.treeview.scroll_to_cell((len(self.data) - 1,))
            self.treeview.set_sensitive(False)
        #=======================================================================
        # Show all the widgets
        #=======================================================================
        if self.neww:
            self.window.show_all()
        else:
            self.scrn.show_all()

    def doFormat(self, data, cols):
        if cols[3] == "CB":
            pass
        elif type(cols[3]) in (list, tuple):
            if cols[3][0] == "XX":
                data = cols[3][1][int(data)-1][0][:int(cols[2])]
            elif cols[3][0] == "xx":
                data = cols[3][1][int(data)-1][1][:int(cols[2])]
        elif data == None:
            data = ""
        elif cols[3] == "TX":
            data = CCD(data, "NA", cols[2]).disp
        else:
            data = CCD(data, cols[3], cols[2]).disp
        return data

    def colToggled(self, cell, path, model):
        model[path][0] = not model[path][0]
        sel = [int(path)] + self.data[int(path)]
        if model[path][0] == True:
            self.selection.append(sel)
        else:
            self.selection.remove(sel)
        return

    def doSearch(self, *args):
        self.treeview.set_search_column(args[1])
        self.treeview.emit("start_interactive_search")

    def doSelect(self, *args):
        self.posn = int(args[1][0])
        self.selection = self.getSelData(self.posn)
        if self.comm:
            self.comm(self.selection)
        else:
            self.closeProcess()

    def doButton(self, widget, args):
        if args and args[0]:
            if len(args) == 2 and args[1]:
                data = self.getSelData(self.treeview.get_cursor()[0][0])
                self.closeProcess()
                args[0](data)
            else:
                self.closeProcess()
                args[0]()
        else:
            self.closeProcess()

    def getSelData(self, pos):
        selection = [pos]
        model, iter = self.treeview.get_selection().get_selected()
        for x, c in enumerate(self.cols):
            if type(c[3]) in (list, tuple):
                typ = model.get_value(iter, x)
                for i, t in enumerate(c[3][1]):
                    if c[3][0] == "XX" and typ == t[0]:
                        break
                    elif c[3][0] == "xx" and typ == t[1]:
                        break
                data = CCD(i+1, "UI", 2).work
            else:
                data = CCD(model.get_value(iter, x), c[3], c[2]).work
            selection.append(data)
        return selection

    def doFilter(self, *args):
        self.fltcol = []
        self.olddata = self.data[:]
        tit = (("Filter Selection"),)
        fld = []
        cnt = 0
        for num, col in enumerate(self.cols):
            if col[4] in ("F", "Y"):
                self.fltcol.append((cnt, num))
                fld.append((("T",0,cnt,0),"I%s" % col[3],col[2],col[1],col[1],
                    "","Y",None,None,None,None))
                cnt += 1
        if not fld:
            return
        tnd = ((self.doFEnd, "n"), )
        txt = (self.doFExit, )
        self.ft = TartanDialog(self.fltr, newscn=(0, 1),
            tops=True, title=tit, eflds=fld, tend=tnd, txit=txt)

    def doFEnd(self, widget=None):
        self.ft.closeProcess()
        data = []
        for col in self.fltcol:
            exec "flt%s = re.compile('%s', re.IGNORECASE)" % (col[0],
                self.ft.t_work[0][0][col[0]])
        for dat in self.olddata:
            matches = []
            for col in self.fltcol:
                exec "matches.append(flt%s.match(dat[col[1]]))" % col[0]
            if not None in matches:
                data.append(dat)
        if not data:
            self.doFExit()
        else:
            if self.neww:
                self.window.destroy()
            else:
                self.vbox.destroy()
            if len(data) == 1:
                data[0].insert(0, 0)
                self.selection = data[0]
                self.closeProcess()
            else:
                self.drawWindow()
                self.data = data[:]
                self.showData()

    def doFExit(self, widget=None):
        self.ft.closeProcess()

    def doQuit(self, *args):
        self.selection = None
        if args and args[2] == "quit":
            self.closeProcess()

    def closeProcess(self):
        if self.loop:
            # If True quit the main loop
            gtk.main_quit()
        if self.neww:
            self.window.destroy()
        else:
            self.vbox.destroy()
        if self.top:
            # Change the cursor of the toplevel back to normal
            self.top.window.set_cursor(None)

class ScrollText:
    """
    A scrolled text widget
    """
    def __init__(self, **args):
        defaults = {
            "scrn": None,
            "width": 800,
            "height": 600,
            "title": "",
            "mess": None,
            "font": "Courier 10",
            "wrap": False,
            "select": False,
            "butt": None,
            "exit": None,
            "horizontal": True,
            "vertical": True}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        if self.horizontal:
            self.horizontal = gtk.POLICY_AUTOMATIC
        else:
            self.horizontal = gtk.POLICY_NEVER
        if self.vertical:
            self.vertical = gtk.POLICY_AUTOMATIC
        else:
            self.vertical = gtk.POLICY_NEVER
        self.doVariables()
        self.doProcess()
        gtk.main()

    def doVariables(self):
        self.buts = [("gtk-close", self.closeProcess)]
        if self.butt:
            self.buts.extend(self.butt)

    def doProcess(self):
        if self.scrn:
            top = self.scrn.get_toplevel()
        else:
            top = None
        win = MkWindow(trans=top, title=self.title, modal=True,
            size=(self.width, self.height, "fixed"), resiz=False,
            decor=True, frame=True)
        self.window = win.newwin
        vbox = gtk.VBox(False, 0)
        win.newfrm.add(vbox)
        scrolled_window = gtk.ScrolledWindow()
        scrolled_window.set_border_width(10)
        scrolled_window.set_policy(self.horizontal, self.vertical)
        textview = gtk.TextView()
        textview.modify_font(pango.FontDescription(self.font))
        textview.set_editable(False)
        textview.set_sensitive(False)
        if self.wrap:
            textview.set_wrap_mode(gtk.WRAP_WORD)
        if self.mess:
            textbuffer = textview.get_buffer()
            textbuffer.set_text(self.mess)
        scrolled_window.add(textview)
        vbox.pack_start(scrolled_window, True, True, 0)
        # create a hbox to hold the buttons
        hbox = gtk.HBox(False, 0)
        vbox.pack_start(hbox, False, False, 10)
        # create the buttons
        for num, but in enumerate(self.buts):
            exec "but%s = MyButton(stock=but[0])" % num
            exec "but%s.connect('clicked', self.execCmd, num)" % num
            exec "but%s.set_flags(gtk.CAN_DEFAULT)" % num
            exec "hbox.pack_start(but%s, True, False, 0)" % num
            exec "but%s.show()" % num
        self.window.show_all()
        if self.select:
            textview.set_sensitive(True)

    def execCmd(self, widget, *args):
        if self.buts[args[0]][1]:
            self.buts[args[0]][1]()
        self.closeProcess()

    def quitProcess(self, widget, *args):
        return True

    def closeProcess(self, widget=None, *args):
        self.window.destroy()
        gtk.main_quit()
        if self.exit:
            self.exit()

class ScrollGrid:
    """
    A scrolled grid widget

    pbar   - A progressbar object
    scrn   - A widget to display the grid on
    width  - The width of the grid if no scrn
    height - The height of the grid if no scrn
    name   - The file and sheet name to create for exporting
    head   - A list of headings comprising:
                 Text
                 Start Col (default is 0 col)
                 End Col   (default is 0 col)
    cols   - A list of column headers comprising:
                 Text
                 Start Col (default is current col)
                 End Col   (default is current col)
    form   - A list of column format details comprising:
                 Format per CCD
                 Size   per CCD
                 Clickable (True or False) - default is False
                 Negative colour
                 Ignore for ULINE[D/S] and TOTAL - Default is False
    data   - A list of data comprising:
                 A type e.g. HEAD, BODY etc.
                 A list of values
                 A line count
    event  - A tuple as follows: ("signal", function)
    rcdic  - The tartan rcdic
    """
    def __init__(self, **args):
        defaults = {
            "pbar": None,
            "scrn": None,
            "width": 800,
            "height": 600,
            "name": "",
            "head": [],
            "cols": [],
            "form": [],
            "data": [],
            "butt": [],
            "event": None,
            "rcdic": None}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        self.doVariables()
        self.doProcess()
        if not self.oldwin:
            gtk.main()

    def doVariables(self):
        self.rows = len(self.head) + len(self.cols) + len(self.data)
        if type(self.butt) == tuple:
            self.butt = list(self.butt)
        self.butt.append(("Export to Spreadsheet", self.exportSheet))
        self.butt.append(("gtk-close", self.closeProcess))

    def doProcess(self):
        # main vbox
        if self.scrn:
            modal = True
            self.oldwin = self.scrn.get_toplevel()
            self.width, self.height = self.oldwin.get_size()
        else:
            modal = False
            self.oldwin = None
        self.newwin = MkWindow(title=self.name, modal=modal, decor=True,
            size=(self.width, self.height, "fixed"), frame=False,
            remov=True).newwin
        self.newwin.get_settings().set_string_property("gtk-font-size",
            "Courier 10", "")
        newscrn = gtk.VBox(False, 0)
        self.newwin.add(newscrn)
        newscrn.set_size_request(self.width, self.height)
        # main headings
        hfont = pango.FontDescription("Courier 12")
        colr = newscrn.get_style().copy().bg[gtk.STATE_ACTIVE]
        for head in self.head:
            if head in (list, tuple):
                head = head[0]
            ebox = gtk.EventBox()
            ebox.modify_bg(gtk.STATE_NORMAL, colr)
            newscrn.pack_start(ebox, False, False, 0)
            labl = gtk.Label()
            labl.modify_font(hfont)
            ebox.add(labl)
            if head:
                labl.set_markup("<b>%s</b>" % head.replace("&", "&amp;"))
        # create main boxes
        hbox = gtk.HBox(False, 0)
        newscrn.pack_start(hbox, True, True, 0)
        vboxa = gtk.VBox(False, 0)
        hbox.pack_start(vboxa, False, False, 0)
        vboxb = gtk.VBox(False, 0)
        hbox.pack_start(vboxb, True, True, 0)

        # create a detail heading scrolled window
        heada_win = gtk.ScrolledWindow()
        heada_win.set_policy(gtk.POLICY_NEVER, gtk.POLICY_NEVER)
        vboxa.pack_start(heada_win, False, False, 0)
        self.heada_tab = gtk.Table(len(self.cols), columns=1, homogeneous=True)
        heada_win.add_with_viewport(self.heada_tab)

        # create a detail body scrolled window
        bodya_win = gtk.ScrolledWindow()
        self.bodya_adj = bodya_win.get_vadjustment()
        bodya_win.set_policy(gtk.POLICY_NEVER, gtk.POLICY_NEVER)
        vboxa.pack_start(bodya_win, True, True, 0)
        self.bodya_tab = gtk.Table(rows=self.rows, columns=1, homogeneous=True)
        bodya_win.add_with_viewport(self.bodya_tab)

        # create a column heading scrolled window
        headb_win = gtk.ScrolledWindow()
        self.headb_adj = headb_win.get_hadjustment()
        headb_win.set_policy(gtk.POLICY_NEVER, gtk.POLICY_NEVER)
        vboxb.pack_start(headb_win, False, False, 0)
        self.headb_tab = gtk.Table(len(self.cols), columns=len(self.cols[0])-1,
            homogeneous=False)
        headb_win.add_with_viewport(self.headb_tab)

        # create a column body scrolled window
        bodyb_win = gtk.ScrolledWindow()
        bodya_adj = bodyb_win.get_vadjustment()
        bodya_adj.connect("notify::value", self.doVAdjustment)
        bodyb_adj = bodyb_win.get_hadjustment()
        bodyb_adj.connect("notify::value", self.doHAdjustment)
        bodyb_win.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
        vboxb.pack_start(bodyb_win, True, True, 0)
        self.bodyb_tab = gtk.Table(rows=self.rows, columns=len(self.cols[0])-1,
            homogeneous=False)
        bodyb_win.add_with_viewport(self.bodyb_tab)

        # populate the tables
        # column headings
        bfont = pango.FontDescription("Courier 10")
        for y, col in enumerate(self.cols):
            for x, c in enumerate(col):
                if type(c) == str:
                    nc = [c, x, x]
                else:
                    nc = c
                if y == len(self.cols) - 1 and x == len(col) - 1:
                    chrs = int(self.form[x][1]) + 3
                else:
                    chrs = int(self.form[x][1])
                fmt = "%s-%ss" % ("%", chrs)
                head = gtk.Label()
                head.modify_font(bfont)
                head.set_markup(fmt % nc[0].replace("&", "&amp;"))
                if x == 0:
                    self.heada_tab.attach(head, nc[1], nc[2]+1, y, y+1, 0,0,0,0)
                else:
                    self.headb_tab.attach(head, nc[1]-1, nc[2], y, y+1, 0,0,0,0)
                for z in range(nc[1], nc[2]+1):
                    exec "self.hlabl_%s_%s = head" % (y, z)
        self.hlabl_x = z + 1

        # column data
        col1 = gtk.gdk.color_parse("white")
        col2 = newscrn.get_style().copy().bg[gtk.STATE_NORMAL]
        colr = col2
        for y, dat in enumerate(self.data):
            if self.pbar:
                self.pbar.displayProgress()
            if colr == col2:
                colr = col1
            else:
                colr = col2
            for x, d in enumerate(dat[1]):
                ebox = gtk.EventBox()
                ebox.modify_bg(gtk.STATE_NORMAL, colr)
                if type(d) in (str, unicode):
                    val = CCD(d, "NA", self.form[x][1])
                else:
                    val = CCD(d, self.form[x][0], self.form[x][1])
                labl = gtk.Label(val.disp)
                if len(self.form[x]) == 4 and val.work < 0:
                    labl.modify_fg(gtk.STATE_NORMAL,
                        gtk.gdk.color_parse(self.form[x][3]))
                labl.modify_font(bfont)
                ebox.add(labl)
                if self.form[x][2]:
                    if not self.event:
                        ebox.connect("button_release_event", self.execCmd, y, x)
                    else:
                        ebox.connect(self.event[0], self.event[1], dat[2], x)
                if x == 0:
                    self.bodya_tab.attach(ebox, x, x+1, y, y+1, 0, 0, 0, 0)
                else:
                    self.bodyb_tab.attach(ebox, x-1, x, y, y+1, 0, 0, 0, 0)
                exec "self.blabl_%s_%s = labl" % (y, x)
                exec "self.bcell_%s_%s = ebox" % (y, x)
        self.blabl_x = x + 1
        self.blabl_y = y + 1
        # create a hbox to hold the buttons
        hbox = gtk.HBox(False, 0)
        # create a label and spinbox for the font
        newscrn.pack_start(hbox, False, False, 10)
        sbl = gtk.Label("Font Size")
        sbl.set_alignment(0, 0.5)
        hbox.pack_start(sbl, False, True, 0)
        sba = gtk.Adjustment(10, 6, 12, 1, 0, 0)
        sbw = gtk.SpinButton(sba, 0, 0)
        sbw.set_wrap(True)
        hbox.pack_start(sbw, False, True, 0)
        sba.connect("value_changed", self.doSetFont, sbw)
        # create the buttons
        for num, but in enumerate(self.butt):
            exec "but%s = MyButton(stock=but[0])" % num
            exec "but%s.connect('clicked', self.execCmd, num)" % num
            exec "but%s.set_flags(gtk.CAN_DEFAULT)" % num
            exec "hbox.pack_start(but%s, True, True, 0)" % num
            exec "but%s.show()" % num
        self.newwin.show_all()
        if self.pbar:
            self.pbar.closeProgress()

    def doSetFont(self, *args):
        while gtk.events_pending():
            gtk.main_iteration()
        size = args[1].get_value_as_int()
        font = pango.FontDescription("Courier %s" % size)
        for y, col in enumerate(self.cols):
            for x in range(0, self.hlabl_x):
                exec "self.hlabl_%s_%s.modify_font(font)" % (y, x)
        for y in range(0, self.blabl_y):
            for x in range(0, self.blabl_x):
                try:
                    exec "self.blabl_%s_%s.modify_font(font)" % (y, x)
                except:
                    pass
        self.doResize()

    def doVAdjustment(self, *args):
        self.bodya_adj.set_value(args[0].get_value())

    def doHAdjustment(self, *args):
        self.headb_adj.set_value(args[0].get_value())

    def hideCol(self, col):
        for row in range(0, self.blabl_y):
            try:
                exec "self.hlabl_%s_%s.hide()" % (row, col)
            except:
                pass
            try:
                exec "self.bcell_%s_%s.hide()" % (row, col)
            except:
                pass
        self.doResize()

    def showCol(self, col):
        for row in range(0, self.blabl_y):
            try:
                exec "self.hlabl_%s_%s.show()" % (row, col)
            except:
                pass
            try:
                exec "self.bcell_%s_%s.show()" % (row, col)
            except:
                pass
        self.doResize()

    def doResize(self):
        self.heada_tab.queue_resize()
        self.headb_tab.queue_resize()
        self.bodya_tab.queue_resize()
        self.bodyb_tab.queue_resize()

    def execCmd(self, widget, *args):
        if self.butt[args[0]][1]:
            self.butt[args[0]][1]()

    def exportSheet(self, widget=None, *args):
        doWriteExport(xtype="X", name=self.name, sheet=self.name,
            heads=self.head, colsh=self.cols, forms=self.form,
            datas=self.data, rcdic=self.rcdic)

    def closeProcess(self, widget=None, *args):
        self.newwin.destroy()
        if self.oldwin:
            self.oldwin.show()
        # There must be a main loop in the calling class/module
        gtk.main_quit()

class SplashScreen:
    """
    A splash screen widget
    """
    def __init__(self, scrn, text=None):
        self.scrn = scrn
        self.text = text
        if not self.scrn:
            print self.text
        else:
            self.doSplash()

    def doSplash(self):
        self.alw = gtk.Alignment(0.5, 0.5, 0, 0)
        self.frame = gtk.Frame()
        self.alw.add(self.frame)
        self.frame.set_border_width(5)
        self.frame.set_shadow_type(gtk.SHADOW_ETCHED_OUT)
        if not self.text:
            self.text = """Welcome to TARTAN Financial Systems,


Please enjoy yourself while you wait ..."""
        evt, self.label = doLabel(gtk, self.scrn.get_toplevel(),
            text=self.text, font=pango.FontDescription("Sans bold 12"))
        self.frame.add(evt)
        self.alw.show_all()
        self.scrn.pack_start(self.alw, True, True, 0)
        while gtk.events_pending():
            gtk.main_iteration()

    def closeSplash(self):
        if self.scrn:
            self.alw.destroy()
            while gtk.events_pending():
                gtk.main_iteration()

class ProgressBar:
    """
    A progress bar widget where:

        typ -   The type of progressbar
                    P = Printing the Report
                    G = Generating the Report (Pulse Step)
                    F = Updating Files
        stp -   With typ G (Pulse) this is the step rate.
        cnt -   With typ G (Pulse) this is the number of times between pulses.
        max -   With typ P and F this is the maximum number of records.
        esc -   Whether to allow the escape key to stop the action.
        new -   Create a new window.
        col -   Background colour.
    """
    def __init__(self, scrn, **args):
        self.scrn = scrn
        defaults = {
            "typ": "P",
            "stp": .1,
            "cnt": None,
            "max": None,
            "esc": None,
            "new": None,
            "col": "white"}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        if self.new:
            top = self.scrn.get_toplevel()
            win = MkWindow(trans=top, size=(600, -1, "fixed"), resiz=False,
                decor=False, frame=True)
            self.win = win.newwin
            self.scrn = win.newfrm
            self.win.show_all()
        else:
            self.win = self.scrn.get_toplevel()
        if type(self.typ) == str:
            self.typ = [self.typ, ""]
        if self.typ[0] == "G":
            if self.typ[1]:
                self.mess = self.typ[1]
            else:
                self.mess = "Generating the Report"
        elif self.typ[0] == "P":
            if self.typ[1]:
                self.mess = self.typ[1]
            else:
                self.mess = "Printing the Report"
        elif self.typ[0] == "F":
            if self.typ[1]:
                self.mess = self.typ[1]
            else:
                self.mess = "Updating Files"
        else:
            self.mess = self.typ[0]
        self.qty = 0
        self.can = "no"
        self.old = None
        self.alw = gtk.Alignment(0.5, 0.5, 0, 0)
        self.alw.show()
        self.scrn.add(self.alw)
        self.progBar = gtk.ProgressBar()
        self.progBar.set_size_request(600, -1)
        self.progBar.set_orientation(0)  # left to right
        if self.esc:
            try:
                self.win.connect("mywindo", self.doQuit)
                self.mess = "%s, <Esc> to Abort" % self.mess
            except:
                self.esc = False
        self.alw.add(self.progBar)
        self.alw.show_all()
        if self.typ[0] == "G":
            self.progBar.set_pulse_step(self.stp)
        else:
            self.progBar.set_fraction(0)

    def doQuit(self, *args):
        self.can = "yes"

    def displayProgress(self, done=None):
        if self.typ[0] == "G":
            if self.cnt and self.qty % self.cnt:
                self.qty += 1
                return
            self.qty += 1
            self.progBar.set_text("%s, Please Wait ....." % self.mess)
            self.progBar.pulse()
        else:
            frac = int(((done + 1) * 100.0) / self.max)
            cent = round(frac / 100.0, 2)
            if cent > 1.0:
                cent = 1.0
            if cent != self.old:
                self.old = cent
                self.progBar.set_fraction(cent)
                self.progBar.set_text("%s (%s%s)" % (self.mess, str(frac), "%"))
        while gtk.events_pending():
            gtk.main_iteration()
        if self.esc and self.typ[0] != "G":
            return self.can

    def closeProgress(self):
        self.alw.destroy()
        if self.new:
            self.win.destroy()

class MkWindow:
    """
    Valid Arguments:
        name    The name of the window
        icon    The icon for the window
        title   The title of the window
        size    The size of the window and type as a tuple type can be
                    default or fixed
        resiz   Whether the window is resizable
        trans   The window which this window is transient to
        decor   Whether this window has WM decorations
        frame   Whether this window has a frame
        bordr   The border width
        modal   Whether this window is modal
        remov   Whether this window is removable
        place   Where to place the window e.g. float, center, mouse
        arrow   Whether or not to enable the arrow keys (default is disable)
        ipath   The path to additional icons to be added to the icon factory
    """
    def __init__(self, **args):
        for arg in ("name", "icon", "title", "size", "bordr", "resiz", "trans",
                "decor", "frame", "modal", "remov", "place", "arrow", "ipath"):
            if arg in args:
                exec "self.%s = args[arg]" % arg
            else:
                exec "self.%s = None" % arg
        if not self.place:
            self.place = "C"
        if not self.resiz:
            self.resiz = True
        self.drawWindow()
        self.setAttributes()
        self.newwin.realize()

    def drawWindow(self):
        self.newwin = MyWindow(type=gtk.WINDOW_TOPLEVEL)
        self.newwin.connect("destroy", self.doDestroyNo)
        self.newwin.connect("delete_event", self.doDestroyYes)
        self.newwin.connect("key_press_event", self.doKeyPress)

    def setAttributes(self):
        if self.name:
            self.newwin.set_name(self.name)
        if self.icon:
            self.newwin.set_icon_from_file(self.icon)
        if self.trans:
            style = self.trans.get_style().copy()
            self.newwin.set_style(style)
            self.newwin.set_transient_for(self.trans)
            self.newwin.set_destroy_with_parent(True)
        if self.decor:
            self.newwin.set_decorated(self.decor)
            if self.title:
                self.newwin.set_title(self.title)
            else:
                self.newwin.set_title("Tartan Systems")
        else:
            self.newwin.set_decorated(False)
        if self.size:
            if len(self.size) == 3 and self.size[2] == "fixed":
                self.newwin.set_size_request(self.size[0], self.size[1])
                self.newwin.set_resizable(self.resiz)
            else:
                self.newwin.set_default_size(self.size[0], self.size[1])
        else:
            self.newwin.set_resizable(self.resiz)
        if self.bordr:
            self.newwin.set_border_width(self.bordr)
        if self.frame:
            self.newfrm = gtk.Frame()
            self.newfrm.set_border_width(5)
            self.newfrm.set_shadow_type(gtk.SHADOW_OUT)
            self.newwin.add(self.newfrm)
        if self.modal:
            self.newwin.set_modal(True)
        if self.trans:
            self.newwin.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
        elif self.place.lower() == "n":
            self.newwin.set_position(gtk.WIN_POS_NONE)
        elif self.place.lower() == "m":
            self.newwin.set_position(gtk.WIN_POS_MOUSE)
        else:
            self.newwin.set_position(gtk.WIN_POS_CENTER_ALWAYS)
        # Load Customised and New Icons
        items = [
            ("gtk-show", "_Show", 0, 0, None),
            ("gtk-exit", "E_xit", 0, 0, None),
            ("gtk-reset", "_Reset", 0, 0, None),
            ("gtk-accept", "_Accept", 0, 0, None),
            ("gtk-update", "_Update", 0, 0, None),
            ("gtk-drsquery", "_DrsQuery", 0, 0, None),
            ("gtk-strquery", "_StrQuery", 0, 0, None),
            ("gtk-restore", "_Restore Database", 0, 0, None),
            ("gtk-manual", "_Tartan Manual", 0, 0, None),
            ("gtk-reprint", "_Reprint", 0, 0, None),
            ("gtk-backup", "_Backup Database", 0, 0, None)
            ]
        aliases = [
            ("gtk-show", gtk.STOCK_JUSTIFY_FILL),
            ("gtk-exit", gtk.STOCK_QUIT),
            ("gtk-reset", gtk.STOCK_CLEAR),
            ("gtk-accept", gtk.STOCK_APPLY),
            ("gtk-update", gtk.STOCK_APPLY),
            ("gtk-drsquery", gtk.STOCK_DIALOG_QUESTION),
            ("gtk-strquery", gtk.STOCK_DIALOG_QUESTION),
            ("gtk-restore", gtk.STOCK_REVERT_TO_SAVED),
            ("gtk-manual", gtk.STOCK_DIALOG_INFO),
            ("gtk-reprint", gtk.STOCK_PRINT),
            ("gtk-backup", gtk.STOCK_SAVE)
            ]
        gtk.stock_add(items)
        factory = gtk.IconFactory()
        factory.add_default()
        style = self.newwin.get_style().copy()
        for stock, alias in aliases:
            iset = style.lookup_icon_set(alias)
            factory.add(stock, iset)
        if self.ipath:
            newicons = [
                ("gtk-contacts", "_Contacts", "users.png"),
                ("gtk-deladd", "_DelAdd", "truck.png"),
                ("gtk-drsmaint", "_DrsMaint", "edit.png"),
                ("gtk-edit", "_Edit", "edit.png"),
                ("gtk-email", "_Email", "email.png"),
                ("gtk-filter", "_Filter", "filter.png"),
                ("gtk-import", "_Import", "import.png"),
                ("gtk-message", "_Message", "info.png"),
                ("gtk-notes", "_Notes", "edit.png"),
                ("gtk-preview", "_Preview", "preview.png"),
                ("gtk-ribbon", "_Ribbon", "ribbon.png"),
                ("gtk-send", "_Send", "send.png"),
                ("gtk-strmaint", "_StrMaint", "edit.png"),
                ("gtk-users", "_Users", "users.png")]
            for name, text, icon in newicons:
                gtk.stock_add([(name, text, 0, 0, "gtk20")])
                ipath = os.path.join(self.ipath, icon)
                if os.path.isfile(ipath):
                    iset = gtk.IconSet(gtk.gdk.pixbuf_new_from_file(ipath))
                    factory.add(name, iset)

    def doKeyPress(self, widget, event):
        # Disable the Tab, Up and Down keys from progressing to the next widget
        keyname = gtk.gdk.keyval_name(event.keyval)
        if keyname == "Tab":
            return True
        if not self.arrow and keyname in ("Up", "Down"):
            return True

    def doDestroyNo(self, widget=None):
        return True

    def doDestroyYes(self, widget=None, data=None):
        if not self.remov:
            return True
        ok = askQuestion(screen=self.newwin, head="Destroy Window",
            mess="Are you Sure you want to Quit this Window!\n"\
            "It is Possible that Some Data Could Be Lost.\n"\
            "It is Far Better to Exit Normally.", default="no")
        if ok == "no":
            return True
        try:
            sys.stdout.close()
        except:
            pass
        sys.exit()

class ShowEmail:
    def __init__(self, window=False, message=None, embed=None, attach=None):
        mkw = MkWindow(decor=True, modal=True, trans=window, size=(800, -1))
        vbox = gtk.VBox(False, 0)
        mkw.newwin.add(vbox)
        if message or embed:
            mkw.newwin.resize(824, 600)
        if message:
            sw1 = gtk.ScrolledWindow()
            sw1.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            text = gtk.TextView()
            text.set_editable(False)
            text.set_sensitive(False)
            text.set_wrap_mode(gtk.WRAP_WORD)
            buff = text.get_buffer()
            buff.set_text(message)
            sw1.add(text)
            vbox.pack_start(sw1, True, True, 0)
        if embed:
            sw2 = gtk.ScrolledWindow()
            sw2.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            tb = gtk.Table(len(embed), 1)
            for r, image in enumerate(embed):
                si = ShowImage(None, image, msiz=800)
                tb.attach(si.image, 0, 1, r, r+1, 0, gtk.EXPAND|gtk.FILL, 0, 5)
            sw2.add_with_viewport(tb)
            vbox.pack_start(sw2, True, True, 0)
        if attach:
            evt, hed = doLabel(gtk, mkw.newwin, style="out", color=True,
                text="Attachment: %s" % attach)
            vbox.pack_start(evt, False, False, 0)
        butt = MyButton("Close")
        butt.connect("clicked", self.doClose, mkw.newwin)
        vbox.pack_start(butt, False, False, 0)
        mkw.newwin.show_all()
        gtk.main()

    def doClose(self, *args):
        for arg in args:
            arg.destroy()
        gtk.main_quit()

class ShowImage:
    def __init__(self, vbox, flenam, msiz=400, flip=None):
        if flip == "Y":
            flenam = flipImage(flenam)
            if not flenam:
                return
        self.image = gtk.Image()
        pixbuf = gtk.gdk.pixbuf_new_from_file(flenam)
        width = pixbuf.get_width()
        height = pixbuf.get_height()
        relate = float(height / (width * 1.0))
        if width > height:
            width = msiz
            height = int(width * relate)
        else:
            height = msiz
            width = int(height / relate)
        scaled_buf = pixbuf.scale_simple(width, height, gtk.gdk.INTERP_BILINEAR)
        self.image.set_from_pixbuf(scaled_buf)
        if vbox:
            vbox.pack_start(self.image, False, False, 0)
            self.image.show()

class TartanCalendar:
    def __init__(self, screen=None, select=False, quit=True):
        # Create alignment
        self.alw = gtk.Alignment(0.5, 0.5, 0, 0)
        self.alw.set_padding(0, 0, 5, 5)
        if type(screen) == gtk.VBox:
            self.screen = screen
            self.screen.pack_start(self.alw)
        else:
            top = screen.get_toplevel()
            mkw = MkWindow(decor=False, modal=True, trans=top)
            self.screen = mkw.newwin
            self.screen.show_all()
            self.screen.add(self.alw)
        # Create vbox
        vbx = gtk.VBox(False, 0)
        self.alw.add(vbx)
        cal = gtk.Calendar()
        cal.display_options(
            gtk.CALENDAR_SHOW_HEADING | \
            gtk.CALENDAR_SHOW_DAY_NAMES | \
            gtk.CALENDAR_SHOW_WEEK_NUMBERS |
            gtk.CALENDAR_WEEK_START_MONDAY)
        cal.modify_font(pango.FontDescription("Sans 12"))
        if select:
            cal.set_sensitive(True)
            cal.connect("key_press_event", self.selectDay)
            cal.connect("day-selected-double-click", self.selectDay)
        vbx.pack_start(cal, False, False, 0)
        if quit:
            bt2 = MyButton(stock="gtk-quit")
            bt2.connect("clicked", self.closeCalendar)
            vbx.pack_start(bt2)
        self.alw.show_all()
        cal.grab_focus()
        self.date = 0
        gtk.main()

    def selectDay(self, widget=None, keys=None):
        if keys:
            key = gtk.gdk.keyval_name(keys.keyval)
            if key == "Escape":
                self.date = None
                self.closeCalendar()
            if key in ("KP_Enter", "Return"):
                e = gtk.gdk.Event(gtk.gdk.KEY_PRESS)
                e.keyval = gtk.keysyms.space
                widget.emit("key-press-event", e)
            else:
                return
        date = widget.get_date()
        self.date = (date[0] * 10000) + ((date[1] + 1) * 100) + date[2]
        self.closeCalendar()

    def closeCalendar(self, widget=None):
        self.alw.destroy()
        if type(self.screen) != gtk.VBox:
            self.screen.destroy()
        gtk.main_quit()

class TartanConfig:
    """
    This class is used to create and modify the tartanrc file
    """
    def __init__(self, **args):
        defaults = {
            "mf": None,
            "rcfile": None,
            "rcdic": None,
            "level": 3}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        self.setVariables()
        self.drawDialog()
        self.mf.startLoop()

    def setVariables(self):
        if not self.rcdic:
            if self.rcfile:
                self.rcfile = os.path.normpath(self.rcfile)
            self.rcdic = loadRcFile(rcfile=self.rcfile, default=True)
            self.rcfile = os.path.normpath(self.rcdic["name"])
        else:
            self.rcfile = os.path.normpath(self.rcdic["name"])
            self.rcdic = loadRcFile(self.rcfile)
        self.fmnam = self.rcdic["mft"]
        self.fmsiz = self.rcdic["mfs"]
        self.fdnam = self.rcdic["dft"]
        self.fdsiz = self.rcdic["dfs"]
        self.edit = False
        self.click = []
        if self.mf:
            self.nmf = False
        else:
            self.nmf = True
            self.mf = MainFrame(self.rcdic, dbobj=False)

    def drawDialog(self):
        tit = ("Tartan Configuration",)
        dir = {
            "stype":  "F",
            "types":  "dir",
            "initd":  ""}
        if sys.platform == "win32":
            typ = "*.exe"
        else:
            typ = "*"
        fle = {
            "stype":  "F",
            "types":  "fle",
            "ftype":  (("Executable", "%s" % typ),)}
        self.loadFonts()
        aft = {
            "stype": "C",
            "titl": "Available Fonts",
            "data": self.aft}
        fft = {
            "stype": "C",
            "titl": "Available Fonts",
            "data": self.fft}
        self.loadThemes()
        thm = {
            "stype": "C",
            "titl": "Available Themes",
            "data": self.thm}
        self.rcfile = os.path.normpath(self.rcdic["name"])
        self.dbdir = os.path.normpath(self.rcdic["dbdir"])
        r1s = (
            ("PostgreSQL", "P"),
            ("SQLite", "S"),
            ("MySQL", "M"),
            ("Firebird", "F"),
            ("MsSQL 2000", "Q"))
        r2s = (("None", "N"), ("Centre", "C"), ("Mouse", "M"))
        r3s = (("Yes", "Y"), ("No", "N"))
        r4s = (("Yes", "Y"), ("No", "N"), ("Multimedia", "M"))
        fld = (
            (("T",0,0,0),"IFF",50,"Configuration File","",
                self.rcfile,"N",self.doRcFile,None,None,None,None,
                "This is the main configuration file for a Tartan "\
                "Installation. There can be as many of these as you "\
                "desire, each one accessing a different set of choices. "\
                "This file is loaded by the ms0000 module and the default "\
                "file can be altered with the -r option. Please refer "\
                "to the manual for more detailed information."),
            (("T",1,0,0),("IRB",r1s),0,"Database Engine","",
                "S","N",None,None,None,None),
            (("T",1,1,0),"INA",30,"Database Name","",
                self.rcdic["dbname"],"N",None,None,None,("notblank",)),
            (("T",1,2,0),"INA",30,"Host Name","",
                "localhost","N",None,None,None,("notblank",)),
            (("T",1,3,0),"IFF",50,"Files Directory","",
                self.rcdic["dbdir"],"N",self.doDir,dir,None,None,None),
            (("T",1,4,0),"IUI",4,"Port Number","",
                "","N",None,None,None,("efld",)),
            (("T",1,5,0),"INA",30,"Administrator","",
                self.rcdic["dbuser"],"N",None,None,None,("notblank",),None,
                "The Database Administrator."),
            (("T",1,6,0),"IHA",30,"Password","",
                self.rcdic["dbpwd"],"N",None,None,None,None,None,
                "The Database Administrator's Password."),
            (("T",2,0,0),"IFF",50,"Program Path","",
                self.rcdic["prgdir"],"N",self.doPrg,dir,None,("dir",),None,
                "The Directory where the Tartan programs are situated."),
            (("T",2,1,0),"IFF",50,"Archive Path","",
                self.rcdic["arcdir"],"N",self.doDir,dir,None,None,None,
                "The Directory where the archived reports are stored."),
            (("T",2,2,0),"IFF",50,"Backup Path","",
                self.rcdic["bupdir"],"N",self.doDir,dir,None,None,None,
                "The Directory where all backup's are stored."),
            (("T",2,3,0),"IFF",50,"Work Path","",
                self.rcdic["wrkdir"],"N",self.doDir,dir,None,None,None,
                "The Directory where all work files are stored."),
            (("T",2,4,0),"IFF",50,"Upgrade Path","",
                self.rcdic["upgdir"],"N",self.doDir,dir,None,None,None,
                "The Directory where System Upgrades are stored."),
            (("T",2,5,0),"IFF",50,"PDF Viewer","",
                self.rcdic["vwr"],"N",None,fle,None,("cmd",),None,
                "The Path of the program used to view PDF files."),
            (("T",2,6,0),"IFF",50,"Print Command","",
                self.rcdic["prn"],"N",None,fle,None,("cmd",),None,
                "The Path of the program to use for printing reports."),
            (("T",2,7,0),"IFF",50,"PS Converter","",
                self.rcdic["cnv"],"N",None,fle,None,None,None,
                "The Path of the program to use for converting PDF to PS."),
            (("T",2,8,0),"IFF",50,"CSV Reader","",
                self.rcdic["cnv"],"N",self.doExport,fle,None,None,None,
                "The Path of the program to use for reading CSV files e.g. "\
                "c:\Program Files\Microsoft Office\Office10\EXCEL.EXE"),
            (("T",2,9,0),"INA",10,"Screen Geometry","Screen Geometry",
                self.rcdic["geo"],"N",self.doGeometry,None,None,("notblank",)),
            (("T",2,10,0),("IRB",r2s),0,"Screen Placement","Screen Placement",
                self.rcdic["plc"],"N",None,None,None,None),
            (("T",2,11,0),("IRB",r3s),0,"Enforce Confirm","",
                self.rcdic["cnf"].upper(),"N",None,None,None,None),
            (("T",2,12,0),("IRB",r3s),0,"Auto-completion","",
                self.rcdic["acnf"].upper(),"N",None,None,None,None),
            (("T",2,13,0),("IRB",r3s),0,"Tool-tips","",
                self.rcdic["ttip"].upper(),"N",None,None,None,None),
            (("T",2,14,0),("IRB",r3s),0,"Text Viewer","",
                self.rcdic["txtv"].upper(),"N",None,None,None,None),
            (("T",2,15,0),("IRB",r3s),0,"Entry Frame","",
                self.rcdic["efrm"].upper(),"N",None,None,None,None),
            (("T",2,16,0),("IRB",r4s),0,"Error Alarm","",
                self.rcdic["errs"].upper(),"N",None,None,None,None),
            (("T",3,0,0),"INA",30,"Menu Font Name","",
                self.rcdic["mft"],"N",self.doFtNam,aft,None,("in",self.aft)),
            (("T",3,0,0),"IUI",2,"Size","Menu Font Size",
                self.rcdic["mfs"],"N",self.doFtSiz,None,None,("between",7,14)),
            (("T",3,1,0),"OTV",30,"Preview"),
            (("T",3,2,0),"INA",30,"Default Font Name","",
                self.rcdic["dft"],"N",self.doFtNam,fft,None,("in",self.fft)),
            (("T",3,2,0),"IUI",2,"Size","Default Font Size",
                self.rcdic["dfs"],"N",self.doFtSiz,None,None,("between",7,14)),
            (("T",3,4,0),"OTV",30,"Preview"),
            (("T",3,3,0),"INA",30,"Theme","",
                self.rcdic["theme"],"N",self.doTheme,thm,None,("in",self.thm)))
        tag = (
            ("DataBase",None,("T",1,1),("T",3,0)),
            ("General",None,("T",1,1),("T",3,0)),
            ("Dialog",None,("T",1,1),("T",3,0)))
        but = (
            ("Save", None, self.doSave, 1, ("T",1,1), ("T",0,1)),
            ("Cancel", None, self.doCancel, 1, None, None))
        tnd = (
            (self.doEnd, "n"),
            (self.doEnd, "n"),
            (self.doEnd, "n"),
            (self.doEnd, "y"))
        txt = (
            self.doExit,
            self.doExit,
            self.doExit,
            self.doExit)
        self.df = TartanDialog(self.mf, screen=self.mf.body, title=tit,
            eflds=fld, tags=tag, butt=but, tend=tnd, txit=txt,
            clicks=self.doClick)
 
    def doRcFile(self, frt, pag, r, c, p, i, w):
        if self.level < 3 and w != self.rcfile:
            return "Invalid File"
        self.rcfile = w
        self.edit = True

    def doClick(self, *opts):
        if not self.edit:
            return
        if opts == (0, 0):
            # First Field, Configuration File Name
            return
        if self.click:
            self.df.loadEntry("T", self.click[0], self.click[1],
                data=self.df.t_work[self.click[0]][0][self.click[1]])
        self.click = opts[2]
        self.df.focusField("T", self.click[0], (self.click[1] + 1))

    def doDir(self, frt, pag, r, c, p, i, w):
        if not os.path.isdir(w):
            ok = askQuestion(screen=self.mf.body, head="Missing Directory",
                mess="Directory %s Does Not Exist, Would You Like to Create "\
                "it Now?" % w, default="yes")
            if ok == "no":
                return "Invalid Directory"
            try:
                os.makedirs(w)
            except:
                return "Cannot Create %s Directory" % w

    def doPrg(self, frt, pag, r, c, p, i, w):
        if sys.platform == "win32":
            tst1 = os.path.normpath(os.path.join(w, "ms0000.py"))
            tst2 = os.path.normpath(os.path.join(w, "ms0000.exe"))
            if not os.path.isfile(tst1) and not os.path.isfile(tst2):
                return "Missing %s or %s" % (tst1, tst2)

    def doConvrt(self, frt, pag, r, c, p, i, w):
        if w and not os.path.isfile(w):
            return "Invalid Converter"

    def doExport(self, frt, pag, r, c, p, i, w):
        if w and not os.path.isfile(w):
            return "Invalid Exporter"

    def doGeometry(self, frt, pag, r, c, p, i, w):
        geo = w.split("x")
        if len(geo) != 2:
            return "Invalid Geometry"
        if int(geo[0]) >= 1280:
            self.df.loadEntry("T", 3, 1, data=12)
            self.df.loadEntry("T", 3, 4, data=14)
        elif int(geo[0]) >= 1024:
            self.df.loadEntry("T", 3, 1, data=10)
            self.df.loadEntry("T", 3, 4, data=12)
        elif int(geo[0]) >= 800:
            self.df.loadEntry("T", 3, 1, data=8)
            self.df.loadEntry("T", 3, 4, data=10)
        else:
            self.df.loadEntry("T", 3, 1, data=7)
            self.df.loadEntry("T", 3, 4, data=8)

    def doFtNam(self, frt, pag, r, c, p, i, w):
        if c == 1:
            pos = 2
            self.fmnam = w
            size = self.df.t_work[pag][0][c]
            text = "Menu Font - Size %s" % size
        else:
            pos = 5
            self.fdnam = w
            size = self.df.t_work[pag][0][c]
            text = "Body Font - Size %s" % size
        self.df.loadEntry(frt, pag, pos, data=text)
        widget, self.fmsiz = self.df.getEntry(frt, pag, pos)
        widget.modify_font(pango.FontDescription("%s %s" % (w, size)))

    def doFtSiz(self, frt, pag, r, c, p, i, w):
        if c == 2:
            pos = 2
            self.fmsiz = w
            font = self.fmnam
            text = "Menu Font - Size %s" % w
        else:
            pos = 5
            self.fdsiz = w
            font = self.fdnam
            text = "Body Font - Size %s" % w
        self.df.loadEntry(frt, pag, pos, data=text)
        widget, self.fmsiz = self.df.getEntry(frt, pag, pos)
        widget.modify_font(pango.FontDescription("%s %s" % (font, w)))

    def doTheme(self, frt, pag, r, c, p, i, w):
        fontm = "%s %s" % (self.fmnam, self.fmsiz)
        fontd = "%s %s" % (self.fdnam, self.fdsiz)
        gtk.rc_reparse_all_for_settings(setTheme(gtk, w, fontm, fontd), True)

    def doEnd(self):
        if self.df.pag == 0:
            self.rcdic = loadRcFile(self.rcfile, default=True)
            self.loadAllFields()
            self.df.focusField("T", 1, 1)
        elif self.df.pag == 3:
            self.doSave()
        else:
            self.df.selPage(self.df.tags[self.df.pag][0])

    def doExit(self):
        if self.df.pag == 0:
            self.doCancel()
        elif self.df.pag == 1:
            pass
        else:
            self.df.selPage(self.df.tags[self.df.pag - 2][0])

    def loadAllFields(self):
        #DataBase
        if self.rcdic["dbase"] == "FBird":
            dbase = "F"
        elif self.rcdic["dbase"] == "MsSQL":
            dbase = "Q"
        elif self.rcdic["dbase"] == "MySQL":
            dbase = "M"
        elif self.rcdic["dbase"] == "SQLite":
            dbase = "S"
        elif self.rcdic["dbase"] == "PgSQL":
            dbase = "P"
        self.df.loadEntry("T",1,0,dbase)
        self.df.loadEntry("T",1,1,self.rcdic["dbname"])
        self.df.loadEntry("T",1,2,self.rcdic["dbhost"])
        self.df.loadEntry("T",1,3,self.rcdic["dbdir"])
        self.df.loadEntry("T",1,4,self.rcdic["dbport"])
        self.df.loadEntry("T",1,5,self.rcdic["dbuser"])
        self.df.loadEntry("T",1,6,self.rcdic["dbpwd"])
        #General
        self.df.loadEntry("T",2,0,self.rcdic["prgdir"])
        self.df.loadEntry("T",2,1,self.rcdic["arcdir"])
        self.df.loadEntry("T",2,2,self.rcdic["bupdir"])
        self.df.loadEntry("T",2,3,self.rcdic["wrkdir"])
        self.df.loadEntry("T",2,4,self.rcdic["upgdir"])
        self.df.loadEntry("T",2,5,self.rcdic["vwr"])
        self.df.loadEntry("T",2,6,self.rcdic["prn"])
        self.df.loadEntry("T",2,7,self.rcdic["cnv"])
        self.df.loadEntry("T",2,8,self.rcdic["exp"])
        self.df.loadEntry("T",2,9,self.rcdic["geo"])
        self.df.loadEntry("T",2,10,self.rcdic["plc"])
        self.df.loadEntry("T",2,11,self.rcdic["cnf"].upper())
        self.df.loadEntry("T",2,12,self.rcdic["acnf"].upper())
        self.df.loadEntry("T",2,13,self.rcdic["ttip"].upper())
        self.df.loadEntry("T",2,14,self.rcdic["txtv"].upper())
        self.df.loadEntry("T",2,15,self.rcdic["efrm"].upper())
        self.df.loadEntry("T",2,16,self.rcdic["errs"].upper())
        #Font and Theme
        self.df.loadEntry("T",3,0,self.rcdic["mft"])
        self.df.loadEntry("T",3,1,self.rcdic["mfs"])
        self.df.loadEntry("T",3,3,self.rcdic["dft"])
        self.df.loadEntry("T",3,4,self.rcdic["dfs"])
        self.df.loadEntry("T",3,6,self.rcdic["theme"])

    def doSave(self, widget=None):
        frt, pag, col, mes = self.df.doCheckFields()
        if mes:
            self.df.selPage(self.df.tags[pag - 1][0])
            self.df.focusField(frt, pag, (col+1), err=mes)
        else:
            self.doWrite()

    def doWrite(self, widget=None):
        rcf = os.path.dirname(self.rcfile)
        dbd = self.df.t_work[1][0][3]
        self.prgdir = self.df.t_work[2][0][0]
        self.arcdir = self.df.t_work[2][0][1]
        self.bupdir = self.df.t_work[2][0][2]
        self.wrkdir = self.df.t_work[2][0][3]
        self.upgdir = self.df.t_work[2][0][4]
        for name in (rcf, dbd, self.prgdir, self.arcdir, self.bupdir, self.wrkdir, self.upgdir):
            name = self.checkCreateDir(name)
        cc = []
        if self.df.t_work[1][0][0] == "F":
            cc.append(["dbase", "FBird"])
        elif self.df.t_work[1][0][0] == "Q":
            cc.append(["dbase", "MsSQL"])
        elif self.df.t_work[1][0][0] == "M":
            cc.append(["dbase", "MySQL"])
        elif self.df.t_work[1][0][0] == "P":
            cc.append(["dbase", "PgSQL"])
        elif self.df.t_work[1][0][0] == "S":
            cc.append(["dbase", "SQLite"])
        cc.extend([
            ["dbname", self.df.t_work[1][0][1]],
            ["dbhost", self.df.t_work[1][0][2]],
            ["dbdir", os.path.normpath(self.df.t_work[1][0][3])],
            ["dbport", self.df.t_work[1][0][4]],
            ["dbuser", self.df.t_work[1][0][5]],
            ["dbpwd", self.df.t_work[1][0][6]],
            ["prgdir", os.path.normpath(self.prgdir)],
            ["arcdir", os.path.normpath(self.arcdir)],
            ["bupdir", os.path.normpath(self.bupdir)],
            ["wrkdir", os.path.normpath(self.wrkdir)],
            ["upgdir", os.path.normpath(self.upgdir)],
            ["vwr", os.path.normpath(self.df.t_work[2][0][5])],
            ["prn", os.path.normpath(self.df.t_work[2][0][6])]])
        if self.df.t_work[2][0][7]:
            cc.append(["cnv", os.path.normpath(self.df.t_work[2][0][7])])
        else:
            cc.append(["cnv", ""])
        cc.extend([
            ["exp", self.df.t_work[2][0][8]],
            ["geo", self.df.t_work[2][0][9]],
            ["plc", self.df.t_work[2][0][10]],
            ["cnf", self.df.t_work[2][0][11]],
            ["acnf", self.df.t_work[2][0][12]],
            ["ttip", self.df.t_work[2][0][13]],
            ["txtv", self.df.t_work[2][0][14]],
            ["efrm", self.df.t_work[2][0][15]],
            ["errs", self.df.t_work[2][0][16]],
            ["mft", self.df.t_work[3][0][0]],
            ["mfs", self.df.t_work[3][0][1]],
            ["dft", self.df.t_work[3][0][3]],
            ["dfs", self.df.t_work[3][0][4]],
            ["theme", self.df.t_work[3][0][6]]])
        f = open(self.rcfile, "w")
        for a, b in cc:
            d = "['%s'] = '%s'" % (a, b)
            f.write(d + "\n")
            if a == "prgdir":
                self.prgdir = b
            elif a == "arcdir":
                self.arcdir = b
            elif a == "bupdir":
                self.bupdir = b
            elif a == "wrkdir":
                self.wrkdir = b
            elif a == "upgdir":
                self.upgdir = b
            elif a == "vwr":
                self.vwr = b
            elif a == "prn":
                self.prn = b
            elif a == "cnv":
                self.exp = b
            elif a == "exp":
                self.exp = b
            else:
                exec "self.%s = '%s'" % (a, b)
        f.close()
        if sys.platform == "win32":
            try:
                gtkdir = os.environ["GTK_BASEPATH"]
                if not os.path.isdir(gtkdir):
                    raise Exception
            except:
                gtkdir = findFile(start=getPrgPath(), name="gtk", ftyp="d")
                if not gtkdir:
                    gtkdir = findFile(start="C:\\", name="gtk", ftyp="d")
            if gtkdir and os.path.exists(gtkdir):
                try:
                    gtkrc = open(os.path.join(gtkdir, ".gtkrc-2.0"), "w")
                    gtkrc.write('gtk-theme-name = "%s"\n' % \
                        self.df.t_work[3][0][4])
                    gtkrc.close()
                except:
                    pass
        ok = askQuestion(self.mf.body, "Configuration", "Please Note that Any "\
            "Changes made will only take effect once Tartan is Restarted."\
            "Would you Like to Exit Tartan Now?")
        self.closeProcess(ok)

    def doCancel(self, widget=None):
        self.rcfile = None
        self.closeProcess()

    def closeProcess(self, ok="no"):
        self.df.closeProcess()
        self.mf.closeLoop()
        if self.nmf:
            self.mf.window.destroy()
        if ok == "yes":
            sys.exit()

    def checkCreateDir(self, name):
        path = os.path.normpath(name)
        try:
            os.makedirs(path)
        except:
            pass
        return path

    def loadFonts(self):
        self.aft = []
        self.fft = []
        sp = SplashScreen(self.mf.body,
            "Loading List of Fixed Fonts\n\n      Please Wait...")
        fc = gtk.Frame().get_pango_context()
        ff = fc.list_families()
        for f in ff:
            self.aft.append(f.get_name())
            if f.is_monospace():
                self.fft.append(f.get_name())
        self.aft.sort()
        self.fft.sort()
        sp.closeSplash()

    def loadThemes(self):
        self.thm = ["Default"]
        bad = ("Default", )
        thmdir = gtk.rc_get_theme_dir()
        if os.path.isdir(thmdir):
            thm = filter(lambda s: os.path.isdir(os.path.join(thmdir, s)),
                os.listdir(thmdir))
            for th in thm:
                dir = os.path.join(thmdir, th, "gtk-2.0")
                if os.path.exists(dir) and th not in bad:
                    self.thm.append(th)
            self.thm.sort()

class FileDialog:
    def __init__(self, title=None, parent=None, initd=None, ftype=None, multi=False):
        if title:
            self.title = title
        else:
            self.title = "Select ...."
        self.parent = parent
        self.initd = initd
        self.ftype = ftype
        self.multi = multi

    def askopenfilename(self):
        return self.doSelect("f")

    def askdirectory(self):
        return self.doSelect("d")

    def doSelect(self, stype):
        if stype == "f":
            action = gtk.FILE_CHOOSER_ACTION_OPEN
        else:
            action = gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER
        self.dialog = gtk.FileChooserDialog(title=self.title,
            parent=self.parent, action=action, buttons=(gtk.STOCK_CANCEL,
            gtk.RESPONSE_CANCEL, gtk.STOCK_OPEN, gtk.RESPONSE_OK))
        self.dialog.set_default_response(gtk.RESPONSE_OK)
        if self.initd:
            self.dialog.set_current_folder(self.initd)
        if self.multi:
            self.dialog.set_select_multiple(True)
        if stype == "f":
            if self.ftype:
                for typ in self.ftype:
                    filtr = gtk.FileFilter()
                    filtr.set_name(typ[0])
                    filtr.add_pattern(typ[1])
                    self.dialog.add_filter(filtr)
            else:
                filtr = gtk.FileFilter()
                filtr.set_name("All Files")
                filtr.add_pattern("*")
                self.dialog.add_filter(filtr)
        response = self.dialog.run()
        if response == gtk.RESPONSE_OK:
            if self.multi:
                fle = ""
                for f in self.dialog.get_filenames():
                    if not fle:
                        fle = f
                    else:
                        fle = fle + "," + f
            else:
                fle = self.dialog.get_filename()
        else:
            fle = None
        self.dialog.destroy()
        return fle

class RunCommand:
    """
    This class is used to run an external command:
    
    if typ = 'N' then outputting the result to a TextView.
    if typ = 'Y' then running the command in a vte Terminal.
    if typ = 'S' then execute the file with the default command.
    """
    def __init__(self, scrn, cmd, typ="N"):
        self.scrn = scrn
        self.cmd = cmd
        self.typ = typ
        if self.typ == "S":
            if sys.platform == "win32":
                os.startfile(self.cmd)
            else:
                self.cmd = os.path.join(sys.path[0], self.cmd)
                subprocess.call("%s %s" % ("firefox", self.cmd), shell=True)
            return
        if self.typ == "N":
            self.doScreen()
            self.doOutput()
        else:
            self.doTerminal()
        self.scrn.show_all()
        gtk.main()

    def doScreen(self):
        # Scrolled Window
        self.scrn.get_toplevel()
        self.scrn.get_toplevel().realize()
        size = tuple(self.scrn.get_toplevel().allocation)
        self.sw = gtk.ScrolledWindow()
        self.sw.set_size_request(int(size[2]*.90), int(size[3]*.80))
        self.sw.set_policy(gtk.POLICY_NEVER, gtk.POLICY_AUTOMATIC)
        # Text View
        self.view = gtk.TextView()
        self.view.set_editable(False)
        self.view.modify_font(pango.FontDescription("courier"))
        self.sw.add(self.view)
        self.scrn.pack_start(self.sw, True, True, 0)
        # Button
        self.but = MyButton("Close")
        self.but.connect("clicked", self.doClose)
        self.scrn.pack_start(self.but, True, True, 0)

    def doOutput(self):
        buff = self.view.get_buffer()
        proc = subprocess.Popen(self.cmd, shell=True, bufsize=0,
            stdin=subprocess.PIPE, stdout=subprocess.PIPE,
            stderr=subprocess.PIPE, close_fds=True)
        stdin, stdout, stderr = (proc.stdin, proc.stdout, proc.stderr)
        while True:
            line = stdout.readline()
            if not line:
                break
            iter = buff.get_end_iter()
            buff.place_cursor(iter)
            buff.insert(iter, line)
            self.view.scroll_to_mark(buff.get_insert(), 0.1)

    def doTerminal(self):
        import vte
        self.term = vte.Terminal()
        self.term.connect ("show",lambda term: term.feed_child("%s\n"%self.cmd))
        self.term.connect ("child-exited", self.doClose)
        self.term.fork_command()
        self.scrn.add(self.term)

    def doClose(self, widget=None):
        if self.typ == "N":
            self.view.destroy()
            self.but.destroy()
            self.sw.destroy()
        else:
            self.term.destroy()
        gtk.main_quit()

class AgeAll:
    """
    This class is used to Age Debtors, Creditors and Members Transactions where:

    system:     "drs", "crs" or "mem"
    agetyp:     N = Normal  - Selective allocations are allowed.
                H = History - Same as Normal but including paid-up items.
                O = Oldest  - The transaction is automatically allocated against
                              the oldest outstanding transactions, any
                              unallocated amount remains unallocated.
                C = Current - The transaction retains its balance, i.e. not
                              allocated.
    agekey:     drs - [conum, chain, acno, atyp, aref, agedt, amt, dis]
                crs - [conum, acno, atyp, aref, agedt, amt, dis]
                mem - [conum, memno, atyp, aref, agedt, amt, dis]
    """
    def __init__(self, mf, **args):
        self.mf = mf
        defaults = {
            "system": None,
            "agetyp": None,
            "agekey": None}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        if self.system.lower() == "drs":
            idx = 3
        else:
            idx = 2
        self.atyp, self.aref, self.agedt, self.amt, self.dis = self.agekey[idx:]
        if self.agetyp == "C":
            self.data = None
        else:
            self.doGetData()
        self.cancel = False
        if not self.data:
            self.doAgeCurrent()
        elif self.agetyp == "O":
            self.doAgeOldest()
        else:
            self.doAgeTrans()

    def doGetData(self):
        if self.system.lower() == "drs":
            col = ["drt_trdt", "drt_ref1", "drt_type", "drt_tramt"]
            whr = [("drt_cono", "=", self.agekey[0]), ("drt_chain", "=",
                self.agekey[1]), ("drt_acno", "=", self.agekey[2])]
        elif self.system.lower() == "crs":
            col = ["crt_trdt", "crt_ref1", "crt_type", "crt_tramt"]
            whr = [("crt_cono", "=", self.agekey[0]), ("crt_acno", "=",
                self.agekey[1])]
        else:
            col = ["mlt_trdt", "mlt_refno", "mlt_type", "mlt_tramt"]
            whr = [("mlt_cono", "=", self.agekey[0]), ("mlt_memno", "=",
                self.agekey[1])]
        get, all = getTrn(self.mf.dbm, self.system.lower(), whr=whr)
        self.data = []
        for rec in all:
            bal = rec[get.index("balance")]
            if self.agetyp != "H" and not bal:
                continue
            dat = []
            for c in col:
                dat.append(rec[get.index(c)])
            dat.append(rec[get.index("paid")])
            dat.append(bal)
            self.data.append(dat)

    def doAgeCurrent(self, widget=None):
        pass

    def doAgeOldest(self):
        unal = float(ASD(self.amt) + ASD(self.dis))
        for dat in self.data:
            trbal = float(ASD(dat[5]) + ASD(unal))
            if trbal < 0:
                unal = trbal
                trbal = 0
            elif trbal > 0:
                unal = 0
            else:
                unal = 0
                trbal = 0
            alloc = float(ASD(dat[5]) - ASD(trbal))
            if alloc:
                self.doAllocate(dat[2], dat[1], alloc)
        alloc = float(ASD(self.amt) + ASD(self.dis) - ASD(unal))
        if alloc:
            self.doAllocated(alloc)

    def doAgeTrans(self):
        unal = float(ASD(self.amt) + ASD(self.dis))
        self.ageT = AgeTrans(self.mf, self.system.lower(), self.data, unal)
        self.cancel = self.ageT.ccl
        if self.cancel:
            return
        for tr in self.ageT.data:
            if tr[6]:
                self.doAllocate(tr[2], tr[1], tr[6])
        totl = float(ASD(self.amt) + ASD(self.dis))
        alloc = float(ASD(totl) + ASD(self.ageT.total.work))
        if alloc:
            self.doAllocated(alloc)

    def doAllocate(self, typ, ref, amt):
        if self.system.lower() == "drs":
            data = self.agekey[:3] + [typ, ref, self.agedt, self.atyp,
                self.aref, amt, 0]
            sql = Sql(self.mf.dbm, "drsage", prog=__name__)
            sql.insRec("drsage", data=data)
        elif self.system.lower() == "crs":
            data = self.agekey[:2] + [typ, ref, self.agedt, self.atyp,
                self.aref, amt, 0]
            sql = Sql(self.mf.dbm, "crsage", prog=__name__)
            sql.insRec("crsage", data=data)
        else:
            data = self.agekey[:2] + [typ, ref, self.agedt, self.atyp,
                self.aref, amt, 0]
            sql = Sql(self.mf.dbm, "memage", prog=__name__)
            sql.insRec("memage", data=data)

    def doAllocated(self, alloc):
        if not self.dis:
            da = 0.00
            aa = alloc
        elif self.dis and self.dis < alloc:
            da = float(ASD(self.dis) - ASD(alloc))
            aa = 0.00
        else:
            da = self.dis
            aa = float(ASD(alloc) - ASD(self.dis))
        if da:
            self.doAllocate(6, self.aref, da)
        if aa:
            self.doAllocate(self.atyp, self.aref, aa)

class AgeTrans:
    """
    Class for the Distribution of an Amount Over Transactions

    Usage: AgeTrans(mf, system, data, total)

        system = crs or drs
        data = List of lists of all available transactions lists comprising:
             [Date, Reference, Type, Amount, Paid, Balance]
        total = Total Amount to be Distributed (Decimal Value)
        exit = Create an Exit Button to leave the balance unallocated if True

    The class returns as follows:
        self.data = List of lists comprising:
            [Date, Reference, Type, Amount, Paid, Balance, Allocated]
        self.total = Total Amount Not Yet Allocated
    """
    def __init__(self, mf, system, data, total, exit=True):
        self.mf = mf
        self.system = system
        self.data = data
        self.total = CCD(float(ASD(0) - ASD(total)), "SD", 11.2)
        self.exit = exit
        self.doVariables()
        if self.total.work:
            while not self.xit and not self.ccl:
                self.doAllocate()
                if not self.total.work:
                    break
        else:
            self.doAllocate()

    def doVariables(self):
        if self.system == "crs":
            typ = ("XX", crtrtp)
        elif self.system == "drs":
            typ = ("XX", drtrtp)
        elif self.system == "mem":
            typ = ("XX", mltrtp)
        self.cols = (
            (0, "   Date",    10,   "D1", None),
            (1, "Reference",   9,   "Na", None),
            (2, "Typ",         3,   typ,  None),
            (3, "    Amount", 11.2, "SD", None),
            (4, " Allocated", 11.2, "SD", None),
            (5, "   Balance", 11.2, "SD", None))
        self.work = []
        self.xit = False
        self.ccl = False
        self.posn = 0
        for num, rec in enumerate(self.data):
            self.data[num].append(0)
            self.work.append(rec[:6])

    def doAllocate(self):
        # Scrolled List Box
        self.lb = SelectChoice(self.mf.body, "Available Transactions",
            self.cols, self.work, sort=False, loop=False, font=self.mf.fontd,
            comm=self.enterAll, posn=self.posn)
        # Allocation Fields
        hbox = gtk.HBox(True, 0)
        self.lb.vbox.pack_start(hbox, False, False, 0)
        lab = gtk.Label("Allocation Amount")
        hbox.pack_start(lab, False, False, 0)
        self.allAmt = gtk.Entry(max=11)
        self.allAmt.set_width_chars(11)
        self.allAmt.set_sensitive(False)
        self.allAmt.connect("activate", self.enterAmt)
        hbox.pack_start(self.allAmt, False, False, 0)
        lab = gtk.Label("Unallocated")
        hbox.pack_start(lab, False, False, 0)
        self.allBal = gtk.Entry(max=11)
        self.allBal.set_width_chars(11)
        self.allBal.set_editable(False)
        self.allBal.unset_flags(gtk.CAN_FOCUS)
        self.allBal.set_text(self.total.disp)
        hbox.pack_start(self.allBal, False, False, 0)
        bbox = gtk.Table(1, 2, True)
        cpos = 0
        self.lb.vbox.pack_start(bbox, False, False, 0)
        self.allBut = MyButton(stock="gtk-apply")
        self.allBut.connect("clicked", self.enterAmt)
        self.allBut.set_sensitive(False)
        bbox.attach(self.allBut, cpos, cpos + 1, 0, 1)
        cpos += 1
        if self.exit:
            bbox.resize(1, 3)
            xitBut = MyButton(stock="gtk-exit")
            xitBut.connect("clicked", self.exitAge)
            bbox.attach(xitBut, cpos, cpos + 1, 0, 1)
            cpos += 1
        canBut = MyButton(stock="gtk-cancel")
        canBut.connect("clicked", self.cancelAll)
        bbox.attach(canBut, cpos, cpos + 1, 0, 1)
        self.lb.vbox.show_all()
        # Button Tooltips
        if PYGTK_VERSION[1] < 12:
            ttips = gtk.Tooltips()
            ttips.set_tip(self.allBut, "Apply this Allocation")
            if self.exit:
                ttips.set_tip(xitBut, "Leave the Balance as Unallocated")
            ttips.set_tip(canBut, "Cancel All Allocations")
        else:
            self.allBut.set_tooltip_text("Apply this Allocation")
            if self.exit:
                xitBut.set_tooltip_text("Leave the Balance as Unallocated")
            canBut.set_tooltip_text("Cancel All Allocations")
        # Main Loop
        gtk.main()

    def enterAll(self, data):
        self.pos = data[0]
        self.posn = self.lb.posn
        self.allAmt.set_sensitive(True)
        self.allBut.set_sensitive(True)
        bal = self.work[self.pos][5]
        if bal > self.total.work:
            bal = self.total.work
        wrk = CCD(bal, "SD", 11.2)
        self.allAmt.set_text(wrk.disp)
        self.allAmt.set_flags(gtk.CAN_FOCUS)
        self.allAmt.grab_focus()

    def enterAmt(self, widget=None):
        wrk = CCD(self.allAmt.get_text(), "SD", 11.2)
        self.data[self.pos][6] = float(ASD(self.data[self.pos][6]) + \
            ASD(wrk.work))
        all = float(ASD(self.work[self.pos][4]) + ASD(wrk.work))
        bal = float(ASD(self.work[self.pos][3]) - ASD(all))
        self.work[self.pos][4] = all
        self.work[self.pos][5] = bal
        tot = float(ASD(self.total.work) - ASD(wrk.work))
        self.total = CCD(tot,"SD",11.2)
        self.closeProcess()

    def exitAge(self, widget=None):
        self.xit = True
        self.closeProcess()

    def cancelAll(self, widget=None):
        ok = askQuestion(self.mf.body, head="Cancel",
            mess="Are You Certain You Want to Cancel This Entry?")
        if ok == "yes":
            self.ccl = True
            self.closeProcess()

    def closeProcess(self):
        self.lb.closeProcess()
        gtk.main_quit()

class Balances:
    """
    This class is used to return various balances from Debtors, Creditors,
        Stores, Assets and Members tables:

    doAllBals       - Returns Total, Current, 30, 60, 90 and 120+ Balances with
                        or without transactions
    doCrsDrsHist    - Returns the Last 12 Months Sale and Payment Balances
    doStrBals       - Returns Opening Balance Qty and Cost, Movements and
                      Closing Balance Qty and Cost for the Current Month and
                      for the Year to Date, Average and Last cost prices.
                      If trans=True the transactions are returned.
                      If markup=True the markup percentage is returned.
    doStrOrds       - Returns the Qty on Hand, Qty on Order and Qty Back Order
    doStrHist       - Returns the Last 12 Months Receipt and Issue Qty and Value
    doAssBals       - Returns the Opening Book Values for a Financial Period

    system          - ASS, CRS, DRS, MEM or STR
    conum           - Company number
    curdt           - Current period, YYYYMM
    keys            - Assets is the group and code
                      Creditors is the acno
                      Debtors is a list containing (chain, acno)
                      Members is the member number
                      Stores is a list containing:
                          if len(keys) == 1: rep
                          if len(keys) == 2: rep, period
                          if len(keys) == 3: group, code, loc
                          if len(keys) == 4: group, code, loc, period
    """
    def __init__(self, mf, system, conum, curdt, keys):
        self.mf = mf
        self.system = system.upper()
        self.conum = conum
        self.curdt = curdt
        if self.mf.display:
            scrn = self.mf.body
        else:
            scrn = None
        if not keys:
            showError(scrn, "Keys Error", "You Must have a Key Argument")
            sys.exit()
        else:
            if self.system == "CRS" and len(keys) == 1:
                self.sql = Sql(self.mf.dbm, "crstrn", prog=__name__)
                self.acno = keys[0]
            elif self.system == "DRS" and len(keys) == 2:
                self.sql = Sql(self.mf.dbm, "drstrn", prog=__name__)
                self.chain, self.acno = keys
            elif self.system == "MEM" and len(keys) == 1:
                self.sql = Sql(self.mf.dbm, "memtrn", prog=__name__)
                self.memno = keys[0]
            elif self.system == "STR" and len(keys) in (1, 2, 3, 4):
                self.sql = Sql(self.mf.dbm, ["strgrp", "strmf2", "strtrn",
                    "strpom", "strpot", "slsiv1", "slsiv2"], prog=__name__)
                self.group = None
                self.code = None
                self.loc = None
                period = None
                self.rep = None
                if len(keys) == 1:
                    self.rep = keys[0]
                elif len(keys) == 2:
                    self.rep, period = keys
                elif len(keys) == 3:
                    self.group, self.code, self.loc = keys
                elif len(keys) == 4:
                    self.group, self.code, self.loc, period = keys
                if period:
                    if period[0] == "P":
                        s, e, f = getPeriods(self.mf, self.conum, period[1])
                        if not s:
                            self.start = self.curdt
                        else:
                            self.start = s.work / 100
                    else:
                        self.start = period[1]
                else:
                    self.start = curdt
            elif self.system == "ASS" and len(keys) == 2:
                self.sql = Sql(self.mf.dbm, "asstrn", prog=__name__)
                self.group, self.code = keys
                self.start = curdt
            else:
                showError(scrn, "Parameter Error",
                    "Invalid Parameters %s %s" % (self.system, str(keys)))
                sys.exit()

    def doAllBals(self, trans="N"):
        """
        trans:
            "Y" = returns all transactions except those zeroed previously
            "N" = does not return transactions
            "A" = returns all transactions
        """
        ages = [0,0,0,0,0]
        if self.system == "CRS":
            whr = [("crt_cono", "=", self.conum), ("crt_acno", "=", self.acno)]
            w = whr[:]
            w.append(("crt_curdt", "<", self.curdt))
            obal = self.sql.getRec(tables="crstrn",
                cols=["round(sum(crt_tramt), 2)"], where=w, limit=1)
            if not obal or not obal[0]:
                obal = 0.0
            else:
                obal = obal[0]
            w = whr[:]
            w.append(("crt_curdt", "<=", self.curdt))
            cbal = self.sql.getRec(tables="crstrn",
                cols=["round(sum(crt_tramt), 2)"], where=w, limit=1)
            if not cbal or not cbal[0]:
                cbal = 0.0
            else:
                cbal = cbal[0]
            col, trns = getTrn(self.mf.dbm, self.system.lower(), dte=self.curdt,
                whr=whr, zer=trans)
        elif self.system == "DRS":
            whr = [("drt_cono", "=", self.conum), ("drt_chain", "=",
                self.chain), ("drt_acno", "=", self.acno)]
            w = whr[:]
            w.append(("drt_curdt", "<", self.curdt))
            obal = self.sql.getRec(tables="drstrn",
                cols=["round(sum(drt_tramt), 2)"], where=w, limit=1)
            if not obal or not obal[0]:
                obal = 0.0
            else:
                obal = obal[0]
            w = whr[:]
            w.append(("drt_curdt", "<=", self.curdt))
            cbal = self.sql.getRec(tables="drstrn",
                cols=["round(sum(drt_tramt), 2)"], where=w, limit=1)
            if not cbal or not cbal[0]:
                cbal = 0.0
            else:
                cbal = cbal[0]
            col, trns = getTrn(self.mf.dbm, self.system.lower(), dte=self.curdt,
                whr=whr, zer=trans)
        elif self.system == "MEM":
            whr = [("mlt_cono","=",self.conum), ("mlt_memno","=",self.memno)]
            w = whr[:]
            w.append(("mlt_curdt", "<", self.curdt))
            obal = self.sql.getRec(tables="memtrn",
                cols=["round(sum(mlt_tramt), 2)"], where=w, limit=1)
            if not obal or not obal[0]:
                obal = 0.0
            else:
                obal = obal[0]
            w = whr[:]
            w.append(("mlt_curdt", "<=", self.curdt))
            cbal = self.sql.getRec(tables="memtrn",
                cols=["round(sum(mlt_tramt), 2)"], where=w, limit=1)
            if not cbal or not cbal[0]:
                cbal = 0.0
            else:
                cbal = cbal[0]
            col, trns = getTrn(self.mf.dbm, self.system.lower(), dte=self.curdt,
                whr=whr, zer=trans)
        else:
            return
        if trns:
            for trn in trns:
                if self.system == "CRS":
                    age = self.curdt - trn[col.index("crt_curdt")]
                elif self.system == "DRS":
                    age = self.curdt - trn[col.index("drt_curdt")]
                else:
                    age = self.curdt - trn[col.index("mlt_curdt")]
                while age > 88:
                    age = age - 88
                if age < 0:
                    age = 0
                if age > 4:
                    age = 4
                ages[age] = float(ASD(ages[age])+ASD(trn[col.index("balance")]))
        if trans == "N":
            return obal, cbal, ages
        else:
            w = whr[:]
            if self.system == "CRS":
                w.append(("crt_curdt", "=", self.curdt))
                cmth = self.sql.getRec(tables="crstrn", cols=["count(*)"],
                    where=w, limit=1)
            elif self.system == "DRS":
                w.append(("drt_curdt", "=", self.curdt))
                cmth = self.sql.getRec(tables="drstrn", cols=["count(*)"],
                    where=w, limit=1)
            else:
                w.append(("mlt_curdt", "=", self.curdt))
                cmth = self.sql.getRec(tables="memtrn", cols=["count(*)"],
                    where=w, limit=1)
            return obal, cbal, ages, [col, trns, cmth[0]]

    def doCrsDrsHist(self):
        this = [0,0]
        hist = [[0,0,0,0,0,0,0,0,0,0,0,0,0],[0,0,0,0,0,0,0,0,0,0,0,0,0]]
        start = self.curdt - 199
        if (start % 100) == 13:
            start = start + 100 - 12
        if self.system == "DRS":
            col=["drt_curdt", "drt_type", "round(sum(drt_tramt), 2)",
                "round(sum(drt_taxamt), 2)"]
            whr=[("drt_cono", "=", self.conum), ("drt_chain", "=", self.chain),
                ("drt_acno", "=", self.acno), ("drt_curdt", "between",
                start, self.curdt)]
            grp = "drt_curdt, drt_type"
            bals = self.sql.getRec(tables="drstrn", cols=col, where=whr,
                group=grp, order=grp)
        elif self.system == "CRS":
            col=["crt_curdt", "crt_type", "round(sum(crt_tramt), 2)",
                "round(sum(crt_taxamt), 2)"]
            whr=[("crt_cono", "=", self.conum), ("crt_acno", "=", self.acno),
                ("crt_curdt", "between", start, self.curdt)]
            grp = "crt_curdt, crt_type"
            bals = self.sql.getRec(tables="crstrn", cols=col, where=whr,
                group=grp, order=grp)
        for bal in bals:
            if bal[1] in (1, 4):
                x = 0
            elif bal[1] in (2, 5):
                x = 1
            else:
                continue
            this[x] = float(ASD(this[x]) + ASD(bal[2]) - ASD(bal[3]))
            mth = self.curdt - bal[0]
            while mth > 88:
                mth = mth - 88
            if mth > 11:
                mth = 12
            hist[x][mth] = float(ASD(hist[x][mth]) + ASD(bal[2]) - ASD(bal[3]))
        return this, hist

    def doStrBals(self, start=None, trans="N", markup=False):
        if not start:
            start = self.start
        if not self.loc:
            where = [("stt_cono", "=", self.conum), ("stt_group", "=",
                self.group), ("stt_code", "=", self.code)]
        else:
            where = [("stt_cono", "=", self.conum), ("stt_group", "=",
                self.group), ("stt_code", "=", self.code), ("stt_loc", "=",
                self.loc)]
        # Current Period
        whr = where[:]
        whr.append(("stt_curdt", "<", self.curdt))
        obal = self.sql.getRec(tables="strtrn",
            cols=["round(sum(stt_qty), 2)", "round(sum(stt_cost), 2)"],
            where=whr, limit=1)
        if not obal:
            obal = [0, 0]
        if not obal[0]:
            obal[0] = 0
        if not obal[1]:
            obal[1] = 0
        whr = where[:]
        whr.append(("stt_curdt", "between", self.curdt, self.curdt))
        grp = "stt_type"
        move = self.sql.getRec(tables="strtrn", cols=["stt_type",
            "round(sum(stt_qty), 2)", "round(sum(stt_cost), 2)",
            "round(sum(stt_sell), 2)"], where=whr, group=grp, order=grp)
        whr = where[:]
        whr.append(("stt_curdt", "<=", self.curdt))
        cbal = self.sql.getRec(tables="strtrn", cols=[
            "round(sum(stt_qty), 2)", "round(sum(stt_cost), 2)"],
            where=whr, limit=1)
        if not cbal:
            cbal = [0, 0]
        if not cbal[0]:
            cbal[0] = 0
        if not cbal[1]:
            cbal[1] = 0
        data = [obal, move, cbal]
        # Year to Date
        whr = where[:]
        whr.append(("stt_curdt", "<", start))
        obal = self.sql.getRec(tables="strtrn", cols=[
            "round(sum(stt_qty), 2)", "round(sum(stt_cost), 2)"],
            where=whr, limit=1)
        if not obal:
            obal = [0, 0]
        if not obal[0]:
            obal[0] = 0
        if not obal[1]:
            obal[1] = 0
        whr = where[:]
        whr.append(("stt_curdt", "between", start, self.curdt))
        move = self.sql.getRec(tables="strtrn", cols=["stt_type",
            "round(sum(stt_qty), 2)", "round(sum(stt_cost), 2)",
            "round(sum(stt_sell), 2)"], where=whr, group=grp, order=grp)
        if not move:
            move = []
        whr = where[:]
        whr.append(("stt_curdt", "<=", self.curdt))
        cbal = self.sql.getRec(tables="strtrn", cols=["round(sum(stt_qty), 2)",
            "round(sum(stt_cost), 2)"], where=whr, limit=1)
        if not cbal:
            cbal = [0, 0]
        if not cbal[0]:
            cbal[0] = 0
        if not cbal[1]:
            cbal[1] = 0
        data.extend([obal, move, cbal])
        # Last and Average cost prices
        whr = where[:]
        whr.append(("stt_type", "in", (1, 3)))
        lcst = self.sql.getRec(tables="strtrn", cols=["stt_qty",
            "stt_cost"], where=whr, order="stt_trdt desc", limit=1)
        if not lcst:
            lcst = 0
        else:
            qty = CCD(lcst[0], "SD", 11.2)
            cst = CCD(lcst[1], "SD", 11.2)
            if qty.work == 0:
                lcst = 0
            else:
                lcst = round((cst.work / qty.work), 2)
        whr = where[:]
        whr.append(("stt_type", "in", (1, 3, 5)))
        acst = self.sql.getRec(tables="strtrn", cols=["round(sum(stt_qty), 2)",
            "round(sum(stt_cost), 2)"], where=where, limit=1)
        if not acst[0]:
            acst[0] = 0
        if not acst[1]:
            acst[1] = 0
        qty = CCD(acst[0], "SD", 11.2)
        cst = CCD(acst[1], "SD", 11.2)
        if qty.work <= 0:
            acst = lcst
        else:
            acst = round((cst.work / qty.work), 2)
        data.extend([acst, lcst])
        # Transactions
        if trans == "Y":
            whr = where[:]
            whr.append(("stt_curdt", "between", start, self.curdt))
            ord = "stt_trdt, stt_type, stt_ref1"
            data.append((self.sql.getRec(tables="strtrn", where=whr,
                order=ord), self.sql.strtrn_dic))
        # Markup Percentage
        if markup:
            acc = self.sql.getRec(tables="strgrp", cols=["gpm_markup"],
                where=[("gpm_cono", "=", self.conum), ("gpm_group", "=",
                self.group)], limit=1)
            mup = acc[0]
            acc = self.sql.getRec(tables="strmf2", cols=["st2_markup"],
                where=[("st2_cono", "=", self.conum), ("st2_group", "=",
                self.group), ("st2_code", "=", self.code)], limit=1)
            if acc[0]:
                mup = acc[0]
            data.append(mup)
        return data

    def doStrOrds(self):
        # Closing Balance
        whr = [("stt_cono", "=", self.conum)]
        if self.loc:
            whr.append(("stt_loc", "=", self.loc))
        whr.extend([
            ("stt_group", "=", self.group),
            ("stt_code", "=", self.code),
            ("stt_curdt", "<=", self.curdt)])
        cbal = self.sql.getRec(tables="strtrn", cols=[
            "round(sum(stt_qty), 2)"], where=whr, limit=1)
        if not cbal or not cbal[0]:
            data = [CCD(0, "SD", 12.2)]
        else:
            data = [CCD(cbal[0], "SD", 12.2)]
        # Orders on Suppliers
        whr = [("pom_cono", "=", self.conum)]
        if self.loc:
            whr.append(("pom_loc", "=", self.loc))
        whr.extend([
            ("pom_date", "<=", self.curdt),
            ("pom_delno", "<>", "cancel"),
            ("pom_deldt", "=", 0),
            ("pot_cono=pom_cono",),
            ("pot_ordno=pom_ordno",),
            ("pot_group", "=", self.group),
            ("pot_code", "=", self.code)])
        oord = self.sql.getRec(tables=["strpom", "strpot"], cols=[
            "round(sum(pot_qty), 2)"], where=whr, limit=1)
        if not oord or not oord[0]:
            data.append(CCD(0, "SD", 12.2))
        else:
            data.append(CCD(oord[0], "SD", 12.2))
        # Orders by Customers
        whr = [
            ("si1_cono", "=", self.conum),
            ("si1_rtn", "=", "O"),
            ("si1_date", "<=", self.curdt),
            ("si1_invno", "=", ""),
            ("si2_cono=si1_cono",),
            ("si2_rtn=si1_rtn",),
            ("si2_docno=si1_docno",),
            ("si2_group", "=", self.group),
            ("si2_code", "=", self.code)]
        if self.loc:
            whr.append(("si2_loc", "=", self.loc))
        bord = self.sql.getRec(tables=["slsiv1", "slsiv2"], cols=[
            "round(sum(si2_qty), 2)"], where=whr, limit=1)
        if not bord or not bord[0]:
            data.append(CCD(0, "SD", 12.2))
        else:
            data.append(CCD(bord[0], "SD", 12.2))
        return data

    def doStrHist(self, start=None, more=None):
        if not start:
            start = self.start
        this = [[0,0],[0,0],[0,0,0]]
        hist = [
            [[0,0],[0,0],[0,0],[0,0],[0,0],[0,0],
            [0,0],[0,0],[0,0],[0,0],[0,0],[0,0]],
            [[0,0],[0,0],[0,0],[0,0],[0,0],[0,0],
            [0,0],[0,0],[0,0],[0,0],[0,0],[0,0]],
            [[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],
            [0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0],[0,0,0]]]
        whr = [("stt_cono", "=", self.conum)]
        if self.rep:
            whr.append(("stt_rep", "=", self.rep))
        else:
            whr.extend([("stt_group", "=", self.group),
                ("stt_code", "=", self.code)])
            if self.loc:
                whr.append(("stt_loc", "=", self.loc))
        whr.append(("stt_curdt", "between", start, self.curdt))
        if more:
            for state in more:
                whr.append(state)
        grp = "stt_curdt, stt_type"
        bals = self.sql.getRec(tables="strtrn", cols=["stt_curdt",
            "stt_type", "round(sum(stt_qty), 2)", "round(sum(stt_cost), 2)",
            "round(sum(stt_sell), 2)"], where=whr, group=grp, order=grp)
        for bal in bals:
            if bal[1] in (1,3,5):
                x = 0
            elif bal[1] in (2,4,6,7):
                x = 1
            else:
                continue
            this[x][0] = float(ASD(this[x][0]) + ASD(bal[2]))
            this[x][1] = float(ASD(this[x][1]) + ASD(bal[3]))
            if bal[1] == 7:
                this[2][0] = float(ASD(this[2][0]) + ASD(bal[2]))
                this[2][1] = float(ASD(this[2][1]) + ASD(bal[3]))
                this[2][2] = float(ASD(this[2][2]) + ASD(bal[4]))
            mth = self.curdt - bal[0]
            while mth > 88:
                mth = mth - 88
            if mth > 11:
                mth = 11
            hist[x][mth][0] = float(ASD(hist[x][mth][0]) + ASD(bal[2]))
            hist[x][mth][1] = float(ASD(hist[x][mth][1]) + ASD(bal[3]))
            if bal[1] == 7:
                hist[2][mth][0] = float(ASD(hist[2][mth][0]) + ASD(bal[2]))
                hist[2][mth][1] = float(ASD(hist[2][mth][1]) + ASD(bal[3]))
                hist[2][mth][2] = float(ASD(hist[2][mth][2]) + ASD(bal[4]))
        return this, hist

    def doAssBals(self, start=None, end=None, trans="N"):
        # trans: N - No Transactions, C - Current Year, H - All
        if not start:
            start = self.start
        w1 = [("ast_cono", "=", self.conum), ("ast_group", "=", self.group),
            ("ast_code", "=", self.code)]
        w2 = w1[:]
        w2.append(("ast_curdt", "<", start))
        w3 = w2[:]
        w3.append(("ast_mtyp", "in", (3, 5)))
        if self.sql.getRec(tables="asstrn", where=w3):
            # Sold or Written Off Asset, Ignore
            return
        w3 = w2[:]
        w3.append(("ast_mtyp", "in", (1, 2)))
        cap = self.sql.getRec(tables="asstrn",
            cols=["round(sum(ast_amt1), 2)"], where=w3, limit=1)
        w3 = w2[:]
        w3.append(("ast_mtyp", "=", 4))
        dep = self.sql.getRec(tables="asstrn",
            cols=["round(sum(ast_amt1), 2)", "round(sum(ast_amt2), 2)"],
            where=w3, limit=1)
        if not cap[0]:
            cap[0] = 0
        if not dep[0] or not dep[0]:
            dep[0] = 0
        if not dep[0] or not dep[1]:
            dep[1] = 0
        cbl = float(ASD(cap[0]) + ASD(dep[0]))
        rbl = float(ASD(cap[0]) + ASD(dep[1]))
        w3 = w1[:]
        if not end:
            w3.append(("ast_curdt", ">=", start))
        else:
            w3.append(("ast_curdt", "between", start, end))
        mov = self.sql.getRec(tables="asstrn", cols=["ast_mtyp",
            "round(sum(ast_amt1), 2)", "round(sum(ast_amt2), 2)"], where=w3,
            group="ast_mtyp", order="ast_mtyp")
        col = self.sql.asstrn_col
        if trans == "H":
            trn = self.sql.getRec(tables="asstrn", where=w1,
                order="ast_date, ast_type")
            return cap[0], dep[0], dep[1], cbl, rbl, mov, [trn, col]
        elif trans == "C":
            trn = self.sql.getRec(tables="asstrn", where=w3,
                order="ast_date, ast_type")
            return cap[0], dep[0], dep[1], cbl, rbl, mov, [trn, col]
        else:
            return cap[0], dep[0], dep[1], cbl, rbl, mov

class Batches:
    """
    This class is used to maintain batch records.
    """
    def __init__(self, mf, coy, nam, per, sys, rtn, multi=None, glint=None):
        self.mf = mf
        self.conum = coy
        self.conam = nam
        self.per = per
        self.sys = sys.upper()
        self.rtn = rtn
        self.multi = multi
        self.glint = glint
        self.setVariables()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, ["ctlbat", "ctlctl"], prog=__name__)
        t = time.localtime()
        self.sysdt = ((t[0] * 10000) + (t[1] * 100) + t[2])
        t = str(self.sysdt)
        t = (int(t[0:4]),int(t[4:6]), int(t[6:8]),0,0,0,0,0,0)
        self.curdt = ((t[0] * 100) + t[1])
        self.s_per = self.per[1][0] / 100
        self.e_per = self.per[2][0] / 100
        self.bankctl = False
        if self.sys == "ASS":
            self.ttp = artrtp
            self.nam = artrtp[self.rtn-1][1]
            if self.rtn in (1, 2) and self.glint == "Y":
                self.bankctl = True
        elif self.sys == "CRS":
            self.ttp = crtrtp
            self.nam = crtrtp[self.rtn-1][1]
            if self.rtn in (2, 5) and self.glint == "Y":
                self.bankctl = True
        elif self.sys == "DRS":
            self.ttp = drtrtp
            self.nam = drtrtp[self.rtn-1][1]
            if self.rtn in (2, 5) and self.glint == "Y":
                self.bankctl = True
        elif self.sys == "GEN":
            self.ttp = gltrtp
            self.nam = gltrtp[self.rtn-1][1]
            if self.rtn in (2, 6, 7):
                self.bankctl = True
        elif self.sys == "MEM":
            self.ttp = mltrtp
            self.nam = mltrtp[self.rtn-1][1]
            if self.rtn in (2, 5) and self.glint == "Y":
                self.bankctl = True
        elif self.sys == "RTL":
            self.ttp = rttrtp
            self.nam = rttrtp[self.rtn-1][1]
            if self.rtn in (2, 3) and self.glint == "Y":
                self.bankctl = True
        elif self.sys == "SLN":
            self.ttp = sltrtp
            self.nam = sltrtp[self.rtn-1][1]
            if self.rtn in (2, 4) and self.glint == "Y":
                self.bankctl = True
        elif self.sys == "STR":
            self.ttp = sttrtp
            self.nam = sttrtp[self.rtn-1][1]
        self.acc = None
        self.ctl = ""
        if self.bankctl:
            acc = self.sql.getRec(tables=["ctlctl"], cols=["ctl_code"],
                where=[("ctl_cono", "=", self.conum), ("ctl_code", "like",
                "bank_%")])
            if len(acc) == 1:
                self.ckCtl(acc[0][0])

    def doBatchHeader(self):
        tit = ("%s %s - Batch Details"% (finsys[self.sys][0], self.nam),)
        btm = {
            "stype": "R",
            "tables": ("ctlbat",),
            "cols": (
                ("btm_batno", "", 0, "Bat-Num"),
                ("btm_rtyp", ("xx", self.ttp), 20, "Type"),
                ("btm_curdt", "", 0, "Curr-Dt"),
                ("btm_multi", "", 0, "M")),
            "where": [
                ("btm_cono", "=", self.conum),
                ("btm_styp", "=", self.sys),
                ("btm_rtyp", "=", self.rtn),
                ("btm_curdt", "between", self.s_per, self.e_per),
                ("btm_ind", "=", "N")],
            "group": "btm_batno",
            "zero": "0"}
        ctl = {
            "stype": "R",
            "tables": ("ctlctl", "genmst"),
            "cols": (
                ("ctl_code", "", 0, "Ctl-Code"),
                ("ctl_conacc", "", 0, "G/L-Num"),
                ("glm_desc", "", 30, "Description")),
            "where": (
                ("ctl_cono", "=", self.conum),
                ("ctl_code", "like", "bank_%"),
                ("glm_cono=ctl_cono",),
                ("glm_acno=ctl_conacc",))}
        r1s = (("No", "N"), ("Yes", "Y"))
        fld = [
            (("T",0,0,0),"INa",7,"Batch Number","Batch Number",
                "","N",self.doBatNum,btm,None,("notblank",)),
            (("T",0,1,0),"ID1",10,"Capture Date (CCYYMMDD)","Capture Date "\
                "(YYYYMMDD)",self.sysdt,"N",self.doCapdt,None,None,("efld",)),
            (("T",0,2,0),"ID2",7,"Current Period (CCYYMM)","Current Period "\
                "(YYYYMM)",self.curdt,"N",self.doCurdt,None,None,("efld",)),
            (("T",0,3,0),"ISI",6,"Number of Entries","Number of Entries",
                "","N",self.doQty,None,None,("efld",)),
            (("T",0,4,0),"ISD",13.2,"Value of Entries","Value of Entries",
                "","N",self.doVal,None,None,("efld",)),
            (("T",0,5,0),"INA",5,"Batched By","",
                "","N",None,None,None,None),
            (("T",0,6,0),"INA",5,"Captured By","",
                "","N",self.doCapBy,None,None,None),
            (("T",0,7,0),("IRB",r1s),0,"Multiple Date Allocations","",
                "N","N",self.doMulti,None,None,None)]
        if self.bankctl:
            fld.append(
            (("T",0,8,0),"INA",10,"Bank Control","Bank Control Record",
                "","N",self.doCtl,ctl,None,("efld",)))
        tnd = ((self.doHeadEnd, "y"), )
        txt = (self.doHeadExit, )
        self.df = TartanDialog(self.mf, screen=self.mf.body, tops=True,
            title=tit, eflds=fld, tend=tnd, txit=txt)
        self.mf.startLoop()

    def doBatNum(self, frt, pag, r, c, p, i, w):
        self.batno = w
        whr = [
            ("btm_cono", "=", self.conum),
            ("btm_batno", "=", self.batno),
            ("btm_styp", "=", self.sys),
            ("btm_rtyp", "=", self.rtn),
            ("btm_curdt", "between", self.s_per, self.e_per)]
        self.bat = self.sql.getRec(tables="ctlbat", where=whr, limit=1)
        if not self.bat:
            self.new = True
        elif self.bat[self.sql.ctlbat_col.index("btm_ind")] == "Y":
            return "This Batch Has Aready Been Balanced"
        else:
            self.new = False
            end = self.sql.ctlbat_col.index("btm_ctl")
            if self.bankctl:
                end += 1
            for n, d in enumerate(self.bat[4:end]):
                self.df.loadEntry(frt, pag, p+n+1, data=d)
            self.ckCtl(self.bat[self.sql.ctlbat_col.index("btm_ctl")])
            return "sk2"

    def doCapdt(self, frt, pag, r, c, p, i, w):
        pass

    def doCurdt(self, frt, pag, r, c, p, i, w):
        if w < self.s_per or w > self.e_per:
            return "Invalid Period"
        if self.rtn == 7:
            self.df.loadEntry(frt, pag, p+1, data=0)
            self.df.loadEntry(frt, pag, p+2, data=0)
            return "sk2"

    def doQty(self, frt, pag, r, c, p, i, w):
        if not self.new:
            if w != self.bat[self.sql.ctlbat_col.index("btm_trno")]:
                cf = PwdConfirm(self.mf, conum=self.conum, system="MST",
                    code="ChgBatch")
                if cf.flag == "no":
                    return "Invalid Change of Batch Quantity"

    def doVal(self, frt, pag, r, c, p, i, w):
        if not self.new:
            if w != self.bat[self.sql.ctlbat_col.index("btm_trval")]:
                cf = PwdConfirm(self.mf, conum=self.conum, system="MST",
                    code="ChgBatch")
                if cf.flag == "no":
                    return "Invalid Change of Batch Value"
            return "nd"

    def doCapBy(self, frt, pag, r, c, p, i, w):
        if self.sys not in ("ASS", "CRS", "DRS", "GEN", "MEM", "RTL", "SLN"):
            if not self.bankctl:
                return "nd"
            elif self.acc:
                self.df.loadEntry(frt, pag, p+2, self.ctl)
                return "nd"
            else:
                return "sk1"
        if self.multi:
            self.df.loadEntry(frt, pag, p+1, data=self.multi.upper())
            if not self.bankctl:
                return "nd"
            elif self.acc:
                self.df.loadEntry(frt, pag, p+2, self.ctl)
                return "nd"
            else:
                return "sk1"

    def doMulti(self, frt, pag, r, c, p, i, w):
        if w == "Y":
            cf = PwdConfirm(self.mf, conum=self.conum, system="MST",
                code="MultiDate")
            if cf.flag == "no":
                return "Invalid Multiple Date Allocations"
        if not self.bankctl:
            return "nd"
        elif self.acc:
            self.df.loadEntry(frt, pag, p+1, self.ctl)
            return "nd"

    def doCtl(self, frt, pag, r, c, p, i, w):
        if w[:4] != "bank":
            self.acc =  None
            self.ctl = ""
        else:
            self.ckCtl(w)
        if not self.acc:
            return "Invalid Bank Control Code"

    def ckCtl(self, ctl):
        if not ctl:
            self.acc = None
            self.ctl = ""
            return
        self.ctl = ctl
        gc = GetCtl(self.mf)
        ctlctl = gc.getCtl("ctlctl", self.conum)
        if not ctlctl:
            self.acc = None
        else:
            if gc.chkRec(self.conum, ctlctl, [self.ctl]):
                self.acc = None
            else:
                self.acc = ctlctl[self.ctl]

    def doHeadEnd(self):
        self.capdt = self.df.t_work[0][0][1]
        self.curdt = self.df.t_work[0][0][2]
        self.trno = self.df.t_work[0][0][3]
        self.trval = self.df.t_work[0][0][4]
        self.batby = self.df.t_work[0][0][5]
        self.indby = self.df.t_work[0][0][6]
        self.multi = self.df.t_work[0][0][7]
        self.getTotals()
        if self.new:
            self.sql.insRec("ctlbat", data=[self.conum, self.sys, self.rtn,
                self.batno, self.capdt, self.curdt, self.trno, self.trval,
                self.batby, self.indby, self.multi, self.ctl, "N"])
        else:
            self.sql.updRec("ctlbat", cols=["btm_trno", "btm_trval"],
                data=[self.trno, self.trval], where=[("btm_cono", "=",
                self.conum), ("btm_styp", "=", self.sys), ("btm_rtyp", "=",
                self.rtn), ("btm_batno", "=", self.batno), ("btm_curdt", "=",
                self.curdt)])
        self.mf.dbm.commitDbase()
        self.df.closeProcess()
        self.mf.closeLoop()

    def getTotals(self):
        if self.sys == "ASS":
            tab = "asstrn"
            col = ["count(*)", "round(sum(ast_amt1), 2)",
                "round(sum(ast_vat), 2)"]
            whr = [("ast_cono", "=", self.conum), ("ast_type", "=", self.rtn),
                ("ast_batch", "=", self.batno), ("ast_curdt", "between",
                self.s_per, self.e_per), ("ast_batind", "=", "N")]
        elif self.sys == "CRS":
            tab = "crstrn"
            col = ["count(*)", "round(sum(crt_tramt), 2)"]
            whr = [("crt_cono", "=", self.conum), ("crt_type", "=", self.rtn),
                ("crt_batch", "=", self.batno), ("crt_curdt", "between",
                self.s_per, self.e_per), ("crt_batind", "=", "N")]
        elif self.sys == "DRS":
            tab = "drstrn"
            col = ["count(*)", "round(sum(drt_tramt), 2)"]
            whr = [("drt_cono", "=", self.conum), ("drt_type", "=", self.rtn),
                ("drt_batch", "=", self.batno), ("drt_curdt", "between",
                self.s_per, self.e_per), ("drt_batind", "=", "N")]
        elif self.sys == "GEN":
            tab = "gentrn"
            col = ["count(*)", "round(sum(glt_tramt), 2)"]
            whr = [("glt_cono", "=", self.conum)]
            if self.rtn == 7:
                whr.append(("glt_type", "in", (2, 4, 6)))
            else:
                whr.append(("glt_type", "=", self.rtn))
            whr.extend([("glt_batch", "=", self.batno), ("glt_curdt", "between",
                self.s_per, self.e_per), ("glt_batind", "=", "N")])
        elif self.sys == "MEM":
            tab = "memtrn"
            col = ["count(*)", "round(sum(mlt_tramt), 2)"]
            whr = [("mlt_cono", "=", self.conum), ("mlt_type", "=", self.rtn),
                ("mlt_batch", "=", self.batno), ("mlt_curdt", "between",
                self.s_per, self.e_per), ("mlt_batind", "=", "N")]
        elif self.sys == "RTL":
            tab = "rtltrn"
            col = ["count(*)", "round(sum(rtt_tramt), 2)"]
            whr = [("rtt_cono", "=", self.conum), ("rtt_type", "=", self.rtn),
                ("rtt_batch", "=", self.batno), ("rtt_curdt", "between",
                self.s_per, self.e_per), ("rtt_batind", "=", "N")]
        elif self.sys == "SLN":
            tab = "wagltf"
            col = ["count(*)", "round(sum(wlt_amt), 2)"]
            whr = [("wlt_cono", "=", self.conum), ("wlt_type", "=", self.rtn),
                ("wlt_batch", "=", self.batno), ("wlt_curdt", "between",
                self.s_per, self.e_per), ("wlt_batind", "=", "N")]
        elif self.sys == "STR":
            tab = "strtrn"
            col = ["count(*)", "round(sum(stt_cost), 2)"]
            whr = [("stt_cono", "=", self.conum), ("stt_type", "=", self.rtn),
                ("stt_batch", "=", self.batno), ("stt_curdt", "between",
                self.s_per, self.e_per), ("stt_batind", "=", "N")]
        sql = Sql(self.mf.dbm, tab, prog=__name__)
        a = sql.getRec(tables=tab, cols=col, where=whr, limit=1)
        if a[0]:
            self.batqty = CCD(a[0], "SI", 6).work
            if self.sys == "ASS":
                self.batval = CCD(float(ASD(a[1]) + ASD(a[2])), "SD", 13.2).work
            else:
                self.batval = CCD(a[1], "SD", 13.2).work
        else:
            self.batqty = CCD(0, "SI", 6).work
            self.batval = CCD(0, "SD", 13.2).work

    def doHeadExit(self):
        self.batno = None
        self.df.closeProcess()
        self.mf.closeLoop()

    def doBatchTotal(self, imp=False):
        self.qty = CCD(self.batqty, "SI", 6)
        self.val = CCD(self.batval, "SD", 13.2)
        self.qex = CCD(self.trno, "SI", 6)
        self.vex = CCD(self.trval, "SD", 13.2)
        self.qdf = CCD((self.qex.work - self.qty.work), "SI", 6)
        self.vdf = CCD(float(ASD(self.vex.work)-ASD(self.val.work)), "SD", 13.2)
        if imp:
            self.sql.updRec("ctlbat", cols=["btm_trno", "btm_trval",
                "btm_ind"], data=[self.qty.work, self.val.work, "Y"],
                where=[("btm_cono", "=", self.conum), ("btm_batno", "=",
                self.batno), ("btm_styp", "=", self.sys), ("btm_rtyp", "=",
                self.rtn), ("btm_curdt", "between", self.s_per, self.e_per)])
            self.flagTrans()
        elif not self.qty.work and not self.val.work:
            self.sql.delRec("ctlbat", where=[("btm_cono", "=", self.conum),
                ("btm_batno", "=", self.batno), ("btm_styp", "=", self.sys),
                ("btm_rtyp", "=", self.rtn), ("btm_curdt", "between",
                self.s_per, self.e_per)])
        elif self.qdf.work or self.vdf.work:
            self.doDispTotal()
            self.doLoadDetail()
            self.mf.startLoop()
        elif self.qty.work or self.val.work:
            self.sql.updRec("ctlbat", cols=["btm_ind"], data=["Y"],
                where=[("btm_cono", "=", self.conum), ("btm_batno", "=",
                self.batno), ("btm_styp", "=", self.sys), ("btm_rtyp", "=",
                self.rtn), ("btm_curdt", "between", self.s_per, self.e_per)])
            self.flagTrans()
        self.mf.dbm.commitDbase()

    def doDispTotal(self):
        tit = ("Batch Totals",)
        fld = (
            (("T",0,0,0),"ONa",7,"Batch Number"),
            (("T",0,1,0),"OSI",6,"Number of Entries Expected"),
            (("T",0,2,0),"OSI",6,"                  Entered"),
            (("T",0,3,0),"OSI",6,"                  Difference"),
            (("T",0,4,0),"OSD",13.2,"Value of Entries  Expected"),
            (("T",0,5,0),"OSD",13.2,"                  Entered"),
            (("T",0,6,0),"OSD",13.2,"                  Difference"))
        but = (("Continue",None,self.doTotExit,1,None,None),)
        tnd = ((self.doTotExit, "n"),)
        txt = (self.doTotExit,)
        self.df = TartanDialog(self.mf, screen=self.mf.body, tops=True,
            title=tit, eflds=fld, butt=but, tend=tnd, txit=txt)

    def doLoadDetail(self):
        self.df.loadEntry("T", 0, 0, data=self.batno)
        self.df.loadEntry("T", 0, 1, data=self.qex.work)
        self.df.loadEntry("T", 0, 2, data=self.qty.work)
        self.df.loadEntry("T", 0, 3, data=self.qdf.work)
        self.df.loadEntry("T", 0, 4, data=self.vex.work)
        self.df.loadEntry("T", 0, 5, data=self.val.work)
        self.df.loadEntry("T", 0, 6, data=self.vdf.work)
        self.mf.updateStatus("Batch Does Not Balance")
        self.df.B0.set_flags(gtk.CAN_FOCUS)
        self.df.B0.grab_focus()

    def flagTrans(self):
        if self.sys == "ASS":
            tab = "asstrn"
            col = ["ast_batind"]
            whr = [("ast_cono", "=", self.conum), ("ast_type", "=", self.rtn),
                ("ast_batch", "=", self.batno), ("ast_curdt", "between",
                self.s_per, self.e_per), ("ast_batind", "=", "N")]
        elif self.sys == "CRS":
            tab = "crstrn"
            col = ["crt_batind"]
            whr = [("crt_cono", "=", self.conum), ("crt_type", "=", self.rtn),
                ("crt_batch", "=", self.batno), ("crt_curdt", "between",
                self.s_per, self.e_per), ("crt_batind", "=", "N")]
        elif self.sys == "DRS":
            tab = "drstrn"
            col = ["drt_batind"]
            whr = [("drt_cono", "=", self.conum), ("drt_type", "=", self.rtn),
                ("drt_batch", "=", self.batno), ("drt_curdt", "between",
                self.s_per, self.e_per), ("drt_batind", "=", "N")]
        elif self.sys == "GEN":
            tab = "gentrn"
            col = ["glt_batind"]
            whr = [("glt_cono", "=", self.conum), ("glt_type", "=", self.rtn),
                ("glt_batch", "=", self.batno), ("glt_curdt", "between",
                self.s_per, self.e_per), ("glt_batind", "=", "N")]
        elif self.sys == "MEM":
            tab = "memtrn"
            col = ["mlt_batind"]
            whr = [("mlt_cono", "=", self.conum), ("mlt_type", "=", self.rtn),
                ("mlt_batch", "=", self.batno), ("mlt_curdt", "between",
                self.s_per, self.e_per), ("mlt_batind", "=", "N")]
        elif self.sys == "RTL":
            tab = "rtltrn"
            col = ["rtt_batind"]
            whr = [("rtt_cono", "=", self.conum), ("rtt_type", "=", self.rtn),
                ("rtt_batch", "=", self.batno), ("rtt_curdt", "between",
                self.s_per, self.e_per), ("rtt_batind", "=", "N")]
        elif self.sys == "SLN":
            tab = "wagltf"
            col = ["wlt_batind"]
            whr = [("wlt_cono", "=", self.conum), ("wlt_type", "=", self.rtn),
                ("wlt_batch", "=", self.batno), ("wlt_curdt", "between",
                self.s_per, self.e_per), ("wlt_batind", "=", "N")]
        elif self.sys == "STR":
            tab = "strtrn"
            col = ["stt_batind"]
            whr = [("stt_cono", "=", self.conum), ("stt_type", "=", self.rtn),
                ("stt_batch", "=", self.batno), ("stt_curdt", "between",
                self.s_per, self.e_per), ("stt_batind", "=", "N")]
        sql = Sql(self.mf.dbm, tab, prog=__name__)
        sql.updRec(tab, cols=col, data=["Y"], where=whr)

    def doTotExit(self, widget=None):
        self.df.closeProcess()
        self.mf.closeLoop()

class CCD:
    """
    Data Types
    ============================================================================
    CB = Check Box
    D0 = Date (CCYYMMDD)
    D1 = Date (CCYYMMDD)
    d1 = Date (CCYYMMDD) Alowing Zero
    D2 = Current Period Date (CCYYMM)
    d2 = Current Period Date (CCYYMM) Allowing Zero
    D3 = Date (DD-MMM-YYYY)
    DT = Date and Time (DD-MMM-YYYY-HH:MM)
    FF = File or Directory Format
    LA = Lower Case Alphanumeric (Must also have 'size' set)
    La = Lower Case Alphanumeric Right Justified (Must also have 'size' set)
    HA = Normal Alphanumeric Hidden as for Passwords
    NA = Normal Alphanumeric (Must also have 'size' set)
    Na = Normal Alphanumeric Right Justified (Must also have 'size' set)
    RW = Raw Data, No formatting or checking
    SD = Signed Decimal (size includes sign and decimals)
    Sd = Convert Integer to Signed Decimal (size includes sign and decimals)
    SI = Signed Integer (size includes sign)
    SL = Signed Long Integer (size includes sign)
    TM = Time (HH:MM)
    tM = Time (HH:MM) Allowing Zero
    TS = Timestamp (CCYY-MM-DD HH:MM:SS)
    TV = Text View - This is only used with TartanDialog for input purposes
    Tv = Text View - This is only used with TartanDialog for input purposes
    TX = Text or Blob Alphanumeric Variable Length - Storing Text Views etc.
    UA = Upper Case Alphanumeric (Must also have 'size' set)
    Ua = Upper Case Alphanumeric Right Justified (Must also have 'size' set)
    UD = Unsigned Decimal (size includes decimals)
    Ud = Convert Integer to Unsigned Decimal (size includes decimals)
    UI = Unsigned Integer
    UL = Unsigned Long Integer
    US = Unsigned Serial
    """
    def __init__(self, data, types, size=0):
        self.mthnum = {
            "JAN":1, "FEB":2, "MAR":3, "APR":4, "MAY":5, "JUN":6,
            "JUL":7, "AUG":8, "SEP":9, "OCT":10, "NOV":11, "DEC":12}
        #=======================================================================
        # Work around data types from TartanDialog which are 3 characters long
        #=======================================================================
        if len(types) == 3:
            types = types[1:]
        #=======================================================================
        # Work around for None
        #=======================================================================
        if data == None:
            if types[1].lower() in ("a", "v", "w", "x"):
                self.data = ""
            else:
                self.data = "0"
        #=======================================================================
        # Raw Data Type
        #=======================================================================
        if  types == "RW":
            self.err = ""
            self.work = self.disp = self.data = data
            return
        #=======================================================================
        # Timestamp
        #=======================================================================
        if  types == "TS":
            self.err = ""
            self.work = self.disp = self.data = data
            return
        #=======================================================================
        # Work around for previous format error
        #=======================================================================
        elif not types[0] == "T" and types[1].upper() != "A" and data == "None":
            self.data = "0"
        #=======================================================================
        else:
            self.data = str(data)
        self.data = self.data.strip()
        self.types = types
        self.size = size
        if types in ("D1","d1","D2","d2","DT","dT","TM","tM"):
            if self.data[-1:] == "L":
                self.data = self.data[:-1]
        elif types in ("Sd", "Ud"):
            self.data = str(float(long(self.data) / 100.0))
        elif types[1] in ("I","L"):
            self.data = self.data.split(".")[0]
            if self.data[-1:] == "L":
                self.data = self.data[:-1]
        elif self.types[1].lower() in ("a", "v", "x"):
            self.data = self.data.replace("\\", "/")
        self.processData()

    def processData(self):
        self.quo = int(self.size / 1)
        self.rem = int(round((self.size % 1) * 10))
        self.disp = None
        self.work = None
        self.err = ""
        if self.types in ("D0","d0"):
            #===================================================================
            # Work around for previous format error
            #===================================================================
            if self.data[3:6].upper() in self.mthnum:
                self.chkD3_Date()
            #===================================================================
            else:
                self.chkD0_Date()
        elif self.types in ("D1","d1"):
            self.chkD1_Date()
        elif self.types in ("D2","d2"):
            self.chkD2_Date()
        elif self.types in ("D3","d3"):
            self.chkD3_Date()
        elif self.types in ("DT","dT"):
            self.chkDT_Date()
        elif self.types in ("TM","tM"):
            self.chkTM_Time()
        elif self.types[1] in ("A","a"):
            self.data = self.remUnicodeErrors(self.data)
            dat = self.data[:self.quo]
            if self.types[1] == "A":
                if not self.data:
                    self.work = dat
                    self.disp = " " * self.quo
                else:
                    if self.types[0] == "L":
                        dat = dat.lower()
                    elif self.types[0] == "U":
                        dat = dat.upper()
                    self.work = dat
                    self.disp = dat.ljust(self.quo)
            elif self.types[1] == "a":
                try:
                    if dat[-1:] == "L":
                        raise exception
                    dat = str(long(dat))
                    if dat[-1:] == "L":
                        dat = dat[:-1]
                    self.work = dat.rjust(self.quo)
                    self.disp = dat.rjust(self.quo)
                except:
                    if self.types[0] == "L":
                        dat = dat.lower()
                    elif self.types[0] == "U":
                        dat = dat.upper()
                    self.work = dat
                    self.disp = dat.ljust(self.quo)
        elif self.types == "UI":
            self.unsignedInteger()
        elif self.types == "US":
            self.unsignedInteger()
        elif self.types == "SI":
            self.signedInteger()
        elif self.types == "UL":
            self.unsignedLong()
        elif self.types == "SL":
            self.signedLong()
        elif self.types in ("UD", "Ud"):
            self.unsignedDecimal()
        elif self.types in ("SD", "Sd"):
            self.signedDecimal()
        elif self.types in ("FF", "TV", "Tv", "TX"):
            self.data = self.remUnicodeErrors(self.data)
            self.work = self.data.strip()
            self.disp = self.data.strip()
        else:
            self.err = "Invalid Input, Retry (%s)" % self.types

    def remUnicodeErrors(self, data):
        #=======================================================================
        # Remove unicode errors
        #=======================================================================
        if data:
            try:
                text = unicode(data, errors="replace")
            except:
                text = ""
                for ch in data:
                    if ord(ch) > 127:
                        text = text + "?"
                    else:
                        text = text + ch
        else:
            text = data
        return str(text)

    def chkD0_Date(self):
        date = self.data.replace("-", "")
        self.disp = date
        try:
            self.work = int(date.replace("-", ""))
        except:
            self.work = date

    def chkD1_Date(self):
        date = self.data.replace("-", "")
        if date in ("", "0", "00000000") and (self.types == "d1"):
            self.work = 0
            self.disp = "0000-00-00"
        else:
            try:
                if len(date) < 5 or len(date) > 8:
                    raise Exception
                yy = int(date[0:(len(date) - 4)])
                mm = int(date[(len(date) - 4):(len(date) - 2)])
                dd = int(date[(len(date) - 2):len(date)])
                if yy < 100 and yy > 30:
                    yy += 1900
                elif yy < 100:
                    yy += 2000
                if mm < 1 or mm > 12:
                    self.err = "Invalid Date (D1-1)"
                    return
                chkmth = copy.deepcopy(mthnam)
                if mm == 2:
                    if (yy % 4) == 0:
                        chkmth[2][2] = 29
                    else:
                        chkmth[2][2] = 28
                if dd < 1 or dd > chkmth[mm][2]:
                    self.err = "Invalid Date (D1-2)"
                    return
                self.work = (yy * 10000) + (mm * 100) + dd
                self.disp = "%04i-%02i-%02i" % (yy, mm, dd)
                if len(self.disp) < self.size:
                    self.disp = self.disp + \
                    (" " * (int(self.size/1) - len(self.disp)))
            except:
                self.err = "Invalid Date (D1-3)"

    def chkD2_Date(self):
        date = self.data.replace("-", "")
        if date in ("", "0", "000000") and (self.types == "d2"):
            self.work = 0
            self.disp = "0000-00"
        else:
            try:
                if len(date) < 3 or len(date) > 6:
                    raise Exception
                yy = int(date[0:(len(date) - 2)])
                mm = int(date[(len(date) - 2):len(date)])
                if yy < 100 and yy > 30:
                    yy += 1900
                elif yy < 100:
                    yy += 2000
                if mm < 1 or mm > 12:
                    self.err = "Invalid Month (D2-1)"
                else:
                    self.work = (yy * 100) + mm
                    self.disp = "%04i-%02i" % (yy, mm)
                    if len(self.disp) < self.size:
                        self.disp = self.disp + \
                        (" " * (int(self.size/1) - len(self.disp)))
            except:
                self.err = "Invalid Current Date (D2-2)"

    def chkD3_Date(self):
        date = self.data
        self.work = date
        self.disp = date
        try:
            yy = int(date[7:11])
            mm = self.mthnum[date[3:6].upper()]
            dd = int(date[0:2])
            if yy < 100 and yy > 30:
                yy += 1900
            elif yy < 100:
                yy += 2000
            chkmth = copy.deepcopy(mthnam)
            if mm == 2:
                if (yy % 4) == 0:
                    chkmth[2][2] = 29
                else:
                    chkmth[2][2] = 28
            if mm < 1 or mm > 12:
                self.err = "Invalid Month (D3-1)"
            elif dd < 1 or dd > chkmth[mm][2]:
                self.err = "Invalid Day (D3-2)"
            else:
                self.isoDate = (yy * 10000) + (mm*100) + dd
        except:
            self.err = "Invalid Date (D3-3)"

    def chkDT_Date(self):
        date = self.data.replace(" ", "").replace("-", "").replace(":", "")
        if date in ("", "0", "000000000000") and (self.types == "dT"):
            self.work = 0
            self.disp = "0000-00-00 00:00"
            return
        try:
            if len(date) < 9 or len(date) > 12:
                raise Exception
            yy = int(date[0:(len(date) - 8)])
            mm = int(date[(len(date) - 8):(len(date) - 6)])
            dd = int(date[(len(date) - 6):(len(date) - 4)])
            hh = int(date[(len(date) - 4):(len(date) - 2)])
            tt = int(date[(len(date) - 2):len(date)])
            if yy < 100 and yy > 30:
                yy += 1900
            elif yy < 100:
                yy += 2000
            if mm < 1 or mm > 12:
                self.err = "Invalid Date (D1-1)"
                return
            chkmth = copy.deepcopy(mthnam)
            if mm == 2:
                if (yy % 4) == 0:
                    chkmth[2][2] = 29
                else:
                    chkmth[2][2] = 28
            if dd < 1 or dd > chkmth[mm][2]:
                self.err = "Invalid Date (D1-2)"
                return
            self.work = (yy*100000000)+(mm*1000000)+(dd*10000)+(hh*100)+tt
            self.disp = "%04i-%02i-%02i %02i:%02i" % (yy, mm, dd, hh, tt)
            if len(self.disp) < self.size:
                self.disp = self.disp + \
                (" " * (int(self.size/1) - len(self.disp)))
        except:
            self.err = "Invalid Date (DT)"

    def chkTM_Time(self):
        time = self.data.replace(":", "")
        if time in ("", "0", "0000") and (self.types == "tM"):
            self.work = 0
            self.disp = "00:00"
            return
        if len(time) > 4:
            time = time[:-2]
        try:
            if len(time) < 1:
                raise Exception
            time = int(time)
            hh = time / 100
            mm = time % 100
            self.work = (hh*100)+mm
            self.disp = "%02i:%02i" % (hh, mm)
            if len(self.disp) < self.size:
                self.disp = self.disp+(" "*(int(self.size/1)-len(self.disp)))
        except:
            self.err = "Invalid Time (%s)" % self.types

    def unsignedInteger(self):
        if not self.data:
            self.data = "0"
        try:
            if self.data.find("-") != -1:
                self.err = "Invalid UnSigned Integer (UI-)"
            else:
                self.work = int(self.data)
                self.disp = self.data.rjust(self.quo)
                if len(self.disp) > self.quo:
                    self.err = "Invalid UnSigned Integer (UIs)"
        except:
            self.err = "Invalid UnSigned Integer (UI)"

    def signedInteger(self):
        if not self.data:
            self.data = "0"
        try:
            s = self.data.find("-")
            if s == (len(self.data) - 1):
                self.data = self.data.replace("-", "")
                self.data = "-" + self.data
            self.work = int(self.data)
            s = self.data.find("-")
            if s != -1:
                self.data = self.data.replace("-", "")
                self.data = self.data + "-"
            else:
                self.data = self.data + " "
            self.disp = self.data.rjust(self.quo)
            if len(self.disp) > self.quo:
                self.err = "Invalid UnSigned Integer (UIs)"
        except:
            self.err = "Invalid Signed Integer (SI)"

    def unsignedLong(self):
        if not self.data:
            self.data = "0"
        try:
            if self.data.find("-") != -1:
                self.err = "Invalid UnSigned Long (UI-)"
            else:
                self.work = long(self.data)
                self.disp = self.data.rjust(self.quo)
                if len(self.disp) > self.quo:
                    self.err = "Invalid UnSigned Long (ULs)"
        except:
            self.err = "Invalid UnSigned Long (UL)"

    def signedLong(self):
        if not self.data:
            self.data = "0"
        try:
            s = self.data.find("-")
            if s == (len(self.data) - 1):
                self.data = self.data.replace("-", "")
                self.data = "-" + self.data
            self.work = long(self.data)
            s = self.data.find("-")
            if s != -1:
                self.data = self.data.replace("-", "")
                self.data = self.data + "-"
            else:
                self.data = self.data + " "
            self.disp = self.data.rjust(self.quo)
            if len(self.disp) > self.quo:
                self.err = "Invalid UnSigned Long (ULs)"
        except:
            self.err = "Invalid Signed Long (SL)"

    def unsignedDecimal(self):
        if not self.data:
            self.data = "0.0"
        try:
            if self.data.find("-") != -1:
                self.err = "Invalid UnSigned Decimal (UD-)"
            else:
                self.work = round(float(self.data), self.rem)
                self.data = str(self.work)
                dec = self.data.split(".")
                if len(dec[0]) > self.quo - (self.rem + 1):
                    self.err = "Invalid Unsigned Decimal (UDs)"
                elif len(dec[1]) > self.rem:
                    self.err = "Invalid Unsigned Decimal (UDd)"
                else:
                    self.data = self.data + ((self.rem - len(dec[1]))*"0")
                    self.disp = self.data.rjust(self.quo)
                    d = len(self.disp) - self.quo
                    if d > 0:
                        self.disp = self.disp[d:]
        except:
            self.err = "Invalid UnSigned Decimal (UD)"

    def signedDecimal(self):
        if not self.data:
            self.data = "0.0"
        try:
            if self.data[-1:] == "-":
                self.data = "-" + self.data[:-1]
            self.work = round(float(self.data), self.rem)
            if not self.work:
                self.work = 0.0
            self.data = str(self.work)
            dec = self.data.split(".")
            if len(dec[0]) > self.quo - (self.rem + 1):
                self.err = "Invalid Signed Decimal (SDs)"
            elif len(dec[1]) > self.rem:
                self.err = "Invalid Signed Decimal (SDd)"
            else:
                self.data = self.data + ((self.rem - len(dec[1]))*"0")
                if self.work < 0:
                    self.data = self.data.replace("-", "")
                    self.data = self.data + "-"
                else:
                    self.data = self.data + " "
                self.disp = self.data.rjust(self.quo)
                d = len(self.disp) - self.quo
                if d > 0:
                    self.disp = self.disp[d:]
        except:
            self.err = "Invalid Signed Decimal (SD)"

class ASD:
    """
    Used for Adding and Subtracting Decimal Amounts as follows:

    answer = float(ASD(amount1) + ASD(amount2))
    answer = float(ASD(amount1) - ASD(amount2))
    """
    def __init__(self, amount, decimals=2):
        self.decimals = decimals
        self.multiply = 1.0
        for x in range(decimals):
            self.multiply = self.multiply * 10
        self.amt = long(round((float(amount) * self.multiply), 0))

    def __add__(self, other):
        return ASD((self.amt + other.amt) / self.multiply)

    def __sub__(self, other):
        return ASD((self.amt - other.amt) / self.multiply)

    def __float__(self):
        return(round((self.amt / self.multiply), self.decimals))

class CsvCvt:
    """
    Used for parsing delimiter separated and fixed length files

    types = Type of file CSV, FLF or OUT
    delm  = Delimiter COM, SEM, TAB or SPC
    alpha = Character to enclose alpha fields SIC or DIC
    flds  = A list of field sizes for fixed length files
    """
    def __init__(self, line, **args):
        defaults = {
            "types": "CSV",
            "delm": "COM",
            "alpha": "DIC",
            "flds": None}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "%s = defaults[nam]" % nam
        if delm == "TAB":
            self.delm = "    "
        elif delm == "SEM":
            self.delm = ";"
        elif delm == "SPC":
            self.delm = " "
        else:
            self.delm = ","
        if alpha == "SIC":
            self.alpha = "'"
        else:
            self.alpha = '"'
        self.output = []
        if types == "CSV":
            self.read_csv(line)
        elif types == "FLF":
            self.read_fff(line, flds)

    def read_csv(self, line):
        """
        Method: CsvCvt.read_csv(line)
        Desc: parses a 'line' of delimiter separated values into 'self.output'
        """
        if line.find("\032") == -1 and len(line) != 1:
            infld = False
            field = ""
            for nn, ch in enumerate(line):
                if ch in (self.delm, chr(10), chr(13)):
                    if not infld:
                        self.output.append(field)
                        field = ""
                    else:
                        field = field + ch
                    if ch in (chr(10), chr(13)):
                        break
                elif ch == self.alpha:
                    if infld:
                        #if (nn + 1) == len(line):
                        #    infld = False
                        if line[nn + 1] in (chr(10), chr(13)):
                            infld = False
                        elif line[nn + 1] == self.delm:
                            infld = False
                        elif self.alpha == "'":
                            field = field + " "
                        elif self.alpha == '"':
                            field = field + "'"
                    else:
                        infld = True
                elif ch == "'":
                    field = field + " "
                else:
                    field = field + ch

    def read_fff(self, line, flds):
        """
        Method: CsvCvt.read_fff(line)
        Desc:   Parses a 'line' of fixed length field format into 'self.output'
        """
        if flds != None and line.find("\032") == -1 and len(line) != 1:
            if self.alpha != None:
                line = line.replace(alpha, "")
            line = line.replace("'", "")
            line = line.replace(chr(10), "")
            line = line.replace(chr(13), "")
            start = 0
            for f in flds:
                end = start + f
                self.output.append(line[start:end])
                start = end

    def write_csv(self, data):
        """
        data  = A list of data in CCD format to convert to csv
        """
        out = ""
        for d in data:
            if d.types[1].lower() == "a":
                if not out:
                    out = self.alpha + d.work + self.alpha
                else:
                    out = out + self.delm + self.alpha + d.work + self.alpha
            else:
                if not out:
                    out = str(d.work)
                else:
                    out = out + self.delm + str(d.work)
        return out + "\n"

class DBCreate:
    """
    This is used to setup a database and create necessary tables.
    The command line options are as follows:

    -c  = Do as Follows:
            i = Initial setup, i.e. create the database as well
            f = Create ftable and ffield Only, Unpopulated
            p = Create and Populate ftable and ffield Only, needs -s or -t
            t = Create ftable, ffield and Tables (Default), needs -s or -t
            g = Grant Privileges to group for all tables.
            u = Create Database User Only
    -d  = The path where the '.csv' and '.sys' files reside
    -i  = Interactive, True or False (False)
    -l  = Display a splash screen, True or False (False)
    -p  = Admin Password
    -r  = Auto creation of frelat and wagrcv data, True or False (True)
    -s  = System Name e.g. parlor referring to a '.sys' file
                       or
    -t  = String of comma separated Table Names if not a system
            referring to '.csv' files
    -u  = Admin User
    -w  = Commit transactions, True or False (True)
    -x  = Close the database on completion, True or False (True)
    """
    def __init__(self, mf=None, opts=None):
        if not mf:
            sys.exit()
        else:
            self.mf = mf
        self.act = False
        self.csvdir = os.path.join(self.mf.prgdir, "csv")
        self.put = "t"
        self.usr = mf.dbuser
        self.pwd = mf.dbpwd
        self.sys = ""
        self.tab = ""
        self.cmt = True
        self.xit = True
        self.rcv = True
        self.spl = False
        if opts:
            for o, v in opts:
                if o == "-i":
                    self.act = v
                elif o == "-c":
                    self.put = v.lower()
                elif o == "-d":
                    self.csvdir = v
                elif o == "-u":
                    self.usr = v
                elif o == "-p":
                    self.pwd = v
                elif o == "-r":
                    self.rcv = v
                elif o == "-s":
                    self.sys = v
                elif o == "-t":
                    self.tab = v.split(",")
                elif o == "-w":
                    self.cmt = v
                elif o == "-x":
                    self.xit = v
                elif o == "-l":
                    self.spl = v
        if self.mf.display:
            self.scrn = self.mf.body
        else:
            self.scrn = None
        if self.act:
            self.doInterActive()
            self.mf.startLoop()
            if self.put:
                self.spl = True
                if self.doCreate():
                    if self.cmt:
                        self.mf.dbm.commitDbase()
                    if self.xit:
                        self.mf.dbm.closeDbase()
        else:
            if self.tab:
                chk = self.doCheckTables(self.tab)
                if chk:
                    print "Csv for Table %s Does Not Exist" % chk
                    return
            if self.doCreate():
                if self.cmt:
                    self.mf.dbm.commitDbase()
                if self.xit:
                    self.mf.dbm.closeDbase()

    def doInterActive(self):
        tit = ("Database Creation Program",)
        mtd = {
            "stype": "C",
            "titl": "Table Creation Methods",
            "data": [
                ["i", "Initial setup, i.e. create the database as well"],
                ["f", "Create ftable and ffield tables only"],
                ["g", "Grant all privileges to database group"],
                ["p", "Create ftable, ffield and other tables"],
                ["t", "Create other Tables"],
                ["u", "Create user only"],
                ],
            "index": 0}
        sel = {
            "stype": "C",
            "head": "",
            "data": [],
            "mode": "",
            "retn": "D"}
        fle = {
            "stype": "F",
            "types": "dir",
            "initd": self.mf.prgdir}
        r1s = (("System","s"),("Tables","t"))
        fld = (
            (("T",0,0,0),"INA",30,"Administrator","Database Admin Username",
                self.usr,"N",self.doUsr,None,None,("notblank",)),
            (("T",0,1,0),"IHA",30,"Password","Password",
                self.pwd,"N",self.doPwd,None,None,None),
            (("T",0,2,0),"INA",1,"Method","Method",self.put,"N",
                self.doMethod,mtd,None,("in",("i","f","p","t","g","u"))),
            (("T",0,3,0),"INA",30,"User Name","User Name",
                "","N",self.doUsr,None,None,("notblank",)),
            (("T",0,4,0),"IHA",30,"Password","Password",
                "","N",self.doPwd,None,None,None),
            (("T",0,5,0),"IFF",50,"Path","Path to csv Files",
                self.csvdir,"N",self.doDir,fle,None,("dir",)),
            (("T",0,6,0),("IRB",r1s),0,"Type","System or Table Names",
                "t","N",self.doType,None,None,None),
            (("T",0,7,0),"INA",50,"Name","System or Table Names",
                "","N",self.doFiles,sel,None,("notblank",),None,
                "The system name or a comma delimited string of table names"))
        tnd = ((self.doEnd, "y"), )
        txt = (self.doExit, )
        self.df = TartanDialog(self.mf, screen=self.mf.body, title=tit,
            eflds=fld, tend=tnd, txit=txt)

    def doUsr(self, frt, pag, r, c, p, i, w):
        if c == 1:
            self.usr = w
        else:
            self.newusr = w

    def doPwd(self, frt, pag, r, c, p, i, w):
        if c == 2:
            self.pwd = w
            if self.put == "i":
                self.df.loadEntry(frt, pag, p+1, data=self.put)
                return "sk1"
        else:
            self.newpwd = w
            return "sk3"

    def doMethod(self, frt, pag, r, c, p, i, w):
        self.put = w
        if self.put == "g":
            return "nd"
        elif self.put != "u":
            return "sk2"

    def doDir(self, frt, pag, r, c, p, i, w):
        if os.path.isdir(w) == 0:
            return "You Must Enter a Valid Directory!"
        self.csvdir = w
        if self.put not in ("p", "t"):
            if self.put == "i":
                self.typ = "s"
                self.df.loadEntry(frt, pag, p+1, data=self.typ)
                self.df.loadEntry(frt, pag, p+2, data=self.sys)
            else:
                self.typ = "t"
                self.df.loadEntry(frt, pag, p+1, data=self.typ)
                self.df.loadEntry(frt, pag, p+2, data="")
            return "sk2"

    def doType(self, frt, pag, r, c, p, i, w):
        self.typ = w
        if self.typ == "s":
            self.df.topf[0][5][8]["head"] = "Available Systems"
            self.df.topf[0][5][8]["data"] = (
                ("tartan", "Tartan"),
                ("bwlcmp", "Bowling Clubs"),
                ("member", "Member's Ledger"))
            self.df.topf[0][5][8]["mode"] = "S"
            self.df.topf[0][5][8]["comnd"] = None
        else:
            data = []
            list = glob.glob(os.path.join(self.csvdir, "*_idx.csv"))
            list.sort()
            for flenam in list:
                nm = os.path.basename(flenam)
                nm = nm.split("_")
                ip = open(flenam, "r")
                li = ip.readline()
                ip.close()
                li = li.split(",")
                data.append([nm[0], li[0].replace('"', '')])
            self.df.topf[0][5][8]["titl"] = "Available Tables"
            self.df.topf[0][5][8]["data"] = data
            self.df.topf[0][5][8]["mode"] = "M"
            self.df.topf[0][5][8]["comnd"] = self.doFleCmd

    def doFleCmd(self, frt, pag, r, c, p, i, w):
        f = ""
        for t in w:
            f = "%s%s," % (f, t[0])
        if f:
            f = f[:-1]
        self.df.loadEntry(frt, pag, p, data=f)

    def doFiles(self, frt, pag, r, c, p, i, w):
        if self.typ == "s":
            self.sys = w
            self.tab = []
            if os.path.isfile(os.path.join(self.csvdir, self.sys+".sys")) == 0:
                return "System File Does Not Exist"
        else:
            self.sys = ""
            self.tab = w.replace(" ", "")
            self.tab = self.tab.split(",")
            chk = self.doCheckTables(self.tab)
            if chk:
                return "Csv for Table %s Does Not Exist" % chk

    def doEnd(self):
        self.df.closeProcess()
        self.mf.closeLoop()

    def doExit(self):
        self.put = None
        self.df.closeProcess()
        self.mf.closeLoop()

    def doCheckTables(self, tables):
        for tab in tables:
            fle = os.path.join(self.csvdir, tab.strip() + "_fld.csv")
            if os.path.isfile(fle) == 0:
                return tab

    def doCreate(self):
        self.setConstants()
        if self.put == "i":
            if self.mf.dbm.checkDbase():
                answer = askQuestion(self.mf.body, "Drop Database?",
                    "The database already exists, do you want to Drop it?",
                    default="no")
                if answer == "yes":
                    self.mf.dbm.dropDbase()
                    self.mf.dbm.createDbase()
                else:
                    return
            else:
                self.mf.dbm.createDbase()
            if not self.sys:
                self.sys = "tartan"
        self.mf.dbm.openDbase(dbuser=self.usr, dbpwd=self.pwd, levl=0)
        if self.spl:
            sp = SplashScreen(self.mf.body,
                "Creating Format and Data Tables\n\nPlease Wait...")
        if self.put == "i":
            self.createUser(self.usr, self.pwd)
            self.createInitialTables()
        elif self.put == "f":
            self.createInitialTables()
        elif self.put == "g":
            all = self.mf.sqlRec("Select ft_tabl from ftable group by ft_tabl")
            for tab in all:
                if self.mf.dbase == "FBird":
                    self.mf.sqlRec("Grant ALL on %s to PUBLIC" % tab[0])
                elif self.mf.dbase == "PgSQL":
                    self.mf.sqlRec("Grant ALL on %s to group %s" % (tab[0],
                        self.mf.dbname))
                    seq = self.mf.sqlRec("Select ff_name from ffield where "\
                        "ff_tabl = '%s' and ff_type = 'US'" % tab[0])
                    if seq:
                        self.mf.sqlRec("Grant ALL on %s_%s_seq to group %s" %
                            (tab[0], seq[0][0], self.mf.dbname))
        elif self.put == "u":
            self.createUser(self.newusr, self.newpwd)
        if self.put in ("i", "p", "t"):
            tables = []
            if self.sys:
                fl = open(os.path.join(self.csvdir, self.sys + ".sys"), "r")
                dat = fl.readlines()
                for tab in dat:
                    tables.append(tab.replace("\012", ""))
            elif self.tab:
                for tab in self.tab:
                    tables.append(tab.strip())
            self.setTableFields(["ftable","ffield"])
            self.createTables(tables)
        if self.spl:
            sp.closeSplash()
        return True

    def setConstants(self):
        if self.mf.dbase == "FBird":
            self.dtm = "timestamp"
            self.key = "varchar"
            self.var = "varchar"
            self.txt = "blob sub_type text"
            self.int = "integer"
            self.lng = "bigint"
            self.dec = "decimal"
            self.flt = "numeric"
            self.ser = "bigint"
            self.blb = "blob"
        elif self.mf.dbase == "MsSQL":
            self.dtm = "varchar"
            self.key = "varchar"
            self.var = "varchar"
            self.txt = "ntext"
            self.int = "int"
            self.lng = "bigint"
            self.dec = "decimal"
            self.flt = "float"
            self.ser = "integer not null identity(1, 1) primary key"
            self.blb = "sql_variant"
        elif self.mf.dbase == "MySQL":
            self.dtm = "date"
            self.key = "varchar"
            self.var = "char"
            self.txt = "text"
            self.int = "int"
            self.lng = "bigint"
            self.dec = "decimal"
            self.flt = "float"
            self.ser = "int unsigned not null auto_increment primary key"
            self.blb = "blob"
        elif self.mf.dbase == "PgSQL":
            self.dtm = "timestamp"
            self.key = "varchar"
            self.var = "varchar"
            self.txt = "text"
            self.int = "int4"
            self.lng = "int8"
            self.dec = "numeric"
            self.flt = "numeric"
            self.ser = "serial"
            self.blb = "varchar"
        elif self.mf.dbase == "SQLite":
            self.dtm = "varchar"
            self.key = "varchar"
            self.var = "varchar"
            self.txt = "blob"
            self.int = "integer"
            self.lng = "long"
            self.dec = "float"
            self.flt = "float"
            self.ser = "integer primary key"
            self.blb = "blob"

    def createUser(self, user, pwd):
        try:
            if self.mf.dbase == "FBird":
                from kinterbasdb import services
                conn = services.connect(host='localhost', user='sysdba',
                   password='masterkey')
                if conn.userExists(user):
                    return
                user = services.User(user)
                if pwd:
                    user.password = pwd
                conn.addUser(user)
            elif self.mf.dbase == "MsSQL":
                pass
            elif self.mf.dbase == "MySQL":
                self.mf.sqlRec("Use mysql")
                use = self.mf.sqlRec(("Select User from user where "\
                    "User = %s", (user,)), limit=1)
                if not use:
                    if pwd:
                        cmd = "Grant delete, insert, select, update on %s.* "\
                            "to %s@'%s' identified by '%s'" % (self.mf.dbname,
                            self.mf.dbuser, "%", self.mf.dbpwd)
                    else:
                        cmd = "Grant delete, insert, select, update on %s.* "\
                            "to %s@'%s'" % (self.mf.dbname,
                            self.mf.dbuser, "%")
                    self.mf.sqlRec(cmd)
                    self.mf.sqlRec("Flush Privileges")
                self.mf.sqlRec("Use %s" % self.mf.dbname)
            elif self.mf.dbase == "PgSQL":
                use = self.mf.sqlRec(("Select usename from pg_user where "\
                    "usename = %s", (user,)), limit=1)
                if not use:
                    if pwd:
                        cmd = "Create user %s with password '%s' "\
                            "in group %s" % (user, pwd, self.mf.dbname)
                    else:
                        cmd = "Create user %s "\
                            "in group %s" % (user, self.mf.dbname)
                    self.mf.sqlRec(cmd)
            elif self.mf.dbase == "SQLite":
                pass
        except:
            showException(self.scrn, self.mf.rcdic["wrkdir"],
                "(Error Creating User %s)" % user, dbm=self.mf.dbm)
            sys.exit()

    def createInitialTables(self):
        if self.put != "i":
            for table in ("ftable", "ffield"):
                self.mf.dbm.dropTable(table)
        sql = "Create table ftable ("\
            "ft_tabl %s(6),"\
            "ft_desc %s(30),"\
            "ft_seq  %s,"\
            "ft_type %s(1),"\
            "ft_key0 %s(20),"\
            "ft_key1 %s(20),"\
            "ft_key2 %s(20),"\
            "ft_key3 %s(20),"\
            "ft_key4 %s(20),"\
            "ft_key5 %s(20),"\
            "ft_key6 %s(20),"\
            "ft_key7 %s(20),"\
            "ft_key8 %s(20),"\
            "ft_key9 %s(20))" % (
            self.key, self.key, self.int, self.key, self.key, self.key,
            self.key, self.key, self.key, self.key, self.key, self.key,
            self.key, self.key)
        self.mf.sqlRec(sql)
        if self.mf.dbase == "FBird":
            self.mf.dbm.commitDbase()
        sql = "Create unique index ftable_key1 on ftable %s"
        sql = sql % "(ft_tabl, ft_seq)"
        self.mf.sqlRec(sql)
        if self.mf.dbase == "FBird":
            self.mf.sqlRec("Grant ALL on ftable to PUBLIC")
        elif self.mf.dbase == "PgSQL":
            self.mf.sqlRec("Grant ALL on ftable to group %s" % self.mf.dbname)
        if self.mf.dbase == "FBird":
            sql = "Create table ffield ("\
                "ff_tabl %s(6),"\
                "ff_seq  %s,"\
                "ff_name %s(20),"\
                "ff_type %s(2),"\
                "ff_size %s(6,2),"\
                "ff_desc %s(30),"\
                "ff_head %s(30))" % (self.key, self.int, self.key, self.key,
                    self.flt, self.key, self.key)
        else:
            sql = "Create table ffield ("\
                "ff_tabl %s(6),"\
                "ff_seq  %s,"\
                "ff_name %s(20),"\
                "ff_type %s(2),"\
                "ff_size %s,"\
                "ff_desc %s(30),"\
                "ff_head %s(30))" % (self.key, self.int, self.key, self.key,
                    self.flt, self.key, self.key)
        self.mf.sqlRec(sql)
        if self.mf.dbase == "FBird":
            self.mf.dbm.commitDbase()
        sql = "Create unique index ffield_key1 on ffield %s"
        sql = sql % "(ff_name)"
        self.mf.sqlRec(sql)
        if self.mf.dbase == "FBird":
            self.mf.sqlRec("Grant ALL on ffield to PUBLIC")
        elif self.mf.dbase == "PgSQL":
            self.mf.sqlRec("Grant ALL on ffield to group %s" % self.mf.dbname)
        key = (
            ("ftable","Tables",1,"U","ft_name","ft_seq"),
            ("ffield","Fields",1,"U","ff_tabl","ff_name"))
        det = (
            ("ftable",0,"ft_tabl","NA","6.0","Table Name","None"),
            ("ftable",1,"ft_desc","NA","30.0","Description","None"),
            ("ftable",2,"ft_seq","UI","1.0","Key Sequence","None"),
            ("ftable",3,"ft_type","NA","1.0","Key Type","None"),
            ("ftable",4,"ft_key0","NA","20.0","First Key","None"),
            ("ftable",5,"ft_key1","NA","20.0","Second Key","None"),
            ("ftable",6,"ft_key2","NA","20.0","Third Key","None"),
            ("ftable",7,"ft_key3","NA","20.0","Fourth Key","None"),
            ("ftable",8,"ft_key4","NA","20.0","Fifth Key","None"),
            ("ftable",9,"ft_key5","NA","20.0","Sixth Key","None"),
            ("ftable",10,"ft_key6","NA","20.0","Seventh Key","None"),
            ("ftable",11,"ft_key7","NA","20.0","Eighth Key","None"),
            ("ftable",12,"ft_key8","NA","20.0","Ninth Key","None"),
            ("ftable",13,"ft_key9","NA","20.0","Tenth Key","None"),
            ("ffield",0,"ff_tabl","NA","6.0","Table Name","None"),
            ("ffield",1,"ff_seq","UI","3.0","Table Seq","None"),
            ("ffield",2,"ff_name","NA","20.0","Field Seq","None"),
            ("ffield",3,"ff_type","NA","2.0","Field Type","None"),
            ("ffield",4,"ff_size","UD","6.2","Field Size","None"),
            ("ffield",5,"ff_desc","NA","30.0","Field Description","None"),
            ("ffield",6,"ff_head","NA","30.0","Field Heading","None"))
        sql = "Insert into ftable (ft_tabl,ft_desc,ft_seq,ft_type,ft_key0,"\
            "ft_key1,ft_key2,ft_key3,ft_key4,ft_key5,ft_key6,ft_key7,"\
            "ft_key8,ft_key9) values ('%s','%s',%s,'%s','%s','%s','',"\
            "'','','','','','','')"
        for x in range(0, len(key)):
            self.mf.sqlRec(sql % key[x])
        sql = "Insert into ffield (ff_tabl,ff_seq,ff_name,ff_type,ff_size,"\
            "ff_desc,ff_head) values ('%s',%s,'%s','%s',%s,'%s','%s')"
        for x in range(0, len(det)):
            self.mf.sqlRec(sql % det[x])
        if self.cmt:
            self.mf.dbm.commitDbase()

    def createTables(self, tables):
        for table in tables:
            if self.put == "t":
                self.mf.dbm.dropTable(table)
            self.tableCreate(table)

    def tableCreate(self, table):
        self.mf.sqlRec("Delete from ftable where ft_tabl = '%s'" % table)
        self.mf.sqlRec("Delete from ffield where ff_tabl = '%s'" % table)
        pth = os.path.join(self.csvdir, table)
        if os.path.isfile("%s_fld.csv" % pth) == 0:
            if not self.mf.display:
                print "FLD Error %s" % table
            else:
                showError(self.mf.body, "Error", "FLD Not Found: %s" % table)
            sys.exit()
        fl = open("%s_fld.csv" % pth, "r")
        ll = fl.readlines()
        ff = []
        for x in range(0, len(ll)):
            data = [table,]
            data.extend(CsvCvt(ll[x]).output)
            ff.append(data)
        if os.path.isfile("%s_idx.csv" % pth) == 0:
            if not self.mf.display:
                print "IDX Error %s" % table
            else:
                showError(self.mf.body, "Error", "IDX Not Found: %s" % table)
            sys.exit()
        fk = open("%s_idx.csv" % pth, "r")
        lk = fk.readlines()
        fk = []
        for x in range(0, len(lk)):
            line = lk[x].replace("\012", "")
            cols = line.split(",")
            data = [table,]
            for y in range(0, len(cols)):
                col = cols[y].replace('"', "")
                if y == 1:
                    data.append(int(col))
                else:
                    data.append(col.strip())
            for y in range(13-len(cols)):
                data.append("")
            fk.append(data)
        x = 0
        self.sc1_array = makeArray(5, 200, 1, "S")
        for tab, seq, nam, typ, siz, des, hed in ff:
            self.sc1_array[x][0] = nam
            self.sc1_array[x][1] = typ
            self.sc1_array[x][2] = siz
            self.sc1_array[x][3] = des
            self.sc1_array[x][4] = hed
            x += 1
        seq_1 = 0
        for nam, typ, siz, des, hed in self.sc1_array:
            if not nam:
                break
            if hed == None:
                hed = ""
            data = (self.ffield_l1, table, seq_1, nam, typ, siz, des, hed)
            self.mf.sqlRec( "Insert into ffield (%s) values ('%s',%s,"\
                "'%s','%s',%s,'%s','%s')" % data)
            seq_1 += 1
        x = 0
        self.sc2_array = makeArray(13,10,1,"S")
        for tab, dsc, seq, typ, k0, k1, k2, k3, k4, k5, k6, k7, k8, k9 in fk:
            self.sc2_array[x][0] = dsc
            self.sc2_array[x][1] = seq
            self.sc2_array[x][2] = typ
            self.sc2_array[x][3] = k0
            self.sc2_array[x][4] = k1
            self.sc2_array[x][5] = k2
            self.sc2_array[x][6] = k3
            self.sc2_array[x][7] = k4
            self.sc2_array[x][8] = k5
            self.sc2_array[x][9] = k6
            self.sc2_array[x][10] = k7
            self.sc2_array[x][11] = k8
            self.sc2_array[x][12] = k9
            x += 1
        seq_2 = 0
        for nam,seq,typ,k1,k2,k3,k4,k5,k6,k7,k8,k9,k10 in self.sc2_array:
            if not seq:
                if seq_2 != 0:
                    break
                else:
                    seq = 1
            if seq == 1:
                data = (self.ftable_l1,table,nam,seq,typ,
                k1,k2,k3,k4,k5,k6,k7,k8,k9,k10)
            else:
                data = (self.ftable_l1,table,"",seq,typ,
                k1,k2,k3,k4,k5,k6,k7,k8,k9,k10)
            self.mf.sqlRec( "Insert into ftable (%s) values ('%s','%s',%s,"\
                "'%s','%s','%s','%s','%s','%s','%s','%s','%s','%s','%s')" % \
                data)
            seq_2 += 1
        if seq_1 != 0 and seq_2 != 0:
            if self.put in ("i", "t"):
                self.tableRestCreate(table)
            if self.cmt:
                self.mf.dbm.commitDbase()
        else:
            self.mf.dbm.rollbackDbase()

    def tableRestCreate(self, table):
        self.setTableFields([table])
        keys = self.mf.sqlRec("Select %s from ftable where "\
            "ft_tabl = '%s' order by ft_seq" % (self.ftable_l1, table))
        fields = self.mf.sqlRec( "Select %s from ffield where "\
            "ff_tabl = '%s' order by ff_seq" % (self.ffield_l1, table))
        key = []
        for x in range(0, len(keys)):
            k = ""
            for y in range(4, len(keys[x])):
                if keys[x][y]:
                    k = k + keys[x][y] + ","
            if k:
                l = k.rfind(",")
                key.append(k[:l])
        query = "Create table %s(" % table
        sequence = ""
        for t, q, n, y, s, d, e in fields:
            query = query + n
            if y[0] in ("D","d"):
                if y[1] in ("1","2"):
                    query = query + " %s," % self.int
                elif y[1] == "T":
                    query = query + " %s," % self.lng
            elif y[0] in ("H","N","U") and y[1] in ("A","a"):
                if s:
                    s = str(int(s))
                    query = query + " %s(" % self.var + s + "),"
                else:
                    query = query + " %s," % self.var
            elif y[0] in ("S","T","U"):
                if y[1] in ("D","d"):
                    if y[0] == "S":
                        s = s - 1
                    query = query + " %s(" % self.dec + str(s) + "),"
                elif y[1] == "L":
                    query = query + " %s," % self.lng
                elif y[1] in ("I","M"):
                    query = query + " %s," % self.int
                elif y[1] == "S":
                    sequence = n
                    query = query + " %s," % self.ser
                elif y[1] == "X":
                    query = query + " %s," % self.txt
            else:
                return "Invalid Data Type"
        query = query.replace(".", ",")
        l = query.rfind(",")
        query = query[:l] + ")"
        self.mf.sqlRec(query)
        if self.mf.dbase == "FBird":
            self.mf.dbm.commitDbase()
        for x in range(0, len(key)):
            if sys.platform == "win32" or self.mf.dbase in ("FBird", "MySQL"):
                if table in ("memkon", "emllog"):
                    continue
            if keys[x][3] == "U":
                sql = "Create unique index %s_key%s on %s (%s)" % \
                    (table, str(keys[x][2]), table, key[x])
            elif keys[x][3] == "N":
                sql = "Create index %s_key%s on %s (%s)" % \
                    (table, str(keys[x][2]), table, key[x])
            self.mf.sqlRec(sql)
        if self.mf.dbase == "FBird":
            self.mf.sqlRec("Grant ALL on %s to PUBLIC" % table)
            if sequence:
                self.mf.sqlRec("Create generator gen_%s_id" % table)
                self.mf.sqlRec("Create trigger %s_bi for %s active before "\
                    "insert position 0 as begin if (new.%s = 0 or new.%s "\
                    "is null) then new.%s = gen_id(gen_%s_id,1); end" % (table,
                    table, sequence, sequence, sequence, table))
                self.mf.sqlRec("Set generator gen_%s_id to 0" % table)
            self.mf.dbm.commitDbase()
        elif self.mf.dbase == "PgSQL":
            self.mf.sqlRec("Grant ALL on %s to group %s" % \
                (table, self.mf.dbname))
            if sequence:
                self.mf.sqlRec("Grant ALL on %s_%s_seq to group %s" % \
                    (table, sequence, self.mf.dbname))
        if table == "ctlpwu" and self.put == "i":
            sql = Sql(self.mf.dbm, table, prog=__name__)
            sql.insRec(table, data=["admin","System Administrator","",0,"",9])
        elif table in ("frelat", "wagrcv") and self.rcv:
            sql = Sql(self.mf.dbm, table, prog=__name__)
            dat = "%s.dat" % os.path.join(self.csvdir, table)
            if os.path.isfile(dat) == 0:
                pass
            else:
                fle = open(dat, "r")
                lin = fle.readlines()
                for l in lin:
                    if l[0] == "#":
                        continue
                    l = l.replace(chr(10), "")
                    l = l.replace(chr(13), "")
                    l = l.replace(chr(34), "")
                    l = l.replace(chr(39), "")
                    l = l.split("|")
                    if table == "wagrcv":
                        l[0] = int(l[0])
                    sql.insRec(table, data=l)

    def setTableFields(self, tables):
        for tab in tables:
            fd, k1, k2, k3, l1, l2, l3 = self.getFields(tab)
            exec "self.%s_l1 = l1" % tab

    def getFields(self, tab):
        data = self.mf.sqlRec("Select ft_tabl, ft_desc, ft_seq, ft_type, "\
            "ft_key0, ft_key1, ft_key2, ft_key3, ft_key4, ft_key5, ft_key6, "\
            "ft_key7, ft_key8, ft_key9 from ftable where ft_tabl = '%s' "\
            "order by ft_seq" % tab)
        key1 = data[0][4]
        key2 = str(data[0][4]) + "=?"
        key3 = [data[0][4],]
        for x in range(5, len(data[0])):
            if data[0][x]:
                key1 = key1 + "," + data[0][x]
                key2 = key2 + " and " + data[0][x] + "=?"
                key3.append(data[0][x])
        data = self.mf.sqlRec("Select ff_name, ff_type, ff_size, ff_seq "\
            "from ffield where ff_tabl = '%s' order by ff_seq" % tab)
        dic1 = {}
        dic1[data[0][0]] = [tab,1,data[0][1],data[0][2]]
        lst1 = data[0][0]
        lst2 = data[0][0] + "=?"
        lst3 = [data[0][0],]
        for x in range(1, len(data)):
            lst1 = lst1 + "," + data[x][0]
            lst2 = lst2 + "," + data[x][0] + "=?"
            lst3.append(data[x][0])
            dic1[data[x][0]] = [tab,x+1,data[x][1],data[x][2]]
        return (dic1, key1, key2, key3, lst1, lst2, lst3)

class FinReport:
    """
    This class generates a list comprising:

        Type of line i.e. H, L, G, T, S, U, C, P etc.
        Account Type (B,P,O)
        Print Highlighted (y or n)
        Formfeed (y/n)
        Account Number in the case of 'L'
        Description i.e. Description or Single/Double/Blank for underlines
        Line count
        Number of lines to skip
        List of Amounts (N/A,Lyr,Cyr,Bud,Ytd,Btd) for type "N"
        List of Monthly Amounts (Balances or Values) x 14 i.e.
            Opening Balance + 12 Monthly Balances + Closing Balance for type 'M'
        Chart Label i.e.
            Whether to include this total in a bar or line chart
        Page Heading i.e. Balance Sheet, Profit and Loss or Other
        Normal sign (+/-)

    Using the report file as follows:
    
    Num  Field       FF  Size  Description                     Heading
    ---  ----------  --  ----  ------------------------------  -----------
      0  glr_cono    UI   3.0  Company Number                  Coy
      1  glr_repno   UI   3.0  Report Number                   No
      2  glr_seq     UI   5.0  Sequence                        Seq
      3  glr_type    UA   1.0  Type                            T
      4  glr_desc    NA  30.0  Description                     Description
      5  glr_high    UA   1.0  Highlight (y/n)                 H
      6  glr_ffeed   UA   1.0  Form Feed (y/n)                 F
      7  glr_ignore  UA   1.0  Ignore Account Type (y/n)       I
      8  glr_from    UI   7.0  From Account                    From-Ac
      9  glr_to      UI   7.0  To Account                      To-Ac
     10  glr_obal    UA   1.0  Include Opening Balance (y/n)   B
     11  glr_accum   UA   1.0  Accumulate Month Values (y/n)   A
     12  glr_print   UA   1.0  Print Values (y/n/+/-)          P
     13  glr_norm    UA   1.0  Normal Sign (+/-)               S
     14  glr_acbal   UA   1.0  Add/Sub/Ignore Balance (a/s/i)  A
     15  glr_store   UA   1.0  Store Balances (y/n)            S
     16  glr_snum1   UI   2.0  Storage Number 1                N1
     17  glr_snum2   UI   2.0  Storage Number 2                N2
     18  glr_snum3   UI   2.0  Storage Number 3                N3
     19  glr_acstr   UA   1.0  Add/Sub/Ignore Storage (a/s/i)  A
     20  glr_group   UI   3.0  Group Number                    Grp
     21  glr_total   UI   1.0  Total Level                     T
     22  glr_clear   UA   1.0  Clear Total/Storage Level(s)    C
     23  glr_strper  UD   6.2  Percentage of Stored Amount     %-tage
     24  glr_uline   UA   1.0  Underline (b/d/s)               U
     25  glr_cbase   UA   1.0  Calculation Base (p/a/s)        B
     26  glr_ctype   UA   1.0  Calculation Type (+ - * /)      C
     27  glr_camnt   SD  13.2  Percent or Amount               Amount
     28  glr_label   NA  10.0  Chart Label, Space = None       Label

    as well as the following parameters:

    conum    = The Company to be used
    period   = The Financial Period to be used
    repco    = The report layout company number or 0 for a general report
    curdt    = The current financial period of the report e.g. 200302
    vcode    = The value code to be used:
                V = Values
                B = Budgets
                C = Combination of Values and Budgets
                X = Variance to Budgets
                D = Details
    dcode    = The detail code to be used if vcode = "D"
    consol   = A tuple or list of Company Numbers to Consolidate or "Y" for All
                e.g. (1,2,3,4,5,6) or "D" for Departmental or "N" for None
    rtype    = The basis for the printing parameter to be based on
                Y = Use the YTD value or
                M = Use the current period value
    """
    def __init__(self, mf, conum, period, repco, repno, curdt, **args):
        self.mf = mf
        self.conum = conum
        self.period = period
        self.repco = repco
        self.repno = repno
        self.curdt = curdt
        defaults = {
            "vcode": "V",
            "dcode": None,
            "rtype": "Y",
            "consol": "N"}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        no = self.setVariables()
        if not no:
            self.pb = ProgressBar(self.mf.body, typ="G")
            self.mainProcess()
            self.pb.closeProgress()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, ["ctlmst", "genbal", "genbud", "gendtm",
            "gendtt", "genmst", "genrpt", "gentrn"], prog=__name__)
        self.s_per = self.period[1][0] / 100
        self.e_per = self.period[2][0] / 100
        if self.period[0] > 0:
            self.s_lyr, e_lyr, fin = getPeriods(self.mf, self.conum,
                (self.period[0] - 1))
        self.cmth = self.curdt % 100
        self.lines = 1
        self.count = 0
        self.grpind = False
        self.exit = False
        self.clearData()
        self.allFields = []
        self.val_tot = makeArray(6, 10, 1)
        self.mth_tot = makeArray(14, 10, 1)
        self.val_str = makeArray(6, 100, 1)
        self.mth_str = makeArray(14, 100, 1)
        if type(self.consol) != list:
            if self.consol == "Y":
                con = "("
                coy = self.sql.getRec(tables="ctlmst", cols=["ctm_cono"],
                    order="ctm_cono")
                if not coy:
                    return "no"
                self.consol = []
                for c in coy:
                    self.consol.append(c[0])
            else:
                self.consol = None

    def mainProcess(self):
        if self.vcode == "D":
            self.getDcode()
        self.rep_num = 0
        self.rpt_lst = self.sql.getRec(tables="genrpt", where=[("glr_cono",
            "=", self.repco), ("glr_repno", "=", self.repno)], order="glr_seq")
        if self.rpt_lst and self.rpt_lst[self.rep_num][3] in ("P", "B", "O"):
            self.s_type = self.rpt_lst[0][3]
            self.atype = self.s_type
            if self.s_type == "P":
                self.stanam = "Profit and Loss"
            elif self.s_type == "B":
                self.stanam = "Balance Sheet"
            elif self.s_type == "O":
                self.stanam = self.rpt_lst[self.rep_num][4]
            while not self.exit:
                self.processRecords()

    def getDcode(self):
        acc = self.sql.getRec(tables=["gendtm", "gendtt"],
            cols=["gdm_desc", "sum(gdt_value)"], where=[("gdm_cono", "=",
            self.conum), ("gdm_code", "=", self.dcode), ("gdt_cono=gdm_cono",),
            ("gdt_cono=gdm_cono",), ("gdt_code=gdm_code",),
            ("gdt_curdt", "between", self.s_per, self.curdt)],
            group="gdm_desc", limit=1)
        self.ddet = [acc[0], CCD(acc[1], "SD", 13.2).disp]

    def processRecords(self):
        if self.grpind:
            self.grpind = False
        else:
            self.nextRepRecord()
            self.storeRepRecord()
            if self.rpt_dic == {}:
                self.exit = True
        if not self.exit:
            if self.s_type == "H":
                self.doHead()
            elif self.s_type == "L":
                self.doLedger()
            elif self.s_type == "G":
                self.doGroup()
            elif self.s_type == "T":
                self.doTotal()
            elif self.s_type == "S":
                self.doStore()
            elif self.s_type == "U":
                self.doUnderline()
            elif self.s_type == "P":
                self.doPercent()
            elif self.s_type == "C":
                self.doCalc()

    def doHead(self):
        self.pb.displayProgress()
        if self.rpt_dic["glr_ignore"] == "Y":
            if self.atype == "P":
                self.checkAccountType("B")
            else:
                self.checkAccountType("P")
        self.printLine()
        self.clearData()

    def doLedger(self):
        accs = self.readGenmst()
        for acc in accs:
            self.pb.displayProgress()
            self.lstdicGenmst(acc)
            self.s_cono = self.gen_dic["glm_cono"]
            self.s_acno = self.gen_dic["glm_acno"]
            self.s_desc = self.gen_dic["glm_desc"]
            self.accumData()
            if self.s_accum == "Y":
                # Accumulate progressive monthly totals
                for mth in range(1, 13):
                    self.mth[mth] = float(ASD(self.mth[mth]) +
                        ASD(self.mth[mth-1]))
            if self.rtype == "Y":
                amt = self.val[4]
            else:
                amt = self.mth[self.cmth]
            if self.s_print == "+" and amt < 0:
                self.clearData()
                continue
            elif self.s_print == "-" and amt >= 0:
                self.clearData()
                continue
            elif self.s_print == "N":
                pass
            else:
                self.printLine()
            self.accumTotals(self.s_acbal)
            if self.s_store == "Y":
                self.storeBalances()
            self.clearData()

    def doGroup(self):
        self.grpind = True
        while self.rpt_dic and self.rpt_dic["glr_type"] == "G" and \
                        self.rpt_dic["glr_group"] == self.s_group:
            accs = self.readGenmst()
            for acc in accs:
                self.pb.displayProgress()
                self.lstdicGenmst(acc)
                self.s_cono = self.gen_dic["glm_cono"]
                self.s_acno = self.gen_dic["glm_acno"]
                self.accumData()
            self.nextRepRecord()
        if self.s_accum == "Y":
            # Accumulate progressive monthly totals
            for mth in range(1, 13):
                self.mth[mth] = float(ASD(self.mth[mth]) + ASD(self.mth[mth-1]))
        if self.rtype == "Y":
            amt = self.val[4]
        else:
            amt = self.mth[self.cmth]
        if self.s_print == "+" and amt < 0:
            pass
        elif self.s_print == "-" and amt >= 0:
            pass
        else:
            if self.s_print == "Y":
                self.printLine()
            self.accumTotals(self.s_acbal)
            if self.s_store == "Y":
                self.storeBalances()
        self.clearData()
        self.storeRepRecord()

    def doTotal(self):
        self.pb.displayProgress()
        for x in range(0, 6):
            self.val[x] = self.val_tot[self.s_total][x]
        for x in range(0, 14):
            self.mth[x] = self.mth_tot[self.s_total][x]
        if self.rtype == "Y":
            amt = self.val[4]
        else:
            amt = self.mth[self.cmth]
        if self.s_print == "+" and amt < 0:
            self.clearData()
            return
        elif self.s_print == "-" and amt >= 0:
            self.clearData()
            return
        elif self.s_print == "N":
            pass
        else:
            self.printLine()
        if self.s_store == "Y":
            self.storeBalances()
        self.clearData()
        if self.s_clear == "Y":
            for x in range(0, self.s_total+1):
                for y in range(0, 6):
                    self.val_tot[x][y] = 0
                for y in range(0, 14):
                    self.mth_tot[x][y] = 0

    def doStore(self):
        self.pb.displayProgress()
        for x in range(0, 6):
            if self.val_str[self.s_snum1][x] == 0:
                self.val[x] = 0
            else:
                self.val[x] = round((self.val_str[self.s_snum1][x] *
                self.s_strper / 100.0), 2)
        for x in range(0, 13):
            if self.mth_str[self.s_snum1][x] == 0:
                self.mth[x] = 0
            else:
                self.mth[x] = round((self.mth_str[self.s_snum1][x] *
                    self.s_strper / 100.0), 2)
                self.mth[13] = float(ASD(self.mth[13]) + ASD(self.mth[x]))
        if self.rtype == "Y":
            amt = self.val[4]
        else:
            amt = self.mth[self.cmth]
        if self.s_print == "+" and amt < 0:
            self.clearData()
            return
        elif self.s_print == "-" and amt >= 0:
            self.clearData()
            return
        elif self.s_print == "N":
            pass
        else:
            self.printLine()
        self.accumTotals(self.s_acbal)
        self.clearData()
        if self.s_clear == "Y":
            for x in range(0, 6):
                self.val_str[self.s_snum1][x] = 0
            for x in range(0, 14):
                self.mth_str[self.s_snum1][x] = 0

    def doUnderline(self):
        self.pb.displayProgress()
        if self.rpt_dic["glr_uline"] == "B":
            self.s_desc = "Blank"
        elif self.rpt_dic["glr_uline"] == "S":
            self.s_desc = "Single"
        else:
            self.s_desc = "Double"
        self.printLine()
        self.clearData()

    def doPercent(self):
        self.pb.displayProgress()
        for x in range(0, 6):
            if self.val_str[self.s_snum1][x] == 0 or \
                self.val_str[self.s_snum2][x] == 0:
                self.val[x] = 0
            else:
                self.val[x] = round((self.val_str[self.s_snum1][x] / \
                self.val_str[self.s_snum2][x]) * 100.0, 2)
        for x in range(0, 14):
            if self.mth_str[self.s_snum1][x] == 0 or \
                self.mth_str[self.s_snum2][x] == 0:
                    self.mth[x] = 0
            else:
                self.mth[x] = round((self.mth_str[self.s_snum1][x] / \
                self.mth_str[self.s_snum2][x]) * 100.0, 2)
        self.printLine()
        self.clearData()

    def doCalc(self):
        self.pb.displayProgress()
        #if self.rtype == "Y":
        #   amt = self.val[4]
        #else:
        #   amt = self.mth[self.cmth]
        #if self.s_print == "+" and amt < 0:
        #    self.clearData()
        #    return
        #elif self.s_print == "-" and amt >= 0:
        #    self.clearData()
        #    return
        #elif self.s_print == "N":
        #   pass
        #else:
        #   self.printLine()
        #self.clearData()

    def readGenmst(self):
        if not self.consol:
            if not self.rpt_dic["glr_to"]:
                accs = self.sql.getRec(tables="genmst", where=[("glm_cono", "=",
                    self.conum), ("glm_acno", "=", self.rpt_dic["glr_from"])])
            else:
                accs = self.sql.getRec(tables="genmst", where=[("glm_cono", "=",
                    self.conum), ("glm_acno", ">=", self.rpt_dic["glr_from"]),
                    ("glm_acno", "<=", self.rpt_dic["glr_to"])],
                    order="glm_acno")
            if not accs:
                accs = []
        else:
            accs = []
            if not self.rpt_dic["glr_to"]:
                acs = self.sql.getRec(tables="genmst", cols=["glm_acno"],
                    where=[("glm_cono", "in", self.consol), ("glm_acno", "=",
                    self.rpt_dic["glr_from"])], group="glm_acno")
            else:
                acs = self.sql.getRec(tables="genmst", cols=["glm_acno"],
                    where=[("glm_cono", "in", self.consol), ("glm_acno", ">=",
                    self.rpt_dic["glr_from"]), ("glm_acno", "<=",
                    self.rpt_dic["glr_to"])], group="glm_acno",
                    order="glm_acno")
            if acs:
                for ac in acs:
                    accs.append(self.sql.getRec(tables="genmst",
                    where=[("glm_cono", "in", self.consol),
                    ("glm_acno", "=", ac[0])], limit=1))
        return accs

    def lstdicGenmst(self, acc):
        self.gen_lst = acc
        self.gen_dic = {}
        num = 0
        for fld in self.sql.genmst_col:
            self.gen_dic[fld] = acc[num]
            num += 1
        if self.rpt_dic["glr_ignore"] == "N":
            self.checkAccountType(self.gen_dic["glm_type"])

    def clearData(self):
        self.val = [0,0,0,0,0,0,0,0]
        self.mth = [0,0,0,0,0,0,0,0,0,0,0,0,0,0]

    def accumData(self):
        curdt = self.s_per
        if self.rpt_dic["glr_obal"] == "Y" and self.vcode not in ("B", "X"):
            o_lyr, o_cyr = self.getObal()
            self.val[1] = float(ASD(self.val[1]) + ASD(o_lyr))
            self.val[4] = float(ASD(self.val[4]) + ASD(o_cyr))
            self.mth[0] = float(ASD(self.mth[0]) + ASD(o_cyr))
            self.mth[13] = float(ASD(self.mth[13]) + ASD(o_cyr))
        if self.vcode == "C":
            cut = "n"
        for mth in range(1, 13):
            lyr, cyr, bud = self.getMbal(curdt)
            if self.vcode == "B" or (self.vcode == "C" and cut == "y"):
                self.mth[mth] = float(ASD(self.mth[mth]) + ASD(bud))
                self.mth[13] = float(ASD(self.mth[13]) + ASD(bud))
            elif self.vcode == "C":
                self.mth[mth] = float(ASD(self.mth[mth]) + ASD(cyr))
                self.mth[13] = float(ASD(self.mth[13]) + ASD(cyr))
            elif self.vcode == "X":
                self.mth[mth] = float(ASD(self.mth[mth]) + ASD(bud))
                self.mth[13] = float(ASD(self.mth[13]) + ASD(bud))
                self.mth[mth] = float(ASD(self.mth[mth]) - ASD(cyr))
                self.mth[13] = float(ASD(self.mth[13]) - ASD(cyr))
            else:
                self.val[1] = float(ASD(self.val[1]) + ASD(lyr))
                if curdt == self.curdt:
                    self.val[2] = float(ASD(self.val[2]) + ASD(cyr))
                    self.val[3] = float(ASD(self.val[3]) + ASD(bud))
                self.val[4] = float(ASD(self.val[4]) + ASD(cyr))
                self.val[5] = float(ASD(self.val[5]) + ASD(bud))
                self.mth[mth] = float(ASD(self.mth[mth]) + ASD(cyr))
                self.mth[13] = float(ASD(self.mth[13]) + ASD(cyr))
            if curdt == self.curdt:
                if self.vcode == "C":
                    cut = "y"
                else:
                    break
            y = curdt / 100
            m = curdt % 100
            m += 1
            if m > 12:
                y += 1
                m -= 12
            curdt = (y * 100) + m

    def getObal(self):
        if not self.consol:
            if not self.period[0]:
                lyr = None
            else:
                lyr = self.sql.getRec(tables="genbal",
                    cols=["round(sum(glo_cyr), 2)"], where=[("glo_cono", "=",
                    self.s_cono), ("glo_acno", "=", self.s_acno), ("glo_trdt",
                    "=", self.s_lyr.work)], limit=1)[0]
            cyr = self.sql.getRec(tables="genbal",
                cols=["round(sum(glo_cyr), 2)"], where=[("glo_cono", "=",
                self.s_cono), ("glo_acno", "=", self.s_acno), ("glo_trdt",
                "=", self.period[1][0])], limit=1)[0]
        else:
            if not self.period[0]:
                lyr = None
            else:
                lyr = self.sql.getRec(tables="genbal",
                    cols=["round(sum(glo_cyr), 2)"], where=[("glo_cono", "in",
                    self.consol), ("glo_acno", "=", self.s_acno), ("glo_trdt",
                    "=", self.s_lyr.work)], limit=1)[0]
            cyr = self.sql.getRec(tables="genbal",
                cols=["round(sum(glo_cyr), 2)"], where=[("glo_cono", "in",
                self.consol), ("glo_acno", "=", self.s_acno), ("glo_trdt",
                "=", self.period[1][0])], limit=1)[0]
        if not lyr:
            lyr = 0
        else:
            lyr = float(lyr)
        if not cyr:
            cyr = 0
        else:
            cyr = float(cyr)
        return (lyr, cyr)

    def getMbal(self, curdt):
        if not self.consol:
            lyr = self.sql.getRec(tables="gentrn",
                cols=["round(sum(glt_tramt), 2)"], where=[("glt_cono", "=",
                self.s_cono), ("glt_acno", "=", self.s_acno), ("glt_curdt",
                "=", (curdt - 100))], limit=1)[0]
            cyr = self.sql.getRec(tables="gentrn",
                cols=["round(sum(glt_tramt), 2)"], where=[("glt_cono", "=",
                self.s_cono), ("glt_acno", "=", self.s_acno), ("glt_curdt",
                "=", curdt)], limit=1)[0]
            bud = self.sql.getRec(tables="genbud", cols=["sum(glb_tramt)"],
                where=[("glb_cono", "=", self.s_cono), ("glb_acno", "=",
                self.s_acno), ("glb_curdt", "=", curdt)], limit=1)[0]
        else:
            lyr = self.sql.getRec(tables="gentrn",
                cols=["round(sum(glt_tramt), 2)"], where=[("glt_cono", "in",
                self.consol), ("glt_acno", "=", self.s_acno), ("glt_curdt",
                "=", (curdt - 100))], limit=1)[0]
            cyr = self.sql.getRec(tables="gentrn",
                cols=["round(sum(glt_tramt), 2)"], where=[("glt_cono", "in",
                self.consol), ("glt_acno", "=", self.s_acno), ("glt_curdt",
                "=", curdt)], limit=1)[0]
            bud = self.sql.getRec(tables="genbud", cols=["sum(glb_tramt)"],
                where=[("glb_cono", "in", self.consol), ("glb_acno", "=",
                self.s_acno), ("glb_curdt", "=", curdt)], limit=1)[0]
        if not lyr:
            lyr = 0
        else:
            lyr = float(lyr)
        if not cyr:
            cyr = 0
        else:
            cyr = float(cyr)
        if not bud:
            bud = 0
        else:
            bud = int(bud)
        return (lyr, cyr, bud)

    def printLine(self):
        self.signChkChg()
        if self.vcode == "D":
            val, mth = self.doDcode()
        else:
            val = self.val[:]
            mth = self.mth[:]
        if not self.s_type == "L":
            self.s_acno = 0
        self.allFields.append([self.s_type, self.atype, self.s_high,
            self.s_ffeed, self.s_acno, self.s_desc, self.count, self.lines,
            val, mth, self.s_label, self.stanam, self.s_norm])
        self.signChkChg()
        self.count += 1
        self.lines = 1

    def doDcode(self):
        curdt = self.s_per
        val_d = [0,0,0]
        mth_d = [0,0,0,0,0,0,0,0,0,0,0,0,0,0]
        val_p = [0,0,0,0,0,0]
        mth_p = [0,0,0,0,0,0,0,0,0,0,0,0,0,0]
        for mth in range(1, 13):
            lyr, cyr = self.getDbal(curdt)
            val_d[0] = float(ASD(val_d[0]) + ASD(lyr))
            if curdt == self.curdt:
                val_d[1] = cyr
            val_d[2] = float(ASD(val_d[2]) + ASD(cyr))
            if self.s_accum == "Y":
                mth_d[mth] = float(ASD(mth_d[mth-1]) + \
                    ASD(mth_d[mth]) + ASD(cyr))
                mth_d[13] = mth_d[mth]
            else:
                mth_d[mth] = float(ASD(mth_d[mth]) + ASD(cyr))
                mth_d[13] = float(ASD(mth_d[13]) + ASD(cyr))
            if curdt == self.curdt:
                break
            else:
                y = curdt / 100
                m = curdt % 100
                m += 1
                if m > 12:
                    y += 1
                    m = m - 12
                curdt = (y * 100) + m
        if val_d[0] == 0:
            val_p[1] = 0
        else:
            val_p[1] = round((self.val[1] / val_d[0]) * 100, 2)
        if val_d[1] == 0:
            val_p[2] = 0
            val_p[3] = 0
        else:
            val_p[2] = round((self.val[2] / val_d[1]) * 100, 2)
            val_p[3] = round((self.val[3] / val_d[1]) * 100, 2)
        if val_d[2] == 0:
            val_p[4] = 0
            val_p[5] = 0
        else:
            val_p[4] = round((self.val[4] / val_d[2]) * 100, 2)
            val_p[5] = round((self.val[5] / val_d[2]) * 100, 2)
        for x in range(0, 14):
            if mth_d[x] == 0:
                mth_p[x] = 0
            else:
                mth_p[x] = round((self.mth[x] / mth_d[x]) * 100, 2)
        return (val_p, mth_p)

    def getDbal(self, curdt):
        lyr = self.sql.getRec(tables="gendtt", cols=["gdt_value"],
            where=[("gdt_cono", "=", self.conum), ("gdt_code", "=", self.dcode),
            ("gdt_curdt", "=", (curdt - 100))], limit=1)
        cyr = self.sql.getRec(tables="gendtt", cols=["gdt_value"],
            where=[("gdt_cono", "=", self.conum), ("gdt_code", "=", self.dcode),
            ("gdt_curdt", "=", curdt)], limit=1)
        if not lyr:
            lyr = 0
        else:
            lyr = lyr[0]
        if not cyr:
            cyr = 0
        else:
            cyr = cyr[0]
        return (lyr, cyr)

    def signChkChg(self):
        if self.s_type in ("C","G","L","S","T") and self.s_norm == "N":
            for x in range(0, 6):
                self.val[x] = float(ASD(0) - ASD(self.val[x]))
            for x in range(0, 14):
                self.mth[x] = float(ASD(0) - ASD(self.mth[x]))
        
    def accumTotals(self, acbal):
        for x in range(1, 10):
            if acbal == "A":
                for y in range(0, 6):
                    self.val_tot[x][y] = \
                        float(ASD(self.val_tot[x][y]) + ASD(self.val[y]))
                for y in range(0, 14):
                    self.mth_tot[x][y] = \
                        float(ASD(self.mth_tot[x][y]) + ASD(self.mth[y]))
            elif acbal == "S":
                for y in range(0, 6):
                    self.val_tot[x][y] = \
                        float(ASD(self.val_tot[x][y]) - ASD(self.val[y]))
                for y in range(0, 14):
                    self.mth_tot[x][y] = \
                        float(ASD(self.mth_tot[x][y]) - ASD(self.mth[y]))

    def storeBalances(self):
        if self.s_acstr == "A":
            for x in range(0, 6):
                self.val_str[self.s_snum1][x] = \
                    float(ASD(self.val_str[self.s_snum1][x]) + ASD(self.val[x]))
            for x in range(0, 14):
                self.mth_str[self.s_snum1][x] = \
                    float(ASD(self.mth_str[self.s_snum1][x]) + ASD(self.mth[x]))
        elif self.s_acstr == "S":
            for x in range(0, 6):
                self.val_str[self.s_snum1][x] = \
                    float(ASD(self.val_str[self.s_snum1][x]) - ASD(self.val[x]))
            for x in range(0, 14):
                self.mth_str[self.s_snum1][x] = \
                    float(ASD(self.mth_str[self.s_snum1][x]) - ASD(self.mth[x]))

    def nextRepRecord(self):
        self.rpt_dic = {}
        if self.rep_num == len(self.rpt_lst) - 1:
            self.exit = True
        else:
            self.rep_num += 1
            num = 0
            for fld in self.sql.genrpt_col:
                self.rpt_dic[fld] = self.rpt_lst[self.rep_num][num]
                num += 1

    def storeRepRecord(self):
        self.s_acno = 0
        for key in self.rpt_dic:
            k = key.split("_")
            exec "self.s_%s = self.rpt_dic[key]" % k[1]

    def checkAccountType(self, glm_type):
        if glm_type == self.atype or self.atype == "O":
            pass
        else:
            if glm_type == "P" and self.atype == "P":
                pass
            elif glm_type == "B" and self.atype == "B":
                pass
            else:
                self.count = 66
            self.atype = glm_type
            if self.atype == "P":
                self.stanam = "Profit and Loss"
            else:
                self.stanam = "Balance Sheet"

class GetCtl:
    """
    Used for extracting control records
    """
    def __init__(self, mf):
        self.mf = mf
        self.doVariables()

    def doVariables(self):
        self.errs = {
            "ctlsys": """Missing ctlsys Record, Please Run

Control -> System Record Maintenance

and Create It.""",
            "assctl": """Missing assctl Record for Company %s, Please Run

Asset's Register -> File Maintenance -> Control Record

and Create It.""",
            "crsctl": """Missing crsctl Record for Company %s, Please Run
            
Creditor's Ledger -> File Maintenance -> Control Record
            
and Create It.""",
            "ctlctl": """Missing ctlctl Record for Company %s, Please Run
            
General Ledger -> File Maintenance -> Control Accounts
            
and Create Them.""",
            "ctlmst": """Missing ctlmst Record for Company %s, Please Run
            
Control -> Company Records Maintenance
            
and Create It.""",
            "drsctl": """Missing drsctl Record for Company %s, Please Run
            
Debtor's Ledger -> File Maintenance -> Control Record
            
and Create It.""",
            "genctl": """Missing genctl Record for Company %s, Please Run
            
General Ledger -> File Maintenance -> Control Record
            
and Create It.""",
            "memctl": """Missing memctl Record for Company %s, Please Run
            
Member's Ledger -> File Maintenance -> Control Record
            
and Create It.""",
            "rcactl": """Missing rcactl Record for Company %s, Please Run
            
Rental's Ledger -> File Maintenance -> Control Record
            
and Create It.""",
            "rtlctl": """Missing rtlctl Record for Company %s, Please Run
            
Rental's Ledger -> File Maintenance -> Control Record
            
and Create It.""",
            "slsctl": """Missing slsctl Record for Company %s, Please Run
            
Sales Invoicing -> File Maintenance -> Control Record
            
and Create It.""",
            "strctl": """Missing strctl Record for Company %s, Please Run
            
Stores Ledger -> File Maintenance -> Control Record
            
and Create It.""",
            "wagctl": """Missing wagctl Record for Company %s, Please Run
            
Salaries and Wages -> File Maintenance -> Control Record
            
and Create It."""}

    def getCtl(self, table, conum=None, error=True):
        sql = Sql(self.mf.dbm, ["ffield", table], prog=__name__, errs=error)
        if sql.error:
            return None
        dict = {}
        if table == "ctlctl":
            ctl = sql.getRec(tables=[table], cols=["ctl_code", "ctl_conacc"],
                where=[("ctl_cono", "=", conum)])
            if not ctl:
                if error:
                    if not self.mf.display:
                        print self.errs[table] % conum
                    else:
                        showError(self.mf.body, "Control Record Error",
                            self.errs[table] % conum)
                return None
            for c in ctl:
                dict["%s" % c[0]] = c[1]
            return dict
        else:
            fld = sql.getRec(tables="ffield", cols=["ff_name"],
                where=[("ff_tabl", "=", table)])
            if not fld:
                if error:
                    if not self.mf.display:
                        print "Invalid Table %s" % table
                    else:
                        showError(self.mf.body, "Control Record Error",
                            "Invalid Table %s" % table)
                return None
            col = []
            pos = None
            for f in fld:
                col.append(f[0])
                if f[0].count("_cono"):
                    pos = f[0]
            if not pos:
                all = sql.getRec(tables=table, cols=col)
            else:
                all = sql.getRec(tables=table, cols=col,
                    where=[(pos, "=", conum)])
            if not all:
                if error:
                    if not conum:
                        showError(self.mf.body, "Control Record Error",
                            self.errs[table])
                    else:
                        showError(self.mf.body, "Control Record Error",
                            self.errs[table] % conum)
                return None
            for x in range(0, len(col)):
                dict["%s" % col[x].lower()] = all[0][x]
            return dict

    def chkRec(self, conum, ctlctl, ctls):
        chk = "ok"
        sql = Sql(self.mf.dbm, "genmst", prog=__name__)
        for key in ctls:
            try:
                if key in ctlctl:
                    acc = ctlctl[key]
                    tst = sql.getRec(tables="genmst", cols=["count(*)"],
                        where=[("glm_cono", "=", conum), ("glm_acno", "=",
                        int(acc))], limit=1)
                    if not tst[0]:
                        chk = "no"
                else:
                    chk = "no"
            except:
                chk = "no"
            if chk == "no":
                break
        if chk == "no":
            showError(self.mf.body, "Control Record Error",
                """Missing Control Record %s for Company %s, Please Run

General Ledger -> File Maintenance -> Control Records

and Create It.""" % (key, conum))
            return key

class LoanInterest:
    """
    Used to calculate interest on a loan:

    period - Start of financial period
    coy    - Company number
    empno  - Employee number
    loan   - Loan number
    rate   - Interest rate
    start  - Starting date of the loan
    trdate - Transaction date
    batch  - Batch number
    curdt  - Current date
    update - Update tables, Y or N
    glctls = G/L control accounts, [wag_slc, wag_sli]
    """
    def __init__(self, mf, period, coy, empno, loan, rate, start, trdate, batch, curdt, update="N", glctls=None):
        self.mf = mf
        self.period = period
        self.conum = coy
        self.empno = empno
        self.loan = loan
        self.rate = rate
        self.start = start
        self.trdate = trdate
        self.batch = batch
        self.curdt = curdt
        self.update = update
        self.glctls = glctls
        self.setVariables()
        self.doProcess()

    def setVariables(self):
        t = time.localtime()
        self.sysdtw = (t[0] * 10000) + (t[1] * 100) + t[2]
        self.sql = Sql(self.mf.dbm, ["wagltf", "gentrn"], prog=__name__)
        self.idt = 0
        self.imt = 0
        self.bal = 0
        self.cap = 0

    def doProcess(self):
        wlt = self.sql.getRec(tables="wagltf", cols=["wlt_trdt", "wlt_type",
            "wlt_amt"], where=[("wlt_cono", "=", self.conum), ("wlt_empno",
            "=", self.empno), ("wlt_loan", "=", self.loan)],
            order="wlt_trdt, wlt_type")
        for x in range(0, len(wlt)):
            dat = CCD(wlt[x][0], "d1", 10)
            typ = CCD(wlt[x][1], "UI", 2)
            amt = CCD(wlt[x][2], "SD", 13.2)
            self.bal = float(ASD(self.bal) + ASD(amt.work))
            self.cap = float(ASD(self.cap) + ASD(amt.work))
            if typ.work == 1:
                self.idt = dat.work
                if dat.work >= self.period and dat.work >= self.start:
                    self.cap = float(ASD(self.cap) - ASD(amt.work))
        if self.idt == 0:
            dt1 = self.start
        else:
            dt1 = self.idt
        dt2 = self.trdate
        rat = self.rate
        days = dateDiff(dt1, dt2, "days")
        self.imt = round((self.cap * rat * days / 36500), 2)
        if self.imt < 0:
            self.imt = 0
        self.bal = float(ASD(self.bal) + ASD(self.imt))
        if self.imt and self.update == "Y":
            desc = str(dt1)[2:] + " - " + str(dt2)[2:] + " on " + str(self.cap)
            data = [self.conum, self.empno, self.loan, self.batch, 1, dt2,
                "AutoInt", self.imt, self.cap, 0, rat, self.curdt,
                desc[:30], "", self.sysdtw, 0]
            self.sql.insRec("wagltf", data=data)
            if self.glctls:
                credit = float(ASD(0) - ASD(self.imt))
                vals = ((self.glctls[0], self.imt), (self.glctls[1], credit))
                for val in vals:
                    whr = [("glt_cono", "=", self.conum), ("glt_acno", "=",
                        val[0]), ("glt_curdt", "=", self.curdt), ("glt_trdt",
                        "=", dt2), ("glt_type", "=", 4), ("glt_refno", "=",
                        "AutoInt"), ("glt_batch", "=", self.batch)]
                    data = self.sql.getRec(tables="gentrn", where=whr, limit=1)
                    if data:
                        tramt = data[self.sql.gentrn_col.index("glt_tramt")]
                        tramt = float(ASD(tramt) + ASD(val[1]))
                        data[self.sql.gentrn_col.index("glt_tramt")] = tramt
                        self.sql.updRec("gentrn", where=whr, data=data)
                    else:
                        data = [self.conum, val[0], self.curdt, dt2,
                            4, "AutoInt", self.batch, val[1], 0.00,
                            "Interest on Staff Loans", "", "", 0,
                            self.sysdtw, 0]
                        self.sql.insRec("gentrn", data=data)

class PrintOrder:
    """
    Used to print stores orders using
    the ctlmst, crsmst, strpom and strpot tables
    """
    def __init__(self, mf, conum, docno, **args):
        self.mf = mf
        self.conum = conum
        self.docno = CCD(docno, "UI", 9.0)
        defaults = {
            "prtnam": "Default",
            "email": "n",
            "copy": "n"}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        ok = self.doVariables()
        if ok == None:
            self.doProcess()

    def doVariables(self):
        self.sql = Sql(self.mf.dbm, ["ctlmst", "strpom", "strpot", "crsmst"],
            prog=__name__)
        ctl3 = self.sql.ctlmst_col
        crm3 = self.sql.crsmst_col
        self.pom3 = self.sql.strpom_col
        self.pag = 0
        self.cnt = 66
        self.head = []
        self.totabl = 0
        self.totnot = 0
        self.totvat = 0
        self.invtot = 0
        self.pom = self.sql.getRec(tables="strpom", where=[("pom_cono", "=",
            self.conum), ("pom_ordno", "=", self.docno.work)], limit=1)
        if not self.pom:
            return "Invalid Document"
        acc = self.pom[self.pom3.index("pom_acno")]
        our = ""
        cus = self.pom[self.pom3.index("pom_cusord")]
        job = self.pom[self.pom3.index("pom_jobnum")]
        per = self.pom[self.pom3.index("pom_contact")]
        vtn = self.pom[self.pom3.index("pom_vatno")]
        ord = self.pom[self.pom3.index("pom_ordby")]
        crm = self.sql.getRec(tables="crsmst", where=[("crm_cono", "=",
            self.conum), ("crm_acno", "=", acc)], limit=1)
        ctl = self.sql.getRec(tables="ctlmst", where=[("ctm_cono", "=",
            self.conum)], limit=1)
        self.logo = ctl[ctl3.index("ctm_logo")]
        self.flip = ctl[ctl3.index("ctm_flip")]
        ########################################################################
        # Start of Order Layout
        ########################################################################
        if not self.logo or not os.path.exists(self.logo):
            self.logo = None
        elif self.flip == "N":
            self.logo = flipImage(self.logo)
        if not self.logo:
            self.head.append("%1s%-30s" % ("", ctl[ctl3.index("ctm_name")]))
        else:
            self.head.append(" ")
        if self.copy.lower() == "a":
            doctyp = "AMENDED PURCHASE ORDER"
        elif self.copy.lower() == "d":
            doctyp = "CANCELLED PURCHASE ORDER"
        elif self.copy.lower() == "y":
            doctyp = "COPY PURCHASE ORDER"
        else:
            doctyp = "PURCHASE ORDER"
        ########################################################################
        # Print Company Details
        ########################################################################
        if self.logo:
            self.head.append("%89s%22s" % ("", doctyp))
            self.head.append(" ")
            self.head.append(" ")
            self.head.append(" ")
            self.head.append(" ")
        else:
            ####################################################################
            # Company Details
            ####################################################################
            addr = [ctl[ctl3.index("ctm_add1")]]
            if ctl[ctl3.index("ctm_add2")]:
                addr.append(ctl[ctl3.index("ctm_add2")])
            if ctl[ctl3.index("ctm_add3")]:
                addr.append(ctl[ctl3.index("ctm_add3")])
            if ctl[ctl3.index("ctm_pcode")]:
                addr.append(ctl[ctl3.index("ctm_pcode")])
            num = 4 - len(addr)
            if num:
                addr.extend([""] * num)
            self.head.append("%2s%-30s%2s%-30s%25s%22s" % ("", addr[0], "",
                ("RegNo: " + ctl[ctl3.index("ctm_regno")]), "", doctyp))
            self.head.append("%2s%-30s%2s%-30s" % ("", addr[1], "",
                ("TaxNo: " + ctl[ctl3.index("ctm_taxno")])))
            self.head.append("%2s%-30s%2s%-30s" % ("", addr[2], "",
                ("TelNo: " + ctl[ctl3.index("ctm_tel")])))
            self.head.append("%2s%-30s%2s%-30s" % ("", addr[3], "",
                ("FaxNo: " + ctl[ctl3.index("ctm_fax")])))
            self.head.append(" ")
        ########################################################################
        # Creditor's Name, Address and Delivery Address
        ########################################################################
        addr = [crm[crm3.index("crm_add1")]]
        if crm[crm3.index("crm_add2")]:
            addr.append(crm[crm3.index("crm_add2")])
        if crm[crm3.index("crm_add3")]:
            addr.append(crm[crm3.index("crm_add3")])
        if crm[crm3.index("crm_pcod")]:
            addr.append(crm[crm3.index("crm_pcod")])
        num = 4 - len(addr)
        if num:
            addr.extend([""] * num)
        dela = [self.pom[self.pom3.index("pom_add1")]]
        if self.pom[self.pom3.index("pom_add2")]:
            dela.append(self.pom[self.pom3.index("pom_add2")])
        if self.pom[self.pom3.index("pom_add3")]:
            dela.append(self.pom[self.pom3.index("pom_add3")])
        if self.pom[self.pom3.index("pom_add4")]:
            dela.append(self.pom[self.pom3.index("pom_add4")])
        num = 4 - len(dela)
        if num:
            dela.extend([""] * num)
        self.head.append("%1s%-30s%5s%-30s%36s%9s" % ("",
            crm[crm3.index("crm_name")], "", dela[0], "", self.docno.disp))
        self.head.append("%1s%-30s%5s%-30s" % ("", addr[0], "", dela[1]))
        self.head.append("%1s%-30s%5s%-30s%38s%s" % ("", addr[1], "", dela[2],
            "", "%-7s"))
        self.head.append("%1s%-30s%5s%-30s" % ("", addr[2], "", dela[3]))
        date = CCD(self.pom[self.pom3.index("pom_date")], "D1", 10.0)
        self.head.append("%1s%-30s%70s%10s" % ("", addr[3], "", date.disp))
        self.head.append(
            "%1s%-9s%1s%-18s%1s%-7s%1s%-37s%1s%-20s%1s%3s%4s%-7s%1s" % ("",
            our, "", cus, "", job, "", per, "", vtn, "", ord, "", acc, ""))
        ########################################################################
        # End of Order Layout
        ########################################################################
        if type(self.email) == tuple:
            self.email, self.emadd = self.email
        else:
            self.emadd = ""
        if not self.emadd:
            self.emadd = crm[crm3.index("crm_ord_email")]
        if not self.emadd:
            self.emadd = crm[crm3.index("crm_acc_email")]
        self.pdfnam = getModName(self.mf.rcdic["wrkdir"], "PurOrd",
            "%s_%s" % (self.conum, self.docno.work), ext="pdf", wait=1)
        self.c, self.cpi, self.inch, self.suc, self.duc = \
            doReportlab(self.mf.prgdir, self.pdfnam, orient="L")
        # Width in pixels of each character
        self.cwidth = self.c.stringWidth(("*"), "Tartan", self.cpi)
        # Width in pixels of the form
        self.twidth = self.cwidth * 112

    def doProcess(self):
        pot3 = self.sql.strpot_col
        pot = self.sql.getRec(tables="strpot", where=[("pot_cono", "=",
            self.conum), ("pot_ordno", "=", self.docno.work)])
        if not pot:
            self.doPrintHead()
        else:
            for line in pot:
                if self.cnt == 66:
                    self.doPrintHead()
                if self.cnt == 37:
                    self.doPrintCfwd()
                grp = CCD(line[pot3.index("pot_group")], "NA", 3.0)
                cod = CCD(line[pot3.index("pot_code")], "NA", 20.0)
                des = CCD(line[pot3.index("pot_desc")], "NA", 30.0)
                unt = CCD(line[pot3.index("pot_units")], "SD", 10.0)
                qty = CCD(line[pot3.index("pot_qty")], "SD", 13.2)
                pri = CCD(line[pot3.index("pot_price")], "UD", 12.2)
                dis = CCD(line[pot3.index("pot_disper")], "UD", 6.2)
                rat = CCD(line[pot3.index("pot_vatrat")], "UD", 6.2)
                if unt.work == 0:
                    price = pri.work
                else:
                    price = round((pri.work / unt.work), 2)
                excpri = round((price * 1), 2)                  #S/Price
                inc = float(ASD(100) + ASD(rat.work))
                incpri = round((price * inc / 100.0), 4)        #S/Price+VAT
                net = float(ASD(100) - ASD(dis.work))
                excamt = round((qty.work * excpri * net / 100.0), 2)
                incamt = round((qty.work * incpri * net / 100.0), 2)
                vatamt = float(ASD(incamt) - ASD(excamt))
                if excamt == incamt:
                    self.totnot = float(ASD(self.totnot) + ASD(excamt))
                else:
                    self.totabl = float(ASD(self.totabl) + ASD(excamt))
                self.totvat = float(ASD(self.totvat) + ASD(vatamt))
                self.invtot = float(ASD(self.invtot) + ASD(incamt))
                val = CCD(incamt, "SD", 13.2)
                body="%1s%3s%1s%20s%1s%30s%1s%13s%1s%12s%1s%6s%1s%6s%1s%13s%1s"
                self.t.textLine(body % (
                    "", grp.disp, "", cod.disp, "", des.disp, "", qty.disp,
                    "", pri.disp, "", dis.disp, "", rat.disp, "", val.disp,
                    ""))
                self.cnt += 1
        self.doPrintTotal()
        self.c.save()
        if self.prtnam == "view":
            view = "V"
        else:
            view = "P"
        if self.email.lower() == "y" and self.emadd:
            mail = self.emadd
        else:
            mail = "N"
        doPrinter(mf=self.mf, conum=self.conum, rep=self.pdfnam,
            head="Purchase Order %s" % self.docno.disp, view=view,
            prtnam=self.prtnam, mail=mail)

    def doPrintCfwd(self):
        self.t.textLine()
        self.c.line(self.cwidth * .5, (38.5 * (self.inch/6)),
        self.twidth-(self.cwidth * .5), (38.5 * (self.inch/6)))
        self.t.textLine("%1s%62s%15s" % ("", "", "Carried Forward"))
        self.c.drawText(self.t)
        self.c.showPage()
        self.doPrintHead()

    def doPrintHead(self):
        self.c.translate(self.inch, 0)    # Set x and y margins
        if self.copy.lower() in ("a", "d", "y"):
            self.c.setFillGray(.7)
            x = self.cwidth * 111.5
            if self.copy.lower() == "a":
                doctyp = "AMENDED ORDER"
            elif self.copy.lower() == "d":
                doctyp = "CANCELLED ORDER"
            elif self.copy.lower() == "y":
                doctyp = "COPY OF ORDER"
            self.c.setFont("Tartan-Bold", 70)
            self.c.drawCentredString(x/2, self.c._pagesize[1]/1.7, doctyp)
            self.c.setFont("Tartan", self.cpi)
            self.c.setFillGray(0)
        self.doDrawOrder()
        self.t = self.c.beginText()
        self.t.setTextOrigin(0, 22)
        self.t.setFont("Tartan-Bold", 18)
        self.t.textLine(self.head[0])
        self.t.setFont("Tartan-Bold", self.cpi)
        self.t.textLine(self.head[1])
        self.t.textLine(self.head[2])
        self.t.textLine(self.head[3])
        self.t.textLine(self.head[4])
        self.t.setFont("Tartan", self.cpi)
        self.t.textLine()
        self.t.textLine(self.head[5])
        self.t.textLine(self.head[6])
        self.t.textLine(self.head[7])
        self.pag += 1
        page = CCD(self.pag, "UI", 7.0)
        self.t.textLine(self.head[8] % page.disp)
        self.t.textLine(self.head[9])
        self.t.textLine(self.head[10])
        self.t.textLine()
        self.t.textLine()
        self.t.textLine(self.head[11])
        self.t.textLine()
        self.t.textLine()
        self.cnt = 18

    def doDrawOrder(self):
        if self.logo:
            self.c.drawImage(self.logo, 0, 10, 400, 70, None)
        ########################################################################
        # Vertical left and right margins
        ########################################################################
        self.c.line(self.cwidth * .5, (14 * (self.inch / 6)),
            self.cwidth * .5, (38.5 * (self.inch / 6)))
        self.c.line(self.cwidth * 111.5, (14 * (self.inch / 6)),
            self.cwidth * 111.5, (38.5 * (self.inch / 6)))
        ########################################################################
        # Draw Charge To and Deliver To rectangles
        ########################################################################
        self.c.setFillGray(.9)
        x = self.cwidth * .5
        w = self.cwidth * 31
        self.c.rect(x, (7.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (7.75 * (self.inch / 6)), w, (6 * (self.inch / 6)),
            fill=0)
        x = self.cwidth * 35.5
        self.c.rect(x, (7.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (7.75 * (self.inch / 6)), w, (6 * (self.inch / 6)),
            fill=0)
        ########################################################################
        # Fill rectangles with details
        ########################################################################
        self.c.setFillGray(0)
        self.c.setFont("Tartan-Bold", 8)
        self.c.drawString(self.cwidth * 1, (8.45 * (self.inch / 6)),
            "Supplier:")
        self.c.drawString(self.cwidth * 36, (8.45 * (self.inch / 6)),
            "Deliver To:")
        self.c.setFont("Tartan", self.cpi)
        ########################################################################
        # Draw rectangles for Document, Page Number and Date
        ########################################################################
        self.c.setFillGray(.9)
        x = self.cwidth * 100.5
        w = self.cwidth * 11
        self.c.rect(x, (7.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (8.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=0)
        self.c.rect(x, (9.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (10.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=0)
        self.c.rect(x, (11.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (12.75 * (self.inch / 6)), w, (1.25 * (self.inch / 6)),
            fill=0)
        self.c.setFillGray(0)
        ########################################################################
        # Fill rectangles with details
        ########################################################################
        self.c.setFont("Tartan-Bold", 8)
        self.c.drawCentredString(self.twidth-(self.cwidth * 5.5),
            (8.45 * (self.inch / 6)), "Order No")
        self.c.drawCentredString(self.twidth-(self.cwidth * 5),
            (10.45 * (self.inch / 6)), "Page No")
        self.c.drawCentredString(self.twidth-(self.cwidth * 5),
            (12.45 * (self.inch / 6)), "Date")
        self.c.setFont("Tartan", self.cpi)
        ########################################################################
        # Draw rectangles for ribbon line
        ########################################################################
        self.c.setFillGray(.9)
        self.c.rect(self.cwidth * .5, (14 * (self.inch / 6)),
            (self.cwidth * 111), (1.5 * (self.inch / 6)), fill=1)
        self.c.setFillGray(0)
        for num in (10.5, 29.5, 37.5, 75.5, 96.5, 100.5):
            col = self.cwidth * num
            self.c.line(col, (14 * (self.inch / 6)), col, 17 * (self.inch / 6))
        ########################################################################
        # Fill rectangles with details
        ########################################################################
        txt = "Del-Note  Cus-Ord-Number     Job-Num "\
            "Contact-Person                        "\
            "V.A.T. Number        Ord    Acc-Num"
        self.c.setFont("Tartan-Bold", self.cpi)
        self.c.drawString(self.cwidth * 1, (15 * (self.inch / 6)), txt)
        self.c.setFont("Tartan", self.cpi)
        ########################################################################
        # Draw rectangles for body headings
        ########################################################################
        self.c.setFillGray(.9)
        self.c.rect(self.cwidth * .5, (17 * (self.inch / 6)),
            (self.cwidth * 111), (1.5 * (self.inch / 6)), fill=1)
        self.c.setFillGray(0)
        for num in (4.5, 25.5, 56.5, 70.5, 83.5, 90.5, 97.5):
            col = self.cwidth * num
            self.c.line(col, (17 * (self.inch / 6)), col, 38.5 * (self.inch/6))
        txt = "Grp Product Code         Description                    "\
            "    Quantity         Price Disc-%  Tax-%        Value "
        self.c.setFont("Tartan-Bold", self.cpi)
        self.c.drawString(self.cwidth * 1, (18 * (self.inch / 6)), txt)
        self.c.setFont("Tartan", self.cpi)

    def doPrintTotal(self):
        lines = 38 - self.cnt
        for x in range(lines):
            self.t.textLine()
        ########################################################################
        # Draw rectangles for totals
        ########################################################################
        self.c.setFillGray(.9)
        y = 38.5 * (self.inch / 6)
        for x, w, f in ((.5, 13, 1), (13.5, 14, 0), (27.5, 14, 1),
            (41.5, 14, 0), (55.5, 14, 1), (69.5, 14, 0), (83.5, 14, 1),
            (97.5, 14, 0)):
            self.c.rect((self.cwidth * x), y, (self.cwidth * w),
                (1.5 * (self.inch / 6)), fill=f)
        self.c.setFillGray(0)
        abl = CCD(self.totabl, "SD", 13.2)
        exc = CCD(self.totnot, "SD", 13.2)
        vat = CCD(self.totvat, "SD", 13.2)
        amt = CCD(self.invtot, "SD", 13.2)
        tots = "%1s%-12s%1s%13s%1s%-13s%1s%13s%1s%-13s%1s%13s%1s%-13s%1s%13s%1s"
        self.t.setFont("Tartan-Bold", self.cpi)
        self.t.textLine(tots % (
            "", "Taxable", "", abl.disp, "", "Non-Taxable", "", exc.disp,
            "", "Total Tax", "", vat.disp, "", "Total Value", "", amt.disp, ""))
        self.t.setFont("Tartan", self.cpi)
        if not self.pom[self.pom3.index("pom_mes1")]:
            pass    #No Message
        else:
            self.t.textLine()
            self.t.textLine()
            ####################################################################
            # Draw Message rectangles
            ####################################################################
            self.c.setFillGray(.9)
            x = self.cwidth * .5
            w = self.cwidth * 31
            self.c.rect(x, (40.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
                fill=1)
            self.c.rect(x, (40.75 * (self.inch / 6)), w, (6 * (self.inch / 6)),
                fill=0)
            self.c.setFillGray(0)
            self.c.setFont("Tartan-Bold", 8)
            self.c.drawString(self.cwidth * 1, (41.45 * (self.inch / 6)),
                "Message:")
            self.c.setFont("Tartan", self.cpi)
            for x in range(1, 6):
                exec "mess = '%1s%-30s' % ('', "\
                    "self.pom[self.pom3.index('pom_mes%s' % x)])"
                self.t.textLine(mess)
        self.c.drawText(self.t)
        self.c.showPage()

class PrintInvoice:
    """
    Used to print sales invoices, credit notes, orders and quotations using
    the ctlmst, drsmst, slsiv1 and slsiv2 tables
    """
    def __init__(self, mf, conum, dtype, docno, **args):
        self.mf = mf
        self.conum = conum
        self.dtype = dtype
        self.docno = CCD(docno, "UI", 9.0)
        defaults = {
            "prtnam": "Default",
            "useadd": "p",
            "dnote": "n",
            "dvals": "n",
            "email": "n",
            "copy": "n"}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        no = self.doVariables()
        if not no:
            self.doProcess()
            if self.dnote.lower() == "y":
                self.dtype = "D"
                self.doVariables()
                self.doProcess()

    def doVariables(self):
        if self.dtype != "D":
            self.sql = Sql(self.mf.dbm, ["ctlmst", "ctlmes", "drsmst",
                "slsiv1", "slsiv2"], prog=__name__)
            self.iv1 = self.sql.getRec(tables="slsiv1", where=[("si1_cono",
                "=", self.conum), ("si1_rtn", "=", self.dtype), ("si1_docno",
                "=", self.docno.work)], limit=1)
            if not self.iv1:
                return "no"
            self.iv2 = self.sql.getRec(tables="slsiv2", where=[("si2_cono",
                "=", self.conum), ("si2_rtn", "=", self.dtype), ("si2_docno",
                "=", self.docno.work)])
        ctl3 = self.sql.ctlmst_col
        drm3 = self.sql.drsmst_col
        self.iv13 = self.sql.slsiv1_col
        self.pag = 0
        self.cnt = 66
        self.head = []
        self.totabl = 0
        self.totnot = 0
        self.totvat = 0
        self.invtot = 0
        chn = self.iv1[self.iv13.index("si1_chain")]
        acc = self.iv1[self.iv13.index("si1_acno")]
        our = self.iv1[self.iv13.index("si1_our_ord")]
        cus = self.iv1[self.iv13.index("si1_cus_ord")]
        job = self.iv1[self.iv13.index("si1_jobno")]
        per = self.iv1[self.iv13.index("si1_contact")]
        vtn = self.iv1[self.iv13.index("si1_vatno")]
        rep = self.iv1[self.iv13.index("si1_rep")]
        drm = self.sql.getRec(tables="drsmst", where=[("drm_cono", "=",
            self.conum), ("drm_chain", "=", chn), ("drm_acno", "=", acc)],
            limit=1)
        ########################################################################
        # Company Banking and Logo Details
        ########################################################################
        ctl = self.sql.getRec(tables="ctlmst", where=[("ctm_cono", "=",
            self.conum)], limit=1)
        if ctl[ctl3.index("ctm_b_name")]:
            self.bank = [ctl[ctl3.index("ctm_b_name")]]
        else:
            self.bank = []
        if ctl[ctl3.index("ctm_b_branch")]:
            self.bank.append("Branch: %s"%ctl[ctl3.index("ctm_b_branch")])
        if ctl[ctl3.index("ctm_b_ibt")]:
            self.bank.append("Branch Code: %s"%ctl[ctl3.index("ctm_b_ibt")])
        if ctl[ctl3.index("ctm_b_acno")]:
            self.bank.append("Account Number: %s"%ctl[ctl3.index("ctm_b_acno")])
        for x in range(len(self.bank), 6):
            self.bank.append("")
        self.logo = ctl[ctl3.index("ctm_logo")]
        self.flip = ctl[ctl3.index("ctm_flip")]
        ########################################################################
        # Terms and Conditions
        ########################################################################
        mess = self.sql.getRec(tables="ctlmes", cols=["mss_detail"],
            where=[("mss_system", "=", "CON"), ("mss_message", "=", 1)],
            limit=1)
        if mess:
            self.term = textFormat(mess[0], width=30)
        else:
            self.term = []
        for x in range(len(self.term), 6):
            self.term.append("")
        ########################################################################
        # Start of Invoice Layout
        ########################################################################
        if not self.logo or not os.path.exists(self.logo):
            self.logo = None
        elif self.flip == "N":
            self.logo = flipImage(self.logo)
        if not self.logo:
            self.head.append("%1s%-30s" % ("", ctl[ctl3.index("ctm_name")]))
        else:
            self.head.append(" ")
        if self.copy.lower() == "y":
            cpy = "COPY"
        else:
            cpy = ""
        if ctl[ctl3.index("ctm_taxno")]:
            cpy = "%s TAX" % cpy
        if self.dtype == "I":
            self.doctyp = "%s INVOICE" % cpy
        elif self.dtype == "C":
            self.doctyp = "%s CREDIT NOTE" % cpy
        elif self.dtype == "O":
            self.doctyp = "%s SALES ORDER" % cpy
        elif self.dtype == "Q":
            self.doctyp = "%s QUOTATION" % cpy
        elif self.dtype == "D":
            self.doctyp = "DELIVERY NOTE"
        else:
            return "Invalid Document Type"
        ########################################################################
        # Print Company Details
        ########################################################################
        if self.logo:
            self.head.append("%92s%19s" % ("", self.doctyp))
            self.head.append(" ")
            self.head.append(" ")
            self.head.append(" ")
            self.head.append(" ")
        else:
            ####################################################################
            # Extract Required Address
            ####################################################################
            if self.useadd.lower() == "p":
                addr = [ctl[ctl3.index("ctm_add1")]]
                if ctl[ctl3.index("ctm_add2")]:
                    addr.append(ctl[ctl3.index("ctm_add2")])
                if ctl[ctl3.index("ctm_add3")]:
                    addr.append(ctl[ctl3.index("ctm_add3")])
                if ctl[ctl3.index("ctm_pcode")]:
                    addr.append(ctl[ctl3.index("ctm_pcode")])
            else:
                addr = [ctl[ctl3.index("ctm_sadd1")]]
                if ctl[ctl3.index("ctm_sadd2")]:
                    addr.append(ctl[ctl3.index("ctm_sadd2")])
                if ctl[ctl3.index("ctm_sadd3")]:
                    addr.append(ctl[ctl3.index("ctm_sadd3")])
                if ctl[ctl3.index("ctm_spcode")]:
                    addr.append(ctl[ctl3.index("ctm_spcode")])
            num = 4 - len(addr)
            if num:
                addr.extend([""] * num)
            self.head.append("%2s%-30s%2s%-30s%28s%19s" % ("", addr[0], "",
                ("RegNo: " + ctl[ctl3.index("ctm_regno")]), "", self.doctyp))
            self.head.append("%2s%-30s%2s%-30s" % ("", addr[1], "",
                ("TaxNo: " + ctl[ctl3.index("ctm_taxno")])))
            self.head.append("%2s%-30s%2s%-30s" % ("", addr[2], "",
                ("TelNo: " + ctl[ctl3.index("ctm_tel")])))
            self.head.append("%2s%-30s%2s%-30s" % ("", addr[3], "",
                ("FaxNo: " + ctl[ctl3.index("ctm_fax")])))
            self.head.append(" ")
        ########################################################################
        # Debtor's Name, Address and Delivery Address
        ########################################################################
        addr = [drm[drm3.index("drm_add1")]]
        if drm[drm3.index("drm_add2")]:
            addr.append(drm[drm3.index("drm_add2")])
        if drm[drm3.index("drm_add3")]:
            addr.append(drm[drm3.index("drm_add3")])
        if drm[drm3.index("drm_pcod")]:
            addr.append(drm[drm3.index("drm_pcod")])
        num = 4 - len(addr)
        if num:
            addr.extend([""] * num)
        dela = [self.iv1[self.iv13.index("si1_add1")]]
        if self.iv1[self.iv13.index("si1_add2")]:
            dela.append(self.iv1[self.iv13.index("si1_add2")])
        if self.iv1[self.iv13.index("si1_add3")]:
            dela.append(self.iv1[self.iv13.index("si1_add3")])
        if self.iv1[self.iv13.index("si1_add4")]:
            dela.append(self.iv1[self.iv13.index("si1_add4")])
        num = 4 - len(dela)
        if num:
            dela.extend([""] * num)
        self.head.append("%1s%-30s%5s%-30s%36s%9s" % ("",
            drm[drm3.index("drm_name")], "", dela[0], "", self.docno.disp))
        self.head.append("%1s%-30s%5s%-30s" % ("", addr[0], "", dela[1]))
        self.head.append("%1s%-30s%5s%-30s%38s%s" % ("", addr[1], "", dela[2],
            "", "%-7s"))
        self.head.append("%1s%-30s%5s%-30s" % ("", addr[2], "", dela[3]))
        date = CCD(self.iv1[self.iv13.index("si1_date")], "D1", 10.0)
        self.head.append("%1s%-30s%70s%10s" % ("", addr[3], "", date.disp))
        self.head.append(
            "%1s%-9s%1s%-18s%1s%-7s%1s%-37s%1s%-20s%1s%3s%4s%-7s%1s" % ("",
            our, "", cus, "", job, "", per, "", vtn, "", rep, "", acc, ""))
        ########################################################################
        # End of Invoice Layout
        ########################################################################
        if self.dtype in ("C", "I"):
            pfx = "InvCrn"
        elif self.dtype == "O":
            pfx = "SlsOrd"
        elif self.dtype == "Q":
            pfx = "SlsQte"
        elif self.dtype == "D":
            return
        if type(self.email) == str:
            self.emadd = ""
        else:
            self.email, self.emadd = self.email
        if self.dtype == "Q":
            # If no email address supplied, use sales address else accounts
            if not self.emadd:
                self.emadd = drm[drm3.index("drm_sls_email")]
            if not self.emadd:
                self.emadd = drm[drm3.index("drm_acc_email")]
        else:
            # If no email address supplied, use accounts address else sales
            if not self.emadd:
                self.emadd = drm[drm3.index("drm_acc_email")]
            if not self.emadd:
                self.emadd = drm[drm3.index("drm_sls_email")]
        self.pdfnam = getModName(self.mf.rcdic["wrkdir"], pfx,
            "%s_%s" % (self.conum, self.docno.work), ext="pdf", wait=1)
        self.c, self.cpi, self.inch, self.suc, self.duc = \
            doReportlab(self.mf.prgdir, self.pdfnam, orient="L")
        # Width in pixels of each character
        self.cwidth = self.c.stringWidth(("*"), "Tartan", self.cpi)
        # Width in pixels of the form
        self.twidth = self.cwidth * 112

    def doProcess(self):
        iv23 = self.sql.slsiv2_col
        if not self.iv2:
            self.doPrintHead()
        else:
            for line in self.iv2:
                if self.cnt == 66:
                    self.doPrintHead()
                if self.cnt == 37:
                    self.doPrintCfwd()
                grp = CCD(line[iv23.index("si2_group")], "NA", 3.0)
                cod = CCD(line[iv23.index("si2_code")], "NA", 20.0)
                fmt = textFormat(line[iv23.index("si2_desc")], width=30)
                if not fmt:
                    fmt = [""]
                uoi = CCD(line[iv23.index("si2_uoi")], "NA", 10.0)
                unt = CCD(line[iv23.index("si2_units")], "UI", 7.0)
                qty = CCD(line[iv23.index("si2_qty")], "SD", 13.2)
                pri = CCD(line[iv23.index("si2_price")], "UD", 12.2)
                dis = CCD(line[iv23.index("si2_disc_per")], "UD", 6.2)
                rat = CCD(line[iv23.index("si2_vat_rate")], "UD", 6.2)
                if unt.work == 0:
                    price = pri.work
                else:
                    price = round((pri.work / unt.work), 2)
                excpri = round((price * 1), 2)                  #S/Price
                incrat = float(ASD(100.0) + ASD(rat.work))
                incpri = round((price * incrat / 100.0), 4)     #S/Price+VAT
                net = float(ASD(100.0) - ASD(dis.work))
                excamt = round((qty.work * excpri * net / 100.0), 2)
                incamt = round((qty.work * incpri * net / 100.0), 2)
                vatamt = float(ASD(incamt) - ASD(excamt))
                if excamt == incamt:
                    self.totnot = float(ASD(self.totnot) + ASD(excamt))
                else:
                    self.totabl = float(ASD(self.totabl) + ASD(excamt))
                self.totvat = float(ASD(self.totvat) + ASD(vatamt))
                self.invtot = float(ASD(self.invtot) + ASD(incamt))
                val = CCD(incamt, "SD", 13.2)
                for num, lin in enumerate(fmt):
                    if self.cnt == 37:
                        self.doPrintCfwd()
                    if num == 0 and num + 1 == len(fmt):
                        if self.dtype == "D" and self.dvals == "N":
                            body=" %3s %20s %-30s %13s %12s %6s %6s %13s"
                            self.t.textLine(body % (grp.disp, cod.disp, lin,
                                qty.disp, "", dis.disp, rat.disp, ""))
                        else:
                            body=" %3s %20s %-30s %13s %12s %6s %6s %13s"
                            self.t.textLine(body % (grp.disp, cod.disp, lin,
                                qty.disp, pri.disp, dis.disp, rat.disp,
                                val.disp))
                    elif num == 0:
                        body=" %3s %20s %-30s"
                        self.t.textLine(body % (grp.disp, cod.disp, lin))
                    elif num + 1 == len(fmt):
                        if self.dtype == "D" and self.dvals == "N":
                            body="%25s %-30s %13s %12s %6s %6s %13s"
                            self.t.textLine(body % ("", lin, qty.disp, "",
                                dis.disp, rat.disp, ""))
                        else:
                            body="%25s %-30s %13s %12s %6s %6s %13s"
                            self.t.textLine(body % ("", lin, qty.disp, pri.disp,
                                dis.disp, rat.disp, val.disp))
                    else:
                        body="%25s %-30s"
                        self.t.textLine(body % ("", lin))
                    self.cnt += 1
        self.doPrintTotal()
        self.c.drawText(self.t)
        self.c.showPage()
        if self.dtype == "I" and self.dnote.lower() == "y":
            return
        self.c.save()
        if self.prtnam == "view":
            view = "V"
        else:
            view = "P"
        if self.email.lower() == "y" and self.emadd:
            mail = self.emadd
        else:
            mail = "N"
        if self.dtype in ("I", "D"):
            head = "Invoice %s" % self.docno.disp
        elif self.dtype == "C":
            head = "Credit Note %s" % self.docno.disp
        elif self.dtype == "O":
            head = "Sales Order %s" % self.docno.disp
        elif self.dtype == "Q":
            head = "Quotation %s" % self.docno.disp
        doPrinter(mf=self.mf, conum=self.conum, rep=self.pdfnam,
            head=head, view=view, prtnam=self.prtnam, mail=mail)

    def doPrintCfwd(self):
        self.t.textLine()
        self.c.line(self.cwidth*.5, (38.5*(self.inch / 6)),
        self.twidth-(self.cwidth*.5), (38.5*(self.inch / 6)))
        self.t.textLine("%1s%62s%15s" % ("", "", "Carried Forward"))
        self.c.drawText(self.t)
        self.c.showPage()
        self.doPrintHead()

    def doPrintHead(self):
        self.c.translate(self.inch*1, 0)    # Set x and y margins
        if self.copy.lower() == "y":
            self.c.setFillGray(.7)
            x = self.cwidth * 111.5
            self.c.setFont("Tartan-Bold", 50)
            self.c.drawCentredString(x/2, self.c._pagesize[1]/1.7, self.doctyp)
            self.c.setFont("Tartan-Bold", self.cpi)
            self.c.setFillGray(0)
        self.doDrawInvoice()
        self.t = self.c.beginText()
        self.t.setTextOrigin(0, 22)
        self.t.setFont("Tartan-Bold", 18)
        self.t.textLine(self.head[0])
        self.t.setFont("Tartan-Bold", self.cpi)
        self.t.textLine(self.head[1])
        self.t.textLine(self.head[2])
        self.t.textLine(self.head[3])
        self.t.textLine(self.head[4])
        self.t.setFont("Tartan", self.cpi)
        self.t.textLine()
        self.t.textLine(self.head[5])
        self.t.textLine(self.head[6])
        self.t.textLine(self.head[7])
        self.pag += 1
        page = CCD(self.pag, "UI", 7.0)
        self.t.textLine(self.head[8] % page.disp)
        self.t.textLine(self.head[9])
        self.t.textLine(self.head[10])
        self.t.textLine()
        self.t.textLine()
        self.t.textLine(self.head[11])
        self.t.textLine()
        self.t.textLine()
        self.cnt = 18

    def doDrawInvoice(self):
        if self.logo:
            self.c.drawImage(self.logo, 0, 10, 400, 70, None)
        ########################################################################
        # Vertical left and right margins
        ########################################################################
        self.c.line(self.cwidth * .5, (14 * (self.inch / 6)),
            self.cwidth * .5, (38.5 * (self.inch / 6)))
        self.c.line(self.cwidth * 111.5, (14 * (self.inch / 6)),
            self.cwidth * 111.5, (38.5 * (self.inch / 6)))
        ########################################################################
        # Draw Charge To and Deliver To rectangles
        ########################################################################
        self.c.setFillGray(.9)
        x = self.cwidth * .5
        w = self.cwidth * 31
        self.c.rect(x, (7.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (7.75 * (self.inch / 6)), w, (6 * (self.inch / 6)),
            fill=0)
        x = self.cwidth * 35.5
        self.c.rect(x, (7.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (7.75 * (self.inch / 6)), w, (6 * (self.inch / 6)),
            fill=0)
        ########################################################################
        # Fill rectangles with details
        ########################################################################
        self.c.setFillGray(0)
        self.c.setFont("Tartan-Bold", 8)
        self.c.drawString(self.cwidth * 1, (8.45 * (self.inch / 6)),
            "Charge To:")
        self.c.drawString(self.cwidth * 36, (8.45 * (self.inch / 6)),
            "Deliver To:")
        self.c.setFont("Tartan", self.cpi)
        ########################################################################
        # Draw rectangles for Document, Page Number and Date
        ########################################################################
        self.c.setFillGray(.9)
        x = self.cwidth * 100.5
        w = self.cwidth * 11
        self.c.rect(x, (7.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (8.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=0)
        self.c.rect(x, (9.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (10.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=0)
        self.c.rect(x, (11.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (12.75 * (self.inch / 6)), w, (1.25 * (self.inch / 6)),
            fill=0)
        self.c.setFillGray(0)
        ########################################################################
        # Fill rectangles with details
        ########################################################################
        if self.dtype == "I":
            txt = "Invoice No"
        elif self.dtype == "C":
            txt = "Cr-Note No"
        elif self.dtype == "O":
            txt = "Order No"
        elif self.dtype == "Q":
            txt = "Quote No"
        elif self.dtype == "D":
            txt = "D-Note No"
        self.c.setFont("Tartan-Bold", 8)
        self.c.drawCentredString(self.twidth-(self.cwidth * 5.5),
            (8.45 * (self.inch / 6)), txt)
        self.c.drawCentredString(self.twidth-(self.cwidth * 5),
            (10.45 * (self.inch / 6)), "Page No")
        self.c.drawCentredString(self.twidth-(self.cwidth * 5),
            (12.45 * (self.inch / 6)), "Date")
        self.c.setFont("Tartan", self.cpi)
        ########################################################################
        # Draw rectangles for ribbon line
        ########################################################################
        self.c.setFillGray(.9)
        self.c.rect(self.cwidth * .5, (14 * (self.inch / 6)),
            (self.cwidth * 111), (1.5 * (self.inch / 6)), fill=1)
        self.c.setFillGray(0)
        for num in (10.5, 29.5, 37.5, 75.5, 96.5, 100.5):
            col = self.cwidth * num
            self.c.line(col, (14 * (self.inch / 6)), col, 17 * (self.inch / 6))
        ########################################################################
        # Fill rectangles with details
        ########################################################################
        txt = "Del-Numbr Cus-Ord-Number     Job-Num "\
            "Contact-Person                        "\
            "V.A.T. Number        Rep    Acc-Num"
        self.c.setFont("Tartan-Bold", self.cpi)
        self.c.drawString(self.cwidth * 1, (15 * (self.inch / 6)), txt)
        self.c.setFont("Tartan", self.cpi)
        ########################################################################
        # Draw rectangles for body headings
        ########################################################################
        self.c.setFillGray(.9)
        self.c.rect(self.cwidth * .5, (17 * (self.inch / 6)),
            (self.cwidth * 111), (1.5 * (self.inch / 6)), fill=1)
        self.c.setFillGray(0)
        for num in (4.5, 25.5, 56.5, 70.5, 83.5, 90.5, 97.5):
            col = self.cwidth * num
            self.c.line(col, (17 * (self.inch / 6)), col, 38.5 * (self.inch/6))
        txt = "Grp Product Code         Description                    "\
            "    Quantity         Price Disc-%  Tax-%        Value "
        self.c.setFont("Tartan-Bold", self.cpi)
        self.c.drawString(self.cwidth * 1, (18 * (self.inch / 6)), txt)
        self.c.setFont("Tartan", self.cpi)

    def doPrintTotal(self):
        lines = 38 - self.cnt
        for x in range(lines):
            self.t.textLine()
        y = 38.5 * (self.inch / 6)
        if self.dtype == "D" and self.dvals == "N":
            self.c.setFillGray(0)
            self.c.line(self.cwidth * .5, y, self.cwidth * 111.5, y)
            self.t.textLine()
        else:
            self.c.setFillGray(.9)
            for x, w, f in ((.5, 13, 1), (13.5, 14, 0), (27.5, 14, 1),
                (41.5, 14, 0), (55.5, 14, 1), (69.5, 14, 0), (83.5, 14, 1),
                (97.5, 14, 0)):
                self.c.rect((self.cwidth * x), y, (self.cwidth * w),
                    (1.5 * (self.inch / 6)), fill=f)
            self.c.setFillGray(0)
            abl = CCD(self.totabl, "SD", 13.2)
            exc = CCD(self.totnot, "SD", 13.2)
            vat = CCD(self.totvat, "SD", 13.2)
            amt = CCD(self.invtot, "SD", 13.2)
            self.t.setFont("Tartan-Bold", self.cpi)
            self.t.textLine("%1s%-12s%1s%13s%1s%-13s%1s%13s%1s%-13s%1s%13s"\
                "%1s%-13s%1s%13s%1s" % ("", "Taxable", "", abl.disp, "",
                "Non-Taxable", "", exc.disp, "", "Total Tax", "", vat.disp,
                "", "Total Value", "", amt.disp, ""))
        self.t.textLine()
        self.t.textLine()
        mess = self.iv1[self.iv13.index("si1_mes1")]
        self.t.setFont("Tartan", self.cpi)
        ########################################################################
        # Draw rectangles: rect(x, y, w, h, fill)
        ########################################################################
        self.c.setFillGray(.9)
        x = self.cwidth * .5
        w = self.cwidth * 31
        self.c.rect(x, (40.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (40.75 * (self.inch / 6)), w, (6 * (self.inch / 6)),
            fill=0)
        x = self.cwidth * 40.5
        w = self.cwidth * 31
        self.c.rect(x, (40.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (40.75 * (self.inch / 6)), w, (6 * (self.inch / 6)),
            fill=0)
        x = self.cwidth * 80.5
        self.c.rect(x, (40.75 * (self.inch / 6)), w, (1 * (self.inch / 6)),
            fill=1)
        self.c.rect(x, (40.75 * (self.inch / 6)), w, (6 * (self.inch / 6)),
            fill=0)
        self.c.setFillGray(0)
        self.c.setFont("Tartan-Bold", 8)
        if self.dtype == "D":
            self.c.drawString(self.cwidth * 1,
            (41.45 * (self.inch / 6)), "Received By:")
        else:
            self.c.drawString(self.cwidth * 1,
            (41.45 * (self.inch / 6)), "Message:")
        self.c.drawString(self.cwidth * 41,
            (41.45 * (self.inch / 6)), "Banking Details:")
        self.c.drawString(self.cwidth * 81,
            (41.45 * (self.inch / 6)), "Terms and Conditions:")
        self.c.setFont("Tartan", self.cpi)
        for x in range(0, 5):
            if self.dtype == "D":
                # Leave the Received By Box empty
                mess = "%40s" % ""
            else:
                # Populate the Message Box
                exec "mm = self.iv1[self.iv13.index('si1_mes%s' % (x+1))]"
                mess = "%1s%-30s%9s" % ("", mm, "")
            mess = "%s%1s%-30s%9s" % (mess, "", self.bank[x], "")
            mess = "%s%1s%-30s%9s" % (mess, "", self.term[x], "")
            self.t.textLine(mess)

class PrintPayslip:
    """
    Used to print payslips

    conum  : Company Number
    empnos : List of Employees
    rundt  : Run Date
    progs  : Print Progressives (y/n)
    runtp  : c=Copy, o=Original, p=Preview
    prtnam : Printer Name
    email  : Email Payslips (y/n) or an email address
    printm : Whether to View/Print emailed payslips
    message: Message to Print
    export : The open file object for Best Upload
    bestac : Standard Bank Best Account Code
    """
    def __init__(self, mf, conum, empnos, rundt, **args):
        self.mf = mf
        self.conum = conum
        self.empnos = empnos
        self.rundt = CCD(rundt, "D1", 10.0)
        defaults = {
            "progs": "n",
            "runtp": "c",
            "prtnam": "Default",
            "email": "n",
            "printm": "n",
            "message": None,
            "export": None,
            "bestac": None}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        self.no = self.doVariables()
        if not self.no:
            if self.email.lower() == "n":
                self.key = "%s_all" % self.conum
                self.doPrintSetup()
            for self.empno in self.empnos:
                if self.email.lower() == "y":
                    self.key = "%s_%s" % (self.conum, self.empno)
                    self.doPrintSetup()
                self.doEmpVariables()
                self.doProcess()
            if self.email.lower() == "n":
                self.doPrintSlip()
        else:
            showError(self.mf.body, "Variable Error", self.no)

    def doVariables(self):
        self.sql = Sql(self.mf.dbm, ["ctlmst", "wagedc", "wagmst", "wagtf1",
            "wagtf2"], prog=__name__)
        ctl = self.sql.getRec(tables="ctlmst", where=[("ctm_cono", "=",
            self.conum)], limit=1)
        ctl3 = self.sql.ctlmst_col
        self.conam = ctl[ctl3.index("ctm_name")]
        self.logo = ctl[ctl3.index("ctm_logo")]
        self.flip = ctl[ctl3.index("ctm_flip")]
        ########################################################################
        # Payslip header
        ########################################################################
        if self.runtp.lower() == "c":
            typ = "Copy"
        elif self.runtp.lower() == "p":
            typ = "Preview"
        else:
            typ = "Original"
        doc = "%s Payslip" % typ
        if not self.logo or not os.path.exists(self.logo):
            self.logo = None
        elif self.flip == "N":
            self.logo = flipImage(self.logo)
        if self.logo:
            self.hedr = [" ", " ", " ", " ", " ", " "]
        else:
            ####################################################################
            # Company Details
            ####################################################################
            self.hedr = [""]
            ctl = self.sql.getRec(tables="ctlmst", where=[("ctm_cono", "=",
                self.conum)], limit=1)
            ctl3 = self.sql.ctlmst_col
            self.hedr.append("%1s%-30s" % ("", ctl[ctl3.index("ctm_name")]))
            addr = [ctl[ctl3.index("ctm_add1")]]
            if ctl[ctl3.index("ctm_add2")]:
                addr.append(ctl[ctl3.index("ctm_add2")])
            if ctl[ctl3.index("ctm_add3")]:
                addr.append(ctl[ctl3.index("ctm_add3")])
            if ctl[ctl3.index("ctm_pcode")]:
                addr.append(ctl[ctl3.index("ctm_pcode")])
            num = 4 - len(addr)
            if num:
                addr.extend([""] * num)
            self.hedr.append("%2s%-30s%2s%-30s" % ("", addr[0], "",
                ("RegNo: " + ctl[ctl3.index("ctm_regno")])))
            self.hedr.append("%2s%-30s%2s%-30s" % ("", addr[1], "",
                ("TaxNo: " + ctl[ctl3.index("ctm_taxno")])))
            self.hedr.append("%2s%-30s%2s%-30s" % ("", addr[2], "",
                ("TelNo: " + ctl[ctl3.index("ctm_tel")])))
            self.hedr.append("%2s%-30s%2s%-30s" % ("", addr[3], "",
                ("FaxNo: " + ctl[ctl3.index("ctm_fax")])))
        self.hedr.append("")
        self.hedr.append("")
        self.hedr.append("%24s%16s%24s" % ("", doc, ""))
        self.hedr.append("")
        self.hedr.append("")
        ########################################################################
        # Start date and email details
        ########################################################################
        y = self.rundt.work / 10000
        m = (self.rundt.work / 100) % 100
        if m < 3:
            y = y - 1
        self.start = (y * 10000) + 301
        self.etotal = 0
        if self.email.count("@"):
            self.emadd = self.email
            self.email = "n"
        else:
            self.emadd = ""

    def doEmpVariables(self):
        self.cnt = 66
        self.erns = {}
        self.deds = {}
        self.head = self.hedr[:]
        self.payern = 0
        self.payded = 0
        wgm = self.sql.getRec(tables="wagmst", where=[("wgm_cono", "=",
            self.conum), ("wgm_empno", "=", self.empno)], limit=1)
        if not wgm:
            return "Invalid Employee %s %s" % (self.conum, self.empno)
        wgm3 = self.sql.wagmst_col
        wt13 = self.sql.wagtf1_col
        wt1 = self.sql.getRec(tables="wagtf1", where=[("wt1_cono", "=",
            self.conum), ("wt1_empno", "=", self.empno), ("wt1_date", "=",
            self.rundt.work)], limit=1)
        if not wt1:
            return "No Payslip for %s %s %s" % (self.conum, self.empno,
                wgm[wgm3.index("wgm_sname")])
        pag = wt1[wt13.index("wt1_page")]
        bch = wt1[wt13.index("wt1_branch")]
        dep = wt1[wt13.index("wt1_dept")]
        ini = wgm[wgm3.index("wgm_fname")].split()
        nam = ""
        for i in ini:
            nam = nam + i + " "
        nam = nam + wgm[wgm3.index("wgm_sname")]
        self.head.append("%1s%-38s%1s%5s%1s%3s%1s%1s%1s%10s%1s%5s" % ("", nam,
            "", self.empno, "", bch, "", dep, "", self.rundt.disp, "", pag))
        self.head.append(" ")
        self.ptype = wgm[wgm3.index("wgm_ptype")]
        self.btype = wgm[wgm3.index("wgm_btype")]
        self.bbranch = wgm[wgm3.index("wgm_bbranch")]
        self.bacno = wgm[wgm3.index("wgm_bacno")]
        self.bname = wgm[wgm3.index("wgm_bname")]
        if self.email.lower() == "y":
            self.emadd = wgm[wgm3.index("wgm_emadd")]

    def doPrintSetup(self):
        ########################################################################
        # Document Layout
        ########################################################################
        self.pdfnam = getModName(self.mf.rcdic["wrkdir"], "PaySlip",
            self.key, ext="pdf")
        self.c, self.cpi, self.inch, self.suc, self.duc = \
            doReportlab(self.mf.prgdir, self.pdfnam, orient="P")
        # Width in pixels of each character
        self.cwidth = self.c.stringWidth(("*"), "Tartan", self.cpi)
        # Width in pixels of the form
        self.twidth = self.cwidth * 93

    def doProcess(self):
        self.notional = False
        wt23 = self.sql.wagtf2_col
        wt2 = self.sql.getRec(tables="wagtf2", where=[("wt2_cono", "=",
            self.conum), ("wt2_empno", "=", self.empno), ("wt2_date", "=",
            self.rundt.work)], order="wt2_type desc, wt2_code")
        if not wt2:
            return
        ltp = ""
        for line in wt2:
            if self.cnt == 66:
                self.doPrintHead()
            if self.cnt == 37:
                self.doPrintCfwd()
            typ = CCD(line[wt23.index("wt2_type")], "UA", 1.0)
            tax = CCD(line[wt23.index("wt2_taxcode")], "UA", 1.0)
            if ltp and typ.work != ltp:
                self.doTypeTotal(ltp)
            ltp = typ.work
            cod = CCD(line[wt23.index("wt2_code")], "UI", 3.0)
            det = self.sql.getRec(tables="wagedc", where=[("ced_cono", "=",
                self.conum), ("ced_type", "=", typ.work), ("ced_code", "=",
                cod.work)], limit=1)
            if not det:
                continue
            des = det[self.sql.wagedc_col.index("ced_desc")]
            hrs = CCD(line[wt23.index("wt2_hrs")], "SD", 13.2)
            if hrs.work == 0:
                hrs.disp = ""
            emt = CCD(line[wt23.index("wt2_eamt")], "SD", 13.2)
            if emt.work == 0:
                continue
            ind = ""
            if typ.work == "E" and tax.work != "T":
                self.payern = float(ASD(self.payern) + ASD(emt.work))
            elif tax.work != "T":
                self.payded = float(ASD(self.payded) + ASD(emt.work))
            else:
                self.notional = True
                ind = " **"
            body = "%1s%-40s%1s%13s%1s%13s%s" % ("", des, "", hrs.disp, "",
                emt.disp, ind)
            self.t.textLine(body)
            self.cnt += 1
        self.doTypeTotal(ltp)
        self.doGrandTotal()
        if self.email.lower() == "y":
            self.doPrintSlip()

    def doPrintSlip(self):
        self.c.save()
        if self.prtnam == "view":
            view = "V"
        else:
            view = "P"
        if self.email.lower() == "n" and not self.emadd:
            mail = "N"
        else:
            mail = self.emadd
        doPrinter(mf=self.mf, conum=self.conum, rep=self.pdfnam,
            head="Payslip for %s" % self.rundt.disp, view=view,
            prtnam=self.prtnam, mail=mail, printm=self.printm)

    def doPrintCfwd(self):
        self.t.textLine()
        self.cnt += 1
        self.c.line(self.cwidth*.5, (38.5*(self.inch / 6)),
            self.twidth-(self.cwidth*.5), (38.5*(self.inch / 6)))
        self.t.textLine("%1s%62s%15s" % ("", "", "Carried Forward"))
        self.cnt += 1
        self.c.drawText(self.t)
        self.c.showPage()
        self.doPrintHead()

    def doPrintHead(self):
        self.c.translate(self.inch*1, self.inch*.5)    # Set x and y margins
        self.doDrawPayslip()
        self.t = self.c.beginText()
        self.t.setTextOrigin(0, 0)
        self.cnt = 0
        for num, line in enumerate(self.head):
            if num == 1:
                self.t.setFont("Tartan-Bold", 18)
            elif num in (1, 8):
                self.t.setFont("Tartan-Bold", self.cpi)
            else:
                self.t.setFont("Tartan", self.cpi)
            self.t.textLine(line)
            self.cnt += 1
        self.t.setFont("Tartan", self.cpi)

    def doDrawPayslip(self):
        if self.logo:
            self.c.drawImage(self.logo, 0, 10, 400, 70, None)
        ########################################################################
        # Draw rectangles for ribbon line
        ########################################################################
        self.c.setFillGray(.9)
        self.c.rect(self.cwidth*.5, (9.50*(self.inch / 6)),
            (self.cwidth*68), (1.25*(self.inch / 6)), fill=1)
        self.c.rect(self.cwidth*.5, (10.75*(self.inch / 6)),
            (self.cwidth*68), (1.75*(self.inch / 6)), fill=0)
        self.c.setFillGray(0)
        for num in (39.5, 45.5, 49.5, 51.5, 62.5):
            col = self.cwidth*num
            self.c.line(col, (9.5*(self.inch / 6)), col, 12.5*(self.inch / 6))
        ########################################################################
        # Fill rectangles with details
        ########################################################################
        txt = "Employee Name                          "\
            "EmpNo Bch D    Date    SlpNo"
        self.c.setFont("Tartan-Bold", self.cpi)
        self.c.drawString(self.cwidth*1, (10.3*(self.inch / 6)), txt)
        self.c.setFont("Tartan", self.cpi)

    def doTypeTotal(self, ttype):
        self.doDrawLine()
        if ttype == "E":
            des = "Gross Pay"
            amt = CCD(self.payern, "SD", 13.2)
        else:
            des = "Total Deductions"
            amt = CCD(self.payded, "SD", 13.2)
        self.t.textLine("%1s%-55s%13s" % ("", des, amt.disp))
        self.t.textLine("")
        self.cnt = self.cnt + 2

    def doGrandTotal(self):
        tot = CCD(float(ASD(self.payern) - ASD(self.payded)), "SD", 9.2)
        self.doDrawLine()
        self.t.setFont("Tartan-Bold", self.cpi)
        self.t.textLine("%1s%-55s%13s" % ("", "Total Net Pay", tot.disp))
        self.t.setFont("Tartan", self.cpi)
        self.cnt += 1
        self.doDrawLine(ltype="d")
        if self.progs.lower() == "y":
            self.doAccumProgs()
            self.t.textLine()
            self.cnt += 1
            self.c.setFillGray(.9)
            self.c.rect(self.cwidth*.5, (self.cnt*(self.inch / 6)),
                (self.cwidth*68), (1.25*(self.inch / 6)), fill=1)
            self.c.setFillGray(0)
            self.t.setFont("Tartan-Bold", self.cpi)
            self.t.textLine("%1s%-55s" % ("", "Progressive Totals"))
            self.t.setFont("Tartan", self.cpi)
            self.cnt += 1
            self.t.textLine()
            self.cnt += 1
            keys = self.erns.keys()
            keys.sort()
            for key in keys:
                amt = CCD(self.erns[key][1], "SD", 13.2)
                self.t.textLine("%1s%-55s%13s" % \
                    ("", self.erns[key][0], amt.disp))
                self.cnt += 1
            keys = self.deds.keys()
            keys.sort()
            for key in keys:
                amt = CCD(self.deds[key][1], "SD", 13.2)
                self.t.textLine("%1s%-55s%13s" % \
                    ("", self.deds[key][0], amt.disp))
                self.cnt += 1
        self.t.textLine()
        self.cnt += 1
        if self.message:
            fmt = textFormat(self.message[0], width=30)
            ####################################################################
            # Draw Message rectangles
            ####################################################################
            self.c.setFillGray(.9)
            x = self.cwidth*.5
            w = self.cwidth*31
            self.c.rect(x, (self.cnt*(self.inch / 6)), w, (1*(self.inch / 6)),
                fill=1)
            self.c.rect(x, (self.cnt*(self.inch / 6)), w, (6*(self.inch / 6)),
                fill=0)
            self.c.setFillGray(0)
            self.c.setFont("Tartan-Bold", 8)
            self.c.drawString(self.cwidth*1, ((self.cnt+.75)*(self.inch / 6)),
                "Message:")
            self.c.setFont("Tartan", self.cpi)
            self.t.textLine()
            self.cnt += 1
            for x in range(0, len(fmt)):
                exec "mess = '%1s%-30s'" % ("", fmt[x])
                self.t.textLine(mess)
                self.cnt += 1
            for y in range(x, 5):
                self.t.textLine()
                self.cnt += 1
            self.t.textLine()
            self.cnt += 1
        self.t.setFont("Tartan-Bold", self.cpi)
        if self.notional:
            self.t.textLine("** Denotes Amounts Used for Tax Calculation "\
                "Purposes Only")
            self.t.textLine()
            self.cnt += 1
        if self.ptype == "C":
            typ = "Cash"
        elif self.ptype == "E":
            typ = "Electronic"
        else:
            typ = "Cheque"
        self.t.textLine("%1s%-55s" % ("", "Method of Payment: %s" % typ))
        self.t.setFont("Tartan", self.cpi)
        self.cnt += 1
        self.c.drawText(self.t)
        self.c.showPage()
        if self.ptype == "E" and self.export and self.bestac:
            value = long(round((tot.work * 100), 0))
            self.export.write("%1s%4s%06u%07u%019u%1s%1s%011u%-20s%10s"\
                "%-15s%1s\r\n" % (2, self.bestac, self.bbranch,
                int(self.empno), long(self.bacno), "", self.btype,
                value, self.bname, "", self.conam[:15], ""))
            self.etotal = float(ASD(self.etotal) + ASD(tot.work))

    def doDrawLine(self, ltype="s"):
        self.c.setFont("Tartan", self.cpi)
        x1 = 56*self.cwidth
        x2 = 68*self.cwidth
        y = self.cnt*(self.inch / 6)
        self.c.line(x1, y, x2, y)
        if ltype == "d":
            y = y + 2
            self.c.line(x1, y, x2, y)

    def doAccumProgs(self):
        wt2 = self.sql.getRec(tables="wagtf2", cols=["wt2_type", "wt2_code",
            "round(sum(wt2_hrs), 2)", "round(sum(wt2_eamt), 2)"],
            where=[("wt2_cono", "=", self.conum), ("wt2_empno", "=",
            self.empno), ("wt2_date", "between", self.start, self.rundt.work)],
            group="wt2_type, wt2_code", order="wt2_type desc, wt2_code")
        for line in wt2:
            typ = CCD(line[0], "UA", 1.0)
            cod = CCD(line[1], "UI", 3.0)
            det = self.sql.getRec(tables="wagedc", cols=["ced_desc"],
                where=[("ced_cono", "=", self.conum), ("ced_type", "=",
                typ.work), ("ced_code", "=", cod.work)], limit=1)
            if not det:
                des = "Unknown"
            else:
                des = det[0]
            hrs = CCD(line[2], "SD", 13.2)
            emt = CCD(line[3], "SD", 13.2)
            if typ.work == "E":
                if cod.work in self.erns:
                    self.erns[cod.work][1] = float(ASD(self.erns[cod.work][1]) +
                        ASD(emt.work))
                else:
                    self.erns[cod.work] = [des, emt.work]
            else:
                if cod.work in self.deds:
                    self.deds[cod.work][1] = float(ASD(self.deds[cod.work][1]) +
                        ASD(emt.work))
                else:
                    self.deds[cod.work] = [des, emt.work]

class PwdConfirm:
    """
    conum       Company Number
    scrn        The widget to use for the container
    code        The system code as per the 'pwctrl' tuple in tartanWork.py i.e.
    system      The system code i.e.
                    MST - Control Modules
                    GEN - General Ledger
                    CRS - Creditor's Ledger
                    DRS - Debtor's Ledger
                    MEM - Member's Ledger
                    RCA - Rental's Ledger (Extended)
                    RTL - Rental's Ledger (Basic)
                    STR - Store's Ledger
                    INV - Sales Invoicing
    code        The routine as per the 'pwctrl' tuple in tartanWork.py e.g.
                    NoPost - Allow Postings to Locked Accounts
    passwd      The non interactive password
    product     In the case of STR, the group, code and description of the
                product

    Examples:
    =========
    GEN NoPost Allows Postings to Locked G/Ledger Accounts
    DRS NewAcc Allows you to Create a New Debtor's Account
    STR ExQty  Allows you to Override a Quantity Issue
    """
    def __init__(self, mf, **args):
        self.mf = mf
        defaults = {
            "conum": None,
            "scrn": None,
            "system": None,
            "code": None,
            "passwd": None,
            "product": None}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        if self.scrn:
            self.newscn = None
        else:
            self.newscn = (0, 1)
        self.setVariables()
        self.loadPasswords()
        if not self.pwd and not self.sup:
            self.flag = "ok"
        elif self.passwd:
            if self.passwd in (self.pwd, self.sup):
                self.flag = "ok"
            else:
                self.flag = "no"
        else:
            self.mainProcess()
            self.mf.startLoop()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, "ctlpwr", prog=__name__)
        self.tries = 0

    def loadPasswords(self):
        pwd = self.sql.getRec(tables="ctlpwr", cols=["pwd_pass", "pwd_desc"],
            where=[("pwd_cono", "=", 0), ("pwd_sys", "=", self.system),
            ("pwd_code", "=", self.code)], limit=1)
        if not pwd:
            pwd = self.sql.getRec(tables="ctlpwr", cols=["pwd_pass",
                "pwd_desc"], where=[("pwd_cono", "=", self.conum), ("pwd_sys",
                "=", self.system), ("pwd_code", "=", self.code)], limit=1)
        if not pwd:
            self.pwd = None
            self.desc = self.code
        else:
            self.pwd = base64.decodestring(pwd[0])
            self.desc = pwd[1]
        sup = self.sql.getRec(tables="ctlpwr", cols=["pwd_pass"],
            where=[("pwd_cono", "=", 0), ("pwd_sys", "=", self.system),
            ("pwd_code", "=", "Super")], limit=1)
        if not sup:
            sup = self.sql.getRec(tables="ctlpwr", cols=["pwd_pass"],
                where=[("pwd_cono", "=", self.conum), ("pwd_sys", "=",
                self.system), ("pwd_code", "=", "Super")], limit=1)
        if not sup:
            self.sup = None
        else:
            self.sup = base64.decodestring(sup[0])

    def mainProcess(self):
        tit = (("%s" % self.desc),)
        if self.product:
            fld = [
                (("T",0,0,0),"ONA",3,"Product Group"),
                (("T",0,0,0),"ONA",20," Code "),
                (("T",0,1,0),"ONA",30,"Description"),
                (("T",0,2,0),"IHA",30,"Password","Password",
                    "","Y",self.doPwd,None,None,None)]
        else:
            fld = [
                (("T",0,3,0),"IHA",30,"Password","Password",
                    "","Y",self.doPwd,None,None,None)]
        but = (("Cancel",None,self.doCancel,1,None, None),)
        tnd = ((self.doExit, "n"), )
        txt = (self.doCancel, )
        self.df = TartanDialog(self.mf, screen=self.scrn, newscn=self.newscn,
            tops=True, title=tit, eflds=fld, butt=but, tend=tnd, txit=txt)
        if self.product:
            for x in range(3):
                self.df.loadEntry("T", 0, x, data=self.product[x])

    def doPwd(self, frt, pag, r, c, p, i, w):
        if w not in (self.pwd, self.sup, self.mf.override):
            self.tries += 1
            if self.tries == 3:
                return "xt"
            else:
                return "Invalid Password"
        self.flag = "ok"
        return "nc"

    def doCancel(self, widget=None):
        self.flag = "no"
        self.doExit()

    def doExit(self):
        self.df.closeProcess()
        self.mf.closeLoop()

class BankImport:
    """
    Used for importing statement files

        mf     - The Mainframe object.
        impfmt - The format of the file to import:
                     O - OFX Open Finacial Exchange Format
                     Q - QIF format for Quiken.
                     S - Standard Bank Best file 133 bytes.
        bankac - The bank account number.
        dtefmt - The date format of the file:
                     A - CCYYMMDD
                     B - DDMMCCYY
                     C - MMDDCCYY or MMDDYY
    """
    def __init__(self, mf, impfmt, bankac, dtefmt):
        self.mf = mf
        self.impfmt = impfmt
        self.bankac = bankac
        self.dtefmt = dtefmt
        self.doVariables()
        while not self.fname:
            self.selectFile()
            if self.fname == None:
                return
        for fname in self.fname:
            if self.impfmt == "O":
                self.importOfxFile(fname)
            elif self.impfmt == "Q":
                self.importQifFile(fname)
            elif self.impfmt == "S":
                self.importStdFile(fname)

    def doVariables(self):
        self.trans = []
        self.fname = []

    def selectFile(self):
        if self.impfmt == "O":
            ftype = [("OFX Files", "*.ofx")]
        elif self.impfmt == "Q":
            ftype = [("QIF Files", "*.qif")]
        else:
            ftype = []
        if sys.platform == "win32":
            lastdir = os.path.join(self.mf.rcdic["wrkdir"], "bankdir")
        else:
            lastdir = os.path.join(self.mf.rcdic["wrkdir"], ".bankdir")
        if os.path.exists(lastdir):
            infle = open(lastdir, "r")
            name = infle.readline().replace("\n", "")
            infle.close()
        else:
            name = self.mf.rcdic["wrkdir"]
        dialog = FileDialog(title="Bank Account Number %s" % self.bankac,
            initd=name, ftype=ftype, multi=True)
        self.fname = dialog.askopenfilename()
        if self.fname == None:
            return
        self.fname = self.fname.split(",")
        for fname in self.fname:
            if fname and self.bankac and not fname.count(self.bankac):
                showError(self.mf.body, "File Error %s" % fname,
                    "The File Name Must Contain the Bank Account Number.")
                self.fname = []
        try:
            infle = open(lastdir, "w")
            infle.write(os.path.dirname(os.path.normpath(self.fname[0])))
            infle.close()
        except:
            pass

    def importOfxFile(self, fname):
        infle = open(fname, "r")
        data = infle.readlines()
        infle.close()
        #version = None
        #for dat in data:
        #    dat = dat.replace(chr(10),"").replace(chr(13),"")
        #    if dat.count("<OFX>"):
        #        version = 200
        #        break
        #    elif dat.count("VERSION:"):
        #        version = dat.split(":")[1]
        #        break
        account = None
        start = False
        self.doClear()
        for line in data:
            line = line.rstrip()
            if not account and line.count("<ACCTID>"):
                account = self.removeTag("ACCTID", line)
                if not account.count(self.bankac):
                    showError(self.mf.body, "File Error %s" % fname,
                        "The File Must Contain the Bank Account Number.")
                    return
                else:
                    continue
            if line.count("<BANKTRANLIST>"):
                # Beginning of Transactions
                start = True
                continue
            if not start:
                continue
            if line.count("<DTEFF>"):
                # Version 102 Date
                if self.date.work:
                    self.addTrans()
                    self.doClear()
                if self.getDate(self.removeTag("DTEFF", line)):
                    return
                continue
            elif line.count("<DTPOSTED>"):
                # Version 200 Date
                if self.date.work:
                    self.addTrans()
                    self.doClear()
                if self.getDate(self.removeTag("DTPOSTED", line)):
                    return
                continue
            elif line.count("<FITID>"):
                # Version 200 Reference
                ref = self.removeTag("FITID", line)
                if len(ref) > 9:
                    ref = ref[-9:]
                self.refno = CCD(ref, "Na", 9)
            elif line.count("<MEMO>"):
                # Version 102 Description
                self.memo = self.removeTag("MEMO", line, compress=True)
            elif line.count("<NAME>"):
                # Version 200 Description
                self.memo = self.removeTag("NAME", line, compress=True)
            elif line.count("<TRNAMT>"):
                # Transaction Amount
                self.amount = CCD(self.removeTag("TRNAMT", line), "SD", 13.2)
        if self.date.work:
            self.addTrans()
        self.trans.sort()

    def removeTag(self, tag, data, compress=False):
        data = data.replace("<%s>" % tag, "").replace("</%s>" % tag, "")
        if compress:
            newdata = data.split()
            data = ""
            for d in newdata:
                if not data:
                    data = d
                else:
                    data = "%s %s" % (data, d)
        return data

    def addTrans(self):
        if self.payee and not self.memo:
            self.memo = self.payee
        if self.memo and not self.payee:
            self.payee = self.memo
        self.trans.append([self.date.work, self.memo, self.refno.work,
            self.payee, self.amount.work])

    def importQifFile(self, fname):
        infle = open(fname, "r")
        data = infle.readlines()
        infle.close()
        self.doClear()
        for line in data:
            line = line.rstrip()
            if line[:1] == "^":
                self.trans.append([self.date.work, self.memo,
                    self.refno.work, self.payee, self.amount.work])
                self.doClear()
                continue
            if line[:1] == "!" and line[1:10] != "Type:Bank":
                showError(self.mf.body, "Import Error",
                    "Invalid Type %s" % line)
                self.trans = []
                return
            if line[:1] == "D":
                if self.getDate(line[1:]):
                    return
            elif line[:1] == "M":
                memo = line[1:].split()
                self.memo = ""
                for m in memo:
                    if not self.memo:
                        self.memo = m.strip()
                    else:
                        self.memo = "%s %s" % (self.memo, m.strip())
            elif line[:1] == "N":
                self.refno = CCD(line[1:], "Na", 9)
            elif line[:1] == "P":
                self.payee = line[1:]
            elif line[:1] == "T":
                self.amount = CCD(line[1:].replace(",", ""), "SD", 13.2)
        self.trans.sort()

    def doClear(self):
        self.date = CCD(0, "d1", 10)
        self.memo = ""
        self.refno = CCD("", "Na", 9)
        self.payee = ""
        self.amount = CCD(0, "SD", 13.2)

    def getDate(self, dt):
        if dt.count("-"):
            dt = dt.split("-")
        elif dt.count("/"):
            dt = dt.split("/")
        error = False
        if type(dt) == list:
            if self.dtefmt == "A":                          # CCYYMMDD
                date = (int(dt[0])*10000)+(int(dt[1])*100)+int(dt[2])
            elif self.dtefmt == "B":                        # DDMMCCYY
                date = (int(dt[2])*10000)+(int(dt[1])*100)+int(dt[0])
            elif self.dtefmt == "C" and len(dt[2]) == 4:    # MMDDCCYY
                date = (int(dt[2])*10000)+(int(dt[0])*100)+int(dt[1])
            elif self.dtefmt == "C" and len(dt[2]) == 2:    # MMDDYY
                date = ((int(dt[2])+2000)*10000)+(int(dt[0])*100)+int(dt[1])
            else:
                error = True
        else:
            if self.dtefmt == "A" and len(dt) == 8:         # CCYYMMDD
                date = (int(dt[:4])*10000)+(int(dt[4:6])*100)+int(dt[6:])
            elif self.dtefmt == "B" and len(dt) == 8:       # DDMMCCYY
                date = (int(dt[4:])*10000)+(int(dt[2:4])*100)+int(dt[:2])
            elif self.dtefmt == "C" and len(dt) == 8:       # MMDDCCYY
                date = (int(dt[4:])*10000)+(int(dt[:2])*100)+int(dt[2:4])
            elif self.dtefmt == "C" and len(dt) == 6:       # MMDDYY
                date = ((int(dt[4:])+2000)*10000)+(int(dt[:2])*100)+int(dt[2:4])
            else:
                error = True
        if not error:
            self.date = CCD(date, "D1", 10)
        if error or self.date.err:
            showError(self.mf.body, "Import Error", "Invalid Date Format")
            self.trans = []
            return "error"

    def importStdFile(self, fname):
        infle = open(fname, "r")
        data = infle.readlines()
        infle.close()
        for line in data:
            err = None
            if sys.platform == "win32":
                max = 132
            else:
                max = 133
            if len(line) != max:
                err = "Invalid Line Length (S/B %s is %s)" % (max, len(line))
            if not err:
                data = CsvCvt(line).output
                if data[2] == "BRANCH":
                    continue
                elif data[2] in ("ACC-NO", "ACC NO"):
                    if int(self.bankac) == int(data[1]):
                        continue
                    else:
                        err = "Invalid Bank Account Number"
                elif data[2] in ("OPEN  ", "CLOSE "):
                    continue
            if not err:
                date = CCD(int(data[1]), "D1", 10)
                err = date.err
            if not err:
                text = data[4] + data[5]
                text = text.split()
                memo = ""
                for m in text:
                    if not memo:
                        memo = m.strip()
                    else:
                        memo = "%s %s" % (memo, m.strip())
                try:
                    ref = int(data[2])
                except:
                    ref = ""
                refno = CCD(ref, "Na", 9)
                err = refno.err
            if not err:
                payee = data[5].strip()
                amount = CCD(data[3], "SD", 17.2)
                err = amount.err
            if err:
                showError(self.mf.body, "Import Error", err)
                self.trans = []
                return
            self.trans.append([date.work, memo, refno.work, payee, amount.work])

class RepPrt:
    """
    This is a Report Printer, following are the required parameters:

    name    :   The report or program name
    tables  :   A list of lists of tables or select statements with as clause
                e.g. [["ombank", "N"], ["omlegr", "I"] or
                     [select sum(glt_tramt) from gentrn as value"]
                or a list of lists of data. Type option below must be 'D'
    joins   :   A list of tables to join as follows:
                    Type   : I = Inner join
                             L = Left Outer join
                             R = Right Outer join
                             F = Full Outer join
                    Table  : Table to join
                    Conditions: String statement of columns to use for the join
                e.g. ["L", "gentrn", "glt_cono=glm_cono and glt_acno=glm_acno"]
    heads   :   A list of up to 3 Headings, if heading is a list or tuple then
                it comprises text and a font size but only for heads[0] if no
                page numbering.
                Examples:
                    ["Old Mutual Bank Account",]
                    [("Old Mutual Bank Account", 12),]
    cols    :   A list of lists of columns to be printed, any expression MUST
                be renamed using the "as" clause, as follows:
                    name The column or expression name
                    type The data type as per CCD
                    size The data display size as per CCD
                    head The field heading
                    disp The disp field, either y or n indicating whether or
                         not to display the column or just use it for selection
                         purposes. It defaults to y.
                    chgs Whether or not the sign must be changed.
                    just Whether to justify left or right, defaults to left
                Use a blank list do print all columns of a table.
                e.g. [["omb_state",             "NA",  5,   "State", "y"],
                      ["omb_date",              "D1", 10,   "Date",  "y"],
                      ["sum(omb_amt) as value", "SD", 13.2, "Value", "y"]]
                                        OR
                A single list of columns to be printed where the column data
                is extracted from the ffield table.
                                        OR
                A combination of the two options.
    stots   :   A list of lists of columns, in least significant order, to be
                breaks for sub-totalling as follows:
                [] = No sub-totals
                    column or "as" name
                    description
                    flag denoting new page after each sub-total (Y/N)
                e.g. [["omb_state","Total Statement Value","Y"],]
                Note: stots depends on gtots, without gtots, stots are ignored
    gtots   :   A list of columns to be totalled containing:
                        column or "as" name
                Use a blank list for no totals.
                e.g. ["value",]
    where   :   A list of tuples of where statements e.g.
                [("omb_recon", "=", "Y"), ("omb_recon", "=", "Y")]
                None = No where statement
    group   :   A string for the "group" statement
                e.g. "omb_date, omb_state, omb_recon"
                None = No group statement
    order   :   A string of the "order by" statement e.g.
                "omb_state asc, omb_date desc"
                None = No order statement
    pdffl   :   The name to override the default name of the pdf file
    opts    :   A string indicating the options selected e.g. "Sort-D Csv-Y"
                None = No Options
    conum   :   An integer for the Company Number e.g. 001
                None = No Company
    conam   :   A string indicating the Company Name e.g. "Tartan Systems"
                None = No Company Name
    ttype   :   A string indicating whether the tables variable contains Table
                    Names or existing Data
                "T" = Containing Table Names
                "D" = Containing Data
    prtdia  :   Whether to have a printer and mail dialog, default is False
    view    :   A string indicating whether to display the report after
                    generating it.  "V" = View, "P" = Print or "C/X" = Export
    prtnam  :   The printers name if not viewing
    mail    :   A list or string indicating who to email the report to after
                    generating it.
    lines   :   A number of body lines to Print Between Page Breaks
    margin  :   A number of lines to use as a Bottom Margin for Page Breaks
    pbar    :   Progress Bar Message:
                "P" = Generating a Report
                "F" = Updating Files
                or Any Other String
                or None for No Progressbar
    refmt   :   Reformat the data as per the column information, default is True
    blank   :   Blank zero values (only if refmt = True)
    pages   :   Whether to number pages, defaults to True
    tails   :   List of lines to print at the end of the report
    """
    def __init__(self, mf, **args):
        self.mf = mf
        defaults = {
            "name": "report",
            "tables": [],
            "joins": [],
            "heads": [],
            "tails": [],
            "cols": [],
            "stots": [],
            "gtots": [],
            "where": [],
            "group": None,
            "order": None,
            "pdffl": None,
            "opts": None,
            "conum": None,
            "conam": None,
            "ttype": "T",
            "prtdia": False,
            "view": "V",
            "prtnam": None,
            "mail": None,
            "lines": None,
            "blank": False,
            "margin": 5,
            "pbar": "P",
            "refmt": True,
            "pages": True,
            "trtp": None,
            "wrkdir": self.mf.rcdic["wrkdir"]}
        for nam in args:
            defaults[nam] = args[nam]
        for nam in defaults:
            exec "self.%s = defaults[nam]" % nam
        if self.heads:
            heads = []
            self.fonts = []
            for n, h in enumerate(self.heads):
                if type(h) in (tuple, list):
                    heads.append(h[0])
                    self.fonts.append(h[1])
                else:
                    heads.append(h)
                    self.fonts.append(None)
            self.heads = heads
            if self.prtdia:
                self.doPrtDialog()
            else:
                self.exit = False
            if not self.exit:
                self.doVariables()
                self.processReport()

    def doPrtDialog(self):
        tit = ("Printer/Email Selection",)
        self.pr = TartanDialog(self.mf, screen=self.mf.body, tops=True,
            title=tit, eflds=[], tend=((self.doPrtEnd,"y"),),
            txit=(self.doPrtExit,), view=self.prtdia[0], mail=self.prtdia[1])
        self.mf.startLoop()

    def doPrtEnd(self):
        self.exit = False
        self.pr.closeProcess()
        self.view = self.pr.repprt[0][1]
        self.prtnam = self.pr.repprt[0][2]
        self.mail = self.pr.repprt[1][2]
        self.mf.closeLoop()

    def doPrtExit(self):
        self.exit = True
        self.pr.closeProcess()
        self.mf.closeLoop()

    def doVariables(self):
        try:
            user = getpass.getuser()
        except:
            user = "unknown"
        t = time.localtime()
        self.sysdttm = "(Printed on: %i/%02i/%02i at %02i:%02i by %s)" % (t[0],
            t[1], t[2], t[3], t[4], user)
        if self.gtots == []:
            self.stots = []
        if self.ttype == "T":
            alltab = self.tables[:]
            if self.joins:
                for join in self.joins:
                    alltab.append(join[1])
            if self.cols == []:
                for tab in alltab:
                    tb = self.removeTables(tab)
                    sql = Sql(self.mf.dbm, tb, prog=__name__)
                    exec "dic = sql.%s_dic" % tb
                    exec "col = sql.%s_col" % tb
                    for c in col:
                        self.cols.append([c, dic[c][2], dic[c][3], dic[c][5]])
            else:
                dic = {}
                for tab in alltab:
                    tb = self.removeTables(tab)
                    sql = Sql(self.mf.dbm, tb, prog=__name__)
                    exec "dic = dict(dic, **sql.%s_dic)" % tb
                for n, c in enumerate(self.cols):
                    if type(c) == str:
                        self.cols[n] = [c, dic[c][2], dic[c][3], dic[c][5]]
        else:
            self.joins = []

    def processReport(self):
        self.setConstants()
        self.setTableFields()
        if self.ttype == "T":
            all = self.getData()
        else:
            all = self.tables
        if all:
            self.formLayout()
            self.printReport(all)

    def setConstants(self):
        self.tot = "s"
        self.subsum = []
        self.gndsum = []

    def setTableFields(self):
        self.fd = {}
        self.l1 = ""                # string of all columns
        self.l3 = []                # list of all columns
        self.o1 = ""                # string of all order columns
        self.pr = []                # list of all display columns
        self.cg = []                # list of all sign change columns
        self.gp = []                # list of all group columns
        func = 0
        for num, col in enumerate(self.cols):
            agg = chkAggregate(col[0])
            if agg and col[1][1:] in ("D", "d"):
                if col[0].count("Sum(") and not col[0].count("Round("):
                    col[0] = col[0].replace("Sum(", "Round(Sum(")
                elif col[0].count("sum(") and not col[0].count("round("):
                    col[0] = col[0].replace("sum(", "round(sum(")
                b = col[0].find(")") + 1
                d = int(round(col[2] * 10) % 10)
                col[0] = col[0][:b] + ",%s)" % d + col[0][b:]
            else:
                d = 0
            if num:
                self.l1 = self.l1 + "," + col[0]
            else:
                self.l1 = self.l1 + col[0]
            nam = removeFunctions(col[0], d)
            c = nam.split(" as ")
            nam = c[len(c) - 1]
            if len(c) == 1:
                func += 1
                if func != 1:
                    self.o1 = self.o1 + "," + nam
                else:
                    self.o1 = self.o1 + nam
            self.l3.append(nam)
            self.fd[nam] = [col[1], col[2], col[3]]
            if len(col) == 4:
                self.pr.append(nam)
            if len(col) > 4 and col[4].lower() == "y":
                self.pr.append(nam)
            if len(col) > 5 and col[5].lower() == "y":
                self.cg.append(nam)
            if not agg:
                self.gp.append(nam)

    def formLayout(self):
        fn = self.pr
        cf = ""
        hf = ""
        tf = ""
        hd =  []
        sd =  []
        ds = 0
        self.end = "n"
        self.expcolsh = [[]]
        self.expforms = []
        for x in range(0, len(fn)):
            nam = fn[x]
            if self.trtp and nam == self.trtp[0]:
                fsiz = 3
                d = "Typ"
                cf = cf + '"%s",'
                hf = hf + "%-3s "
                hd.append(d)
                self.expcolsh[0].append(("Typ", ))
                self.expforms.append(("NA", fsiz))
            else:
                fsiz = int(self.fd[nam][1])
                t = self.fd[nam][0][1]
                if t == "A":
                    cf = cf + '"%s"'
                else:
                    cf = cf + "%s"
                hf = hf + "%-" + "%d" % fsiz + "s"
                if x != len(fn) - 1:
                    cf = cf + ","
                    hf = hf + " "
                else:
                    cf = cf + "\n"
                d = str(self.fd[nam][2])
                if not d:
                    n = fn[x].split("_", 1)
                    d = n[len(n)-1]
                d = d[:fsiz]
                if self.fd[nam][0] == "UI" or self.fd[nam][0] == "UD":
                    d = d.rjust(fsiz)
                elif self.fd[nam][0] == "SI" or self.fd[nam][0] == "SD":
                    d = d.rjust(fsiz - 1)
                hd.append(d)
                self.expcolsh[0].append((d,))
                self.expforms.append((self.fd[nam][0], self.fd[nam][1]))
            if self.gtots:
                cnt = 0
                for c in range(0, len(self.gtots)):
                    if nam == self.gtots[c]:
                        sd.append(d)
                        if ds == 0:
                            if x == 0:
                                # First field is a Total therefore the
                                # Description 'Grand Total' must be
                                # appended after the last field.
                                self.end = "y"
                        else:
                            tf = tf + "%-" + "%d" % (ds-1) + "s "
                            ds = 0
                        tf = tf + "%-" + "%d" % fsiz + "s"
                        if c != len(self.gtots) - 1:
                            tf = tf + " "
                        cnt += 1
                if not cnt:
                    sd.append("")
                    ds = ds + fsiz + 1
        self.h5hf = hf
        self.head4 = hf % tuple(hd)
        if sd:
            self.head5 = hf % tuple(sd)
        else:
            self.head5 = self.head4
        self.tf = tf
        self.cf = cf
        self.max = len(self.head4)
        plus = len(self.sysdttm) + 47
        if self.max <= plus:
            self.max = plus + 1
            self.head4 = self.head4 + (" " * (self.max - len(self.head4)))
        fill = int(self.max - plus)
        if len(self.heads) == 4:
            self.head0 = self.heads.pop(0)
        elif not self.conum:
            hf = "%-34s %"+str(fill)+"s %"+str(len(self.sysdttm))+"s %10s"
            self.head0 = hf % \
                ("TARTAN Financial Systems", "", self.sysdttm, self.name)
        else:
            hf = "%03s %-30s %"+str(fill)+"s %"+str(len(self.sysdttm))+"s %10s"
            self.head0 = hf % \
                (self.conum, self.conam, "", self.sysdttm, self.name)
        fill = int(self.max - len(self.heads[0]) - 12)
        if not self.pages:
            self.head1 = "%-s" % self.heads[0]
        else:
            hf = "%-"+str(len(self.heads[0]))+"s %"+str(fill)+"s"
            self.head1 = hf % (self.heads[0], "") + " Page %5s"
        self.head2 = ""
        self.head3 = ""
        if len(self.heads) > 1:
            self.head2 = (self.heads[1])
        if len(self.heads) == 3:
            self.head3 = (self.heads[2])
        self.expheads = []
        for x in range(4):
            if x == 1:
                h = self.heads[0]
            else:
                exec "h = self.head%s" % x
            if not h:
                continue
            if self.expheads:
                self.expheads.append("")
            self.expheads.append(h)
        if self.opts:
            self.expheads.append("")
            self.expheads.append("Options: %s" % self.opts)
        self.expheads.append("")

    def getData(self):
        # Check table relationships
        rels = []
        for t1 in self.tables:
            for t2 in self.tables:
                if t2 == t1:
                    continue
                all = self.mf.sqlRec("Select rel_col1, rel_col2 from "\
                    "frelat where rel_tab1 = '%s' and rel_tab2 = '%s'" % \
                    (t1, t2))
                if not all:
                    all = self.mf.sqlRec("Select rel_col1, rel_col2 from "\
                        "frelat where rel_tab1 = '%s' and rel_tab2 = '%s'" % \
                        (t2, t1))
                if all:
                    for rel in all:
                        if rels.count(rel) == 0:
                            rels.append(rel)
        if len(self.tables) > 1 and not rels:
            showError(self.mf.body, "Relationship Error",
                "Some Relationships Between Tables %s Missing" % self.tables)
            return "error"
        for rel in rels:
            self.where.append(("%s=%s" % (rel[0], rel[1]),))
        # Create cols
        cols = []
        for col in self.cols:
            cols.append(col[0])
        # Create join statements
        join = ""
        for jon in self.joins:
            if jon[0] == "I":
                join = "%s inner join %s on %s," % (join, jon[1], jon[2])
            elif jon[0] == "L":
                join = "%s left outer join %s on %s," % (join, jon[1], jon[2])
            elif jon[0] == "R":
                join = "%s right outer join %s on %s," % (join, jon[1], jon[2])
            elif jon[0] == "F":
                join = "%s full outer join %s on %s," % (join, jon[1], jon[2])
        if not self.group and self.stots and self.o1:
            self.group = self.o1
        if self.group:
            for col in self.gp:
                if not self.group.count(col):
                    self.group = "%s, %s" % (self.group, col)
        if not self.order and self.o1:
            order = self.o1
        else:
            order = self.order
        sql = Sql(self.mf.dbm, self.tables, prog=__name__)
        return sql.getRec(tables=self.tables, cols=cols, join=join,
            where=self.where, group=self.group, order=order)

    def printReport(self, all):
        if self.mf.display and self.pbar:
            pb = ProgressBar(self.mf.body, max=len(all), typ=self.pbar,
                esc=True)
        if self.pdffl:
            self.pdfnam = os.path.join(self.wrkdir, self.pdffl + ".pdf")
            self.expnam = os.path.join(self.wrkdir, self.pdffl)
        else:
            if not self.conum:
                coy = 0
            else:
                coy = self.conum
            self.pdfnam = getModName(self.wrkdir, self.name, coy, ext="pdf")
            self.expnam = self.pdfnam.replace(".pdf", "")
        self.c, self.ogn, self.cpi, self.ldg, self.lpp, self.suc, self.duc = \
            doReportlab(self.mf.prgdir, self.pdfnam, self.head4)
        for num, font in enumerate(self.fonts):
            if not font:
                self.fonts[num] = self.cpi
        self.uline = ("%s" % (self.suc * self.max))
        self.dline = ("%s" % (self.duc * self.max))
        if self.lines:
            lin = 4                        # head0 and head4
            for h in self.heads:
                if h:
                    lin = lin + 2
            self.lpp = self.lines + lin    # lin = Headings
            if self.margin:
                self.lpp = self.lpp + self.margin
        if self.margin:
            self.lpp = self.lpp - self.margin
        for sub in self.stots:
            exec "self.%s_str = ''" % sub[0]
        for tot in self.gtots:
            exec "self.%s_gtot = 0" % tot
            idx = 0
            for sub in self.stots:
                exec "self.%s_%s_stot = 0" % (tot, idx)
                idx += 1
        self.pgnum = 0
        self.pglin = 99
        self.expdatas = []
        for x in range(0, len(all)):
            if self.mf.display and self.pbar:
                can = pb.displayProgress(x)
                if can == "yes":
                    break
            else:
                can = "no"
            if self.ttype == "D" and all[x] == ["BLANK"]:
                if self.prtnam != "export":
                    self.t.textLine()
                    self.pglin += 1
                else:
                    self.expdatas.append(all[x])
                continue
            fn = self.l3
            num = 0
            fld = []
            cld = []
            for nam in fn:
                t = self.fd[nam][0]
                if t == "TX" and self.prtnam != "export":
                    t = "NA"
                s = self.fd[nam][1]
                if not all[x][num]:
                    d = ""
                else:
                    d = all[x][num]
                if t == "HA":
                    d = base64.decodestring(all[x][num])
                if self.refmt:
                    if self.trtp and nam == self.trtp[0]:
                        exec "vars = %s" % self.trtp[1]
                        d = vars[int(d) - 1][0]
                        exec "%s = CCD(d, '%s', %s)" % (nam, "NA", 3)
                    else:
                        if nam in self.cg:
                            if not d:
                                d = 0
                            elif type(d) == int:
                                d = 0 - d
                            else:
                                d = float(ASD(0) - ASD(d))
                        if self.blank and t[1] in ("D", "I") and not d:
                            exec "%s = CCD('', 'NA', %s)" % (nam, s)
                        else:
                            exec "%s = CCD(d, '%s', %s)" % (nam, t, s)
                    if nam in self.pr:
                        exec "fld.append(%s.disp)" % nam
                        exec "cld.append(%s.work)" % nam
                elif nam in self.pr:
                    fld.append(d)
                    cld.append(d)
                num += 1
            st = "n"
            for sq1 in range((len(self.stots) - 1), -1, -1):
                exec "w = self.%s_str" % self.stots[sq1][0]
                exec "f = %s.work" % self.stots[sq1][0]
                if not w:
                    w = f
                if f != w and st == "n":
                    st = "y"
                    for sq2 in range(0, sq1+1):
                        self.subTotal(self.stots[sq2])
                exec "self.%s_str = f" % self.stots[sq1][0]
            if self.prtnam != "export" and self.pglin >= self.lpp:
                if self.pgnum:
                    self.c.drawText(self.t)
                    self.c.showPage()
                self.pageHeading()
            for gt in self.gtots:
                idx = 0
                if self.fd[gt][0][1] == "D":
                    # Decimal Amount
                    exec "self.%s_gtot = float(ASD(self.%s_gtot) + "\
                        "ASD(%s.work))" % (gt, gt, gt)
                    for s in self.stots:
                        exec "self.%s_%s_stot = float(ASD(self.%s_%s_stot) + "\
                            "ASD(%s.work))" % (gt, idx, gt, idx, gt)
                        idx += 1
                else:
                    # Integer Amount
                    exec "self.%s_gtot = self.%s_gtot + "\
                        "%s.work" % (gt, gt, gt)
                    for s in self.stots:
                        exec "self.%s_%s_stot = self.%s_%s_stot + "\
                            "%s.work" % (gt, idx, gt, idx, gt)
                        idx += 1
            if self.prtnam != "export":
                self.t.textLine(self.h5hf % tuple(fld))
                self.pglin += 1
            else:
                self.expdatas.append(["BODY", cld])
        if self.mf.display and self.pbar:
            pb.closeProgress()
        if can == "yes":
            return
        self.grandTotal()
        try:
            if self.prtnam == "export":
                err = doWriteExport(rcdic=self.mf.rcdic, xtype=self.view,
                    name=self.expnam, sheet=self.name, heads=self.expheads,
                    colsh=self.expcolsh, forms=self.expforms,
                    datas=self.expdatas)
                if err:
                    showError(self.mf.body, "Export Error", err)
            else:
                if self.tails:
                    self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
                    for tail in self.tails:
                        self.t.textLine(tail)
                self.c.drawText(self.t)
                self.c.showPage()
                self.c.save()
                doPrinter(mf=self.mf, conum=self.conum, rep=self.pdfnam,
                    head=self.heads[-1], view=self.view, prtnam=self.prtnam,
                    mail=self.mail)
        except:
            pass

    def pageHeading(self, htype=None):
        if self.prtnam == "export":
            return
        self.t = self.c.beginText(self.ogn[0], self.ogn[1])
        self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
        self.pgnum += 1
        self.t.textLine(self.head0)
        self.t.textLine()
        if self.pages:
            self.t.setFont("Tartan-Bold", self.fonts[0], self.ldg)
            self.t.textLine(self.head1 % self.pgnum)
        else:
            self.t.setFont("Tartan-Bold", self.fonts[0], self.ldg)
            self.t.textLine(self.head1)
        self.pglin = 3
        if self.head2:
            self.t.textLine()
            self.t.setFont("Tartan-Bold", self.fonts[1], self.ldg)
            self.t.textLine(self.head2)
            self.pglin += 2
        if self.head3:
            self.t.textLine()
            self.t.setFont("Tartan-Bold", self.fonts[2], self.ldg)
            self.t.textLine(self.head3)
            self.pglin += 2
        if self.opts:
            self.t.textLine()
            self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
            self.t.textLine("%-10s%s" % ("Options:", self.opts))
            self.pglin += 2
        self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
        self.t.textLine()
        self.pglin += 1
        if not htype:
            self.t.textLine(self.head4)
        elif self.head5[:20] == " " * 20:
            head = "%s%s" % ("Totals Summary", self.head5[14:])
            self.t.textLine(head)
        else:
            self.t.textLine("Totals Summary")
            self.pglin += 1
            self.t.textLine(self.head5)
        self.pglin += 1
        self.t.textLine(self.uline)
        self.pglin += 1
        self.t.setFont("Tartan", self.cpi, self.ldg)

    def subTotal(self, col):
        if self.prtnam == "export":
            self.expdatas.append(["ULINES"])
        else:
            self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
            self.t.textLine(self.uline)
            self.t.setFont("Tartan", self.cpi, self.ldg)
            self.pglin += 1
        snum = self.stots.index(col)
        exec "sdet = self.%s_str" % col[0]
        self.printTotals("s", snum, sdet)
        if self.prtnam != "export" and col[2] == "Y":
            self.pglin = 66
        elif self.tot != "g":
            if self.prtnam == "export":
                self.expdatas.append(["BLANK", [""]])
            else:
                self.t.textLine()
                self.pglin += 1

    def grandTotal(self):
        self.tot = "g"
        for stot in self.stots:
            self.subTotal(stot)
        if self.gtots:
            if self.prtnam == "export":
                self.expdatas.append(["ULINES"])
            else:
                self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
                self.t.textLine(self.uline)
            self.printTotals("g")
            if self.prtnam == "export":
                self.expdatas.append(["ULINED"])
            else:
                self.t.textLine(self.dline)
                self.t.setFont("Tartan", self.cpi, self.ldg)
                self.pglin = self.pglin + 2
        if self.prtnam != "export" and self.stots:
            self.c.drawText(self.t)
            self.c.showPage()
            self.printSummary()

    def printTotals(self, typ, snum=None, sdet=None):
        if typ == "h":
            clr = flg = "n"
            if self.stots:
                typ = "s"
            else:
                typ = "g"
        else:
            clr = flg = "y"
        fn = self.l3
        fld = []
        cld = []
        seq = 0
        for tc in range(0, len(fn)):
            cldf = False
            n = fn[tc]
            for c in range(0, len(self.gtots)):
                if n == self.gtots[c]:
                    cldf = True
                    if typ == "s":
                        exec "j = CCD(self.%s_%s_stot, '%s', %s)" % \
                            (n, snum, self.fd[n][0], float(self.fd[n][1]))
                    else:
                        exec "j = CCD(self.%s_gtot, '%s', %s)" % \
                            (n, self.fd[n][0], float(self.fd[n][1]))
                    if not j.err:
                        fld.append(j.disp)
                        cld.append(j.work)
                        if j.work != 0:
                            flg = "y"
                    else:
                        fld.append("0.00")
                        cld.append(0.00)
                    if clr == "y":
                        if typ == "s":
                            exec "self.%s_%s_stot = 0" % (n, snum)
                        else:
                            exec "self.%s_gtot = 0" % n
                elif seq:
                    fld.append("")
            if not cldf:
                cld.append("")
        if self.tf and flg == "y":
            # Only print Totals if there are some Non Zero Totals
            if self.end == "y":
                if typ == "s":
                    dsc = self.stots[snum][1] + " %s" % sdet
                else:
                    dsc = "Grand Total"
                siz = len(dsc)
                self.tf = self.tf + " %-" + "%d" % siz + "s"
                fld.append(dsc)
                cld.append(dsc)
            elif typ == "s":
                fld.insert(0, self.stots[snum][1] + " %s" % sdet)
                cld[0] = self.stots[snum][1] + " %s" % sdet
            else:
                fld.insert(0, "Grand Total")
                cld[0] = "Grand Total"
            if typ == "s":
                self.subsum.append(fld)
            else:
                self.gndsum.append(fld)
            if self.prtnam == "export":
                self.expdatas.append(["TOTAL", cld])
            else:
                self.t.textLine(self.tf % tuple(fld))
                self.pglin += 1

    def printSummary(self):
        fn = self.l3
        fld = []
        seq = 0
        if not self.subsum:
            return
        self.pageHeading("summary")
        for summ in self.subsum:
            self.t.textLine(self.tf % tuple(summ))
            self.pglin += 1
            if self.pglin >= self.lpp:
                self.c.drawText(self.t)
                self.c.showPage()
                self.pageHeading("summary")
        self.t.setFont("Tartan-Bold", self.cpi, self.ldg)
        self.t.textLine(self.uline)
        self.t.textLine(self.tf % tuple(self.gndsum[0]))
        self.t.textLine(self.dline)
        self.t.setFont("Tartan", self.cpi, self.ldg)

    def removeTables(self, tab):
        n = tab.split(" as ")
        if len(n) == 1:
            n = n[0]
        else:
            n = n[1]
        return(n)

class TabPrt:
    """
    This is a front-end to the RepPrt class for printing table details
        mf     = The mainframe class
        conum  = The company number
        conam  = The company name
        name   = The calling program
        head   = The report heading
        tabs   = The table name or list of tables which must also have cols
        cols   = The columns to print else None for all
        wher   = The where statement
        keys   = The Balances class key
        pdia   = The Print Dialog, defaults to True
        prtnam = The Printer Name
        mail   = The Email Address
        export = Return the details
    """
    def __init__(self, mf, conum=None, conam=None, name=None, head=None, tabs=None, cols=None, wher=None, keys=None, pdia=True, prtnam=None, mail=None, export=False):
        self.mf = mf
        self.conum = conum
        self.conam = conam
        self.name = name
        self.head = head
        if type(tabs) == str:
            self.tabs = [tabs]
        else:
            self.tabs = tabs
        self.cols = cols
        self.wher = wher
        self.keys = keys
        self.pdia = pdia
        self.prtnam = prtnam
        self.mail = mail
        self.export = export
        no = self.setVariables()
        if not no:
            self.printReport()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, self.tabs + ["ftable"], prog=__name__)
        self.dics = {}
        for tab in self.tabs:
            exec "dic = self.sql.%s_dic" % tab
            for key in dic:
                self.dics[key] = dic[key]
        if not self.cols:
            self.cols = []
            for tab in self.tabs:
                exec "col = self.sql.%s_col" % tab
                self.cols.extend(col)
        t = time.localtime()
        self.sysdt = ((t[0] * 10000) + (t[1] * 100) + t[2])
        self.curdt = self.sysdt / 100
        if not self.head:
            self.head = self.sql.getRec(tables="ftable", cols=["ft_desc"],
                where=[("ft_tabl", "=", self.tabs[0])], limit=1)[0]

    def printReport(self):
        all = self.sql.getRec(tables=self.tabs, cols=self.cols,
            where=self.wher, limit=1)
        if not all:
            return
        max1, max2 = 0, 0
        fld = []
        for c in self.cols:
            if max1 < len(self.dics[c][4]):
                max1 = len(self.dics[c][4])
            if max2 < int(self.dics[c][3]):
                max2 = int(self.dics[c][3])
            fld.append([self.dics[c][4], CCD(all[self.cols.index(c)],
                self.dics[c][2], self.dics[c][3]).disp])
        if self.tabs.count("drsmst") and self.keys:
            bals = Balances(self.mf, "DRS", self.conum, self.curdt, self.keys)
            more = bals.doAllBals()
        elif self.tabs.count("crsmst") and self.keys:
            bals = Balances(self.mf, "CRS", self.conum, self.curdt, self.keys)
            more = bals.doAllBals()
        elif self.tabs[0] == "strmf1" and self.keys:
            bals = Balances(self.mf, "STR", self.conum, self.curdt, self.keys)
            more = bals.doStrBals()
        else:
            more = None
        if more and (self.tabs.count("drsmst") or self.tabs.count("crsmst")):
            fld.append(["", ""])
            fld.append(["Current",  CCD(more[2][0], "SD", 13.2).disp])
            fld.append([" 30-Days", CCD(more[2][1], "SD", 13.2).disp])
            fld.append([" 60-Days", CCD(more[2][2], "SD", 13.2).disp])
            fld.append([" 90-Days", CCD(more[2][3], "SD", 13.2).disp])
            fld.append(["120-Days", CCD(more[2][4], "SD", 13.2).disp])
            fld.append(["", ""])
            fld.append(["Balance", CCD(more[1], "SD", 13.2).disp])
        elif more and self.tabs[0] == "strmf1":
            fld.append(["", ""])
            fld.append(["Last    Cost    ", CCD(more[6],    "SD", 13.2).disp])
            fld.append(["Average Cost    ", CCD(more[7],    "SD", 13.2).disp])
            fld.append(["", ""])
            fld.append(["Quantity Balance", CCD(more[5][0], "SD", 13.2).disp])
            fld.append(["Value    Balance", CCD(more[5][1], "SD", 13.2).disp])
        cols = [
            ["a", "NA", (max1+5), "Field Name"],
            ["b", "NA",     max2,    "Details"]]
        if not self.export and (self.pdia or self.prtnam or self.mail):
            if not self.pdia or type(self.pdia) in (list, tuple):
                prtdia = self.pdia
            else:
                prtdia = (("Y","V"),("Y","N"))
            rp = RepPrt(self.mf, name=self.name, tables=fld, heads=[self.head],
                cols=cols, conum=self.conum, conam=self.conam, ttype="D",
                prtdia=prtdia, prtnam=self.prtnam, mail=self.mail, refmt=False)
            self.view = rp.view
            self.mail = rp.mail
            self.exit = rp.exit
        else:
            self.view = None
            self.mail = None
            self.exit = False
            self.mess = ""
            for f in fld:
                if not self.mess:
                    self.mess = "%-20s %30s" % (f[0], f[1])
                else:
                    self.mess = "%s\n%-20s %30s" % (self.mess, f[0], f[1])
            if not self.export:
                showInfo(self.mf.body, self.head, self.mess)

class TartanUser:
    """
    This class is used to add, delete and amend Tartan users
    """
    def __init__(self, mf, sc, sys, mod):
        self.mf = mf
        self.scrn = sc
        self.systems = sys
        self.mod = mod
        self.doVariables()
        self.doUser()
        self.mf.startLoop()

    def doVariables(self):
        self.sql = Sql(self.mf.dbm, ["ctlsys", "ctlmst", "ctlpwu", "ctlpwm"],
            prog=__name__)
        ctlsys = self.sql.getRec(tables="ctlsys", limit=1)
        self.pwmust, self.pwsize, self.pwlife = ctlsys[:3]

    def doUser(self):
        susr = {
            "stype": "R",
            "tables": ("ctlpwu",),
            "cols": (
                ("usr_name", "", 0, "User-Name"),
                ("usr_fnam", "", 0, "Full-Name", "Y"),
                ("usr_coy", "", 0, "Coy"),
                ("usr_lvl", "", 0, "L"))}
        scoy = {
            "stype": "R",
            "tables": ("ctlmst",),
            "cols": (
                ("ctm_cono", "", 0, "Coy"),
                ("ctm_name", "", 0, "Name", "Y"))}
        lvls = [
            [0, "= Enquiries Only"],
            [1, "= 0 plus Reporting"],
            [2, "= 1 plus Data Capture"],
            [3, "= 2 plus File Maintenance"],
            [4, "= 3 plus Month and Year End Routines"],
            [5, "= 4 plus Control Routines"],
            [6, "= 5 plus Financial Year End Routine"],
            [7, "= 6 plus User and Module Password Maintenance"],
            [9, "= Supervisor level, Everything"]]
        slvl = {
            "stype": "C",
            "titl": "Available Levels",
            "data": lvls,
            "index": 0}
        ssys = {
            "stype": "C",
            "titl": "Available Systems",
            "data": self.systems,
            "index": 1}
        sprg = {
            "stype": "C",
            "titl": "Available Modules",
            "head": ("L", "Code", "Description"),
            "data": [],
            "index": 1}
        smod = {
            "stype": "R",
            "tables": ("ctlpwm",),
            "cols": (
                ("mpw_coy", "", 0, "Coy"),
                ("mpw_sys", "", 0, "Sy"),
                ("mpw_prg", "", 0, "Prog")),
            "whera": (("T", "mpw_usr", 0),),
            "order": "mpw_coy, mpw_sys, mpw_prg",
            "comnd": self.doSelRet}
        tit = ("User Maintenance",)
        r1s = (("Yes", "Y"), ("No", "N"))
        fld = (
            (("T",0,0,0),"INA",20,"User Name","",
                "","Y",self.doUserName,susr,None,("notblank",)),
            (("T",0,1,0),"INA",30,"Full Name","",
                "","N",self.doFullName,susr,self.doDelUser,("notblank",)),
            (("T",0,2,0),"IHA",10,"User Password","",
                "","N",self.doUserPwd1,None,None,None),
            (("T",0,2,29),"IHA",10," Check ","Check Password",
                "","N",self.doUserPwd2,None,None,None),
            (("T",0,3,0),("IRB",r1s),0,"Copy Existing User","",
                "N","N",self.doCopyUser1,None,None,None),
            (("T",0,4,0),"INA",20,"Copy User Name","User Name to Copy",
                "","N",self.doCopyUser2,susr,None,None),
            (("T",0,5,0),"INA",30,"Valid Companies","",
                "All","N",self.doValCoy,None,None,None,None,
                "Comma Seperated List of Valid Company Numbers e.g. 1,2,3"),
            (("T",0,5,0),"IUI",1,"Security Level","",
                0,"N",self.doSecLvl,slvl,None,("between",0,9)),
            (("C",0,0,0),"IUI",3,"Coy","Company",
                "r","N",self.doCoy,scoy,None,None),
            (("C",0,0,1),"INA",2,"SS","System Code",
                "r","N",self.doSys,ssys,None,None),
            (("C",0,0,2),"INA",4,"Prog","Module",
                "","N",self.doMod,sprg,self.doDelSys,None),
            (("C",0,0,3),"IHA",20,"Password","",
                "","N",self.doPw1,None,self.doDelMod,None),
            (("C",0,0,4),"IHA",20,"Check Password","Re-Enter Password",
                "","N",self.doPw2,None,None,None))
        row = (15,)
        but = (
            ("Show All Modules",smod,None,0,("C",0,1),(("T",0,1),("C",0,2))),
            ("Quit",None,self.doUserExit,1,None,None))
        tnd = ((self.doWriteUser,"y"),)
        txt = (self.doUserExit,)
        cnd = ((self.doWriteModule,"y"),)
        cxt = (self.doUserNext,)
        self.df = TartanDialog(self.mf, screen=self.scrn, title=tit, eflds=fld,
            rows=row, butt=but, tend=tnd, txit=txt,
            cend=cnd, cxit=cxt)

    def doUserName(self, frt, pag, r, c, p, i, nam):
        self.unam = nam
        usr = self.sql.getRec(tables="ctlpwu", where=[("usr_name", "=", nam)],
            limit=1)
        if not usr:
            self.new_usr = "y"
            if self.unam == "admin":
                self.fnam = "System Administrator"
                self.df.loadEntry(frt, pag, p+1, data=self.fnam)
        else:
            self.new_usr = "n"
            self.fnam = usr[self.sql.ctlpwu_col.index("usr_fnam")]
            self.pwd = usr[self.sql.ctlpwu_col.index("usr_pwd")]
            self.coy = usr[self.sql.ctlpwu_col.index("usr_coy")]
            self.lvl = usr[self.sql.ctlpwu_col.index("usr_lvl")]
            self.df.loadEntry(frt, pag, p+1, data=self.fnam)
            self.df.loadEntry(frt, pag, p+2, data=base64.decodestring(self.pwd))
            self.df.loadEntry(frt, pag, p+3, data=base64.decodestring(self.pwd))
            if not self.coy:
                self.df.loadEntry(frt, pag, p+6, data="All")
            else:
                self.df.loadEntry(frt, pag, p+6, data=self.coy)
            self.df.loadEntry(frt, pag, p+7, data=self.lvl)
            self.doLoadMods(focus=False)
        if self.unam == "admin":
            return "sk1"

    def doFullName(self, frt, pag, r, c, p, i, nam):
        self.fnam = nam

    def doUserPwd1(self, frt, pag, r, c, p, i, pwd):
        if not pwd and self.pwmust == "Y":
            return "You Must Have a Password"
        if self.pwsize and len(pwd) < self.pwsize:
            return "Minimum of %s Characters Required" % self.pwsize
        self.pwd = pwd

    def doUserPwd2(self, frt, pag, r, c, p, i, pwd):
        if pwd != self.pwd:
            return "Passwords Do Not Match"
        if self.unam == "admin":
            self.cpy = "N"
            self.coy = ""
            self.lvl = 9
            return "sk4"

    def doCopyUser1(self, frt, pag, r, c, p, i, cpy):
        self.cpy = cpy
        if self.cpy == "N":
            return "sk1"

    def doCopyUser2(self, frt, pag, r, c, p, i, nam):
        usr = self.sql.getRec(tables="ctlpwu", where=[("usr_name", "=", nam)],
            limit=1)
        if not usr:
            return "Invalid User Name"
        self.cnam = nam
        self.coy = usr[self.sql.ctlpwu_col.index("usr_coy")]
        self.lvl = usr[self.sql.ctlpwu_col.index("usr_lvl")]
        if not self.coy:
            self.df.loadEntry(frt, pag, p+1, data="All")
        else:
            self.df.loadEntry(frt, pag, p+1, data=self.coy)
        self.df.loadEntry(frt, pag, p+2, data=self.lvl)

    def doValCoy(self, frt, pag, r, c, p, i, coy):
        self.coy = ""
        if coy and coy != "All":
            for c in coy.replace(" ", "").split(","):
                try:
                    chk = self.sql.getRec(tables="ctlmst", where=[("ctm_cono",
                        "=", int(c))], limit=1)
                    if not chk:
                        raise Exception
                    elif self.coy:
                        self.coy = "%s,%s" % (self.coy, c)
                    else:
                        self.coy = "%s" % c
                except:
                    return "Invalid Company Number (%s)" % c
            self.df.loadEntry(frt, pag, i, data=self.coy)
        else:
            self.df.loadEntry(frt, pag, i, data="All")

    def doSecLvl(self, frt, pag, r, c, p, i, lvl):
        self.lvl = lvl

    def doWriteUser(self):
        if self.pwd:
            pwd = base64.encodestring(self.pwd)
            t = time.localtime()
            dte = (t[0] * 10000) + (t[1] * 100) + t[2]
        else:
            pwd = ""
            dte = 0
        data = [self.unam, self.fnam, pwd, dte, self.coy, self.lvl]
        if self.new_usr == "y":
            self.sql.insRec("ctlpwu", data=data)
        else:
            self.sql.updRec("ctlpwu", data=data, where=[("usr_name",
            "=", self.unam)])
        if self.unam == "admin":
            self.mf.dbm.commitDbase()
            self.df.focusField("T", 0, 1)
        else:
            if self.cpy == "Y":
                self.sql.delRec("ctlpwm", where=[("mpw_usr", "=", self.unam)])
                all = self.sql.getRec(tables="ctlpwm", where=[("mpw_usr", "=",
                    self.cnam)], order="mpw_coy, mpw_sys, mpw_prg")
                for rec in all:
                    rec[0] = self.unam
                    self.sql.insRec("ctlpwm", data=rec)
            self.doLoadMods()

    def doDelUser(self):
        if self.unam == "admin":
            return "Not Allowed to Delete admin"
        self.sql.delRec("ctlpwu", where=[("usr_name", "=", self.unam)])
        self.sql.delRec("ctlpwm", where=[("mpw_usr", "=", self.unam)])

    def doCoy(self, frt, pag, r, c, p, i, w):
        self.con = w

    def doSys(self, frt, pag, r, c, p, i, w):
        s = False
        for sys in self.systems:
            if w == sys[1]:
                s = True
        if not s:
            return "Invalid System"
        self.sys = w
        data = []
        for mod in self.mod[self.sys]:
            if mod[0] <= self.lvl:
                data.append(mod)
        self.df.colf[pag][2][8]["data"] = data

    def doMod(self, frt, pag, r, c, p, i, prg):
        self.prg = prg
        if not self.prg:
            if self.con:
                return "Invalid Module, Company Exists"
            else:
                self.df.loadEntry(frt, pag, p+1, data="")
                self.df.loadEntry(frt, pag, p+2, data="")
                return "sk2"
        pwd = self.getPassword()
        if pwd:
            self.df.loadEntry(frt, pag, p+1, data=pwd)
            self.df.loadEntry(frt, pag, p+2, data=pwd)

    def doPw1(self, frt, pag, r, c, p, i, pwd):
        self.pwd = pwd
        if not self.pwd:
            return "sk1"

    def doPw2(self, frt, pag, r, c, p, i, pwd):
        if pwd != self.pwd:
            return "Passwords do Not Match"

    def doSelRet(self, frt, pag, r, c, p, i, sel):
        self.con = sel[0]
        self.sys = sel[1]
        self.prg = sel[2]
        self.pwd = self.getPassword()
        self.df.loadEntry(frt, pag, p, data=self.con)
        self.df.loadEntry(frt, pag, p+1, data=self.sys)
        self.df.loadEntry(frt, pag, p+2, data=self.prg)
        self.df.loadEntry(frt, pag, p+3, data=self.pwd)
        self.df.loadEntry(frt, pag, p+4, data=self.pwd)

    def getPassword(self):
        pwd = self.sql.getRec(tables="ctlpwm", cols=["mpw_pwd"],
            where=[("mpw_usr", "=", self.unam), ("mpw_coy", "=", self.con),
            ("mpw_sys", "=", self.sys), ("mpw_prg", "=", self.prg)],
            limit=1)
        if pwd:
            return base64.decodestring(pwd[0])

    def doDelSys(self):
        self.sql.delRec("ctlpwm", where=[("mpw_usr", "=", self.unam),
            ("mpw_coy", "=", self.con), ("mpw_sys", "=", self.sys)])
        self.doLoadMods()

    def doDelMod(self):
        self.sql.delRec("ctlpwm", where=[("mpw_usr", "=", self.unam),
            ("mpw_coy", "=", self.con), ("mpw_sys", "=", self.sys),
            ("mpw_prg", "=", self.prg)])
        self.doLoadMods()

    def doWriteModule(self):
        whr = [("mpw_usr", "=", self.unam), ("mpw_coy", "=", self.con),
            ("mpw_sys", "=", self.sys), ("mpw_prg", "=", self.prg)]
        usr = self.sql.getRec(tables="ctlpwm", where=whr, limit=1)
        pwd = base64.encodestring(self.pwd)
        dat = [self.unam, self.con, self.sys, self.prg, pwd]
        if not usr:
            self.sql.insRec("ctlpwm", data=dat)
        else:
            self.sql.updRec("ctlpwm", data=dat, where=whr)
        self.doLoadMods()

    def doLoadMods(self, focus=True):
        self.df.clearFrame("C", 0)
        self.pos = 0
        all = self.sql.getRec(tables="ctlpwm", where=[("mpw_usr", "=",
            self.unam)], order="mpw_coy, mpw_sys, mpw_prg")
        for row, rec in enumerate(all):
            for num, dat in enumerate(rec):
                if num in (0, 5):
                    continue
                self.df.loadEntry("C", 0, self.pos, data=dat)
                if num == 4:
                    self.df.loadEntry("C", 0, self.pos, data=dat)
                    self.pos += 1
                self.pos += 1
                if self.pos == (self.df.rows[0] * self.df.colq[0]):
                    self.df.scrollScreen(0)
                    self.pos = self.pos - self.df.colq[0]
        if focus:
            self.df.focusField("C", 0, (self.pos+1))

    def doUserNext(self):
        self.mf.dbm.commitDbase()
        self.df.focusField("T", 0, 1)

    def doUserExit(self, widget=None):
        self.mf.dbm.commitDbase()
        self.df.closeProcess()
        self.mf.closeLoop()

class Dbase:
    """
    A class to manipulate databases.

    rcdic  - A dictionary having at least:
                wrkdir:     A work directory
                dbase:      The database API e.g. PgSQL
                dbname:     The database name
                dbhost:     The database host (Not SQLite)
                dbdir:      The directory containing the database (FBird/SQLite)
                dbport:     The port to be used (Not currently used)
                dbuser:     The database user
                dbpwd:      The database user's password else None
    screen - A screen on which to display error messages else None
    """
    def __init__(self, rcdic=None, screen=None):
        if not rcdic:
            rcdic = loadRcFile()
        if not rcdic or rcdic == "error":
            self.err = True
            return
        for key in rcdic:
            exec "self.%s = rcdic[key]" % key
        self.screen = screen
        if self.dbase in ("FBird", "SQLite"):
            if self.dbdir:
                self.dbdsn = os.path.join(self.dbdir, self.dbname)
            else:
                self.dbdsn = self.dbname
        try:
            if self.dbase == "FBird":
                import kinterbasdb as engine
                self.dbf = "?"
            elif self.dbase == "MsSQL":
                import adodbapi as engine
                self.dbf = "?"
            elif self.dbase == "MySQL":
                import MySQLdb as engine
                self.dbf = "%s"
            elif self.dbase == "PgSQL":
                self.typecast = False
                try:
                    import psycopg2 as engine
                    self.typecast = True
                except:
                    try:
                        import psycopg as engine
                    except:
                        try:
                            import pgdb as engine
                        except:
                            raise Exception("PgSQL Missing Database Module")
                self.dbf = "%s"
            elif self.dbase == "SQLite":
                from pysqlite2 import dbapi2 as engine
                self.dbf = "?"
            else:
                raise Exception("Invalid Database Type (%s)" % self.dbase)
            self.err = False
            self.dbopen = False
            self.engine = engine
        except:
            showException(self.screen, self.wrkdir, "DBAPI Error")
            self.err = True

    def checkDbase(self):
        "Check to see if a database exists"
        exists = False
        if self.dbase == "FBird":
            self.openDbase(err=False)
            if self.db:
                exists = True
        elif self.dbase == "MsSQL":
            self.openDbase(dbname="master", err=False)
            if self.db:
                self.cu.execute("Select count(*) from sysdatabases "\
                    "where name='%s'" % self.dbname)
                if self.cu.fetchone()[0]:
                    exists = True
        elif self.dbase == "MySQL":
            self.openDbase(dbname="mysql", err=False)
            if self.db:
                self.cu.execute("Select count(*) from "\
                    "information_schema.tables where table_schema='%s'" % \
                    self.dbname)
                if self.cu.fetchone()[0]:
                    exists = True
        elif self.dbase == "PgSQL":
            self.openDbase(dbname="template1", err=False)
            if self.db:
                self.cu.execute("Select count(*) from pg_catalog."\
                    "pg_database where datname='%s'" % self.dbname)
                if self.cu.fetchone()[0]:
                    exists = True
        elif self.dbase == "SQLite":
            if os.path.exists(self.dbdsn):
                self.openDbase(err=False)
                if self.db:
                    exists = True
        if exists:
            self.closeDbase()
        return exists

    def createDbase(self):
        "Create a database if it does not exist"
        try:
            if self.checkDbase():
                raise Exception("Database Already Exists")
            if self.dbase == "FBird":
                db = self.engine.create_database("Create database '%s' "\
                    "user '%s' password '%s'" % (self.dbdsn, self.dbuser,
                    self.dbpwd))
                db.close()
            elif self.dbase == "MsSQL":
                self.openDbase(dbname="master")
                self.cu.execute("Create database %s" % self.dbname)
                self.closeDbase()
            elif self.dbase == "MySQL":
                self.openDbase(dbname="mysql")
                self.cu.execute("Create database %s character set utf8 "\
                    "collate utf8_bin" % self.dbname)
                self.db.commit()
                self.closeDbase()
            elif self.dbase == "PgSQL":
                self.openDbase(dbname="template1", levl=0)
                self.cu.execute("Create database %s template template0 "\
                    "encoding 'utf-8'" % self.dbname)
                self.cu.execute("Select * from pg_catalog.pg_group "\
                    "where groname = '%s'" % self.dbname)
                chk = self.cu.fetchall()
                if not chk:
                    self.cu.execute("Create group %s" % self.dbname)
                self.db.commit()
                self.closeDbase()
            elif self.dbase == "SQLite":
                # Create the empty file
                try:
                    os.umask(0000)
                except:
                    pass
                db = open(self.dbdsn, "w")
                db.close()
                self.openDbase()
                if not self.db:
                    return
                self.closeDbase()
        except:
            showException(self.screen, self.wrkdir, "Database %s" % self.dbname)

    def openDbase(self, dbname=None, dbuser=None, dbpwd=None, levl=1, err=True):
        """
        Open a database using the following:

        dbname - The name of the database to open if not self.dbname
        dbuser - The user of the database to open if not self.dbuser
        dbpwd  - The pwd  of the database to open if not self.dbpwd
        levl   - The isolation level for postgresql
        err    - Whether or not to exit on an error
        """
        if not dbname:
            dbname = self.dbname
        try:
            if self.dbase == "FBird":
                self.db = self.engine.connect(host=self.dbhost,
                    database=self.dbdsn, user=self.dbuser, password=self.dbpwd)
                    #charset="UTF8")
            elif self.dbase == "MsSQL" and sys.platform == "win32":
                cs = "Provider=sqloledb;Data Source=%s;Initial Catalog=%s;"\
                    "User Id=%s;" % (self.dbhost, dbname, self.dbuser)
                if self.dbpwd:
                    cs = "%sPassword=%s;" % (cs, self.dbpwd)
                self.db = self.engine.connect(cs)
            elif self.dbase == "MySQL":
                if not self.dbpwd:
                    self.db=self.engine.connect(host=self.dbhost,
                        db=dbname, user=self.dbuser,
                        init_command="set autocommit=0")
                else:
                    self.db=self.engine.connect(host=self.dbhost,
                        db=dbname, user=self.dbuser, passwd=self.dbpwd,
                        init_command="set autocommit=0")
            elif self.dbase == "PgSQL":
                dsn = "host=%s dbname=%s" % (self.dbhost, dbname)
                if int(self.dbport):
                    dsn += " port=%s" % self.dbport
                if self.dbuser:
                    dsn += " user=%s" % self.dbuser
                if self.dbpwd:
                    dsn += " password=%s" % self.dbpwd
                self.db = self.engine.connect(dsn)
            elif self.dbase == "SQLite":
                v1 = self.engine.version
                db = open(self.dbdsn, "r")
                vv = db.readline()
                db.close()
                if not vv:
                    # Empty database
                    self.db = self.engine.connect(database=self.dbdsn)
                    self.db.text_factory = str
                elif v1 >= "1.1.0" and vv[:15] == "SQLite format 3":
                    # sqlite version 2 format 3
                    self.db = self.engine.connect(database=self.dbdsn)
                    self.db.text_factory = str
                else:
                    raise Exception("""This is an incorrectly formatted file,
Please convert it to version 3 before continuing!""")
            self.cu = self.db.cursor()
            if self.dbase == "PgSQL":
                self.setIsolation(levl)
                if self.typecast:
                    # Typecast the numeric datatype to be Float and not Decimal
                    pgx = self.engine.extensions
                    NUM = pgx.new_type(pgx.DECIMAL.values, "NUM",
                        lambda d, c: float(d) if d is not None else None)
                    pgx.register_type(NUM, self.db)
            self.dbopen = True
        except:
            self.db = self.cu = None
            if not err and not "TARTANDB" in os.environ:
                return False
            showException(self.screen, self.wrkdir, "Database %s" % dbname)
            sys.exit()

    def commitDbase(self, ask=False, rback=True, mess=None, default="yes"):
        """
        Commit a transaction using the following:

        ask     - Ask for confirmation to commit
        rback   - Whether to perform a rollback if not committing with ask
        mess    - The message to use for confirmation
        default - The default button to highlight, yes or no
        """
        if not ask:
            self.commit = "yes"
        else:
            if not mess:
                mess = "Would you like to COMMIT All Changes?"
            self.commit = askQuestion(self.screen, "Commit?", mess,
                default=default)
        if self.commit == "yes":
            self.db.commit()
        elif rback:
            self.db.rollback()

    def rollbackDbase(self):
        self.db.rollback()

    def vacuumDbase(self, table=None):
        """
        table - A table to vacuum or None for all
        """
        try:
            self.db.commit()
        except:
            pass
        if self.dbase == "PgSQL":
            self.setIsolation(0)
        if table:
            self.cu.execute("Vacuum %s" % table)
        else:
            self.cu.execute("Vacuum")
        if self.dbase == "PgSQL" and not "pgdb" in sys.modules:
            self.db.set_isolation_level(1)

    def setIsolation(self, levl):
        if "pgdb" in sys.modules:
            if levl == 1:
                set = "READ COMMITTED"
            else:
                set = "READ UNCOMMITTED"
            self.cu.execute("Set transaction isolation level %s" % set)
        else:
            self.db.set_isolation_level(levl)

    def closeDbase(self):
        try:
            self.cu.close()
        except:
            pass
        self.db.close()
        self.dbopen = False

    def dropDbase(self):
        try:
            if self.dbase == "FBird":
                self.openDbase()
                self.cu.close()
                self.db.drop_database()
                self.closeDbase()
            elif self.dbase == "MsSQL":
                self.openDbase(dbname="master")
                self.cu.execute("Drop database %s" % self.dbname)
                self.closeDbase()
            elif self.dbase == "MySQL":
                self.openDbase(dbname="mysql")
                self.cu.execute("Drop database %s" % self.dbname)
                self.closeDbase()
            elif self.dbase == "PgSQL":
                self.openDbase(dbname="template1", levl=0)
                self.cu.execute("Drop database %s" % self.dbname)
                self.cu.execute("Drop group %s" % self.dbname)
                self.closeDbase()
            elif self.dbase == "SQLite":
                os.remove(self.dbdsn)
        except:
            showException(self.screen, self.wrkdir, "Database %s" % self.dbname)

    def checkTable(self, table):
        if self.dbase == "FBird":
            self.cu.execute("Select rdb$relation_name from rdb$relations "\
                "where rdb$relation_name in ('%s', '%s')" % (table.lower(),
                table.upper()))
        elif self.dbase == "MsSQL":
            self.cu.execute("Select name from %s.dbo.sysobjects where "\
                "name = '%s'" % (self.dbname, table))
        elif self.dbase == "MySQL":
            self.cu.execute("Select table_name from information_schema.tables "\
                "where table_schema = '%s' and table_name = '%s'" % \
                (self.dbname, table))
        elif self.dbase == "PgSQL":
            self.cu.execute("Select relname from pg_class where "\
                "relname = '%s'" % table)
        elif self.dbase == "SQLite":
            self.cu.execute("Select name from SQLITE_MASTER where "\
                "name = '%s' and type = 'table'" % table)
        return self.cu.fetchone()

    def dropTable(self, table):
        if self.checkTable(table):
            self.cu.execute("Drop table %s" % table)
            if self.dbase == "FBird":
                try:
                    self.cu.execute("Drop generator gen_%s_id" % table)
                except:
                    pass
                try:
                    self.cu.execute("Drop trigger %s_bi" % table)
                except:
                    pass
                self.db.commit()

class Sql:
    """
    This class is used to select, insert, update and delete records.
        table_col  = A list of all column names
        table_fld  = A string of all column names comma delimited
        table_dic  = A dictionary of columns having:
                     {"colname": table, seq, type, size, desc, head}
    """
    def __init__(self, dbm, tables=None, prog=None, errs=True):
        """
        dbm    - The database class
        tables - The tables required for the insRec function
        prog   - The calling module name
        errs   - Whether to display errors
        """
        self.dbm = dbm
        if prog:
            self.prog = prog
        else:
            self.prog = "unknown"
        self.error = []
        if not tables:
            return
        elif type(tables) in (str, unicode):
            self.tables = [tables]
        else:
            self.tables = tables
        for table in self.tables:
            exec "self.%s = table" % table
            cols = self.sqlRec(state=("Select * from ffield where ff_tabl=%s "\
                "order by ff_seq" % self.dbm.dbf, (table,)), excp=errs)
            if cols:
                col = []
                for c in cols:
                    col.append(c[2])
                exec "self.%s_col = col" % table
                fld = ""
                dic = {}
                for n, c in enumerate(cols):
                    fld = "%s%s, " % (fld, c[2])
                    dic[c[2]] = [table, c[1], c[3], c[4], c[5], c[6]]
                fld = fld[:-2]
                exec "self.%s_fld = fld" % table
                exec "self.%s_dic = dic" % table
            elif errs:
                mess = """
The ffield record for %s Does Not Exist

Please Contact Your IT Manager IMMEDIATELY!
""" % table
                showError(None, "ffield Error", mess)
                self.error.append(table)
            else:
                self.error.append(table)

    def sqlRec(self, state=None, limit=None, fetch=False, excp=False):
        """
        state = A sql statement. It can also be a tuple of the statement and
                the data required for the clause e.g.
                    "Select abc from xyz where a=1 and b=2" or
                    ("Select abc from xyz where a=? and b=?", (1,2)) or
                    ("Delete from xyz where a=? and b=?", (1,2)) or
                    ("Insert into xyz values (?,?,?)", (1,2,3))
        limit = The number of records to return. None = All
        fetch = Only fetch from a previous select. True or False
        excp  = Whether or not to show exceptions or just return 'Exception'
        """
        def retList(ret):
            # Adjust to return [] instead of None and convert to list(s)
            if not ret:
                return []
            elif limit and limit == 1:
                return list(ret)
            else:
                return [ list(r) for r in ret ]
        try:
            if not state:
                if fetch:
                    ret = self.dbm.cu.fetchmany(limit)
                    return retList(ret)
                raise Exception("Missing sql statement")
            if type(state) not in (list, tuple):
                comm = state.split()
                self.dbm.cu.execute(state)
            else:
                comm = state[0].split()
                self.dbm.cu.execute(state[0], state[1])
            if comm[0].lower() == "select":
                if fetch:
                    return
                elif limit and limit == 1:
                    ret = self.dbm.cu.fetchone()
                elif limit:
                    ret = self.dbm.cu.fetchmany(limit)
                else:
                    ret = self.dbm.cu.fetchall()
                return retList(ret)
        except:
            if excp and not "TARTANDB" in os.environ:
                return "Exception"
            showException(None, self.dbm.wrkdir,
                "Error Executing SQL Statement\n\n%s\n\nIn module %s" % \
                (str(state), self.prog), dbm=self.dbm)
            try:
                self.dbm.rollbackDbase()
            except:
                pass
            os._exit(0)

    def insRec(self, table, data=[], unique=None, format=True, excp=False):
        """
        table   = The table to insert into
        data    = A list of all the column's data
        unique  = Any column which must be unique e.g. drt_ref1
        format  = Whether or not to format the data before inserting
        excp    = Whether or not to show exceptions or just return 'Exception'
        """
        if table not in self.tables:
            showError(None, "insRec Error",
                "Table %s Not In Sql Tables\n\nIn module %s" % \
                (table, self.prog))
            self.dbm.rollbackDbase()
            sys.exit()
        exec "dic = self.%s_dic" % table
        exec "col = self.%s_col" % table
        temp = list(data[:])
        if len(temp) != len(col):
            showError(None, "insRec Error",
                "Data and CSV Formats Differ\n\n%s\n\nIn module %s" % \
                (table, self.prog))
            self.dbm.rollbackDbase()
            sys.exit()
        ##############################################################
        # Test and Correct for a Column in Unique Keys (e.g. drt_ref1)
        ##############################################################
        if unique:
            whr = []
            cols = self.sqlRec(state=("Select * from ftable where ft_tabl=%s "\
                "and ft_type = 'U' order by ft_seq" % self.dbm.dbf, (table,)))
            if cols:
                for key in cols:
                    k = []
                    d = []
                    q = False
                    for c in key[4:]:
                        if c:
                            k.append(c)
                            d.append(data[col.index(c)])
                            if c == unique:
                                q = True
                    if q:
                        typ = dic[unique][2]
                        siz = dic[unique][3]
                        wrk = CCD(d[k.index(unique)], typ, siz).work
                        quo = int((str(siz).split(".")[0]))
                        seq = 0
                        ok = False
                        while not ok:
                            s = "Select count(*) from %s where" % table
                            for x in k:
                                s = "%s %s=%s and" % (s, x, self.dbm.dbf)
                            get = self.sqlRec((s[:-4], tuple(d)), limit=1)
                            if get[0]:
                                if typ[1].upper() == "A":
                                    try:
                                        wrk = int(wrk) + 1
                                    except:
                                        if len(wrk) == quo:
                                            if wrk[:1] == " ":
                                                dgt = "A"
                                            else:
                                                dgt = chr(ord(wrk[:1]) + 1)
                                            dat = wrk[1:]
                                        else:
                                            dgt = "A"
                                            dat = wrk
                                        fmt = "%1s%" + str((quo - 1)) + "s"
                                        wrk = fmt % (dgt, dat)
                                        if typ == "Na":
                                            wrk = wrk.replace(" ", "0")
                                else:
                                    wrk = wrk + (10**(quo - 1))
                                wrk = CCD(wrk, typ, siz).work
                                d[k.index(unique)] = wrk
                                temp[col.index(unique)] = wrk
                            else:
                                ok = True
        ##############################################################
        fmt = ""
        fld = ""
        skp = 0
        for num, nam in enumerate(col):
            idx = num - skp
            if dic[nam][2] == "US":
                del temp[idx]
                skp += 1
            else:
                if not fld:
                    fld = "%s" % nam
                else:
                    fld = "%s, %s" % (fld, nam)
                if not fmt:
                    fmt = "%s" % self.dbm.dbf
                else:
                    fmt = "%s, %s" % (fmt, self.dbm.dbf)
                if format:
                    if temp[num] == None:
                        if dic[nam][2][1].lower() in ("a", "x"):
                            temp[num] = ""
                        else:
                            temp[num] = 0
                    dat = CCD(temp[num], dic[nam][2], dic[nam][3])
                    if not dat.err:
                        temp[num] = dat.work
        self.sqlRec(("Insert into %s (%s) values (%s)" % (table, fld, fmt),
            tuple(temp)), excp=excp)

    def getRec(self, tables=None, join=None, cols=None, where=None, data=None, group=None, order=None, fetch=False, limit=None, excp=False):
        """
        tables = A string or list of tables
        join   = A join statement
        cols   = A list of columns to return else returns ALL
        where  = A list of tuples which can comprise a combination of:
                    "(", column name, function, data, [data] (and/or) ...
                    functions: = < > <= >= <> in between

                 Examples
                 --------
                 [("glm_acno", "=", 601)]
                 [("glm_acno", "in", (601, 602))]
                 [("glm_acno", "between", 601, 602)]
                 [("(", "glm_acno", "=", 601, "or", "glm_acno", "=", 602, ")")]

                 Note: "and" is the default and therefore will automatically
                    be used to join tuples of statements, "or" must be
                    implicitly stated as in the example above.

        data   = If no where statement then this is a list of ALL the table's
                    columns's data in order to build a where statement
                    e.g. genmst would be: [1, 1, "P", "Hello", "N", "I"]
        group  = A string of columns to group by e.g. "drt_cono, drt_acno"
        order  = A string of columns to order by e.g. "drt_cono, drt_acno"
        fetch  = Only fetch record from a previous select. True or False
        limit  = The number of records to return e.g. 1 else returns all
        excp   = Return the exception immediately
        """
        # check for tables
        if not tables:
            showError(None, "getRec Error",
                "The tables Option Must Exist in a getRec()")
            sys.exit()
        get = "Select"
        # convert tables to list if string
        if type(tables) in (str, unicode):
            tables = [tables]
        # check for only one table if there is data
        if not where and data and len(tables) > 1:
            showError(None, "getRec Error",
                "List of data exists but more than 1 table")
            sys.exit()
        if not cols:
            if len(tables) > 1:
                showError(None, "getRec Error",
                    "No list of cols but more than 1 table")
                sys.exit()
            exec "cols = self.%s_col" % tables[0]
        for col in cols:
            get = "%s %s," % (get, col)
        get = "%s from" % get[:-1]
        for table in tables:
            get = "%s %s," % (get, table)
        get = get[:-1]
        if join:
            get = "%s %s" % (get, join)
        if where:
            whr, dat = self.getSqlWhere(where)
        elif data:
            # data contains the data for all the columns
            whr = ""
            dat = data[:]
            for table in tables:
                exec "cols = self.%s_col" % table
                for col in cols:
                    if not whr:
                        whr = "where %s = %s" % (col, self.dbm.dbf)
                    else:
                        whr = "%s and %s = %s" % (whr, col, self.dbm.dbf)
            limit = 1
        else:
            # all records
            whr = None
            dat = None
        if whr:
            get = "%s %s" % (get, whr)
        if group:
            get = "%s group by %s" % (get, group)
        if order:
            get = "%s order by %s" % (get, order)
        if not dat:
            return self.sqlRec(get, fetch=fetch, limit=limit, excp=excp)
        else:
            return self.sqlRec((get, dat), fetch=fetch, limit=limit, excp=excp)

    def updRec(self, table, where=None, data=None, cols=None, format=True):
        """
        table   = The table to update
        where   = A where statement like the getRec one
        data    = A list of data matching the columns in cols or all if no cols
        cols    = A list of columns matching 'data' else None for all
        format  = Whether or not to format the data before updating
        """
        cmd = ""
        if not where and not data:
            showError(None, "updRec Error", "Missing list of data")
            sys.exit()
        if not cols:
            exec "cols = self.%s_col" % table
        if len(data) != len(cols):
            showError(None, "updRec Error",
                "Length of data not equal to length of cols")
            sys.exit()
        exec "dic = self.%s_dic" % table
        for num, col in enumerate(cols):
            if not cmd:
                cmd = "Update %s set %s = %s" % (table, col, self.dbm.dbf)
            else:
                cmd = "%s, %s = %s" % (cmd, col, self.dbm.dbf)
            if format:
                dat = CCD(data[num], dic[col][2], dic[col][3])
                if not dat.err:
                    data[num] = dat.work
        dat = data[:]
        if where:
            whr, ext = self.getSqlWhere(where)
            dat.extend(ext)
            cmd = "%s %s" % (cmd, whr)
        self.sqlRec((cmd, tuple(dat)))

    def delRec(self, table, cols=None, data=None, where=None):
        """
        table = The table to delete from
        cols  = A list of columns matching 'data' else None for all
        data  = A list of data matching the columns in cols or all columns
        where = A where statement if no data and cols, as per getRec
        """
        if where:
            whr, dat = self.getSqlWhere(where)
            cmd = ("Delete from %s %s" % (table, whr), tuple(dat))
        elif data:
            # Delete where the data matches
            if not cols:
                exec "cols = self.%s_col" % table
            if len(data) != len(cols):
                showError(None, "delRec Error",
                    "Length of data not equal to length of cols")
                sys.exit()
            cmd = ""
            for col in cols:
                if not cmd:
                    cmd = "Delete from %s where %s = %s" % (table, col,
                        self.dbm.dbf)
                else:
                    cmd = "%s and %s = %s" % (cmd, col, self.dbm.dbf)
            cmd = (cmd, tuple(data))
        else:
            # Delete all records
            cmd = "Delete from %s" % table
        self.sqlRec(cmd)

    def getSqlWhere(self, where):
        whr = ""
        dat = []
        for w in where:
            if len(w) == 1:
                if not whr:
                    whr = "where %s" % w[0]
                else:
                    whr = "%s and %s" % (whr, w[0])
                continue
            flag = None
            for c in w:
                if not flag:
                    if not whr:
                        whr = "where %s" % c
                    else:
                        whr = "%s and %s" % (whr, c)
                    if c in ("(", ")"):
                        flag = "a"
                    else:
                        flag = "f"
                elif flag == "a":
                    whr = "%s %s" % (whr, c)
                    flag = "f"
                elif flag == "f":
                    if c.lower() == "ilike" and self.dbm.dbase != "PgSQL":
                        c = "like"
                    whr = "%s %s" % (whr, c)
                    if c.lower() == "between":
                        flag = "b1"
                    elif c.lower() == "not":
                        continue
                    elif c.lower() in ("in", "not in"):
                        flag = "i"
                    else:
                        flag = "b0"
                elif flag in ("b0", "b1"):
                    whr = "%s %s" % (whr, self.dbm.dbf)
                    if type(c) == float:
                        dat.append(str(c))
                    else:
                        dat.append(c)
                    if flag == "b0":
                        flag = "n"
                    else:
                        flag = "b2"
                elif flag == "b2":
                    whr = "%s and %s" % (whr, self.dbm.dbf)
                    if type(c) == float:
                        dat.append(str(c))
                    else:
                        dat.append(c)
                    flag = "n"
                elif flag == "i":
                    whr = "%s (" % whr
                    for x in c:
                        whr = "%s %s," % (whr, self.dbm.dbf)
                        if type(c) == float:
                            dat.append(str(x))
                        else:
                            dat.append(x)
                    whr = whr[:-1] + ")"
                    flag = "n"
                elif flag in ("n", "x"):
                    whr = "%s %s" % (whr, c)
                    if c in ("(", ")"):
                        continue
                    if flag == "n":
                        flag = "x"
                    else:
                        flag = "f"
        if self.dbm.dbase == "PgSQL":
            whr = whr.replace(" % ", " %% ")
        return whr, dat

class TarBckRes:
    """
    Backup and Restore routines for Tartan Systems

        mode - B(ackup), (R)store.
        smtp - The smtp detail.
        ver  - The tartan version.
        sys  - The system file e.g. tartan.sys
    """
    def __init__(self, mf, mode="B", smtp=None, ver=None, sys=None):
        self.mf = mf
        self.mode = mode
        if smtp:
            self.budays = smtp[0]
            self.smtp = smtp[1:]
        else:
            self.budays = 0
            self.smtp = None
        self.ver = ver
        self.sys = sys
        no = self.doVariables()
        if not no:
            if self.mode == "B":
                if self.smtp and self.smtp[1]:
                    self.startBck()
                    self.mf.startLoop()
                else:
                    self.endBckRes()
            else:
                self.startRes()
                self.mf.startLoop()
    
    def doVariables(self):
        t = time.localtime()
        self.sysdtw = ((t[0] * 10000) + (t[1] * 100) + t[2])
        self.bupdir = os.path.join(self.mf.bupdir, self.mf.dbname)
        if self.mode == "R" and not os.path.isdir(self.bupdir):
            return "no"
        self.arcdir = os.path.join(self.bupdir, "arch")
        if self.mode == "R" and not os.path.isdir(self.arcdir):
            return "no"
        self.tmpdir = os.path.join(self.bupdir, "temp")
        # Get all companies
        sql = Sql(self.mf.dbm, "ctlmst", prog=__name__)
        if sql.error:
            return "no"
        self.coys = sql.getRec(tables="ctlmst", cols=["ctm_cono",
            "ctm_name"], order="ctm_cono")
        self.incoac = False
        self.slsinv = False
        if self.coys:
            # Check for Intercompany Facility
            sql = Sql(self.mf.dbm, "genint", errs=False, prog=__name__)
            if not sql.error:
                int = sql.getRec(tables="genint", cols=["count(*)"], limit=1)
                if int[0]:
                    self.incoac = True
            # Check for Sales Invoicing
            sql = Sql(self.mf.dbm, "slsiv1", errs=False, prog=__name__)
            if not sql.error:
                inv = sql.getRec(tables="slsiv1", cols=["count(*)"], limit=1)
                if inv[0]:
                    self.slsinv = True
        # Load System Names and Check for Integration
        sysd = {
            "ASS": ["Asset's Register", "assctl", "cta_glint"],
            "BWL": ["Bowling Clubs"],
            "CRS": ["Creditor's Ledger", "crsctl", "ctc_glint"],
            "CSH": ["Cash Analysis"],
            "DRS": ["Debtor's Ledger", "drsctl", "ctd_glint"],
            "GEN": ["General Ledger"],
            "MEM": ["Member's Ledger", "memctl", "mcm_glint"],
            "RCA": ["Rental's Ledger (Extended)", "rcactl", "cte_glint"],
            "RTL": ["Rental's Ledger (Basic)", "rtlctl", "ctr_glint"],
            "STR": ["Store's Ledger", "strctl", "cts_glint"],
            "WAG": ["Salaries and Wages", "wagctl", "ctw_glint"]}
        self.syss = []
        self.sysi = []
        keys = sysd.keys()
        keys.sort()
        for s in keys:
            det = sysd[s]
            self.syss.append([s, det[0]])
            if len(det) == 1:
                continue
            sql = Sql(self.mf.dbm, det[1], errs=False, prog=__name__)
            if sql.error:
                continue
            int = sql.getRec(tables=det[1], cols=[det[2]])
            if not int:
                continue
            for i in int:
                if i[0] == "Y":
                    self.sysi.append(s)
                    break
        self.syst = []
        for s in self.syss:
            self.syst.append([s[0], s[1]])
        self.wkdays = ["Monday", "Tuesday", "Wednesday", "Thursday",
            "Friday", "Saturday", "Sunday"]
        self.full = "N"
        self.icon = "Y"
        self.isys = "Y"

    def startBck(self):
        tit = ("Database Backup",)
        self.bu = TartanDialog(self.mf, screen=self.mf.body, tops=True,
            title=tit, eflds=[], tend=((self.endBckRes, "y"),),
            txit=(self.noBckRes,), mail=("Y","N","E-Mail Archive"))

    def startRes(self):
        tit = ("Database Restore",)
        day = {
            "stype": "C",
            "titl": "Available Backups",
            "head": [],
            "data": [],
            "index": 0}
        coy = {
            "stype": "C",
            "titl": "",
            "data": self.coys,
            "mode": "M",
            "comnd": self.doCoyCmd}
        sys = {
            "stype": "C",
            "titl": "",
            "data": self.syst,
            "mode": "M",
            "comnd": self.doSysCmd}
        r1s = (("Full","Y"),("Partial","N"))
        r2s = (("Archive","A"),("Weekday","W"))
        r3s = (("Yes","Y"),("Include","I"),("Exclude","E"))
        fld = (
            (("T",0,0,0),("IRB",r1s),0,"Type of Restore","",
                "Y","N",self.doType,None,None,None),
            (("T",0,1,0),("IRB",r2s),0,"Restore From","",
                "A","N",self.doSource,None,None,None),
            (("T",0,2,0),"INA",50,"Archive/Weekday","",
                "","Y",self.doPlace,day,None,None),
            (("T",0,3,0),("IRB",r3s),0,"All Companies","",
                "Y","N",self.doAllCoy,None,None,None),
            (("T",0,4,0),"INA",(30,100),"Companies","",
                "","N",self.doCoySel,coy,None,None),
            (("T",0,5,0),("IRB",r3s),0,"All Systems","",
                "Y","N",self.doAllSys,None,None,None),
            (("T",0,6,0),"INA",(50,100),"Systems","",
                "","N",self.doSysSel,sys,None,None))
        self.bu = TartanDialog(self.mf, screen=self.mf.body, tops=True,
            title=tit, eflds=fld, tend=((self.endBckRes, "y"),),
            txit=(self.noBckRes,))

    def doType(self, frt, pag, r, c, p, i, w):
        self.full = w

    def doSource(self, frt, pag, r, c, p, i, w):
        self.source = w
        data = []
        if self.source == "A":
            head = ("   Date", "  Time", "Archive")
            for arc in glob.glob(os.path.join(self.arcdir, "*.tar")):
                dtt = arc.split("_")[1].split(".")[0]
                dte = CCD(dtt[:8], "D1", 10).disp
                tim = CCD(dtt[8:12], "TM", 5).disp
                data.append([dte, "%s:%s" % (tim, dtt[12:14]),
                    os.path.basename(arc)])
            data.sort()
            data.reverse()
            index = 2
        else:
            head = ("Backup Date", "Backup Day")
            for day in os.listdir(self.bupdir):
                if not day in self.wkdays:
                    continue
                if os.path.isdir(os.path.join(self.bupdir, day)):
                    mt = os.path.getmtime(os.path.join(self.bupdir, day))
                    data.append([mt, time.strftime("%b %d %Y",
                        time.localtime(mt)), day])
            data.sort()
            data.reverse()
            for day in range(len(data)):
                del data[day][0]
            index = 1
        self.bu.topf[pag][p+1][5] = data[0][index]
        self.bu.topf[pag][p+1][8]["head"] = head
        self.bu.topf[pag][p+1][8]["data"] = data
        self.bu.topf[pag][p+1][8]["index"] = index

    def doPlace(self, frt, pag, r, c, p, i, w):
        if self.source == "A":
            if not os.path.isfile(os.path.join(self.arcdir, w)):
                return "Invalid Archive"
            arcfle = w
        else:
            self.daydir = os.path.join(self.bupdir, w)
            if not os.path.isdir(self.daydir):
                return "Invalid Restore Directory"
        try:
            # Delete existing directory and recreate it
            if os.path.exists(self.tmpdir):
                shutil.rmtree(self.tmpdir)
            os.makedirs(self.tmpdir)
        except:
            showException(self.mf.body, self.mf.rcdic["wrkdir"],
                "Permission Error")
            return "Invalid Restore Directory"
        cwd = os.getcwd()
        os.chdir(self.tmpdir)
        if self.source == "A":
            tarfle = tarfile.open(os.path.join(self.arcdir, arcfle), "r")
            try:
                # Only available since 2.5
                tarfle.extractall()
            except:
                for member in tarfle.getmembers():
                    tarfle.extract(member)
            tarfle.close()
        else:
            files = glob.glob(os.path.join(self.daydir, "??????_*.dat"))
            for flenam in files:
                shutil.copy(flenam, self.tmpdir)
        os.chdir(cwd)
        if not os.path.isfile(os.path.join(self.tmpdir, "verupd_0.dat")):
            return
        zipfle = gzip.open(os.path.join(self.tmpdir, "verupd_0.dat"), mode="rb")
        data = zipfle.readlines()
        data = data[0].replace("[[", "").replace("]]", "")
        data = data.split("], [")
        data = list(eval(data[0]))
        if data[0] != self.ver:
            ok = askQuestion(self.mf.body, "Version Error", "The Data "\
                "to be Restored is Version %s which is Not the Same as "\
                "this Version of TARTAN, %s.\n\nOnly Matching Versions "\
                "Should Be Restored!\n\nYou Do So At Your Own Risk!\n\n"\
                "Continue?" % (data[0], self.ver), default="no")
            if ok == "no":
                return "Invalid Version"
        if self.full == "Y":
            return "nd"

    def doAllCoy(self, frt, pag, r, c, p, i, w):
        self.icon = w
        if self.icon == "Y":
            self.bu.loadEntry(frt, pag, i+1, data="")
            return "sk1"
        elif self.icon == "I":
            self.bu.topf[pag][p+1][8]["titl"] = "Select Companies to Include"
        else:
            self.bu.topf[pag][p+1][8]["titl"] = "Select Companies to Exclude"

    def doCoyCmd(self, frt, pag, r, c, p, i, w):
        c = ""
        for co in w:
            if int(co):
                c = "%s%s," % (c, co)
        if c[-1:] == ",":
            c = c[:-1]
        self.bu.loadEntry(frt, pag, p, data=c)

    def doCoySel(self, frt, pag, r, c, p, i, w):
        if not w:
            return "Invalid List of Companies"
        if w[-1:] == ",":
            w = w[:-1]
        self.coy = w.split(",")

    def doAllSys(self, frt, pag, r, c, p, i, w):
        self.isys = w
        if self.isys == "Y":
            self.bu.loadEntry(frt, pag, i+1, data="")
            return "sk1"
        cf = PwdConfirm(self.mf, conum=0, system="MST", code="Systems")
        if cf.flag == "no":
            return "No Authority for Single Systems"
        if self.isys == "I":
            self.bu.topf[pag][p+1][8]["titl"] = "Select Systems to Include"
        else:
            self.bu.topf[pag][p+1][8]["titl"] = "Select Systems to Exclude"

    def doSysCmd(self, frt, pag, r, c, p, i, w):
        self.sys = []
        s = ""
        for sy in w:
            if sy:
                self.sys.append(sy)
                s = "%s%s," % (s, sy)
        if s[-1:] == ",":
            s = s[:-1]
        if s:
            self.bu.loadEntry(frt, pag, p, data=s)

    def doSysSel(self, frt, pag, r, c, p, i, w):
        if not w:
            return "Invalid List of Systems"
        if w[-1:] == ",":
            w = w[:-1]
        self.sys = w.split(",")


    def endBckRes(self):
        if self.mode == "B":
            if self.smtp and self.smtp[1]:
                self.bu.closeProcess()
            self.doBackup()
            if self.smtp and self.smtp[1]:
                if self.bu.repprt[1][2]:
                    sendMail(self.smtp[1], self.smtp[2], self.bu.repprt[1][2],
                        "Tartan Backup for %s" % self.smtp[0],
                        attach=self.arcfle)
                self.mf.closeLoop()
            return
        if self.mf.display:
            self.bu.closeProcess()
        if self.mode == "R" and self.full == "Y":
            self.doFullRestore()
            if self.mf.display:
                self.mf.closeLoop()
            return
        if self.icon == "Y":
            self.con = [0]
            for coy in self.coys:
                self.con.append(coy[0])
            self.con.sort()
        elif self.icon == "I":
            # Included companies
            self.con = self.coy
        elif self.icon == "E":
            # Excluded companies
            self.con = []
            for co in self.coys:
                self.con.append(int(co[0]))
            for co in self.coy:
                del self.con[self.con.index(int(co))]
            self.con.sort()
        if self.isys == "Y":
            # All systems
            self.syn = ["chg", "eml", "fre", "reg", "rpt", "tel", "ver"]
            for sys in self.syss:
                self.syn.append(sys[0].lower())
        elif self.isys == "I":
            self.syn = []
            for sys in self.sys:
                self.syn.append(sys.lower())
        elif self.isys == "E":
            self.syn = []
            for sys in self.syss:
                self.syn.append(sys[0].lower())
            for sys in self.sys:
                self.syn.remove(sys.lower())
        # Check if drs or str and slsinv
        if self.slsinv:
            if "drs" in self.syn:
                if not "str" in self.syn:
                    self.syn.append("str")
                self.syn.append("sls")
            elif "str" in self.syn:
                if not "drs" in self.syn:
                    self.syn.append("drs")
                self.syn.append("sls")
        # Check for integrated systems and exit if not all included
        if self.sysi:
            chk = ["GEN"] + self.sysi
            for c in chk:
                if c.lower() not in self.syn:
                    showError(self.mf.body, "System Error",
                        "Some Systems Are Integrated. You Will Have To "\
                        "Restore the General Ledger System as well as "\
                        "All Other Integrated Systems!")
                    if self.mf.display:
                        self.mf.closeLoop()
                    return
        self.tabs = []
        sql = Sql(self.mf.dbm, "ftable", prog=__name__)
        if self.isys == "Y":
            syss = ["ctl"]
        else:
            syss = []
        syss.extend(self.syn[:])
        for sys in syss:
            tbs = sql.getRec(tables="ftable", cols=["ft_tabl"],
                where=[("ft_tabl", "like", "%s%s" % (sys, "%")),
                ("ft_seq", "=", 1)])
            for tb in tbs:
                self.tabs.append(tb[0])
        self.tabs.sort()
        self.doPartialRestore()
        if self.mf.display:
            self.mf.closeLoop()

    def doBackup(self):
        try:
            os.umask(0000)
        except:
            pass
        try:
            if not os.path.exists(self.tmpdir):
                # Create the backup directory if not existing
                os.makedirs(self.tmpdir)
            if not os.path.exists(self.arcdir):
                # Create the backup archive directory if not existing
                os.makedirs(self.arcdir)
        except:
            showError(self.mf.body, "Backup Error", "The Backup Directory "\
                "%s is Either Not Available or Write Protected! Please "\
                "Rectify this Condition and Retry the Backup!" % self.bupdir)
            return
        if self.budays:
            # Remove old archives
            try:
                for arc in os.listdir(self.arcdir):
                    date = arc.split("_")[1].split(".")[0][:8]
                    if dateDiff(date, self.sysdtw, ptype="days") > self.budays:
                        os.remove(os.path.join(self.arcdir, arc))
            except:
                pass
        tabs = []
        sql = Sql(self.mf.dbm, "ftable", prog=__name__)
        tbs = sql.getRec(tables="ftable", cols=["ft_tabl"], group="ft_tabl")
        for tb in tbs:
            tabs.append(tb[0])
        tabs.sort()
        coys = [[0, ""]] + self.coys
        if self.mf.display:
            p1 = ProgressBar(self.mf.body, max=len(coys))
            p2 = ProgressBar(self.mf.body, max=len(tabs))
        names = []
        for cn, coy in enumerate(coys):
            if self.mf.display:
                p1.mess = "Backing Up Company %s" % coy[0]
                p1.displayProgress(cn)
            else:
                print "Backing Up Company %s ... Please Wait" % coy[0]
            for tn, tab in enumerate(tabs):
                sql = Sql(self.mf.dbm, ["ffield", tab], prog=__name__)
                if self.mf.display:
                    p2.mess = "Backing Up Table %s" % tab
                    p2.displayProgress(tn)
                fld = sql.getRec(tables="ffield", cols=["ff_name"],
                    where=[("ff_tabl", "=", tab), ("ff_seq", "=", 0)],
                    limit=1)
                if fld[0].count("_cono"):
                    whr = [("%s" % fld[0], "=", coy[0])]
                elif coy[0]:
                    continue
                else:
                    whr = None
                flenam = "%s_%s.dat" % (tab, coy[0])
                names.append(flenam)
                zipfle = gzip.open(os.path.join(self.tmpdir, flenam),
                    mode="wb", compresslevel=5)
                sql.getRec(tables=tab, where=whr, fetch=True)
                data = sql.sqlRec(fetch=True, limit=100000)
                while data:
                    zipfle.write(str(data))
                    data = sql.sqlRec(fetch=True, limit=100000)
                zipfle.close()
        cwd = os.getcwd()
        os.chdir(self.tmpdir)
        tme = "%04i%02i%02i%02i%02i%02i" % time.localtime()[:-3]
        self.arcfle = os.path.join(self.arcdir,
            "%s_%s.tar" % (self.mf.dbname, tme))
        tarfle = tarfile.open(self.arcfle, "w")
        for name in names:
            tarfle.add(name)
        tarfle.close()
        os.chdir(cwd)
        if self.mf.display:
            p1.closeProgress()
            p2.closeProgress()

    def doFullRestore(self):
        # Create Database
        self.mf.dbm.closeDbase()
        opts = [
            ("-c", "i"),
            ("-d", "%s" % os.path.join(self.mf.prgdir, "csv")),
            ("-l", True),
            ("-p", self.mf.rcdic["dbpwd"]),
            ("-r", False),
            ("-u", self.mf.rcdic["dbuser"]),
            ("-w", False),
            ("-x", True)]
        if self.sys:
            opts.append(("-s", self.sys))
        else:
            opts.append(("-s", "tartan"))
        create = DBCreate(mf=self.mf, opts=opts)
        if not create.put:
            return
        # Open the database
        self.mf.dbm.openDbase()
        # Restore the Files
        files = glob.glob(os.path.join(self.tmpdir, "??????_*.dat"))
        files.sort()
        if self.mf.display:
            p1 = ProgressBar(self.mf.body, max=len(files),
                typ="Restoring Database %s" % self.mf.dbname)
        for num, fle in enumerate(files):
            if self.mf.display:
                p1.displayProgress(num)
            base = os.path.basename(fle)
            name = base.split("_")[0]
            comp = int(base.split("_")[1].split(".")[0])
            if name in ("ffield", "ftable"):
                continue
            sql = Sql(self.mf.dbm, name, prog=__name__)
            if not comp:
                sql.delRec(name)
            zipfle = gzip.open(fle, mode="rb")
            data = zipfle.readlines()
            zipfle.close()
            if data:
                newd = self.extractData(data)
                if self.mf.display:
                    p2 = ProgressBar(self.mf.body, max=len(newd),
                        typ="Restoring File (%s)" % base)
                for num, dat in enumerate(newd):
                    if self.mf.display:
                        p2.displayProgress(num)
                    sql.insRec(name, data=dat, format=False)
                if self.mf.display:
                    p2.closeProgress()
        if self.mf.display:
            p1.closeProgress()
        self.mf.dbm.commitDbase()

    def doPartialRestore(self):
        if self.mf.display:
            p1 = ProgressBar(self.mf.body, max=len(self.con))
            p2 = ProgressBar(self.mf.body, max=len(self.tabs))
        for cn, coy in enumerate(self.con):
            if self.mf.display:
                p1.mess = "Restoring Company (%s)" % coy
                p1.displayProgress(cn)
            else:
                print "Restoring Company (%s) .... Please Wait" % coy
            sql = Sql(self.mf.dbm, self.tabs, prog=__name__)
            for tn, tab in enumerate(self.tabs):
                if self.mf.display:
                    p2.mess = "Restoring Table (%s)" % tab
                    p2.displayProgress(tn)
                flenam = os.path.join(self.tmpdir, "%s_%s.dat" % (tab, coy))
                if os.path.exists(flenam):
                    ############################
                    # Delete existing records
                    ############################
                    if not coy:
                        whr = None
                    else:
                        fld = self.mf.sqlRec(("Select ff_name from ffield "\
                            "where ff_tabl = %s and ff_seq = %s" % \
                            (self.mf.dbm.dbf, self.mf.dbm.dbf), (tab, 0)),
                            limit=1)
                        # Ignore tables without _cono columns
                        if not fld[0].count("_cono"):
                            continue
                        whr = [("%s" % fld[0], "=", coy)]
                    sql.delRec(tab, where=whr)
                    ############################
                    zipfle = gzip.open(flenam, mode="rb")
                    data = zipfle.readlines()
                    zipfle.close()
                    if data:
                        newd = self.extractData(data)
                        if self.mf.display:
                            p3 = ProgressBar(self.mf.body, max=len(newd),
                                typ="Restoring Records")
                        for num, dat in enumerate(newd):
                            if self.mf.display:
                                p3.displayProgress(num)
                            sql.insRec(tab, data=dat, format=False)
                        if self.mf.display:
                            p3.closeProgress()
        if self.mf.display:
            p2.closeProgress()
            p1.closeProgress()
        self.mf.dbm.commitDbase()

    def noBckRes(self):
        self.bu.closeProcess()
        self.mf.closeLoop()

    def extractData(self, data):
        data = data[0].replace("]][[", "], [")
        data = data[2:-2].split("], [")
        newd = []
        for d in data:
            d = list(eval(d))
            newd.append(d)
        return newd

class CreateChart:
    """
    mf       - The mainframe class
    conum    - The company number
    conam    - The company name
    periods  - A list of the starting and ending period
    title    - A list of the chart's title
    achart   - A list of automatic chart values
    mchart   - A list of manual chart values
    xlab     - The x-axis's lable
    ylab     - The y-axis's lable
    """
    def __init__(self, mf, conum, conam, periods, title, achart, mchart, xlab=None, ylab=None, args=None):
        if not PYCHART:
            return
        self.mf = mf
        self.conum = conum
        self.conam = conam
        self.periods = periods
        self.title = title
        self.achart = achart
        self.mchart = mchart
        if xlab:
            self.xlab = xlab
        else:
            self.xlab = "Months"
        if ylab:
            self.ylab = ylab
        else:
            self.ylab = "Values"
        self.args = args
        self.doVariables()
        if self.achart and self.args:
            self.act = "A"
            self.chart = self.achart[:]
            self.cht = self.args[0]
            self.col = self.args[1]
            self.doChtEnd()
        else:
            self.doProcess()
            self.mf.startLoop()

    def doVariables(self):
        self.s_per = CCD(self.periods[0], "D2", 7)
        self.e_per = CCD(self.periods[1], "D2", 7)

    def doProcess(self):
        tit = ("Pie, Bar and Line Charts",)
        if self.achart:
            r1s = [("Automatic","A")]
        else:
            r1s = []
        r1s.extend([("Manual","M"),("Exit","E")])
        r2s = (
            ("Pie Chart","P"),
            ("Normal Bar","B"),
            ("Stacked Bar","S"),
            ("Line","L"))
        r3s = (
            ("Yes", "Y"),
            ("No", "N"))
        fld = (
            (("T",0,0,0),("IRB",r1s),0,"Action","Select Action",
                "M","N",self.doAct,None,None,None,None),
            (("T",0,1,0),("IRB",r2s),0,"Select Chart","Select Chart",
                "L","N",self.doCht,None,None,None,None),
            (("T",0,2,0),("IRB",r3s),0,"Use Colour","Use Colour (y/n)",
                "Y","N",self.doCol,None,None,None,None))
        self.dc = TartanDialog(self.mf, screen=self.mf.body, tops=True,
            title=tit, eflds=fld, tend=((self.doChtEnd, "y"),),
            txit=(self.doChtExit,), view=("N","V"), mail=("Y","N"))

    def doAct(self, frt, pag, r, c, p, i, w):
        self.act = w
        if self.act == "E":
            return "xt"
        elif self.act == "A":
            self.chart = self.achart[:]
        else:
            titl = "Available Values"
            lines = []
            for line in self.mchart:
                lines.append([line[0], line[1]])
            state = self.dc.disableButtons()
            sc = SChoice(self.mf, scrn=self.mf.body, titl=titl, data=lines,
                rows=None, retn="I", mode="M")
            self.dc.enableButtons(state=state)
            if sc.selection in (None, []):
                return "rf"
            self.chart = []
            for idx in sc.selection:
                self.chart.append(self.mchart[idx])

    def doCht(self, frt, pag, r, c, p, i, w):
        if w == "P" and len(self.chart) > 1:
            return "Only One Range of Values Allowed for a Pie Chart"
        self.cht = w

    def doCol(self, frt, pag, r, c, p, i, w):
        self.col = w

    def doChtEnd(self):
        if self.args:
            view = "V"
            prtnam = "view"
            mail = "N"
        else:
            view = self.dc.repprt[0][1]
            prtnam = self.dc.repprt[0][2]
            mail = self.dc.repprt[1][2]
        if self.col == "Y":
            pychart_theme.use_color = 1
        else:
            pychart_theme.use_color = 0
        pychart_theme.reinitialize()
        self.chtnam = getModName(self.mf.wrkdir, __name__, self.conum,
            ext="pdf")
        cnv = pychart_canvas.init(fname=self.chtnam, format="pdf")
        if self.cht == "P":
            cnv.show(250, 550,"/14/hC/vT%s" % self.title[0][0])
            cnv.show(250, 530, "/12/hC/vT%s For Period %s to %s (%s)" % \
                (self.title[0][1], self.s_per.disp, self.e_per.disp,
                self.title[1]))
            cnv.show(250, 510, "/12/hC/vT%s" % self.chart[0][1])
        else:
            cnv.show(200, 550,"/14/hC/vT%s" % self.title[0][0])
            cnv.show(200, 530, "/12/hC/vT%s For Period %s to %s (%s)" % \
                (self.title[0][1], self.s_per.disp, self.e_per.disp,
                self.title[1]))
        labs = []
        zers = True
        negs = True
        for line in self.chart:
            labs.append(line[1])
            for num in line[2:]:
                if num:
                    zers = False
                if num > 0:
                    negs = False
                if num < 0 and self.cht == "P":
                    showError(self.mf.body, "Chart Error",
                        "You Cannot Have Negative Values in a Pie Chart!\n\n"\
                        "Please Select an Alternate Chart Type.")
                    self.dc.focusField("T",0,1)
                    return
        if zers:
            showError(self.mf.body, "Chart Error",
                "All the Items Selected Have Zero Values!\n\n"\
                "Please Select Items which Have Non Zero Values.")
            self.dc.focusField("T",0,1)
            return
        data = []
        max = 0
        min = 0
        tot = 0
        y = (self.s_per.work / 100) % 100
        m = self.s_per.work % 100
        for x in range(0, 12):
            data.append(["%s '%02i" % (mthnam[m][0], y)])
            for z in range(0, len(self.chart)):
                amt = self.chart[z][x+2]
                data[x].append(amt)
                if self.cht == "S":
                    tot = tot + amt
                elif amt < 0 and amt < min:
                    min = amt
                elif amt > max:
                    max = amt
            if self.cht == "S":
                if tot < 0 and tot < min:
                    min = tot
                elif tot > max:
                    max = tot
                tot = 0
            if m == (self.e_per.work % 100):
                break
            m += 1
            if  m > 12:
                y += 1
                m = 1
        if self.cht == "P":
            are = area.T(
                size = (500,500),
                legend = None,
                x_grid_style = None,
                y_grid_style = None)
            plot = pie_plot.T(
                data = data,
                center = (250, 250),
                radius = 190,
                arc_offsets = [],
                label_offset = 25,
                arrow_style = arrow.a3)
            are.add_plot(plot)
        else:
            tot = max + (0 - min)
            int = 100
            num = tot / int
            while num > 9:
                int += 100
                num = tot / int
            if max:
                max = (((max / int) + 1) * int)
            if min:
                min = (((min / (0 - int)) - 1) * int)
            if self.cht in ("B", "S"):
                are = area.T(
                    size = (400, 500),
                    loc = (10, 0),
                    x_coord = category_coord.T(data, 0),
                    y_range = (min, max),
                    x_axis = axis.X(label=self.xlab),
                    y_axis = axis.Y(label=self.ylab, tic_interval=int),
                    y_grid_interval = int)
                chart_object.set_defaults(bar_plot.T, data=data, width=5)
                t = len(labs)
                for n, d in enumerate(labs):
                    if self.cht == "B":
                        plot = bar_plot.T(label=d, hcol=n+1, cluster=(n, t-1))
                    else:
                        if not n:
                            plot = bar_plot.T(label=d, hcol=n+1)
                        else:
                            stak = plot
                            plot = bar_plot.T(label=d, hcol=n+1, stack_on=stak)
                    are.add_plot(plot)
            elif self.cht == "L":
                are = area.T(
                    size = (400, 500),
                    loc = (0, 0),
                    x_coord = category_coord.T(data, 0),
                    y_range = (min, max),
                    x_axis = axis.X(label=self.xlab),
                    y_axis = axis.Y(label=self.ylab, tic_interval=int),
                    y_grid_interval = int)
                chart_object.set_defaults(line_plot.T, data=data)
                for n, d in enumerate(labs):
                    plot = line_plot.T(label=d, ycol=n+1)
                    are.add_plot(plot)
        are.draw(cnv)
        cnv.close()
        doPrinter(mf=self.mf, conum=self.conum, rep=self.chtnam,
            head=self.title[0], view=view, prtnam=prtnam, mail=mail)
        if not self.args:
            self.dc.focusField("T",0,1)

    def doChtExit(self):
        self.dc.closeProcess()
        self.mf.closeLoop()

class NotesCreate:
    def __init__(self, mf, conum, conam, user, sys, key, commit=True):
        self.mf = mf
        self.conum = conum
        self.conam = conam
        self.user = user
        self.sys = sys
        self.key = str(key)
        self.commit = commit
        self.setVariables()
        self.mainProcess()
        self.mf.startLoop()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, ["ctlnot", "ctlpwu"], prog=__name__)
        t = time.localtime()
        self.sysdtw = ((t[0] * 10000) + (t[1] * 100) + t[2])

    def mainProcess(self):
        tit = ["%s Notes for - %s" % (finsys[self.sys][0], self.key.strip())]
        if self.conum:
            tit.insert(0, "%03i %s" % (self.conum, self.conam))
        cal = {
            "stype": "D"}
        self.ntm = {
            "stype": "R",
            "tables": ("ctlnot",),
            "title": "Notes",
            "cols": (
                ("not_date", "", 0, "Cap-Date"),
                ("not_user", "", 0, "User"),
                ("not_aflag", "", 0, "F"),
                ("not_adate", "", 0, "Action-Dte"),
                ("not_auser", "", 0, "Action-User"),
                ("not_desc", "", 0, "Description"),
                ("not_seq", "", 0, "Sequence")),
            "where": [("not_cono", "=", self.conum), ("not_sys", "=",
                self.sys), ("not_key", "=", self.key)],
            "order": "not_seq",
            "comnd": self.selectNote}
        r1s = (("Normal", "N"), ("Urgent", "U"))
        fld = (
            (("T",0,0,0),("IRB",r1s),0,"Action Flag","Action Flag",
                "N","N",self.doActFlag,None,None,None),
            (("T",0,1,0),"I@not_adate",0,"","",
                self.sysdtw,"N",self.doActDate,cal,None,("efld",)),
            (("T",0,2,0),"I@not_auser",0,"","",
                self.user,"N",self.doActUser,None,None,("efld",)),
            (("T",0,3,0),"ITV",50,"Details","Details",
                "","N",self.doNote,None,None,("notblank",)))
        tnd = ((self.doEnd,"y"), )
        txt = (self.doExit, )
        but = (
            ("Save",None,self.doSave,0,("T",0, 3),("T",0,0)),
            ("Show",None,self.doShow,0,("T",0, 1),("T",0,2)),
            ("Quit",None,self.doExit,1,None,None),
            ("Cancel",None,self.doExit,0,("T",0, 2),None))
        self.df = TartanDialog(self.mf, tops=True, title=tit, eflds=fld,
            tend=tnd, txit=txt, butt=but)

    def doActFlag(self, frt, pag, r, c, p, i, w):
        self.actflag = w

    def doActDate(self, frt, pag, r, c, p, i, w):
        if w and w < self.sysdtw:
            return "Action Date in the Past"
        self.actdate = w

    def doActUser(self, frt, pag, r, c, p, i, w):
        usr = self.sql.getRec(tables="ctlpwu", where=[("usr_name", "=", w)],
            limit=1)
        if not usr:
            return "Invalid User Name"
        self.actuser = w

    def doShow(self, widget=None):
        self.df.mstAlw.hide()
        if not self.df.selectItem(0, self.ntm):
            self.df.mstAlw.show()
            self.df.focusField("T", 0, 1)

    def selectNote(self, frt, pag, r, c, p, i, w):
        state = self.df.disableButtons()
        self.chgflag = w[2]
        self.chgdate = w[3]
        self.chguser = w[4]
        self.nseq = w[-1:][0]
        tit = ("Notes Editing",)
        r1s = (("Normal","N"),("Urgent","U"),("Completed","C"))
        fld = (
            (("T",0,0,0),"O@not_date",0,""),
            (("T",0,1,0),"O@not_user",0,""),
            (("T",0,2,0),"OTv",50,"Details","Details",
                "","N",None,None,None,("notblank",)),
            (("T",0,3,0),("IRB",r1s),0,"Action Flag (C/N/U)","Action Flag",
                "N","N",self.doChgFlag,None,None,None),
            (("T",0,4,0),"I@not_adate",0,"","",
                "","N",self.doChgDate,None,None,("efld",)),
            (("T",0,5,0),"I@not_auser",0,"","",
                "","N",self.doChgUser,None,None,("efld",)))
        but = (
            ("Accept",None,self.doNEnd,1,None,None),
            ("Cancel",None,self.doNExit,1,None, None))
        tnd = ((self.doNEnd, "n"), )
        txt = (self.doNExit, )
        self.nf = TartanDialog(self.mf, newscn=(0,1), tops=True, title=tit,
            eflds=fld, butt=but, tend=tnd, txit=txt, focus=False)
        self.nf.loadEntry("T", 0, 0, data=w[0])
        self.nf.loadEntry("T", 0, 1, data=w[1])
        self.nf.loadEntry("T", 0, 2, data=w[5])
        self.nf.topEntry[0][2].set_wrap_mode(gtk.WRAP_WORD)
        self.nf.loadEntry("T", 0, 3, data=w[2])
        self.nf.loadEntry("T", 0, 4, data=w[3])
        self.nf.loadEntry("T", 0, 5, data=w[4])
        if self.chgflag == "C":
            self.nf.setWidget(self.nf.B0, "normal")
            self.nf.setWidget(self.nf.B1, "normal")
        else:
            self.nf.focusField("T", 0, 4)
        self.mf.startLoop()
        self.df.enableButtons(state=state)

    def doChgFlag(self, frt, pag, r, c, p, i, w):
        self.chgflag = w
        if self.chgflag == "C":
            return "nd"

    def doChgDate(self, frt, pag, r, c, p, i, w):
        self.chgdate = w

    def doChgUser(self, frt, pag, r, c, p, i, w):
        usr = self.sql.getRec(tables="ctlpwu", where=[("usr_name", "=", w)],
            limit=1)
        if not usr:
            return "Invalid User Name"
        self.chguser = w
        return "nd"

    def doNEnd(self, widget=None):
        self.sql.updRec("ctlnot", cols=["not_aflag", "not_adate", "not_auser"],
            data=[self.chgflag, self.chgdate, self.chguser], where=[("not_seq",
            "=", self.nseq)])
        self.mf.dbm.commitDbase()
        self.doNExit()

    def doNExit(self, widget=None):
        self.nf.closeProcess()
        self.mf.closeLoop()
        self.doShow()

    def doSave(self, widget=None):
        wid, self.note = self.df.getEntry("T", 0, 3)
        self.doEnd()

    def doNote(self, frt, pag, r, c, p, i, w):
        self.note = w

    def doEnd(self, widget=None):
        self.df.closeProcess()
        self.sql.insRec("ctlnot", data=[self.conum, self.sys, self.key,
            self.sysdtw, self.user, self.note, self.actflag, self.actdate,
            self.actuser, 0])
        if self.commit:
            self.mf.dbm.commitDbase()
        self.mf.closeLoop()

    def doExit(self, widget=None):
        self.df.closeProcess()
        self.mf.closeLoop()

class NotesPrint:
    def __init__(self, mf, conum, conam, sys):
        self.mf = mf
        self.conum = conum
        self.conam = conam
        self.sys = sys
        self.setVariables()
        self.mainProcess()
        self.mf.startLoop()

    def setVariables(self):
        self.sql = Sql(self.mf.dbm, "ctlnot", prog=__name__)
        t = time.localtime()
        self.sysdtw = ((t[0] * 10000) + (t[1] * 100) + t[2])
        self.data = []

    def mainProcess(self):
        tit = ("%03i %s" % (self.conum, self.conam),
            "%s Masterfile Notes" %  finsys[self.sys][0])
        r1s = (("All", "A"), ("Normal", "N"), ("Urgent", "U"))
        fld = (
            (("T",0,0,0),("IRB",r1s),0,"Action Flag","",
                "A","N",self.doFlag,None,None,None),
            (("T",0,1,0),"Id1",10,"From Capture Date","",
                0,"N",self.doFDate,None,None,("efld",)),
            (("T",0,2,0),"I@not_date",0,"To   Capture Date","",
                self.sysdtw,"N",self.doTDate,None,None,("efld",)),
            (("T",0,3,0),"I@not_adate",0,"From Action Date","",
                "","N",self.doFADate,None,None,("efld",)),
            (("T",0,4,0),"I@not_adate",0,"To   Action Date","",
                "","N",self.doTADate,None,None,("efld",)))
        tnd = ((self.doEnd,"y"), )
        txt = (self.doExit, )
        self.df = TartanDialog(self.mf, title=tit, eflds=fld, tend=tnd,
            txit=txt, view=("N","V"), mail=("Y","N"))

    def doFlag(self, frt, pag, r, c, p, i, w):
        if w == "A":
            self.flag = ""
        else:
            self.flag = w


    def doFDate(self, frt, pag, r, c, p, i, w):
        self.fdate = w

    def doTDate(self, frt, pag, r, c, p, i, w):
        if w < self.fdate:
            return "Invalid Date, Earlier than From Date"
        self.tdate = w

    def doFADate(self, frt, pag, r, c, p, i, w):
        self.fadate = w

    def doTADate(self, frt, pag, r, c, p, i, w):
        if w < self.fadate:
            return "Invalid Date, Earlier than From Date"
        self.tadate = w

    def doEnd(self, widget=None):
        self.view = self.df.repprt[0][1]
        self.pnam = self.df.repprt[0][2]
        self.mail = self.df.repprt[1][2]
        self.df.closeProcess()
        tab = ["ctlnot"]
        whr = [("not_cono", "=", self.conum), ("not_sys", "=", self.sys)]
        if self.flag:
            whr.extend([("not_aflag", "=", self.flag), ("not_date", ">=",
                self.fdate)])
        else:
            whr.append(("not_date", ">=", self.fdate))
        if self.tdate:
            whr.append(("not_date", "<=", self.tdate))
        whr.append(("not_adate", ">=", self.fadate))
        if self.tadate:
            whr.append(("not_adate", "<=", self.tadate))
        ord = "not_key, not_date, not_seq"
        self.data = self.sql.getRec(tables=tab, where=whr, order=ord)
        self.mf.closeLoop()

    def doExit(self, widget=None):
        self.df.closeProcess()
        self.mf.closeLoop()

class AskForData:
    def __init__(self, mf, label=None):
        self.data = ""
        tit = "Please Enter the Required Data"
        fld = ((("T",0,0,0),"ITX",30,label,"",
            "","N",None,None,None,("notblank",)),)
        self.df = TartanDialog(mf, tops=True, title=tit, newscn=(0,0),
            eflds=fld, tend=((self.doEnd,"n"),), txit=((self.doExit),))
        gtk.main()

    def doEnd(self):
        self.data = self.df.t_work[0][0][0]
        self.doExit()

    def doExit(self):
        self.df.closeProcess()
        gtk.main_quit()

class FileImport:
    """
    This is a file import class.

    mf   - The Mainframe class.
    args - Must contain the following:
        
            imptab  - A table to import

                                or

            impcol - A list of lists of field parameters e.g.
                    [[Label, Column Number, Format, Size]] where:
                        Label           - The column name
                        Column Number   - The default column number from 0
                        Format          - The data format as per CCD
                        Size            - The data size as per CCD

           And additionally can contain the following:

            impskp - Fields to skip with table option.
            impfld - Any additional dialog fields.
            impfle - The file name to import else blank to select.
            impdlg - Whether to have a dialog, True or False. Default is True.
    """
    def __init__(self, mf, **args):
        self.mf = mf
        self.impdat = []
        self.impfld = []
        self.impfle = ""
        self.impdlg = True
        self.impign = "N"
        if "impcol" in args:
            self.impcol = args["impcol"]
            sql = Sql(self.mf.dbm, "ffield")
            for num, col in enumerate(self.impcol):
                if col[2][0] == "@":
                    dat = sql.getRec(tables="ffield", cols=["ff_type",
                        "ff_size"], where=[("ff_name", "=", col[2][1:])],
                        limit=1)
                    self.impcol[num][2] = dat[0]
                    self.impcol[num][3] = dat[1]
        elif "imptab" in args:
            self.imptab = args["imptab"]
            self.impcol = []
            if "impskp" in args:
                self.impskp = args["impskp"]
            sql = Sql(self.mf.dbm, "ffield")
            acc = sql.getRec(tables="ffield", where=[("ff_tabl", "=",
                self.imptab)], order="ff_seq")
            if not acc:
                showError(self.mf.body, head="Import Error",
                    mess="Invalid table %s in List of Variables" % self.imptab)
                return
            count = 0
            for a in acc:
                if self.impskp and a[2] in self.impskp:
                    continue
                self.impcol.append([a[5], count, a[3], a[4]])
                count += 1
        else:
            showError(self.mf.body, head="Import Error",
                mess="Missing 'impcol' and/or 'table' in List of Variables")
            return
        if "impfld" in args:
            self.impfld = args["impfld"]
        if "impfle" in args:
            self.impfle = args["impfle"]
        if "impdlg" in args:
            self.impdlg = args["impdlg"]
        if self.impdlg or not self.impfle:
            self.impdlg = True
            self.doDialog()
        else:
            self.doImpEnd()

    def doDialog(self):
        max = 0
        for field in self.impcol:
            if len(field[0]) > max:
                max = len(field[0])
        tit = "Import File Parameters"
        if sys.platform == "win32":
            self.lastdir = os.path.join(self.mf.rcdic["wrkdir"], "impdir")
        else:
            self.lastdir = os.path.join(self.mf.rcdic["wrkdir"], ".impdir")
        if os.path.exists(self.lastdir):
            infle = open(self.lastdir, "r")
            name = infle.readline().replace("\n", "")
            infle.close()
        else:
            name = self.mf.rcdic["wrkdir"]
        if XLWT:
            typ = ("CSV & XLS Files", "*.[c,x][s,l][v,s]")
        else:
            typ = ("CSV Files", "*.csv")
        fle = {
            "stype": "F",
            "types": "fle",
            "initd": name,
            "ftype": (typ,)}
        r1s = (("Yes", "Y"), ("No", "N"))
        idx = len(self.impfld)
        if self.impfle:
            self.impfld.append((("T",0,idx,0),"OFF",50,"File Name"))
        else:
            self.impfld.append((("T",0,idx,0),"IFF",50,"File Name","",
                "","N",self.doImpFle,fle,None,("file",)))
        idx += 1
        self.impfld.append((("T",0,idx,0),("IRB",r1s),0,"Ignore Invalid Lines",
            "","N","N",self.doImpIgn,None,None,None))
        idx += 1
        for num in range(0, len(self.impcol), 2):
            pad = " " * (max - len(self.impcol[num][0]))
            self.impfld.append((("T", 0, idx + num, 0),"IUI",3,
                "%s%s - Column" % (self.impcol[num][0], pad),"",0,"N",
                self.doImpCol,None,None,("efld",)))
            if num != (len(self.impcol) - 1):
                new = (max + 58) - (max + 12)
                pad = " " * (max - len(self.impcol[num+1][0]))
                self.impfld.append((("T", 0, idx + num, new),"IUI",3,
                    "%s%s - Column" % (self.impcol[num+1][0], pad),"",0,"N",
                    self.doImpCol,None,None,("efld",)))
        but = (("Accept",None,self.doImpEnd,0,("T",0,idx+1),("T",0,idx),None),)
        self.ip = TartanDialog(self.mf, title=tit, newscn=(1,1), tops=True,
            eflds=self.impfld, tend=((self.doImpEnd,"n"),),
            txit=(self.doImpExit,), butt=but, focus=False)
        if self.impfle:
            self.ip.loadEntry("T", 0, 0, data=self.impfle)
            self.ip.focusField("T", 0, 2)
        else:
            self.ip.focusField("T", 0, 1)
        self.mf.startLoop()

    def doImpFle(self, frt, pag, r, c, p, i, w):
        self.impfle = w

    def doImpIgn(self, frt, pag, r, c, p, i, w):
        self.impign = w
        for num, col in enumerate(self.impcol):
            self.ip.loadEntry("T", 0, p+num+1, data=col[1])

    def doImpCol(self, frt, pag, r, c, p, i, w):
        self.impcol[p-2][1] = w

    def doImpEnd(self, widget=None):
        if self.impdlg:
            self.ip.closeProcess()
        imperr = False
        if self.impfle.split(".")[1].lower() == "csv":
            try:
                infle = open(self.impfle, "r")
                data = infle.readlines()
                self.impdat = []
                for num, line in enumerate(data):
                    dat = CsvCvt(line).output
                    if not dat:
                        continue
                    try:
                        lin = []
                        for col in self.impcol:
                            d = CCD(dat[col[1]], col[2], col[3])
                            if d.err:
                                raise Exception
                            lin.append(d.work)
                        self.impdat.append(lin)
                    except:
                        if self.impign != "Y":
                            showError(self.mf.body, "Column Error",
                                "Line Number %s is Invalid\n\n%s has "\
                                "Invalid Data %s" % (num, col[0], d.work))
                            raise Exception
            except:
                imperr = True
        elif self.impfle.split(".")[1].lower() in ("xls", "xlsx") and XLRD:
            try:
                workbk = xlrd.open_workbook(self.impfle)
                worksh = workbk.sheet_by_index(0)
                self.impdat = []
                for row in range(0, worksh.nrows):
                    try:
                        lin = []
                        for col in self.impcol:
                            dat = worksh.cell_value(rowx=row, colx=col[1])
                            if col[2] in ("D1", "d1"):
                                try:
                                    dte = xldate_as_tuple(dat, workbk.datemode)
                                    dat = (dte[0]*10000) + (dte[1]*100) + dte[2]
                                except:
                                    dat = 0
                                    col[2] = "d1"
                            dat = CCD(dat, col[2], col[3])
                            if dat.err:
                                raise Exception
                            lin.append(dat.work)
                        self.impdat.append(lin)
                    except:
                        if self.impign != "Y":
                            showError(self.mf.body, "Column Error",
                                "This Line Number, %s, is Invalid" % row)
                            raise Exception
            except:
                imperr = True
        else:
            imperr = True
        if imperr:
            self.doImpExit()
        try:
            infle = open(self.lastdir, "w")
            infle.write(os.path.dirname(os.path.normpath(self.impfle)))
            infle.close()
        except:
            pass
        if self.impdlg:
            self.mf.closeLoop()

    def doImpExit(self, widget=None):
        self.impdat = []
        self.doImpClose()

    def doImpClose(self):
        if self.impdlg:
            self.ip.closeProcess()
            self.mf.closeLoop()
# END
